// -*- Mode:C++ -*-
//Header:
//File: bipredutil.h
//Title: built-in predicate 
//Author: NODA Itsuki
//Date: 1998/6/19
//EndHeader:

//ModifyHistory:
// 1998/6/19: Create this file
//EndModifyHistory:

//Memo:
//EndMemo:

#ifndef _bip_util_h_
#define _bip_util_h_
////////////////////////////////////////////////////////////////////////

extern "C" {
#include <math.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
}
#include <strstream.h>

#include "gem.h"

//======================================================================
// Built-In Predicate Utility
//

//----------------------------------------------------------------------
// Arity Check

inline Bool chkArity(GaeaData& gterm, UInt n) {
   return isGaeaTerm(gterm) && _TermArity(gterm) == n ;
} ;
inline Bool chkArity(GaeaData& gterm, UInt mn, UInt mx) {
   return isGaeaTerm(gterm) 
      && _TermArity(gterm) >= mn 
      && _TermArity(gterm) <= mx ;
} ;
inline Bool chkArityGE(GaeaData& gterm, UInt n) {
   return isGaeaTerm(gterm) && _TermArity(gterm) >= n ;
} ;


inline Bool chkHeadArity(GaeaData& gterm, GaeaData& head, UInt n) {
   return isGaeaTerm(gterm) 
      && _TermHead(gterm) == head.val.symbol
      && _TermArity(gterm) == n ;
} ;
inline Bool chkHeadArity(GaeaData& gterm, GaeaData& head, UInt mn, UInt mx) {
   return isGaeaTerm(gterm) 
      && _TermHead(gterm) == head.val.symbol
      && _TermArity(gterm) >= mn 
      && _TermArity(gterm) <= mx ; 
} ;


//----------------------------------------------------------------------
// Get Argv
//

inline InstGaeaData _fetchedArgv(InstGaeaData& igd, UInt n) {
   InstGaeaData r = fetch(_TermArgv(igd.data, n),igd.binding) ;
   return r ;
} ;

inline GaeaData fetchedArgv(InstGaeaData& igd, UInt n) {
   return _fetchedArgv(igd,n).data ;
} ;

inline GaeaData fetchedValueArgv(InstGaeaData& igd,UInt n,Gem* gem) {
   return fetchValue(InstGaeaData(_TermArgv(igd.data, n),igd.binding),
		     &(gem->uiheap)) ;
} ;

//----------------------------------------------------------------------
// Argv Conversion Macros

#define PICKUP_CURRENT_GOAL(icgoal,cgoal,gem) \
   InstGaeaData icgoal = gem->workingGoal ;\
   GaeaData cgoal = icgoal.data ;

#define PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,arity) \
   InstGaeaData icgoal = gem->workingGoal ;\
   GaeaData cgoal = icgoal.data ;\
   if(!chkArity(cgoal,arity)) return bipNoMatch(gem) ;

#define PICKUP_CURRENT_GOAL_CHECK_ARITY2(icgoal,cgoal,gem,art_min,art_max) \
   InstGaeaData icgoal = gem->workingGoal ;\
   GaeaData cgoal = icgoal.data ;\
   if(!chkArity(cgoal,art_min,art_max)) return bipNoMatch(gem) ;

#define PICKUP_CURRENT_GOAL_CHECK_NOARITY(icgoal,cgoal,gem) \
   InstGaeaData icgoal = gem->workingGoal ;\
   GaeaData cgoal = icgoal.data ;\
   if(!isGaeaSymbol(cgoal)) return bipNoMatch(gem) ;

#define PICKUP_CURRENT_GOAL_CHECK_ARITY_NOorN(icgoal,cgoal,gem,arity) \
   InstGaeaData icgoal = gem->workingGoal ;\
   GaeaData cgoal = icgoal.data ;\
   if(!isGaeaSymbol(cgoal)) {\
      if(!chkArity(cgoal,arity)) return bipNoMatch(gem) ;\
   }\

#define PICKUP_CURRENT_GOAL_CHECK_ARITY_NOorN2(icgoal,cgoal,gem,art_min,art_max) \
   InstGaeaData icgoal = gem->workingGoal ;\
   GaeaData cgoal = icgoal.data ;\
   if(!isGaeaSymbol(cgoal)) {\
      if(!chkArity(cgoal,art_min,art_max)) return bipNoMatch(gem) ;\
   }\

#define ARGVALCONV0(V,gem,icgoal,n,checkfunc,convfunc,errmsg) \
    {\
	GaeaData gdt = fetchedArgv(icgoal,n) ;\
        if(!checkfunc##(gdt)) return bipError(gem,errmsg,n) ;\
        V = convfunc##(gdt) ;\
    }

#define ARGVALCONV(T,V,gem,icgoal,n,checkfunc,convfunc,errmsg) \
    T V ;\
    {\
	GaeaData gdt = fetchedArgv(icgoal,n) ;\
        if(!checkfunc##(gdt)) return bipError(gem,errmsg,n) ;\
        V = convfunc##(gdt) ;\
    }

#define ARGVAL(V,gem,icgoal,n,checkfunc,errmsg) \
    GaeaData V = fetchedArgv(icgoal,n) ;\
    if(!checkfunc##(V)) return bipError(gem,errmsg,n) ;

#define ARGVAL0(V,gem,icgoal,n,checkfunc,errmsg) \
    V = fetchedArgv(icgoal,n) ;\
    if(!checkfunc##(V)) return bipError(gem,errmsg,n) ;

#define ARGVAL_INT(V,gem,icgoal,n) \
    ARGVALCONV(Int,V,gem,icgoal,n,isGaeaInteger,IntVal,"Illegal integer value")
#define ARGVAL0_INT(V,gem,icgoal,n) \
    ARGVALCONV0(V,gem,icgoal,n,isGaeaInteger,IntVal,"Illegal integer value")

#define ARGVAL_TINT(V,gem,icgoal,n) \
    ARGVALCONV(Int,V,gem,icgoal,n,isGNumber,IntVal,"Illegal integer value")
#define ARGVAL0_TINT(V,gem,icgoal,n) \
    ARGVALCONV0(V,gem,icgoal,n,isGNumber,IntVal,"Illegal integer value")

#define ARGVAL_FLT(V,gem,icgoal,n) \
    ARGVALCONV(Float,V,gem,icgoal,n,isGNumber,FltVal,"Illegal float value")
#define ARGVAL0_FLT(V,gem,icgoal,n) \
    ARGVALCONV0(V,gem,icgoal,n,isGNumber,FltVal,"Illegal float value")

#define ARGVAL_SYM(V,gem,icgoal,n) \
    ARGVALCONV(Symbol*,V,gem,icgoal,n,isGaeaSymbol,_SymVal,"Illegal symbol")
#define ARGVAL0_SYM(V,gem,icgoal,n) \
    ARGVALCONV0(V,gem,icgoal,n,isGaeaSymbol,_SymVal,"Illegal symbol")

#define ARGVAL_STR(V,gem,icgoal,n) \
    ARGVALCONV(String,V,gem,icgoal,n,isGaeaString,_StrVal,"Illegal string")
#define ARGVAL0_STR(V,gem,icgoal,n) \
    ARGVALCONV0(V,gem,icgoal,n,isGaeaString,_StrVal,"Illegal string")

#define ARGVAL_CHARPTR(V,gem,icgoal,n) \
    ARGVALCONV(char*,V,gem,icgoal,n,isGaeaString,_StrVal,"Illegal string")
#define ARGVAL0_CHARPTR(V,gem,icgoal,n) \
    ARGVALCONV0(V,gem,icgoal,n,isGaeaString,_StrVal,"Illegal string")

#define ARGVAL_TCHARPTR(V,gem,icgoal,n) \
    ARGVALCONV(char*,V,gem,icgoal,n,isTGaeaString,_TStrVal,"Illegal string")
#define ARGVAL0_TCHARPTR(V,gem,icgoal,n) \
    ARGVALCONV0(V,gem,icgoal,n,isTGaeaString,_TStrVal,"Illegal string")

#define ARGVAL_DAT(V,gem,icgoal,n) \
    GaeaData V = fetchedArgv(icgoal,n) ;

#define ARGVAL_VAR(V,gem,icgoal,n) \
    InstGaeaData V = _fetchedArgv(icgoal,n) ;
#define ARGVAL0_VAR(V,gem,icgoal,n) \
    V = _fetchedArgv(icgoal,n) ;

#define ARGVAL_EXTOBJ(T,V,gem,icgoal,n,checkfunc,errmsg) \
    T V ;\
    {\
	GaeaData gdt = fetchedArgv(icgoal,n) ;\
        if(!checkfunc##(gdt)) return bipError(gem,errmsg,n) ;\
        V = (T)_ExObjVal(gdt) ;\
    }
#define ARGVAL0_EXTOBJ(T,V,gem,icgoal,n,checkfunc,errmsg) \
    {\
	GaeaData gdt = fetchedArgv(icgoal,n) ;\
        if(!checkfunc##(gdt)) return bipError(gem,errmsg,n) ;\
        V = (T)_ExObjVal(gdt) ;\
    }



//----------------------------------------------------------------------
// Unify With Atom
//

inline Bool unifyWith(GaeaData dt, const InstGaeaData& var, Gem* gem) {
   /* dt should be atom */
   return unify(InstGaeaData(dt,gem->currentGoal().binding),var,
		&(gem->uiheap)) ;
} ;
 

//----------------------------------------------------------------------
// Unify With List
//

inline Bool unifyWithListShift(InstGaeaData& list, 
			       const InstGaeaData& car,
			       Gem* gem) {
   Binding* newbinding = gem->uiheap.newBinding() ;
   if(!unify(list,InstGaeaData(ConsTerm,newbinding),&(gem->uiheap)))
      return False ;
   if(!unify(car,InstGaeaData(ConsCarVar,newbinding),&(gem->uiheap)))
      return False ;
   list = InstGaeaData(ConsCdrVar,newbinding) ;
   return True ;
}

inline Bool unifyWithListShift(InstGaeaData& list,
			       const GaeaData& car,
			       Gem* gem) {
   /* car should be atom */
   return unifyWithListShift(list,
			     InstGaeaData(car,gem->currentGoal().binding),
			     gem) ;
}

//----------------------------------------------------------------------
// Return 
//

inline BipStatus bipSuccess0(Gem *gem) {
   return BIP_Success ;
} ;

inline BipStatus bipSuccess(Gem *gem) {
   gem->popAndState() ;
   return bipSuccess0(gem) ;
} ;

inline BipStatus bipNoMatch(Gem *gem) {
   return BIP_NoMatch ;
} ;

inline BipStatus bipCutFail(Gem *gem) {
   gem->popAndState() ;
   return BIP_CutFail ;
} ;

inline BipStatus bipError(Gem *gem, char* msg, InstGaeaData& igd) {
   *(gem->estrm) << "Error: " << msg << ":" << fetchValue(igd,&gem->uiheap) 
		 << endl ;
//   gem->popAndState() ;
   return BIP_Error ;
} ;

inline BipStatus bipError(Gem *gem, char* msg) {
   return bipError(gem,msg,gem->currentGoal()) ;
}

inline BipStatus bipError(Gem *gem, char* msg, Int n) {
   ostrstream ostr ;
   ostr << msg << " (at argument " << n << ")" << ends ;
   return bipError(gem,ostr.str(),gem->currentGoal()) ;
}

//----------------------------------------------------------------------
// List Operation
//

inline Bool isInstConsTerm(const InstGaeaData& idt) {
   GaeaData dt = fetch(idt).data ;
   return isGaeaTerm(dt) && isConsTerm(dt) ;
}

inline InstGaeaData InstConsCarVal(const InstGaeaData& idt) {
   GaeaData dt = fetch(idt).data ;
   Binding* b = fetch(idt).binding ;

   return fetch(_ConsCar(dt),b) ;
}

inline InstGaeaData InstConsCdrVal(const InstGaeaData& idt) {
   GaeaData dt = fetch(idt).data ;
   Binding* b = fetch(idt).binding ;

   return fetch(_ConsCdr(dt),b) ;
}


////////////////////////////////////////////////////////////////////////
#endif


   

