// -*- Mode:C++ -*-
//Header:
//File: bipred_arith.cc
//Title: built-in predicate for arithmetic calculation
//Author: NODA Itsuki
//Date: 1998/7/21
//EndHeader:

//ModifyHistory:
// 1998/7/21: Create this file
//EndModifyHistory:

//Memo:
//EndMemo:

#include "bipred.h"

//======================================================================
// Initialize (external)
//

static Bool bipInit() ;

extern "C" {
   Bool bipredInit_arith() {
      static Bool initializedp = False ;
      if(!initializedp) {
	 bipInit() ;
	 initializedp = True ;
	 return True ;
      } else {
	 return False ;
      }
   } ;
} ;


//======================================================================
// Global Variables
//

GaeaData A_PISymbol ;
GaeaData A_AddSymbol ;
GaeaData A_SubSymbol ;
GaeaData A_MulSymbol ;
GaeaData A_DivSymbol ;
GaeaData A_IRandSymbol ;
GaeaData A_FRandSymbol ;

//======================================================================
// Arithmetic
//

//----------------------------------------------------------------------
// inc

BipStatus bip_inc(Gem* gem) {
   /*	$inc(+V,-IncV)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

   GaeaData val = fetchedArgv(icgoal,0);
   InstGaeaData var = _fetchedArgv(icgoal,1);
   if(isGNumber(val)) {
      if(unifyWith(GaeaInteger(IntVal(val)+1),var,gem))
	 return bipSuccess(gem) ;
      else
	 return bipCutFail(gem) ;
   } 
}

//----------------------------------------------------------------------
// dec

BipStatus bip_dec(Gem* gem) {
   /*	$dec(+V,-DecV)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

   GaeaData val = fetchedArgv(icgoal,0);
   InstGaeaData var = _fetchedArgv(icgoal,1);
   if(isGNumber(val)) {
      if(unifyWith(GaeaInteger(IntVal(val)-1),var,gem))
	 return bipSuccess(gem) ;
      else
	 return bipCutFail(gem) ;
   } 
}


//----------------------------------------------------------------------
// compare

BipStatus bip_comp_gt(Gem* gem) {
   /*	$comp_gt(+Great,+Less)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;
   
   GaeaData x = fetchedArgv(icgoal,0);
   GaeaData y = fetchedArgv(icgoal,1);
   if(isGNumber(x)&&isGNumber(y)) {
      if(FltVal(x)>FltVal(y))
	 return bipSuccess(gem) ;
      else
	 return bipCutFail(gem) ;
   } else {
      return bipError(gem,"Not a Number") ;
   }
}

BipStatus bip_comp_ge(Gem* gem) {
   /*	$comp_ge(+Great,+Less)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

   GaeaData x = fetchedArgv(icgoal,0);
   GaeaData y = fetchedArgv(icgoal,1);
   if(isGNumber(x)&&isGNumber(y)) {
      if(FltVal(x)>=FltVal(y))
	 return bipSuccess(gem) ;
      else
	 return bipCutFail(gem) ;
   } else {
      return bipError(gem,"Not a Number") ;
   }
}

BipStatus bip_comp_lt(Gem* gem) {
   /*	$comp_lt(+Less,+Great)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

   GaeaData x = fetchedArgv(icgoal,0);
   GaeaData y = fetchedArgv(icgoal,1);
   if(isGNumber(x)&&isGNumber(y)) {
      if(FltVal(x)<FltVal(y))
	 return bipSuccess(gem) ;
      else
	 return bipCutFail(gem) ;
   } else {
      return bipError(gem,"Not a Number") ;
   }
}

BipStatus bip_comp_le(Gem* gem) {
   /*	$comp_le(+Less,+Great)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

   GaeaData x = fetchedArgv(icgoal,0);
   GaeaData y = fetchedArgv(icgoal,1);
   if(isGNumber(x)&&isGNumber(y)) {
      if(FltVal(x)<=FltVal(y))
	 return bipSuccess(gem) ;
      else
	 return bipCutFail(gem) ;
   } else {
      return bipError(gem,"Not a Number") ;
   }
}


//======================================================================
// Is
//

GaeaData evalIsForm(InstGaeaData form, Gem* gem) {
   GaeaData f = form.data ;

   if(isGNumber(f)) return f ;
   if(isGaeaSymbol(f)) {
      if(eq(f,A_PISymbol))  return GaeaFloat(M_PI) ;
      else                return ErrorSymbol ;
   }
   /*	Add	*/
   if        (chkHeadArity(f,A_AddSymbol,2)) {
      GaeaData arg0 = evalIsForm(_fetchedArgv(form,0),gem) ;
      GaeaData arg1 = evalIsForm(_fetchedArgv(form,1),gem) ;
      if(!isGNumber(arg0) || !isGNumber(arg1)) 
	 return ErrorSymbol ;
      if(isGaeaInteger(arg0) && isGaeaInteger(arg1)) 
	 return GaeaInteger(IntVal(arg0) + IntVal(arg1)) ;
      else
	 return GaeaFloat(FltVal(arg0) + FltVal(arg1)) ;
   /*	Sub	*/
   } else if (chkHeadArity(f,A_SubSymbol,2)) {
      GaeaData arg0 = evalIsForm(_fetchedArgv(form,0),gem) ;
      GaeaData arg1 = evalIsForm(_fetchedArgv(form,1),gem) ;
      if(!isGNumber(arg0) || !isGNumber(arg1)) 
	 return ErrorSymbol ;
      if(isGaeaInteger(arg0) && isGaeaInteger(arg1)) 
	 return GaeaInteger(IntVal(arg0) - IntVal(arg1)) ;
      else
	 return GaeaFloat(FltVal(arg0) - FltVal(arg1)) ;
   /*	Mul	*/
   } else if (chkHeadArity(f,A_MulSymbol,2)) {
      GaeaData arg0 = evalIsForm(_fetchedArgv(form,0),gem) ;
      GaeaData arg1 = evalIsForm(_fetchedArgv(form,1),gem) ;
      if(!isGNumber(arg0) || !isGNumber(arg1)) 
	 return ErrorSymbol ;
      if(isGaeaInteger(arg0) && isGaeaInteger(arg1)) 
	 return GaeaInteger(IntVal(arg0) * IntVal(arg1)) ;
      else
	 return GaeaFloat(FltVal(arg0) * FltVal(arg1)) ;
   /*	Div	*/
   } else if (chkHeadArity(f,A_DivSymbol,2)) {
      GaeaData arg0 = evalIsForm(_fetchedArgv(form,0),gem) ;
      GaeaData arg1 = evalIsForm(_fetchedArgv(form,1),gem) ;
      if(!isGNumber(arg0) || !isGNumber(arg1)) 
	 return ErrorSymbol ;
      return GaeaFloat(FltVal(arg0) / FltVal(arg1)) ;
   /*	IRand	*/
   } else if (chkHeadArity(f,A_IRandSymbol,1,2)) {
      GaeaData min = GaeaInteger(0) ; 
      GaeaData max = evalIsForm(_fetchedArgv(form,0),gem) ;
      if(chkArity(f,2)) {
	 min = max ;
	 max = evalIsForm(_fetchedArgv(form,1),gem) ;
      }
      if(!isGNumber(min) || !isGNumber(max)) 
	 return ErrorSymbol ;
      return GaeaInteger(IntRand(IntVal(min),IntVal(max))) ;
   /*	FRand	*/
   } else if (chkHeadArity(f,A_FRandSymbol,1,2)) {
      GaeaData min = GaeaFloat(0.0) ; 
      GaeaData max = evalIsForm(_fetchedArgv(form,0),gem) ;
      if(chkArity(f,2)) {
	 min = max ;
	 max = evalIsForm(_fetchedArgv(form,1),gem) ;
      }
      if(!isGNumber(min) || !isGNumber(max)) 
	 return ErrorSymbol ;
      return GaeaFloat(FltRand(FltVal(min),FltVal(max))) ;
   /*	ERROR	*/
   } else {
      return ErrorSymbol ;
   }
}

BipStatus bip_is(Gem* gem) {
   /*	$comp_le(-Var,+Form)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

   InstGaeaData var = _fetchedArgv(icgoal,0) ;
   InstGaeaData form = _fetchedArgv(icgoal,1) ;
   
   GaeaData val = evalIsForm(form, gem) ;
   if(eq(val,ErrorSymbol)) {
      return bipError(gem,"Illegal IS form") ;
   }
   if(unifyWith(val, var, gem)) 
      return bipSuccess(gem) ;
   else 
      return bipCutFail(gem) ;
}

//======================================================================
// misc
//

//----------------------------------------------------------------------
// add integer

BipStatus bip_add_int(Gem* gem) {
   /*	$add_int(-XY,+X,+Y)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,3) ;

   ARGVAL_VAR(xy,gem,icgoal,0) ;
   ARGVAL_INT(x,gem,icgoal,1) ;
   ARGVAL_INT(y,gem,icgoal,2) ;

   if(unifyWith(GaeaInteger(x+y),xy,gem)) {
      return bipSuccess(gem) ;
   }
   else {
      return bipCutFail(gem) ;
   } 
}

//----------------------------------------------------------------------
// sub integer

BipStatus bip_sub_int(Gem* gem) {
   /*	$sub_int(-XY,+X,+Y)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,3) ;

   ARGVAL_VAR(xy,gem,icgoal,0) ;
   ARGVAL_INT(x,gem,icgoal,1) ;
   ARGVAL_INT(y,gem,icgoal,2) ;

   if(unifyWith(GaeaInteger(x-y),xy,gem)) {
      return bipSuccess(gem) ;
   }
   else {
      return bipCutFail(gem) ;
   } 
}




//======================================================================
// Initialize
//

Bool bipInit() {
   _SymVal(GaeaSymbol("$inc"))->setfunc(bip_inc) ;
   _SymVal(GaeaSymbol("$dec"))->setfunc(bip_dec) ;

   _SymVal(GaeaSymbol("$comp_gt"))->setfunc(bip_comp_gt) ;
   _SymVal(GaeaSymbol("$comp_ge"))->setfunc(bip_comp_ge) ;
   _SymVal(GaeaSymbol("$comp_lt"))->setfunc(bip_comp_lt) ;
   _SymVal(GaeaSymbol("$comp_le"))->setfunc(bip_comp_le) ;

   _SymVal(GaeaSymbol("$is"))->setfunc(bip_is) ;

   _SymVal(GaeaSymbol("$add_int"))->setfunc(bip_add_int) ;
   _SymVal(GaeaSymbol("$sub_int"))->setfunc(bip_sub_int) ;

   A_PISymbol = GaeaSymbol("pi") ;
   A_AddSymbol = GaeaSymbol("add") ;
   A_SubSymbol = GaeaSymbol("sub") ;
   A_MulSymbol = GaeaSymbol("mul") ;
   A_DivSymbol = GaeaSymbol("div") ;
   A_IRandSymbol = GaeaSymbol("irand") ;
   A_FRandSymbol = GaeaSymbol("frand") ;
//   A_SinSymbol = GaeaSymbol("sin") ;
//   A_CosSymbol = GaeaSymbol("cos") ;
//   A_TanSymbol = GaeaSymbol("tan") ;
//   A_LogSymbol = GaeaSymbol("log") ;

   return True ;
      
} ;



