// -*- Mode:C++ -*-
//Header:
//File: bipred_ctl.cc
//Title: built-in predicate 
//Author: NODA Itsuki
//Date: 1999/5/3
//EndHeader:

//ModifyHistory:
// 1999/5/3: Create this file
//EndModifyHistory:

//Memo:
//EndMemo:

#include "bipred.h"

//======================================================================
// Initialize (external)
//

static Bool bipInit() ;

extern "C" {
   Bool bipredInit_ctl() {
      static Bool initializedp = False ;
      if(!initializedp) {
	 bipInit() ;
	 initializedp = True ;
	 return True ;
      } else {
	 return False ;
      }
   } ;
} ;

//======================================================================
// True 
//

BipStatus bip_true(Gem* gem) {
   /*	$true	*/

   GaeaData cgoal = gem->workingGoal.data ;

   if(isGaeaSymbol(cgoal)) {
      return bipSuccess(gem) ;
   } else {
      return bipNoMatch(gem) ;
   }
} ;

//======================================================================
// False
//

BipStatus bip_fail(Gem* gem) {
   /*	$fail	*/

   GaeaData cgoal = gem->workingGoal.data ;

   if(isGaeaSymbol(cgoal)) {
      return bipCutFail(gem) ;
   } else {
      return bipNoMatch(gem) ;
   }
} ;

//======================================================================
// And
//

BipStatus bip_and(Gem* gem) {
   /*	$and(+Goal,+Goal,...)	*/

   InstGaeaData icgoal = gem->workingGoal ;
   UInt cutlevel = gem->astate->cutOrLevel ;
   
   if(isGaeaTerm(icgoal.data)) {
      gem->popAndState() ;
      for(UInt i = _TermArity(icgoal.data) ; i > 0 ; i--) {
	 gem->pushAndState(InstGaeaData(_TermArgv(icgoal.data,i-1),
				     icgoal.binding),
			   cutlevel) ;
      }
      return bipSuccess0(gem) ;
   } else {
      return bipNoMatch(gem) ;
   }
} ;

//======================================================================
// Or
//

BipStatus bip_or(Gem* gem) {
   /*	$or(+Goal,+Goal)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

   ARGVAL_VAR(goal1,gem,icgoal,0) ;
   ARGVAL_VAR(goal2,gem,icgoal,1) ;
   

   if(!chkArity(cgoal,2)) return bipNoMatch(gem) ;

   gem->popAndState() ;	// pop $or frame
   gem->pushAndState(goal2,gem->astate->cutOrLevel) ;  // push goal2
   gem->pushOrState() ;	// save goal2 
   gem->popAndState() ; // pop goal2
   gem->pushAndState(goal1,gem->orStateStack.toppos()) ; // push goal1
   // if cut operation discard or clause, then
   // pushAndState(goal1,gem->astate->cutOrLevel) is right.

   return bipSuccess0(gem) ;
}

//======================================================================
// Cut
//

BipStatus bip_cut(Gem* gem) {
   /*	$cut	*/

   InstGaeaData icgoal = gem->workingGoal ;
   GaeaData cgoal = icgoal.data ;
   UInt cutlevel = gem->astate->cutOrLevel ;
   
   if(cutlevel >0) cutlevel-- ;

   gem->popAndState() ;
   gem->orStateStack.setpos(cutlevel) ;
   
   return bipSuccess0(gem) ;
} ;

//======================================================================
// Cut & Fail
//

BipStatus bip_cut_fail(Gem* gem) {
   /*	$cut_fail */

   InstGaeaData icgoal = gem->workingGoal ;
   GaeaData cgoal = icgoal.data ;
   UInt cutlevel = gem->astate->cutOrLevel ;
   
   if(cutlevel >0) cutlevel-- ;

   gem->orStateStack.setpos(cutlevel) ;
   
   return bipCutFail(gem) ;
} ;

//======================================================================
// Cut & And
//

BipStatus bip_cut_and(Gem* gem) {
   /*	$cut_$and(+Goal,+Goal,...)	*/

   InstGaeaData icgoal = gem->workingGoal ;
   GaeaData cgoal = icgoal.data ;
   UInt cutlevel = gem->astate->cutOrLevel ;
   if(cutlevel >0) cutlevel-- ;

   gem->orStateStack.setpos(cutlevel) ;

   
   if(isGaeaTerm(icgoal.data)) {
      gem->popAndState() ;
      for(UInt i = _TermArity(icgoal.data) ; i > 0 ; i--) {
	 gem->pushAndState(InstGaeaData(_TermArgv(icgoal.data,i-1),
				     icgoal.binding),
			   cutlevel) ;
      }
      return bipSuccess0(gem) ;
   } else {
      return bipNoMatch(gem) ;
   }
} ;

//======================================================================
// Cut Level
//

BipStatus bip_save_cutlevel(Gem* gem) {
   /*	$save_cutlevel(-CutLevel) */

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;
   ARGVAL_VAR(vcutlevel,gem,icgoal,0) ;

   UInt cutlevel = gem->orStateStack.toppos() ;

   if(unifyWith(GaeaInteger(cutlevel),vcutlevel,gem)) 
      return bipSuccess(gem) ;
   else
      return bipCutFail(gem) ;
} ;

BipStatus bip_restore_cutlevel(Gem* gem) {
   /*	$restore_cutlevel(+CutLevel) */

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;
   ARGVAL_INT(cutlevel,gem,icgoal,0) ;

   if(cutlevel < 0 || cutlevel >= gem->orStateStack.toppos()) 
      return bipError(gem,"Illegal cutlevel") ;
   
   gem->orStateStack.setpos(cutlevel) ;
   
   return bipSuccess(gem) ;
} ;

//======================================================================
// Call Super
//

BipStatus bip_call_other(Gem* gem) {
   /*	$call_other(+Goal)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;

   UInt cutlevel = gem->astate->cutOrLevel ;

   gem->popAndState() ;
   AndState* new_astate = 
      gem->pushAndState(InstGaeaData(_TermArgv(icgoal.data,0),icgoal.binding),
			cutlevel) ;

   if(cutlevel > 0) {
      OrState* frame_ostate = &(gem->orStateStack.bottomN(cutlevel-1)) ;
      AndState* frame_astate = 
	 &(frame_ostate->andStateStackPointer.stacktop->entity) ;
      new_astate->copyGaeaClausePointer(*frame_astate) ;
   }

   return bipSuccess0(gem) ;
   
} ;

BipStatus bip_call_super(Gem* gem) {
   /*	$call_super(+Goal)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;

   UInt cutlevel = gem->astate->cutOrLevel ;

   gem->popAndState() ;
   AndState* new_astate = 
      gem->pushAndState(InstGaeaData(_TermArgv(icgoal.data,0),icgoal.binding),
			cutlevel) ;

   if(cutlevel > 0) {
      OrState* frame_ostate = &(gem->orStateStack.bottomN(cutlevel-1)) ;
      AndState* frame_astate = 
	 &(frame_ostate->andStateStackPointer.stacktop->entity) ;
      new_astate->copyGaeaClausePointer(*frame_astate) ;
      new_astate->gclpointer.top = 
	 new_astate->gclpointer.topcell->gclauseTable.root->previous ;
   }

   return bipSuccess0(gem) ;
   
} ;

//======================================================================
// Repeat
//

BipStatus bip_repeat(Gem* gem) {
   /*	$repeat	*/

   PICKUP_CURRENT_GOAL_CHECK_NOARITY(icgoal,cgoal,gem) ;

   gem->pushOrState() ;
   return bipSuccess(gem) ;
}

//======================================================================
// While
//

BipStatus bip_while(Gem* gem) {
   /*	$while(Cond)	*/
   
   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;

   ARGVAL_VAR(cond,gem,icgoal,0) ;

   gem->pushOrState() ;
   gem->popAndState() ;
   
   gem->pushAndState(InstGaeaData(FailSymbol,icgoal.binding),
		     gem->astate->cutOrLevel) ;
   gem->pushAndState(InstGaeaData(CutSymbol,icgoal.binding),
		     gem->orStateStack.toppos()) ;
   gem->pushOrState() ;
   gem->popAndState() ;// pop cut
   gem->popAndState() ;// pop fail

   gem->pushAndState(InstGaeaData(CutSymbol,icgoal.binding),
		     gem->orStateStack.toppos()) ;
   gem->pushAndState(cond,gem->orStateStack.toppos()) ;
   
   return bipSuccess0(gem) ;
}

//======================================================================
// DoTimes
//

BipStatus bip_dotimes(Gem* gem) {
   /*	$dotimes(-I,+Max)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2);

   InstGaeaData varc = _fetchedArgv(icgoal,0) ;
   GaeaData max = fetchedArgv(icgoal,1) ;
   if(!isGNumber(max)) return bipError(gem,"Illegal limit number") ;

   Int count ;
   if(isNoData(gem->astate->workdata)) {
      count = 0 ;
   } else {
      count = IntVal(gem->astate->workdata) ;
   }

   if(count >= IntVal(max)) {
      gem->astate->workdata.setNoData() ;	/* This is buggy */
      return bipCutFail(gem) ;
   }

   gem->astate->workdata = GaeaInteger(count+1) ;
   gem->pushOrState() ;

   if(unifyWith(GaeaInteger(count),varc,gem)) {
      return bipSuccess(gem) ;
   } else {
      gem->popOrState() ;
      gem->astate->workdata.setNoData() ;
      return bipCutFail(gem) ;
   }
}

//======================================================================
// For
//

BipStatus bip_for(Gem* gem) {
   /*	$for(-Var,+Init,+Max,+Inc [,+Goal]) */

   PICKUP_CURRENT_GOAL_CHECK_ARITY2(icgoal,cgoal,gem,4,5);

   ARGVAL_VAR(varc,gem,icgoal,0) ;
   ARGVAL_INT(init,gem,icgoal,1) ;
   ARGVAL_INT(max,gem,icgoal,2) ;
   ARGVAL_INT(step,gem,icgoal,3) ;

   Int count ;
   if(isNoData(gem->astate->workdata)) {
      count = init ;
   } else {
      count = IntVal(gem->astate->workdata) ;
   }

   if(chkArity(cgoal,5)) {
      ARGVAL_VAR(goal,gem,icgoal,4) ;
      if(count >= max) {
	 gem->astate->workdata.setNoData() ;	/* This is buggy */
	 return bipSuccess(gem) ;
      }

      gem->astate->workdata = GaeaInteger(count+step) ;
      gem->pushOrState() ;
	 
      if(unifyWith(GaeaInteger(count),varc,gem)) {
	 gem->pushAndState(InstGaeaData(FalseSymbol,
					gem->astate->igdata.binding),
			   gem->orStateStack.toppos()) ;
	 gem->pushAndState(goal,gem->orStateStack.toppos()) ;
	 return bipSuccess0(gem) ;
      } else {
	 gem->popOrState() ;
	 gem->astate->workdata.setNoData() ;
	 return bipCutFail(gem) ;
      }
   } else {
      if(count >= max) {
	 gem->astate->workdata.setNoData() ;	/* This is buggy */
	 return bipCutFail(gem) ;
      }

      gem->astate->workdata = GaeaInteger(count+step) ;
      gem->pushOrState() ;

      if(unifyWith(GaeaInteger(count),varc,gem)) {
	 return bipSuccess(gem) ;
      } else {
	 gem->popOrState() ;
	 gem->astate->workdata.setNoData() ;
	 return bipCutFail(gem) ;
      }
   }
}

//======================================================================
// Eq
//

BipStatus bip_eq(Gem* gem) {
   /*	$eq(X,Y)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

   InstGaeaData x = _fetchedArgv(icgoal,0) ;
   InstGaeaData y = _fetchedArgv(icgoal,1) ;
   if(Unify(x,y,&(gem->uiheap))) {
      return bipSuccess(gem) ;
   } else {
      return bipCutFail(gem) ;
   }
}

//======================================================================
// Trigger for debug
//
BipStatus bip_nop(Gem* gem) {
   /*	$nop	*/

   PICKUP_CURRENT_GOAL_CHECK_NOARITY(icgoal,cgoal,gem) ;

   return bipSuccess(gem) ;
} ;

BipStatus bip_nop1(Gem* gem) {
   /*	$nop1(Arg)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;

   ARGVAL_VAR(cons,gem,icgoal,0) ;
   
   for(InstGaeaData c = cons ; isInstConsTerm(c) ; c = InstConsCdrVal(c)) {
      InstGaeaData car = InstConsCarVal(c) ;
      *(gem->ostrm) << car.data << endl ;
   }

   return bipSuccess(gem) ;
} ;

//======================================================================
// Initialize (external)
//

static Bool bipInit() {

   _SymVal(TrueSymbol)->setfunc(bip_true) ;
   _SymVal(FailSymbol)->setfunc(bip_fail) ;
   _SymVal(AndSymbol)->setfunc(bip_and) ;
   _SymVal(OrSymbol)->setfunc(bip_or) ;
   _SymVal(CutSymbol)->setfunc(bip_cut) ;
   _SymVal(GaeaSymbol("!"))->setfunc(bip_cut) ;
   _SymVal(GaeaSymbol("cut"))->setfunc(bip_cut) ;
   _SymVal(GaeaSymbol("$cut_fail"))->setfunc(bip_cut_fail) ;
   _SymVal(GaeaSymbol("$cut_and"))->setfunc(bip_cut_and) ;

   _SymVal(GaeaSymbol("$save_cutlevel"))->setfunc(bip_save_cutlevel) ;
   _SymVal(GaeaSymbol("$restore_cutlevel"))->setfunc(bip_restore_cutlevel) ;

   _SymVal(GaeaSymbol("$call_other"))->setfunc(bip_call_other) ;
   _SymVal(GaeaSymbol("$call_super"))->setfunc(bip_call_super) ;

   _SymVal(GaeaSymbol("$repeat"))->setfunc(bip_repeat) ;
   _SymVal(GaeaSymbol("$while"))->setfunc(bip_while) ;
   _SymVal(GaeaSymbol("$dotimes"))->setfunc(bip_dotimes) ;
   _SymVal(GaeaSymbol("$for"))->setfunc(bip_for) ;

   _SymVal(GaeaSymbol("$eq"))->setfunc(bip_eq) ;

   _SymVal(GaeaSymbol("$nop"))->setfunc(bip_nop) ;
   _SymVal(GaeaSymbol("$nop1"))->setfunc(bip_nop1) ;

} ;





