// -*- Mode:C++ -*-
//Header:
//File: bipred_data.cc
//Title: built-in predicate 
//Author: NODA Itsuki
//Date: 1999/5/3
//EndHeader:

//ModifyHistory:
// 1999/5/3: Create this file
//EndModifyHistory:

//Memo:
//EndMemo:

#include "bipred.h"

//======================================================================
// Initialize (external)
//

static Bool bipInit() ;

extern "C" {
   Bool bipredInit_data() {
      static Bool initializedp = False ;
      if(!initializedp) {
	 bipInit() ;
	 initializedp = True ;
	 return True ;
      } else {
	 return False ;
      }
   } ;
} ;

//======================================================================
// unify
//

//--------------------------------------------------
// unify

BipStatus bip_unify(Gem* gem) {
   /*	$unify(-Var1,-Var2)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

   ARGVAL_VAR(xvar,gem,icgoal,0) ;
   ARGVAL_VAR(yvar,gem,icgoal,1) ;

   if(unify(xvar,yvar,&(gem->uiheap)))
      return bipSuccess(gem) ;
   else
      return bipCutFail(gem) ;
}

#ifndef ORIGINAL
//--------------------------------------------------
// gunify

BipStatus bip_gunify(Gem* gem) {
   /*	$gunify(-Var1,-Var2)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

   ARGVAL_VAR(xvar,gem,icgoal,0) ;
   ARGVAL_VAR(yvar,gem,icgoal,1) ;

   gem->uiheap.resetGArgStack();
   if(unify(xvar,yvar,&(gem->uiheap))) {
      gem->popAndState() ;
      gem->pushRoleTerms() ;
      return bipSuccess0(gem) ;
   }
   else
      return bipCutFail(gem) ;
}
#endif

//--------------------------------------------------
// unifyable

BipStatus bip_unifiable(Gem* gem) {
   /*	$unifiable(-Var1,-Var2)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

   ARGVAL_VAR(xvar,gem,icgoal,0) ;
   ARGVAL_VAR(yvar,gem,icgoal,1) ;

   gem->pushOrState() ;
   Bool r = unify(xvar,yvar,&(gem->uiheap)) ;
   gem->popOrState() ;

   if(r) 
      return bipSuccess(gem) ;
   else
      return bipCutFail(gem) ;
}

//--------------------------------------------------
// not unify

BipStatus bip_not_unify(Gem* gem) {
   /*	$not_unify(-Var1,-Var2)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

   ARGVAL_VAR(xvar,gem,icgoal,0) ;
   ARGVAL_VAR(yvar,gem,icgoal,1) ;

   gem->pushOrState() ;
   Bool r = unify(xvar,yvar,&(gem->uiheap)) ;
   gem->popOrState() ;

   if(!r) 
      return bipSuccess(gem) ;
   else
      return bipCutFail(gem) ;
}

//======================================================================
// GenSym
//

BipStatus bip_gensym(Gem* gem) {
   /*	$gensym(-Symbol [, +Seed])	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY2(icgoal,cgoal,gem,1,2) ;

   InstGaeaData var = _fetchedArgv(icgoal,0) ;
   GaeaData seed = GenSymSeedSymbol ;
   if(chkArity(cgoal,2)) 
      seed = fetchedArgv(icgoal,1) ;

   if(!isGaeaSymbol(seed))
      return bipError(gem,"Illegal gensym seed") ;

   if(unifyWith(genGaeaSymbol(seed),var,gem)) 
      return bipSuccess(gem) ;
   else
      return bipCutFail(gem) ;
}

//======================================================================
// Make/New Sym
//

Bool generateNameFromList(const InstGaeaData& idt, char* buffer) {
   if(isNullSymbol(idt.data)) {
      *buffer = NULLCHAR ;
      return True ;
   } else {
      if(!isConsTerm(idt.data)) return False ;
      
      InstGaeaData car = fetch(InstGaeaData(_ConsCar(idt.data),
					    idt.binding)) ;
      int l ;
      if(isTGaeaString(car.data)) {
      
	 String name = _TStrVal(car.data) ;
	 strcpy(buffer,(char*)name) ;
	 l = strlen(name) ;

      } else if (isGaeaInteger(car.data)) {
	 sprintf(buffer,"%ld%n",IntVal(car.data),&l) ;
      } else {
		 return False ;
      }
      InstGaeaData cdr = fetch(InstGaeaData(_ConsCdr(idt.data),
					    idt.binding)) ;

      return generateNameFromList(cdr, &(buffer[l])) ;
   }
}

BipStatus bip_makesymbol(Gem* gem) {
   /*	$makesymbol(-Symbol, +NameList)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

   ARGVAL_VAR(vsym,gem,icgoal,0) ;
   ARGVAL_VAR(nlist,gem,icgoal,1) ;

   char tmpname[BUFSIZE] ;

   if(!generateNameFromList(nlist,tmpname)) 
      return bipError(gem,"Illegal name list") ;

   if(unifyWith(GaeaSymbol(tmpname),vsym,gem)) 
      return bipSuccess(gem) ; 
   else
      return bipCutFail(gem) ;
}

BipStatus bip_newsymbol(Gem* gem) {
   /*	$newsymbol(-Symbol, +NameList)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

   ARGVAL_VAR(vsym,gem,icgoal,0) ;
   ARGVAL_VAR(nlist,gem,icgoal,1) ;

   char tmpname[BUFSIZE] ;

   if(!generateNameFromList(nlist,tmpname)) 
      return bipError(gem,"Illegal name list") ;

   if(ExistSymbol(tmpname)) 
      return bipError(gem,"symbols already exists") ;

   if(unifyWith(GaeaSymbol(tmpname),vsym,gem)) 
      return bipSuccess(gem) ; 
   else
      return bipCutFail(gem) ;
}

//======================================================================
// Data Type
//

//--------------------------------------------------
// var
//

BipStatus bip_var(Gem* gem) {
   /*	$var(Var)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;

   InstGaeaData v = _fetchedArgv(icgoal,0) ;

   if(!isGround(v)) return bipSuccess(gem) ;
   else return bipCutFail(gem) ;
}
      
//--------------------------------------------------
// nonvar
//

BipStatus bip_nonvar(Gem* gem) {
   /*	$nonvar(Var)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;

   InstGaeaData v = _fetchedArgv(icgoal,0) ;

   if(isGround(v)) return bipSuccess(gem) ;
   else return bipCutFail(gem) ;
}
      
//--------------------------------------------------
// issymbol
//

BipStatus bip_issymbol(Gem* gem) {
   /*	$issymbol(Val)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;

   GaeaData v = fetchedArgv(icgoal,0) ;

   if(isGaeaSymbol(v)) return bipSuccess(gem) ;
   else return bipCutFail(gem) ;
}
      
//--------------------------------------------------
// isinteger
//

BipStatus bip_isinteger(Gem* gem) {
   /*	$isinteger(Val)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;

   GaeaData v = fetchedArgv(icgoal,0) ;

   if(isGaeaInteger(v)) return bipSuccess(gem) ;
   else return bipCutFail(gem) ;
}
      
//--------------------------------------------------
// isfloat
//

BipStatus bip_isfloat(Gem* gem) {
   /*	$isfloat(Val)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;

   GaeaData v = fetchedArgv(icgoal,0) ;

   if(isGaeaFloat(v)) return bipSuccess(gem) ;
   else return bipCutFail(gem) ;
}
      
//--------------------------------------------------
// isnumber
//

BipStatus bip_isnumber(Gem* gem) {
   /*	$isnumber(Val)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;

   GaeaData v = fetchedArgv(icgoal,0) ;

   if(isGNumber(v)) return bipSuccess(gem) ;
   else return bipCutFail(gem) ;
}
      
//--------------------------------------------------
// isstring
//

BipStatus bip_isstring(Gem* gem) {
   /*	$isstring(Val)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;

   GaeaData v = fetchedArgv(icgoal,0) ;

   if(isGaeaString(v)) return bipSuccess(gem) ;
   else return bipCutFail(gem) ;
}

//--------------------------------------------------
// ispointer
//

BipStatus bip_ispointer(Gem* gem) {
   /*	$ispointer(Val)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;

   GaeaData v = fetchedArgv(icgoal,0) ;

   if(isGaeaPointer(v)) return bipSuccess(gem) ;
   else return bipCutFail(gem) ;
}

//--------------------------------------------------
// istimeval
//

//BipStatus bip_istimeval(Gem* gem) {
//   /*	$istimeval(Val)	*/
//
//   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;
//
//   GaeaData v = fetchedArgv(icgoal,0) ;
//
//   if(isGaeaTimeVal(v)) return bipSuccess(gem) ;
//   else return bipCutFail(gem) ;
//}

//--------------------------------------------------
// isatom
//

BipStatus bip_isatom(Gem* gem) {
   /*	$isatom(Val)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;

   GaeaData v = fetchedArgv(icgoal,0) ;

   if(isGAtom(v)) return bipSuccess(gem) ;
   else return bipCutFail(gem) ;
}

//--------------------------------------------------
// term arity
//

BipStatus bip_term_arity(Gem* gem) {
   /*	$term_arity(Term,Head,Arity)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,3) ;
   
   GaeaData v = fetchedArgv(icgoal,0) ;
   InstGaeaData head = _fetchedArgv(icgoal,1) ;
   InstGaeaData arity = _fetchedArgv(icgoal,2) ;

   if(!isGaeaTerm(v)) return bipCutFail(gem) ;

   if(unifyWith(GaeaSymbol(_TermHead(v)),head,gem) &&
      unifyWith(GaeaInteger(_TermArity(v)),arity,gem))
      return bipSuccess(gem) ;
   else
      return bipCutFail(gem) ;
}
      
//--------------------------------------------------
// term argv
//

BipStatus bip_term_argv(Gem* gem) {
   /*	$term_argv(+Term,+Argc,-Argv)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,3) ;
   
   ARGVAL(v,gem,icgoal,0,isGaeaTerm,"Illegal term") ;
   ARGVAL_INT(argc,gem,icgoal,1) ;
   ARGVAL_VAR(argv,gem,icgoal,2) ;

   if(_TermArity(v) <= argc || argc < 0) ;
      return bipError(gem,"Illegal Argc") ;

   if(unifyWith(_TermArgv(v,argc),argv,gem))
      return bipSuccess(gem) ;
   else
      return bipCutFail(gem) ;
}

//--------------------------------------------------
// term gargv
//

BipStatus bip_term_gargv(Gem* gem) {
   /*	$term_gargv(+Term,+Role,-Value)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,3) ;

   InstGaeaData v = _fetchedArgv(icgoal,0) ;
   GaeaData role = fetchedArgv(icgoal,1) ;
   InstGaeaData argv = _fetchedArgv(icgoal,2) ;

   if(!isGaeaSymbol(role)) 
      return bipError(gem,"Illegal role symbol") ;

   InstGaeaData value = fetchGArg(v,_SymVal(role), &(gem->uiheap)) ;

   if(eq(value.data,UserNeverUseSymbol)) 
      return bipCutFail(gem);

   if(unify(value,argv,&(gem->uiheap)))
      return bipSuccess(gem) ;
   else
      return bipCutFail(gem) ;
}

//--------------------------------------------------
// gen term
//

BipStatus bip_genterm(Gem* gem) {
   /*	$genterm(+Head,+Arity,-Term)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,3) ;

   GaeaData head = fetchedArgv(icgoal,0) ;
   GaeaData arity = fetchedArgv(icgoal,1) ;
   InstGaeaData term = _fetchedArgv(icgoal,2) ;

   if(!isGaeaSymbol(head)) return bipError(gem,"Term-head should be a symbol") ;
   if(!isGaeaInteger(arity)) return bipError(gem,"Arity should be an integer") ;

   UInt art = IntVal(arity) ;
   GaeaData tm = GaeaTerm(head,art) ;
   for(UInt i = 0 ; i < art ; i++){
      _TermArgv(tm,i) = genGaeaVariable() ;
   } ;
   if(unifyWith(tm,term,gem)) return bipSuccess(gem) ;
   else return bipCutFail(gem) ;
} 

BipStatus bip_let_term_argv(Gem* gem) {
   /*	$let_term_argv(+Term,+Argc,+Argv)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,3) ;

   GaeaData term = fetchedArgv(icgoal,0) ;
   GaeaData argc = fetchedArgv(icgoal,1) ;
   GaeaData argv = fetchedArgv(icgoal,2) ;

   if(!isGaeaTerm(term) || !isGNumber(argc)) 
      return bipError(gem,"Illegal term or argc") ;
   if(_TermArity(term) <= IntVal(argc)) 
      return bipError(gem,"Illegal Argc") ;

   _TermArgv(term,IntVal(argc)) = argv ;
   
   return bipSuccess(gem) ;
}
      
//======================================================================
// String
//

//--------------------------------------------------
// length

BipStatus bip_string_length(Gem* gem) {
   /*	$string_length(+String,-Length)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

   ARGVAL_CHARPTR(str,gem,icgoal,0) ;
   ARGVAL_VAR(len,gem,icgoal,1) ;

   if(unifyWith(GaeaInteger(strlen(str)),len,gem)) 
      return bipSuccess(gem) ;
   else
      return bipCutFail(gem) ;
}

//--------------------------------------------------
// string to ascii list

BipStatus bip_string_to_list(Gem* gem) {
   /*	string_to_list(+String,-List) */

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

   ARGVAL_CHARPTR(str,gem,icgoal,0) ;
   ARGVAL_VAR(list,gem,icgoal,1) ;

   InstGaeaData l = list ;

   Int len = strlen(str) ;

   for(Int i = 0 ; i < len ; i++) {
      if(!unifyWithListShift(l,GaeaInteger((Int)(str[i])),gem))
	 return bipCutFail(gem) ;
   }
   if(!unifyWith(NullSymbol,l,gem))
      return bipCutFail(gem) ;

   return bipSuccess(gem) ;
}

//--------------------------------------------------
// ascii list to string 

BipStatus bip_list_to_string(Gem* gem) {
   /*	string_to_list(+List,-String) */

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

   ARGVAL_VAR(list,gem,icgoal,0) ;
   ARGVAL_VAR(str,gem,icgoal,1) ;

   Int len = 0 ;
   for(InstGaeaData l = list ; isInstConsTerm(l) ; l = InstConsCdrVal(l)) {
      if(!isGaeaInteger(InstConsCarVal(l).data))
	 return bipError(gem,"Illegal ASCII code") ;
      len++ ;
   }
   char* b = new char[len+1] ;

   len = 0 ;
   for(InstGaeaData l = list ; isInstConsTerm(l) ; l = InstConsCdrVal(l)) {
      b[len] = (char)IntVal(InstConsCarVal(l).data) ;
      len++ ;
   }

   if(unifyWith(GaeaString(b),str,gem))
      return bipSuccess(gem) ;
   else
      return bipCutFail(gem) ;
}

//--------------------------------------------------
// string to symbol

BipStatus bip_string_to_symbol(Gem* gem) {
   /*	$string_to_symbol(+String,-Symbol)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

   ARGVAL_CHARPTR(str,gem,icgoal,0) ;
   ARGVAL_VAR(sym,gem,icgoal,1) ;

   if(unifyWith(GaeaSymbol(str),sym,gem)) 
      return bipSuccess(gem) ;
   else
      return bipCutFail(gem) ;
}
   
//--------------------------------------------------
// symbol to string 

BipStatus bip_symbol_to_string(Gem* gem) {
   /*	$symbol_to_string(+Symbol,-String)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

   ARGVAL_SYM(sym,gem,icgoal,0) ;
   ARGVAL_VAR(str,gem,icgoal,1) ;

   if(unifyWith(GaeaString(sym->name),str,gem)) 
      return bipSuccess(gem) ;
   else
      return bipCutFail(gem) ;
}
   
//======================================================================
// Stack Object
//

ExtObjTypeID EOT_GaeaDataStack ;
inline Bool isGaeaDataStack(GaeaData& dt) {
   return isGaeaExtObject(EOT_GaeaDataStack,dt) ; 
} ;
inline GaeaDataStack* _GaeaDataStackVal(GaeaData& dt) {
   return (GaeaDataStack*)_ExObjVal(dt) ;
} ;

BipStatus bip_new_stack(Gem* gem) {
   /*	$new_stack(-Stack [,+Size])	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY2(icgoal,cgoal,gem,1,2) ;

   InstGaeaData var = _fetchedArgv(icgoal,0) ;
   Int size = 0 ;
   if(!chkArity(cgoal,2)) {
      GaeaData s = fetchedArgv(icgoal,1) ;
      if(!isGNumber(s) || IntVal(s) < 1) 
	 return bipError(gem,"Illegal array size") ;
      size = IntVal(s) ;
   }

   GaeaDataStack* gdstk ;
   if(size > 0) {
      gdstk = new GaeaDataStack("GaeaDataStack",size) ;
   } else {
      gdstk = new GaeaDataStack("GaeaDataStack") ;
   } ;

   if(unifyWith(GaeaExtObject(EOT_GaeaDataStack,gdstk),var,gem)) {
      return bipSuccess(gem) ;
   } else {
      delete gdstk ;
      return bipCutFail(gem) ;
   }
}

BipStatus bip_push_stack(Gem* gem) {
   /*	$push_stack(+Data,+Stack)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

   GaeaData dt = fetchedValueArgv(icgoal,0,gem) ;
   GaeaData gdstk = fetchedArgv(icgoal,1) ;
   if(!isGaeaDataStack(gdstk)) return bipError(gem,"Illegal gdatastack object") ;

   GaeaDataStack* stk = _GaeaDataStackVal(gdstk) ;
   stk->push(dt) ;

   return bipSuccess(gem) ;
} ;

BipStatus bip_pop_stack(Gem* gem) {
   /*	$pop_stack(-Data,+Stack)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

   InstGaeaData var = _fetchedArgv(icgoal,0) ;
   GaeaData gdstk = fetchedArgv(icgoal,1) ;
   if(!isGaeaDataStack(gdstk)) return bipError(gem,"Illegal gdatastack object") ;
   GaeaDataStack* stk = _GaeaDataStackVal(gdstk) ;
   if(stk->bottomp()) 
      return bipError(gem,"Gdatastack is empty") ;

   GaeaData val = stk->pop() ;
   if(unifyWith(val,var,gem)) 
      return bipSuccess(gem) ;
   else {
      stk->push(val) ;
      return bipCutFail(gem) ;
   }
} 

BipStatus bip_stack_depth(Gem* gem) {
   /*	$stack_depth(-Depth,+Stack)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

   InstGaeaData var = _fetchedArgv(icgoal,0) ;
   GaeaData gdstk = fetchedArgv(icgoal,1) ;
   if(!isGaeaDataStack(gdstk)) return bipError(gem,"Illegal gdatastack object") ;

   GaeaDataStack* stk = _GaeaDataStackVal(gdstk) ;
   Int d = stk->toppos() ;

   if(unifyWith(GaeaInteger(d),var,gem)) 
      return bipSuccess(gem) ;
   else {
      return bipCutFail(gem) ;
   }
} 

BipStatus bip_stack_top(Gem* gem) {
   /*	$stack_top(-Top,+Stack[,+N])	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY2(icgoal,cgoal,gem,2,3) ;

   InstGaeaData var = _fetchedArgv(icgoal,0) ;
   GaeaData gdstk = fetchedArgv(icgoal,1) ;
   if(!isGaeaDataStack(gdstk)) return bipError(gem,"Illegal gdatastack object") ;

   GaeaDataStack* stk = _GaeaDataStackVal(gdstk) ;
   if(stk->bottomp()) 
      return bipError(gem,"Gdatastack is empty") ;

   GaeaData val ;
   if(chkArity(cgoal,3)) {
      Int n = IntVal(fetchedArgv(icgoal,2)) ;
      if(n >= stk->toppos() || n < 0) 
	 return bipError(gem,"N is greater than depth of stack") ;
      val = stk->topN(n) ;
   } else {
      val = stk->top() ;
   }
   
   if(unifyWith(val,var,gem)) 
      return bipSuccess(gem) ;
   else {
      return bipCutFail(gem) ;
   }
} 

BipStatus bip_stack_bottom(Gem* gem) {
   /*	$stack_bottom(-Btm,+Stack,+N)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,3) ;

   InstGaeaData var = _fetchedArgv(icgoal,0) ;
   GaeaData gdstk = fetchedArgv(icgoal,1) ;
   if(!isGaeaDataStack(gdstk)) return bipError(gem,"Illegal gdatastack object") ;
   GaeaData nobj = fetchedArgv(icgoal,2) ;
   if(!isGNumber(nobj)) return bipError(gem,"Illegal number") ;
   Int n = IntVal(nobj) ;
   
   GaeaDataStack* stk = _GaeaDataStackVal(gdstk) ;
   if(stk->bottomp()) 
      return bipError(gem,"Gdatastack is empty") ;

   if(n >= stk->toppos() || n < 0) 
      return bipError(gem,"N is greater than depth of stack") ;
   GaeaData val = stk->bottomN(n) ;
   
   if(unifyWith(val,var,gem)) 
      return bipSuccess(gem) ;
   else {
      return bipCutFail(gem) ;
   }
} 

BipStatus bip_rewind_stack(Gem* gem) {
   /*	$rewind_stack(+Stack,+NewTop)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

   GaeaData gdstk = fetchedArgv(icgoal,0) ;
   if(!isGaeaDataStack(gdstk)) return bipError(gem,"Illegal gdatastack object") ;
   GaeaDataStack* stk = _GaeaDataStackVal(gdstk) ;

   GaeaData nobj = fetchedArgv(icgoal,1) ;
   if(!isGNumber(nobj)) return bipError(gem,"Illegal number") ;
   Int n = IntVal(nobj) ;

   if(n < 0) 
      return bipError(gem,"Newtop pointer should be positive") ;
   
   stk->rewind(n) ;

   return bipSuccess(gem) ;
} 

BipStatus bip_fill_stack(Gem* gem) {
   /*	$fill_stack(+Stack,+From,+To,+Val)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,4) ;

   GaeaData gdstk = fetchedArgv(icgoal,0) ;
   if(!isGaeaDataStack(gdstk)) return bipError(gem,"Illegal gdatastack object") ;
   GaeaDataStack* stk = _GaeaDataStackVal(gdstk) ;

   GaeaData fobj = fetchedArgv(icgoal,1) ;
   if(!isGNumber(fobj)) return bipError(gem,"Illegal number") ;
   Int from = IntVal(fobj) ;

   GaeaData tobj = fetchedArgv(icgoal,2) ;
   if(!isGNumber(tobj)) return bipError(gem,"Illegal number") ;
   Int to = IntVal(tobj) ;

   GaeaData val = fetchedValueArgv(icgoal,3,gem) ;

   if(from < 0 || to < 0) 
      return bipError(gem,"From and Top should be positive") ;
   
   stk->fill(val,from,to) ;

   return bipSuccess(gem) ;
} 



//======================================================================
// Initialize (external)
//

static Bool bipInit() {
   
   _SymVal(GaeaSymbol("$unify"))->setfunc(bip_unify) ;
#ifndef ORIGINAL
   _SymVal(GaeaSymbol("$gunify"))->setfunc(bip_gunify) ;
#endif
   _SymVal(GaeaSymbol("$unifiable"))->setfunc(bip_unifiable) ;
   _SymVal(GaeaSymbol("$not_unify"))->setfunc(bip_not_unify) ;

   _SymVal(GaeaSymbol("$gensym"))->setfunc(bip_gensym) ;
   _SymVal(GaeaSymbol("$makesymbol"))->setfunc(bip_makesymbol) ;
   _SymVal(GaeaSymbol("$newsymbol"))->setfunc(bip_newsymbol) ;

   _SymVal(GaeaSymbol("$var"))->setfunc(bip_var) ;
   _SymVal(GaeaSymbol("$nonvar"))->setfunc(bip_nonvar) ;
   _SymVal(GaeaSymbol("$issymbol"))->setfunc(bip_issymbol) ;
   _SymVal(GaeaSymbol("$isinteger"))->setfunc(bip_isinteger) ;
   _SymVal(GaeaSymbol("$isfloat"))->setfunc(bip_isfloat) ;
   _SymVal(GaeaSymbol("$isnumber"))->setfunc(bip_isnumber) ;
   _SymVal(GaeaSymbol("$isstring"))->setfunc(bip_isstring) ;
   _SymVal(GaeaSymbol("$ispointer"))->setfunc(bip_ispointer) ;
//   _SymVal(GaeaSymbol("$istimeval"))->setfunc(bip_istimeval) ;
   _SymVal(GaeaSymbol("$isatom"))->setfunc(bip_isatom) ;
   _SymVal(GaeaSymbol("$term_arity"))->setfunc(bip_term_arity) ;
   _SymVal(GaeaSymbol("$term_argv"))->setfunc(bip_term_argv) ;
   _SymVal(GaeaSymbol("$term_gargv"))->setfunc(bip_term_gargv) ;
   _SymVal(GaeaSymbol("$genterm"))->setfunc(bip_genterm) ;
   _SymVal(GaeaSymbol("$let_term_argv"))->setfunc(bip_let_term_argv) ;

   _SymVal(GaeaSymbol("$string_length"))->setfunc(bip_string_length) ;
   _SymVal(GaeaSymbol("$string_to_list"))->setfunc(bip_string_to_list) ;
   _SymVal(GaeaSymbol("$list_to_string"))->setfunc(bip_list_to_string) ;
   _SymVal(GaeaSymbol("$string_to_symbol"))->setfunc(bip_string_to_symbol) ;
   _SymVal(GaeaSymbol("$symbol_to_string"))->setfunc(bip_symbol_to_string) ;

   _SymVal(GaeaSymbol("$new_stack"))->setfunc(bip_new_stack) ;
   _SymVal(GaeaSymbol("$push_stack"))->setfunc(bip_push_stack) ;
   _SymVal(GaeaSymbol("$pop_stack"))->setfunc(bip_pop_stack) ;
   _SymVal(GaeaSymbol("$stack_depth"))->setfunc(bip_stack_depth) ;
   _SymVal(GaeaSymbol("$stack_top"))->setfunc(bip_stack_top) ;
   _SymVal(GaeaSymbol("$stack_bottom"))->setfunc(bip_stack_bottom) ;
   _SymVal(GaeaSymbol("$rewind_stack"))->setfunc(bip_rewind_stack) ;
   _SymVal(GaeaSymbol("$fill_stack"))->setfunc(bip_fill_stack) ;

   EOT_GaeaDataStack = DefExtObjType("gdatastack") ;
   
} ;





