// -*- Mode:C++ -*-
//Header:
//File: bipred_env.cc
//Title: built-in predicate 
//Author: NODA Itsuki
//Date: 1999/5/3
//EndHeader:

//ModifyHistory:
// 1999/5/3: Create this file
//EndModifyHistory:

//Memo:
//EndMemo:

#include "bipred.h"

//======================================================================
// Initialize (external)
//

static Bool bipInit() ;

extern "C" {
   Bool bipredInit_env() {
      static Bool initializedp = False ;
      if(!initializedp) {
	 bipInit() ;
	 initializedp = True ;
	 return True ;
      } else {
	 return False ;
      }
   } ;
} ;

//======================================================================
// listing
//

BipStatus bip_listing(Gem* gem) {
   /*	$listing[(+Cell)]	*/

   PICKUP_CURRENT_GOAL(icgoal,cgoal,gem) ;
   
   if(isGaeaSymbol(cgoal)) {
      gem->gcell->listing(*(gem->ostrm),True) ;
      return bipSuccess(gem) ;
   } else if (chkArity(cgoal,1)) {
      GaeaData gcsym = fetchedArgv(icgoal,0);
      if(hasGaeaCell(gcsym)) {
	 _SymGaeaCell(gcsym)->listing(*(gem->ostrm),True,True) ;
	 return bipSuccess(gem) ;
      } else {
	 return bipError(gem,"Invalid Cell Name") ;
      }
   } else {
      return bipNoMatch(gem) ;
   }
}

BipStatus bip_listsymbol(Gem* gem) {
   /*	$listsymbol	*/

   PICKUP_CURRENT_GOAL_CHECK_NOARITY(icgoal,cgoal,gem) ;

   showSymTab() ;
   return bipSuccess(gem) ;
}

BipStatus bip_listprocess(Gem* gem) {
   /*	$listprocess	*/

   PICKUP_CURRENT_GOAL_CHECK_NOARITY(icgoal,cgoal,gem) ;

   showGemList(*(gem->ostrm)) ;
   return bipSuccess(gem) ;
}

BipStatus bip_listprocesstable(Gem* gem) {
   /*	$listprocesstable	*/

   PICKUP_CURRENT_GOAL_CHECK_NOARITY(icgoal,cgoal,gem) ;

   showGemTable(*(gem->ostrm)) ;
   return bipSuccess(gem) ;
}



//======================================================================
// NewCell
//

BipStatus bip_newcell(Gem* gem) {
   /*	$newcell(+CellName)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1);
   
   GaeaData gcsym = fetchedArgv(icgoal,0);
   if(isGaeaSymbol(gcsym) && !hasGaeaCell(gcsym)){
      new GaeaCell(_SymVal(gcsym)) ;
      return bipSuccess(gem) ;
   } else {
      return bipError(gem,"Cell already exists") ;
   }
}

//======================================================================
// checking Cell
//

BipStatus bip_cell_exist(Gem* gem) {
   /*	$cell_exist(+CellName)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1);
   
   GaeaData gcsym = fetchedArgv(icgoal,0);
   if(isGaeaSymbol(gcsym) && hasGaeaCell(gcsym)){
      return bipSuccess(gem) ;
   } else {
      return bipCutFail(gem) ;
   }
}


//======================================================================
// Hide/Unhide cell
//

BipStatus bip_hide_cell(Gem* gem) {
   /*	$hide_cell(+Cell)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1);

   GaeaData gcsym = fetchedArgv(icgoal,0);
   if(hasGaeaCell(gcsym)) {
      _SymGaeaCell(gcsym)->hiddenInListingp = True ;
      return bipSuccess(gem) ;
   } else {
	 return bipError(gem,"Invarid Cell") ;
   }
}

BipStatus bip_unhide_cell(Gem* gem) {
   /*	$unhide_cell(+Cell)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1);

   GaeaData gcsym = fetchedArgv(icgoal,0);
   if(hasGaeaCell(gcsym)) {
      _SymGaeaCell(gcsym)->hiddenInListingp = False ;
      return bipSuccess(gem) ;
   } else {
	 return bipError(gem,"Invarid Cell") ;
   }
}
      
//======================================================================
// PushCell
//
BipStatus bip_push(Gem* gem) {
   /*	$push(+Cell [, +DstCell])	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY2(icgoal,cgoal,gem,1,2);

   if (chkArity(cgoal,1)) {
      GaeaData gcsym = fetchedArgv(icgoal,0);
      if(hasGaeaCell(gcsym)) {
	 gem->gcell->pushGaeaCell(_SymGaeaCell(gcsym)) ;
	 return bipSuccess(gem) ;
      } else {
	 return bipError(gem,"Invarid Cell") ;
      }
   } else if (chkArity(cgoal,2)) {
      GaeaData gcsym = fetchedArgv(icgoal,0) ;
      GaeaData dstcell = fetchedArgv(icgoal,1) ;
      if(hasGaeaCell(gcsym) && hasGaeaCell(dstcell)) {
	 _SymGaeaCell(dstcell)->pushGaeaCell(_SymGaeaCell(gcsym)) ;
	 return bipSuccess(gem) ;
      } else {
	 return bipError(gem,"Invarid Cell") ;
      }
   } else {
      return bipNoMatch(gem) ;
   } ;
}
      
//======================================================================
// PopCell
//
BipStatus bip_pop(Gem* gem) {
   /*	$pop(-Cell [, DstCell ])	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY2(icgoal,cgoal,gem,1,2);

   ARGVAL_VAR(gcvar,gem,icgoal,0) ;
   GaeaCell* dgc = gem->gcell ;
   if(chkArity(cgoal,2)) {
      ARGVAL_DAT(dstcellsym,gem,icgoal,1) ;
      if(!hasGaeaCell(dstcellsym)) 
	 return bipError(gem,"Invalid destination cell") ;
      dgc = _SymGaeaCell(dstcellsym) ;
   }
   
   if(dgc->gcellStack.bottomp()) 
      return bipCutFail(gem) ;

   GaeaCell* gc = dgc->popGaeaCell() ;

   if(unifyWith(gc->gsymbol,gcvar,gem))
      return bipSuccess(gem) ;
   else
      return bipCutFail(gem) ;
}
   

//======================================================================
// RemoveCell
//
BipStatus bip_remove_cell(Gem* gem) {
   /*	$remove_cell(+Cell [, +DstCell])	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY2(icgoal,cgoal,gem,1,2);
   
   if (chkArity(cgoal,1)) {
      GaeaData gcsym = fetchedArgv(icgoal,0);
      if(hasGaeaCell(gcsym)) {
	 UInt r = gem->gcell->removeGaeaCell(_SymGaeaCell(gcsym)) ;
	 if(r > 0) {
	    return bipSuccess(gem) ;
	 } else {
	    return bipCutFail(gem) ;
	 }
      } else {
	 return bipError(gem,"Invarid Cell") ;
      }
   } else if (chkArity(cgoal,2)) {
      GaeaData gcsym = fetchedArgv(icgoal,0) ;
      GaeaData dstcell = fetchedArgv(icgoal,1) ;
      if(hasGaeaCell(gcsym) && hasGaeaCell(dstcell)) {
	 UInt r = _SymGaeaCell(dstcell)->removeGaeaCell(_SymGaeaCell(gcsym)) ;
	 if(r > 0) {
	    return bipSuccess(gem) ;
	 } else {
	    return bipCutFail(gem) ;
	 }
      } else {
	 return bipError(gem,"Invarid Cell") ;
      }
   } else {
      return bipNoMatch(gem) ;
   } ;
}

//======================================================================
// Remove1Cell
//
BipStatus bip_remove_a_cell(Gem* gem) {
   /*	$remove_a_cell(+Cell [, +DstCell])	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY2(icgoal,cgoal,gem,1,2);
   
   if (chkArity(cgoal,1)) {
      GaeaData gcsym = fetchedArgv(icgoal,0);
      if(hasGaeaCell(gcsym)) {
	 UInt r = gem->gcell->removeGaeaCell(_SymGaeaCell(gcsym)) ;
	 if(r > 0) {
	    return bipSuccess(gem) ;
	 } else {
	    return bipCutFail(gem) ;
	 }
      } else {
	 return bipError(gem,"Invalid Cell") ;
      }
   } else if (chkArity(cgoal,2)) {
      GaeaData gcsym = fetchedArgv(icgoal,0) ;
      GaeaData dstcell = fetchedArgv(icgoal,1) ;
      if(hasGaeaCell(gcsym) && hasGaeaCell(dstcell)) {
	 Bool r = _SymGaeaCell(dstcell)->remove1GaeaCell(_SymGaeaCell(gcsym)) ;
	 if(r) {
	    return bipSuccess(gem) ;
	 } else {
	    return bipCutFail(gem) ;
	 }
      } else {
	 return bipError(gem,"Invarid Cell") ;
      }
   } else {
      return bipNoMatch(gem) ;
   } ;
}

//======================================================================
// Environment
//

Bool unifyGaeaCellListSub(const InstGaeaData& gdt, GaeaCell* gcell, UInt n, Gem* gem) ;

Bool unifyGaeaCellListBody(const InstGaeaData& gdt, GaeaCell* gcell, Gem* gem) {
   if(gcell->gcellStack.bottomp()) {
      return unifyWith(gcell->gsymbol,gdt,gem) ;
   } else {
      Binding* newbinding = gem->uiheap.newBinding() ;
      if(!unify(gdt,InstGaeaData(ConsTerm,newbinding),&(gem->uiheap)))
	 return False ;
      if(!unifyWith(gcell->gsymbol,InstGaeaData(ConsCarVar,newbinding),gem))
	 return False ;
      return unifyGaeaCellListSub(InstGaeaData(ConsCdrVar,newbinding),gcell,0,gem) ;
   }
}

Bool unifyGaeaCellListSub(const InstGaeaData& gdt, GaeaCell* gcell, UInt n, Gem* gem) {
   if(gcell->gcellStack.toppos() <= n) {
      return unifyWith(NullSymbol,gdt,gem) ;
   } else {
      Binding* newbinding = gem->uiheap.newBinding() ;
      if(!unify(gdt,InstGaeaData(ConsTerm,newbinding),&(gem->uiheap)))
	 return False ;
      if(!unifyGaeaCellListBody(InstGaeaData(ConsCarVar,newbinding), 
			     gcell->gcellStack.topN(n),gem))
	 return False ;
      return unifyGaeaCellListSub(InstGaeaData(ConsCdrVar,newbinding),
			       gcell,n+1,gem) ;
   }
}

Bool unifyGaeaCellList(InstGaeaData& gdt, GaeaCell* gcell, Gem* gem) {
   BindingState bstate(gdt.binding, &(gem->uiheap)) ;
   Bool r = unifyGaeaCellListBody(gdt,gcell,gem) ;
   if(r) {
      return True ;
   } else {
      rewindBindingState(bstate) ;
      return False ;
   } 
}

BipStatus bip_environment(Gem* gem) {
   /*	$environment(-CellList [,+MCell])	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY2(icgoal,cgoal,gem,1,2);

   InstGaeaData gdt = _fetchedArgv(icgoal,0) ;
   GaeaCell* srccell = gem->gcell ;
   if(chkArity(cgoal,2)) {
      GaeaData mc = fetchedArgv(icgoal,1) ;
      if(!hasGaeaCell(mc)) return bipError(gem,"Illegal cell") ;
      srccell = _SymGaeaCell(mc) ;
   }
   if(unifyGaeaCellList(gdt,srccell,gem)) 
      return bipSuccess(gem) ;
   else 
      return bipCutFail(gem) ;
}

BipStatus bip_in_env(Gem* gem) {
   /*	$in_env(+Cell [, +SrcCell])	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY2(icgoal,cgoal,gem,1,2);

   GaeaData cell = fetchedArgv(icgoal,0) ;
   if(!hasGaeaCell(cell)) return bipError(gem,"Illegal cell") ;
   GaeaCell* src = gem->gcell ;
   if(chkArity(cgoal,2)) {
      GaeaData scell = fetchedArgv(icgoal,1) ;
      if(!hasGaeaCell(scell)) return bipError(gem,"Illegal cell") ;
      src = _SymGaeaCell(scell) ;
   }
   if(src->isInGaeaCellStack(_SymGaeaCell(cell)))
      return bipSuccess(gem) ;
   else
      return bipCutFail(gem) ;
}
BipStatus bip_in_env_recursive(Gem* gem) {
   /*	$in_env_recursive(+Cell [, +SrcCell])	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY2(icgoal,cgoal,gem,1,2);

   GaeaData cell = fetchedArgv(icgoal,0) ;
   if(!hasGaeaCell(cell)) return bipError(gem,"Illegal cell") ;
   GaeaCell* src = gem->gcell ;
   if(chkArity(cgoal,2)) {
      GaeaData scell = fetchedArgv(icgoal,1) ;
      if(!hasGaeaCell(scell)) return bipError(gem,"Illegal cell") ;
      src = _SymGaeaCell(scell) ;
   }
   if(src->isInGaeaCellStackR(_SymGaeaCell(cell)))
      return bipSuccess(gem) ;
   else
      return bipCutFail(gem) ;
}

//======================================================================
// environment 1 (one level cell list)
//

Bool unifyGaeaCellListSub1(const InstGaeaData& gdt, GaeaCell* gcell, UInt n, 
			   Gem* gem) {
   if(gcell->gcellStack.toppos() <= n) {
      return unifyWith(NullSymbol,gdt,gem) ;
   } else {
      Binding* newbinding = gem->uiheap.newBinding() ;
      if(!unify(gdt,InstGaeaData(ConsTerm,newbinding),&(gem->uiheap)))
	 return False ;
      if(!unifyWith(gcell->gcellStack.topN(n)->gsymbol,
		    InstGaeaData(ConsCarVar,newbinding),gem))
	 return False ;
      return unifyGaeaCellListSub1(InstGaeaData(ConsCdrVar,newbinding),
				   gcell,n+1,gem) ;
   }
}

Bool unifyGaeaCellList1(InstGaeaData& gdt, GaeaCell* gcell, Gem* gem) {
   BindingState bstate(gdt.binding, &(gem->uiheap)) ;
   Bool r = unifyGaeaCellListSub1(gdt,gcell,0,gem) ;
   if(r) {
      return True ;
   } else {
      rewindBindingState(bstate) ;
      return False ;
   } 
}

BipStatus bip_environment1(Gem* gem) {
   /*	$environment1(-CellList [,+MCell])	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY2(icgoal,cgoal,gem,1,2);

   InstGaeaData gdt = _fetchedArgv(icgoal,0) ;
   GaeaCell* srccell = gem->gcell ;
   if(chkArity(cgoal,2)) {
      GaeaData mc = fetchedArgv(icgoal,1) ;
      if(!hasGaeaCell(mc)) return bipError(gem,"Illegal cell") ;
      srccell = _SymGaeaCell(mc) ;
   }
   if(unifyGaeaCellList1(gdt,srccell,gem)) 
      return bipSuccess(gem) ;
   else 
      return bipCutFail(gem) ;
}

//======================================================================
// set environment 1 (one level cell list)
//

Bool setGaeaCellList1(const InstGaeaData& idt, GaeaCell* gcell) {
   if(eq(idt.data,NullSymbol)) {
      gcell->gcellStack.reset() ;
      return True ;
   } else {
      if(!isConsTerm(idt.data)) return False ;

      InstGaeaData car = fetch(InstGaeaData(_ConsCar(idt.data),
					    idt.binding)) ;
      if(!hasGaeaCell(car.data)) return False ;

      InstGaeaData cdr = fetch(InstGaeaData(_ConsCdr(idt.data),
					       idt.binding)) ;
      if(!setGaeaCellList1(cdr,gcell)) return False ;

      gcell->pushGaeaCell(_SymGaeaCell(car.data)) ;
      
      return True ;
   }
}

BipStatus bip_set_environment1(Gem* gem) {
   /*	$set_environment1(+CellList [,+MCell])	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY2(icgoal,cgoal,gem,1,2);

   InstGaeaData gdt = _fetchedArgv(icgoal,0) ;
   GaeaCell* dstcell = gem->gcell ;
   if(chkArity(cgoal,2)) {
      GaeaData mc = fetchedArgv(icgoal,1) ;
      if(!hasGaeaCell(mc)) return bipError(gem,"Illegal cell") ;
      dstcell = _SymGaeaCell(mc) ;
   }

   if(setGaeaCellList1(gdt,dstcell))
      return bipSuccess(gem) ;
   else 
      return bipError(gem,"Bad Cell List") ;
}

//======================================================================
// this cell
//

BipStatus bip_this_cell(Gem* gem) {
   /*	$this_cell(-Cell)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1);
   
   InstGaeaData var = _fetchedArgv(icgoal,0) ;

   UInt cutlevel = gem->astate->cutOrLevel ;
   if(cutlevel <= 0) return bipError(gem,"Here is not in cell") ;

   OrState* frame_ostate = &(gem->orStateStack.bottomN(cutlevel-1)) ;
   AndState* frame_astate = 
      &(frame_ostate->andStateStackPointer.stacktop->entity) ;
   
   GaeaCell* cell = frame_astate->gclpointer.topcell ;

   if(unifyWith(GaeaSymbol(cell->name),var,gem))
      return bipSuccess(gem) ;
   else
      return bipCutFail(gem) ;
}

BipStatus bip_remove_this_cell(Gem* gem) {
   /*	$remove_this_cell[(+MotherCell)] */

   PICKUP_CURRENT_GOAL(icgoal,cgoal,gem);

   if(!chkArity(cgoal,1) && !isGaeaSymbol(cgoal)) return bipNoMatch(gem) ;

   GaeaCell* mcell = gem->gcell ;

   if(chkArity(cgoal,1)) {
      GaeaData mcellsym = fetchedArgv(icgoal,0) ;
      if(!hasGaeaCell(mcellsym)) return bipError(gem,"Illegal gcell") ;
      mcell = _SymGaeaCell(mcellsym) ;
   } ;

   UInt cutlevel = gem->astate->cutOrLevel ;
   if(cutlevel <= 0) return bipError(gem,"Here is not in cell") ;

   OrState* frame_ostate = &(gem->orStateStack.bottomN(cutlevel-1)) ;
   AndState* frame_astate = 
      &(frame_ostate->andStateStackPointer.stacktop->entity) ;
   
   GaeaCell* cell = frame_astate->gclpointer.topcell ;

   Bool r = mcell->remove1GaeaCell(cell) ;

   if(r) 
      return bipSuccess(gem) ;
   else
      return bipCutFail(gem) ;
}

//======================================================================
// assert
//

enum BIPAssertMode {
   BIPAM_asserta,
   BIPAM_assertz
} ;

BipStatus bip_assert_fact(Gem* gem, BIPAssertMode mode) {

   PICKUP_CURRENT_GOAL_CHECK_ARITY2(icgoal,cgoal,gem,1,2);

   GaeaData head = fetchedValueArgv(icgoal,0,gem) ;
   GaeaCell* dstcell = gem->agcell() ;
   if(chkArity(cgoal,2)) {
      GaeaData gcl = fetchedArgv(icgoal,1) ;
      if(hasGaeaCell(gcl)) {
	 dstcell = _SymGaeaCell(gcl) ;
      } else {
	 return bipError(gem,"Illegal GaeaCell") ;
      }
   }
   if      (mode == BIPAM_asserta) 
      dstcell->asserta(GaeaClause(head,TrueSymbol)) ;
   else if (mode == BIPAM_assertz)
      dstcell->assertz(GaeaClause(head,TrueSymbol)) ;
   else
      return bipError(gem,"Illegal Assert Mode") ;
   
   return bipSuccess(gem) ;
}

BipStatus bip_assert_rule(Gem* gem, BIPAssertMode mode) {

   PICKUP_CURRENT_GOAL_CHECK_ARITY2(icgoal,cgoal,gem,2,3);

   GaeaData h = _TermArgv(cgoal,0) ;
   GaeaData b = _TermArgv(cgoal,1) ;
   GaeaData c = fetchValue(InstGaeaData(GaeaTerm(NullSymbol,h,b),icgoal.binding),
			&(gem->uiheap)) ;  
				/* this GaeaTerm is just temporal */
   GaeaData head = _TermArgv(c,0) ;
   GaeaData body = _TermArgv(c,1) ;
   GaeaCell* dstcell = gem->agcell() ;
   if(chkArity(cgoal,3)) {
      GaeaData gcl = fetchedArgv(icgoal,2) ;
      if(hasGaeaCell(gcl)) {
	 dstcell = _SymGaeaCell(gcl) ;
      } else {
	 return bipError(gem,"Illegal GaeaCell") ;
      }
   }
   if      (mode == BIPAM_asserta) 
      dstcell->asserta(GaeaClause(head,body)) ;
   else if (mode == BIPAM_assertz)
      dstcell->assertz(GaeaClause(head,body)) ;
   else
      return bipError(gem,"Illegal Assert Mode") ;
   
   return bipSuccess(gem) ;
}

BipStatus bip_asserta_fact(Gem* gem) {
   /*	$asserta_fact(+Head [, +DstCell])	*/

   return bip_assert_fact(gem, BIPAM_asserta) ;
}

BipStatus bip_assertz_fact(Gem* gem) {
   /*	$asserta_fact(+Head [, +DstCell])	*/

   return bip_assert_fact(gem, BIPAM_assertz) ;
}

BipStatus bip_asserta_rule(Gem* gem) {
   /*	$asserta_rule(+Head, +Body [, +DstCell])	*/

   return bip_assert_rule(gem, BIPAM_asserta) ;
}

BipStatus bip_assertz_rule(Gem* gem) {
   /*	$assertz_rule(+Head, +Body [, +DstCell])	*/

   return bip_assert_rule(gem, BIPAM_assertz) ;
}

//======================================================================
// GaeaCellVar
//

BipStatus bip_new_cellvar(Gem* gem) {
   /* $new_cellvar(+Cell,+Symbol [,+InitVal]) */
   
   PICKUP_CURRENT_GOAL_CHECK_ARITY2(icgoal,cgoal,gem,2,3) ;

   GaeaData cell = fetchedArgv(icgoal,0) ;
   if(!hasGaeaCell(cell)) return bipError(gem,"Illegal GaeaCell") ;

   GaeaData sym = fetchedArgv(icgoal,1) ;
   if(!isGaeaSymbol(sym)) return bipError(gem,"Illegal Symbol") ;

   Bool r = (_SymGaeaCell(cell))->newGaeaCellVar(_SymVal(sym)) ;
   if(!r) return bipError(gem,"CellVar already exists") ;
   
   if(chkArity(cgoal,3)) {
      GaeaData val = fetchedArgv(icgoal,2) ;
      GaeaDataBox* gdbox = (_SymGaeaCell(cell))->findGaeaCellVar(_SymVal(sym)) ;
      gdbox->put(val,GaeaDataBox::PT_force) ;
   }
   return bipSuccess(gem) ;
}

//======================================================================
// check existp
//

BipStatus bip_cellvar_exist(Gem* gem) {
   /* $cellvar_exist(+Cell,+Symbol) */
   
   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

   GaeaData cell = fetchedArgv(icgoal,0) ;
   if(!hasGaeaCell(cell)) return bipError(gem,"Illegal GaeaCell") ;

   GaeaData sym = fetchedArgv(icgoal,1) ;
   if(!isGaeaSymbol(sym)) return bipError(gem,"Illegal Symbol") ;

   GaeaDataBox *cv = _SymGaeaCell(cell)->findGaeaCellVar(_SymVal(sym),False) ;

   if(!isNullPtr(cv)) return bipSuccess(gem) ;
   else               return bipCutFail(gem) ;

}

//======================================================================
// get/put 
//

BipStatus _gdbox_get(Gem* gem, GaeaDataBox* gdbox, InstGaeaData& var, 
	       GaeaData& rp, GaeaData& wp) 
{
   Bool retractp ;
   if(eq(rp,YesSymbol)) 
      retractp = True ;
   else if (eq(rp,NoSymbol))
      retractp = False ;
   else 
      return bipError(gem,"Illegal Retract Mode (yes/no)") ;
      
   Bool waitp ;
   if(eq(wp,YesSymbol)) 
      waitp = True ;
   else if (eq(wp,NoSymbol))
      waitp = False ;
   else 
      return bipError(gem,"Illegal Wait Mode (yes/no)") ;

   GaeaData val ;
   Bool r = gdbox->get(val,retractp,waitp) ;
   if(r) {
      if(unifyWith(val,var,gem)) 
	 return bipSuccess(gem) ;
      else
	 return bipCutFail(gem) ;
   } else {
      return bipCutFail(gem) ;
   }
}

BipStatus _gdbox_timedwait_get(Gem* gem, GaeaDataBox* gdbox, InstGaeaData& var, 
			  GaeaData& rp, UInt sec, UInt msec, UInt usec)
{
   Bool retractp ;
   if(eq(rp,YesSymbol)) 
      retractp = True ;
   else if (eq(rp,NoSymbol))
      retractp = False ;
   else 
      return bipError(gem,"Illegal Retract Mode (yes/no)") ;

   GaeaData val ;
   Bool r = gdbox->timedwait_get(val,retractp,sec,msec,usec) ;
   if(r) {
      if(unifyWith(val,var,gem)) 
	 return bipSuccess(gem) ;
      else
	 return bipCutFail(gem) ;
   } else {
      return bipCutFail(gem) ;
   }
}

BipStatus bip_cv_get_r(Gem* gem) {
   /* $cv_get_r(+Symbol,-Val,+RetractMode,+WaitMode) */
   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,4) ;

   GaeaData sym = fetchedArgv(icgoal,0) ;
   InstGaeaData var = _fetchedArgv(icgoal,1) ;
   GaeaData rp = fetchedArgv(icgoal,2) ;
   GaeaData wp = fetchedArgv(icgoal,3) ;

   if(!isGaeaSymbol(sym)) return bipError(gem,"Not a Symbol") ;

   Bool retractp ;
   if(eq(rp,YesSymbol)) 
      retractp = True ;
   else if (eq(rp,NoSymbol))
      retractp = False ;
   else 
      return bipError(gem,"Illegal Retract Mode (yes/no)") ;
      
   Bool waitp ;
   if(eq(wp,YesSymbol)) 
      waitp = True ;
   else if (eq(wp,NoSymbol))
      waitp = False ;
   else 
      return bipError(gem,"Illegal Wait Mode (yes/no)") ;

   GaeaData value ;

   if(gem->gcell->getGaeaCellVarValueR(value,_SymVal(sym),retractp,waitp)) {
      if(unifyWith(value,var,gem))
	 return bipSuccess(gem) ;
      else
	 return bipCutFail(gem) ;
   }
   else
      return bipCutFail(gem) ;
}
   
BipStatus bip_cv_get(Gem* gem) {
   /* $cv_get(+Cell,+Symbol,-Val,+RetractMode,+WaitMode) */

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,5) ;

   GaeaData cell = fetchedArgv(icgoal,0) ;
   GaeaData sym = fetchedArgv(icgoal,1) ;
   InstGaeaData var = _fetchedArgv(icgoal,2) ;
   GaeaData rp = fetchedArgv(icgoal,3) ;
   GaeaData wp = fetchedArgv(icgoal,4) ;

   if(!hasGaeaCell(cell)) return bipError(gem,"Illegal Cell") ;
   if(!isGaeaSymbol(sym)) return bipError(gem,"Not a Symbol") ;

   GaeaDataBox* gdbox = _SymGaeaCell(cell)->findGaeaCellVar(_SymVal(sym)) ;
   if(isNullPtr(gdbox)) return bipError(gem,"Can't find cellvar") ;

   return _gdbox_get(gem,gdbox,var,rp,wp) ;
}

BipStatus bip_cv_timedwait_get(Gem* gem) {
   /* $cv_get(+Cell,+Symbol,-Val,+RetractMode,+Sec [, MSec [ USec ]]) */

   PICKUP_CURRENT_GOAL_CHECK_ARITY2(icgoal,cgoal,gem,5,7) ;

   GaeaData cell = fetchedArgv(icgoal,0) ;
   GaeaData sym = fetchedArgv(icgoal,1) ;
   InstGaeaData var = _fetchedArgv(icgoal,2) ;
   GaeaData rp = fetchedArgv(icgoal,3) ;
   GaeaData sec = fetchedArgv(icgoal,4) ;
   GaeaData msec = GaeaInteger(0) ;
   GaeaData usec = GaeaInteger(0) ;
   if(chkArity(cgoal,6,7)) {
      msec = fetchedArgv(icgoal,5) ;
      if(chkArity(cgoal,7)) {
	 usec = fetchedArgv(icgoal,6) ;
      }
   }

   if(!hasGaeaCell(cell)) return bipError(gem,"Illegal Cell") ;
   if(!isGaeaSymbol(sym)) return bipError(gem,"Not a Symbol") ;

   GaeaDataBox* gdbox = _SymGaeaCell(cell)->findGaeaCellVar(_SymVal(sym)) ;
   if(isNullPtr(gdbox)) return bipError(gem,"Can't find cellvar") ;

   return _gdbox_timedwait_get(gem,gdbox,var,rp,
			       IntVal(sec),IntVal(msec),IntVal(usec)) ;
}

BipStatus _gdbox_put(Gem* gem, GaeaDataBox* gdbox, InstGaeaData& val, GaeaData& wp) {
   GaeaDataBox::PutType ptype ;
   if(eq(wp,YesSymbol)) 
      ptype = GaeaDataBox::PT_wait ;
   else if (eq(wp,NoSymbol))
      ptype = GaeaDataBox::PT_nowait ;
   else if (eq(wp,ForceSymbol))
      ptype = GaeaDataBox::PT_force ;
   else 
      return bipError(gem,"Illegal Wait Mode (yes/no/force)") ;

   GaeaData v = fetchValue(val,&(gem->uiheap)) ;

   Bool r = gdbox->put(v,ptype) ;
   if(r) {
      return bipSuccess(gem) ;
   } else {
      return bipCutFail(gem) ;
   }
}

BipStatus _gdbox_timedwait_put(Gem* gem, GaeaDataBox* gdbox, InstGaeaData& val, 
			  UInt sec, UInt msec, UInt usec) 
{
   GaeaData v = fetchValue(val,&(gem->uiheap)) ;

   Bool r = gdbox->timedwait_put(v,sec,msec,usec) ;
   if(r) {
      return bipSuccess(gem) ;
   } else {
      return bipCutFail(gem) ;
   }
}

BipStatus bip_cv_put(Gem* gem) {
   /* $cv_put(+Cell,+Symbol,+Val,+WriteMode) WriteMode = [yes/no/force] */

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,4) ;

   GaeaData cell = fetchedArgv(icgoal,0) ;
   GaeaData sym = fetchedArgv(icgoal,1) ;
   InstGaeaData val = _fetchedArgv(icgoal,2) ;
   GaeaData wp = fetchedArgv(icgoal,3) ;

   if(!hasGaeaCell(cell)) return bipError(gem,"Illegal Cell") ;
   if(!isGaeaSymbol(sym)) return bipError(gem,"Not a Symbol") ;

   GaeaDataBox* gdbox = _SymGaeaCell(cell)->findGaeaCellVar(_SymVal(sym)) ;
   if(isNullPtr(gdbox)) return bipError(gem,"Can't find cellvar") ;

   return _gdbox_put(gem,gdbox,val,wp) ;
}

BipStatus bip_cv_timedwait_put(Gem* gem) {
   /* $cv_put(+Cell,+Symbol,+Val,+Sec [, MSec [, USec ]]) */

   PICKUP_CURRENT_GOAL_CHECK_ARITY2(icgoal,cgoal,gem,4,6) ;

   GaeaData cell = fetchedArgv(icgoal,0) ;
   GaeaData sym = fetchedArgv(icgoal,1) ;
   InstGaeaData val = _fetchedArgv(icgoal,2) ;
   GaeaData sec = fetchedArgv(icgoal,3) ;
   GaeaData msec = GaeaInteger(0) ;
   GaeaData usec = GaeaInteger(0) ;
   if(chkArity(cgoal,5,6)) {
      msec = fetchedArgv(icgoal,4) ;
      if(chkArity(cgoal,6)) {
	 usec = fetchedArgv(icgoal,5) ;
      }
   }

   if(!hasGaeaCell(cell)) return bipError(gem,"Illegal Cell") ;
   if(!isGaeaSymbol(sym)) return bipError(gem,"Not a Symbol") ;

   GaeaDataBox* gdbox = _SymGaeaCell(cell)->findGaeaCellVar(_SymVal(sym)) ;
   if(isNullPtr(gdbox)) return bipError(gem,"Can't find cellvar") ;

   return _gdbox_timedwait_put(gem,gdbox,val,
			       IntVal(sec),IntVal(msec),IntVal(usec)) ;
}

BipStatus bip_cv_listing(Gem* gem) {
   /* $cv_listing(+Cell) */

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;

   GaeaData cell = fetchedArgv(icgoal,0) ;

   if(!hasGaeaCell(cell)) return bipError(gem,"Illegal Cell") ;

   _SymGaeaCell(cell)->gcellVarTable.listing(*(gem->ostrm)) ;

   return bipSuccess(gem) ;
}


//======================================================================
// GaeaDataBox
//

BipStatus bip_assign_gdbox(Gem* gem) {
   /*	$assign_gdbox(+Symbol)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;

   GaeaData sym = fetchedArgv(icgoal,0) ;
   if(isGaeaSymbol(sym)) {
      Bool r = _SymVal(sym)->assignGDBox() ;
      if(!r)
	 return bipError(gem,"Symbol already has a gdbox") ;
      else
	 return bipSuccess(gem) ;
   } else {
      return bipError(gem,"Not a Symbol") ;
   }
}

BipStatus bip_gdbox_get(Gem* gem) {
   /*	$gdbox_get(+Symbol,-Data,+RetractP,+WaitP)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,4) ;

   GaeaData sym = fetchedArgv(icgoal,0) ;
   InstGaeaData var = _fetchedArgv(icgoal,1) ;
   GaeaData rp = fetchedArgv(icgoal,2) ;
   GaeaData wp = fetchedArgv(icgoal,3) ;

   if(!isGaeaSymbol(sym)) 
      return bipError(gem,"Not a Symbol") ;
   if(!hasGDBox(sym))
      return bipError(gem,"Symbol does not have a gdbox") ;

   GaeaDataBox* gdbox = _SymGDBox(sym) ;

   return _gdbox_get(gem,gdbox,var,rp,wp) ;
}

BipStatus bip_gdbox_timedwait_get(Gem* gem) {
   /*	$gdbox_timedwait_get(+Symbol,-Data,+RetractP,+Sec [,MSec [,USec]]) */

   PICKUP_CURRENT_GOAL_CHECK_ARITY2(icgoal,cgoal,gem,4,6) ;

   GaeaData sym = fetchedArgv(icgoal,0) ;
   InstGaeaData var = _fetchedArgv(icgoal,1) ;
   GaeaData rp = fetchedArgv(icgoal,2) ;
   GaeaData sec = fetchedArgv(icgoal,3) ;
   GaeaData msec = GaeaInteger(0) ;
   GaeaData usec = GaeaInteger(0) ;
   if(chkArity(cgoal,5,6)) {
      msec = fetchedArgv(icgoal,4) ;
      if(chkArity(cgoal,6)) {
	 usec = fetchedArgv(icgoal,5) ;
      }
   }

   if(!isGaeaSymbol(sym)) 
      return bipError(gem,"Not a Symbol") ;
   if(!hasGDBox(sym))
      return bipError(gem,"Symbol does not have a gdbox") ;

   return _gdbox_timedwait_get(gem,_SymGDBox(sym),var,rp,
			       IntVal(sec),IntVal(msec),IntVal(usec)) ;
}

BipStatus bip_gdbox_put(Gem* gem) {
   /*	$gdbox_put(+Symbol,+Data,+PutMode) PutMode = [yes/no/force] */

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,3) ;

   GaeaData sym = fetchedArgv(icgoal,0) ;
   InstGaeaData val = _fetchedArgv(icgoal,1) ;
   GaeaData wp = fetchedArgv(icgoal,2) ;

   if(!isGaeaSymbol(sym)) 
      return bipError(gem,"Not a Symbol") ;
   if(!hasGDBox(sym))
      return bipError(gem,"Symbol does not have a gdbox") ;

   return _gdbox_put(gem,_SymGDBox(sym),val,wp) ;
}

BipStatus bip_gdbox_timedwait_put(Gem* gem) {
   /*	$gdbox_timedwait_put(+Symbol,+Data,+PutMode, +Sec [,+MSec [,+USec]]) */

   PICKUP_CURRENT_GOAL_CHECK_ARITY2(icgoal,cgoal,gem,3,5) ;

   GaeaData sym = fetchedArgv(icgoal,0) ;
   InstGaeaData val = _fetchedArgv(icgoal,1) ;
   GaeaData sec = fetchedArgv(icgoal,2) ;
   GaeaData msec = GaeaInteger(0) ;
   GaeaData usec = GaeaInteger(0) ;
   if(chkArity(cgoal,4,5)) {
      msec = fetchedArgv(icgoal,3) ;
      if(chkArity(cgoal,5)) {
	 usec = fetchedArgv(icgoal,4) ;
      }
   }

   if(!isGaeaSymbol(sym)) 
      return bipError(gem,"Not a Symbol") ;
   if(!hasGDBox(sym))
      return bipError(gem,"Symbol does not have a gdbox") ;

   return _gdbox_timedwait_put(gem,_SymGDBox(sym),val,
			       IntVal(sec),IntVal(msec),IntVal(usec)) ;
}

//======================================================================
// Initialize (external)
//

static Bool bipInit() {
   
/* env */
   
   _SymVal(GaeaSymbol("$listing"))->setfunc(bip_listing) ;
   _SymVal(GaeaSymbol("$listsymbol"))->setfunc(bip_listsymbol) ;
   _SymVal(GaeaSymbol("$listprocess"))->setfunc(bip_listprocess) ;
   _SymVal(GaeaSymbol("$listprocesstable"))->setfunc(bip_listprocesstable) ;

   _SymVal(GaeaSymbol("$newcell"))->setfunc(bip_newcell) ;
   _SymVal(GaeaSymbol("$cell_exist"))->setfunc(bip_cell_exist) ;
   _SymVal(GaeaSymbol("$hide_cell"))->setfunc(bip_hide_cell) ;
   _SymVal(GaeaSymbol("$unhide_cell"))->setfunc(bip_unhide_cell) ;
   _SymVal(GaeaSymbol("$push"))->setfunc(bip_push) ;
   _SymVal(GaeaSymbol("$pop"))->setfunc(bip_pop) ;
   _SymVal(GaeaSymbol("$remove_cell"))->setfunc(bip_remove_cell) ;
   _SymVal(GaeaSymbol("$remove_a_cell"))->setfunc(bip_remove_a_cell) ;
   _SymVal(GaeaSymbol("$environment"))->setfunc(bip_environment) ;
   _SymVal(GaeaSymbol("$environment1"))->setfunc(bip_environment1) ;
   _SymVal(GaeaSymbol("$set_environment1"))->setfunc(bip_set_environment1) ;
   _SymVal(GaeaSymbol("$in_env"))->setfunc(bip_in_env) ;
   _SymVal(GaeaSymbol("$in_env_recursive"))->setfunc(bip_in_env_recursive) ;
   _SymVal(GaeaSymbol("$this_cell"))->setfunc(bip_this_cell) ;
   _SymVal(GaeaSymbol("$remove_this_cell"))->setfunc(bip_remove_this_cell) ;

   _SymVal(GaeaSymbol("$asserta_fact"))->setfunc(bip_asserta_fact) ;
   _SymVal(GaeaSymbol("$assertz_fact"))->setfunc(bip_assertz_fact) ;
   _SymVal(GaeaSymbol("$asserta_rule"))->setfunc(bip_asserta_rule) ;
   _SymVal(GaeaSymbol("$assertz_rule"))->setfunc(bip_assertz_rule) ;

   _SymVal(GaeaSymbol("$new_cellvar"))->setfunc(bip_new_cellvar) ;
   _SymVal(GaeaSymbol("$cellvar_exist"))->setfunc(bip_cellvar_exist) ;
   _SymVal(GaeaSymbol("$cv_get"))->setfunc(bip_cv_get) ;
   _SymVal(GaeaSymbol("$cv_get_r"))->setfunc(bip_cv_get_r) ;
   _SymVal(GaeaSymbol("$cv_put"))->setfunc(bip_cv_put) ;
   _SymVal(GaeaSymbol("$cv_timedwait_get"))->setfunc(bip_cv_timedwait_get) ;
   _SymVal(GaeaSymbol("$cv_timedwait_put"))->setfunc(bip_cv_timedwait_put) ;
   _SymVal(GaeaSymbol("$cv_listing"))->setfunc(bip_cv_listing) ;

   _SymVal(GaeaSymbol("$assign_gdbox"))->setfunc(bip_assign_gdbox) ;
   _SymVal(GaeaSymbol("$gdbox_get"))->setfunc(bip_gdbox_get) ;
   _SymVal(GaeaSymbol("$gdbox_put"))->setfunc(bip_gdbox_put) ;
   _SymVal(GaeaSymbol("$gdbox_timedwait_get"))->setfunc(bip_gdbox_timedwait_get) ;
   _SymVal(GaeaSymbol("$gdbox_timedwait_put"))->setfunc(bip_gdbox_timedwait_put) ;

   
} ;





