// -*- Mode:C++ -*-
//Header:
//File: bipred_io.cc
//Title: built-in predicate 
//Author: NODA Itsuki
//Date: 1999/5/3
//EndHeader:

//ModifyHistory:
// 1999/5/3: Create this file
//EndModifyHistory:

//Memo:
//EndMemo:

#include "bipred.h"

//======================================================================
// Initialize (external)
//

static Bool bipInit() ;

extern "C" {
   Bool bipredInit_io() {
      static Bool initializedp = False ;
      if(!initializedp) {
	 bipInit() ;
	 initializedp = True ;
	 return True ;
      } else {
	 return False ;
      }
   } ;
} ;

//======================================================================
// Write
//

BipStatus bip_write(Gem* gem) {
   /*	$write(+Data [,+Stream])	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY2(icgoal,cgoal,gem,1,2);

   InstGaeaData gdt = _fetchedArgv(icgoal,0);

   gostream *ostr = gem->ostrm ;
   if(chkArity(cgoal,2)) {
      GaeaData strm = fetchedArgv(icgoal,1) ;
      if(!isIOStream(strm))
	 return bipError(gem,"Illegal IOStream object") ;
      ostr = (gostream*)_ExObjVal(strm) ;
   }
   
//   *(gem->ostrm) << fetchValue(gdt,&(gem->uiheap)) ;
   fetchValue(gdt,&(gem->uiheap)).describe(*ostr,False,False,1) ;
   ostr->flush() ;
   return bipSuccess(gem) ;
}

BipStatus bip_nl(Gem* gem) {
   /*	$nl [(+Stream)]	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY_NOorN(icgoal,cgoal,gem,1);

   gostream* ostr = gem->ostrm ;
   if(chkArity(cgoal,1)) {
      GaeaData strm = fetchedArgv(icgoal,0) ;
      if(!isIOStream(strm))
	 return bipError(gem,"Illegal IOStream object") ;
      ostr = (gostream*)_ExObjVal(strm) ;
   }

   *ostr << endl ;
   return bipSuccess(gem) ;
} 

BipStatus bip_tab(Gem* gem) {
   /*	$tab [(+Stream)]	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY_NOorN(icgoal,cgoal,gem,1);

   gostream* ostr = gem->ostrm ;
   if(chkArity(cgoal,1)) {
      GaeaData strm = fetchedArgv(icgoal,0) ;
      if(!isIOStream(strm))
	 return bipError(gem,"Illegal IOStream object") ;
      ostr = (gostream*)_ExObjVal(strm) ;
   }

   *ostr << "\t" ;
   return bipSuccess(gem) ;
} 

BipStatus bip_write_special_char(Gem* gem) {
   /*	$write_special_char(+CharSymbol [, +Stream])	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY2(icgoal,cgoal,gem,1,2);
   
   GaeaData sym = fetchedArgv(icgoal,0) ;

   gostream* ostr = gem->ostrm ;
   if(chkArity(cgoal,2)) {
      GaeaData strm = fetchedArgv(icgoal,1) ;
      if(!isIOStream(strm))
	 return bipError(gem,"Illegal IOStream object") ;
      ostr = (gostream*)_ExObjVal(strm) ;
   }

   if        (isGaeaInteger(sym)) {
      *ostr << (char)(IntVal(sym)) ;
   } else if (eq(sym,SpecialCharSymbol_Newline)) {
      *ostr << endl ;
   } else if (eq(sym,SpecialCharSymbol_Tab)) {
      *ostr << "\t" ;
   } else if (eq(sym,SpecialCharSymbol_Space)) {
      *ostr << " " ;
   } else if (isGaeaSymbol(sym)) {
      *ostr << _SymVal(sym)->name ;
   } else if (isGaeaString(sym)) {
      *ostr << _StrVal(sym) ;
   } else {
      return bipError(gem,"Illegal special charactor symbol") ;
   }
   ostr->flush() ;
   return bipSuccess(gem) ;
}

BipStatus bip_write_ctime(Gem* gem) {
   /*	$write_ctime [(+Stream)]	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY_NOorN(icgoal,cgoal,gem,1);

   gostream* ostr = gem->ostrm ;
   if(chkArity(cgoal,1)) {
      GaeaData strm = fetchedArgv(icgoal,0) ;
      if(!isIOStream(strm))
	 return bipError(gem,"Illegal IOStream object") ;
      ostr = (gostream*)_ExObjVal(strm) ;
   }

   TimeVal ctime ; ctime.update() ;

   ctime.output_normal_form(*ostr) ;

   return bipSuccess(gem) ;
} 


//======================================================================
// Read
//

BipStatus bip_read(Gem* gem) {
   /*	$read(-Data [,+Stream])	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY2(icgoal,cgoal,gem,1,2);

   InstGaeaData var = _fetchedArgv(icgoal,0) ;

   gistream *istr = gem->istrm ;
   if(chkArity(cgoal,2)) {
      GaeaData strm = fetchedArgv(icgoal,1) ;
      if(!isIOStream(strm))
	 return bipError(gem,"Illegal IOStream object") ;
      istr = (gistream*)_ExObjVal(strm) ;
   }
   

//   GaeaData dt = readGaeaData(*istr) ;
   GaeaData dt = readSentence(*istr) ;
   
   if(unifyWith(dt,var,gem)) 
      return bipSuccess(gem) ;
   else
      return bipCutFail(gem) ;

}

//======================================================================
// Read Line
//

BipStatus bip_readline(Gem* gem) {
   /*	$readline(-String [,+Stream])	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY2(icgoal,cgoal,gem,1,2);

   InstGaeaData var = _fetchedArgv(icgoal,0) ;

   gistream *istr = gem->istrm ;
   if(chkArity(cgoal,2)) {
      GaeaData strm = fetchedArgv(icgoal,1) ;
      if(!isIOStream(strm))
	 return bipError(gem,"Illegal IOStream object") ;
      istr = (gistream*)_ExObjVal(strm) ;
   }

   char buf[BUFSIZE] ;
   istr->getline(buf,BUFSIZE) ;

   if(unifyWith(GaeaString(strdup(buf)),var,gem))
      return bipSuccess(gem) ;
   else
      return bipCutFail(gem) ;

}

//======================================================================
// FileAccess
//

BipStatus bip_stdin(Gem* gem) {
   /*	$stdin(-Stream)	*/
   
   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1);

   InstGaeaData var = _fetchedArgv(icgoal,0);

   if(unifyWith(GaeaExtObject(EOT_IOStream,gem->istrm),var,gem))
      return bipSuccess(gem) ;
   else
      return bipCutFail(gem) ;
}

BipStatus bip_stdout(Gem* gem) {
   /*	$stdout(-Stream)	*/
   
   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1);

   InstGaeaData var = _fetchedArgv(icgoal,0);

   if(unifyWith(GaeaExtObject(EOT_IOStream,gem->ostrm),var,gem))
      return bipSuccess(gem) ;
   else
      return bipCutFail(gem) ;
}

BipStatus bip_stderr(Gem* gem) {
   /*	$stderr(-Stream)	*/
   
   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1);

   InstGaeaData var = _fetchedArgv(icgoal,0);

   if(unifyWith(GaeaExtObject(EOT_IOStream,gem->estrm),var,gem))
      return bipSuccess(gem) ;
   else
      return bipCutFail(gem) ;
}

BipStatus bip_fopen(Gem* gem) {
   /*	$fopen(-Stream,+FileName,+Mode)	*/
   
   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,3);

   ARGVAL_VAR(var,gem,icgoal,0) ;

   GaeaData fname = fetchedArgv(icgoal,1) ;
   if(!isGaeaString(fname))
      return bipError(gem,"Illegal filename") ;

   GaeaData mode = fetchedArgv(icgoal,2) ;
   if(!isGaeaSymbol(mode)) 
      return bipError(gem,"Illegal fopen mode") ;

   GaeaData strm ;
   if        (eq(mode,FileReadModeSymbol)) {
      gifstream* istr = new gifstream(_StrVal(fname)) ;
      strm = GaeaExtObject(EOT_IOStream,istr) ;
   } else if (eq(mode,FileWriteModeSymbol)) {
      gofstream* ostr = new gofstream(_StrVal(fname)) ;
      strm = GaeaExtObject(EOT_IOStream,ostr) ;
   } else if (eq(mode,FileAppendModeSymbol)) {
      gofstream* ostr = new gofstream(_StrVal(fname),ios::app) ;
      strm = GaeaExtObject(EOT_IOStream,ostr) ;
   } else {
      return bipError(gem,"Illegal fopen mode") ;
   }

   if(isNullPtr(_ExObjVal(strm))) 
      return bipError(gem,"Can't open file") ;

   if(unifyWith(strm,var,gem)) {
      return bipSuccess(gem) ;
   } else {
      delete ((gfstream*)_ExObjVal(strm)) ;
      return bipCutFail(gem) ;
   }
}

BipStatus bip_fclose(Gem* gem) {
   /*	$fclose(+Stream,+Mode) */

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2);

   GaeaData strm = fetchedArgv(icgoal,0) ;
   if(!isIOStream(strm))
      return bipError(gem,"Illegal IOStream object") ;

   GaeaData mode = fetchedArgv(icgoal,1) ;
   if(!isGaeaSymbol(mode)) 
      return bipError(gem,"Illegal fopen mode") ;

   if        (eq(mode,FileReadModeSymbol)) {
      gifstream* s = ((gifstream*)_ExObjVal(strm)) ;
      delete s ;
   } else if (eq(mode,FileWriteModeSymbol)) {
      gofstream* s = ((gofstream*)_ExObjVal(strm)) ;
      delete s ;
   } else if (eq(mode,FileAppendModeSymbol)) {
      gofstream* s = ((gofstream*)_ExObjVal(strm)) ;
      delete s ;
   } else {
      return bipError(gem,"Illegal fopen mode") ;
   }

   return bipSuccess(gem) ;

}

BipStatus bip_listen(Gem* gem) {
   /* $listen(+Stream) */

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1);

   GaeaData strm = fetchedArgv(icgoal,0) ;
   if(!isIOStream(strm))
      return bipError(gem,"Illegal IOStream object") ;

   gem->istrm = ((gifstream*)_ExObjVal(strm)) ;
   
   return bipSuccess(gem) ;

}

BipStatus bip_listened(Gem* gem) {
   /* $listened(-Stream) */

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1);

   InstGaeaData var = _fetchedArgv(icgoal,0) ;

   GaeaData strm = GaeaExtObject(EOT_IOStream,gem->istrm) ;

   if(unifyWith(strm,var,gem)) {
      gem->istrm = &gin ;
      return bipSuccess(gem) ;
   } else {
      return bipCutFail(gem) ;
   }
}

BipStatus bip_tell(Gem* gem) {
   /* $tell(+Stream) */

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1);

   GaeaData strm = fetchedArgv(icgoal,0) ;
   if(!isIOStream(strm))
      return bipError(gem,"Illegal IOStream object") ;

   gem->ostrm = ((gostream*)_ExObjVal(strm)) ;
   
   return bipSuccess(gem) ;

}

BipStatus bip_told(Gem* gem) {
   /* $told(-Stream) */

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1);

   InstGaeaData var = _fetchedArgv(icgoal,0) ;

   GaeaData strm = GaeaExtObject(EOT_IOStream,gem->ostrm) ;

   if(unifyWith(strm,var,gem)) {
      gem->ostrm = &gout ;
      return bipSuccess(gem) ;
   } else {
      return bipCutFail(gem) ;
   }
}


//======================================================================
// popen
//

BipStatus bip_popen(Gem* gem) {
   /*	$popen(-Strm,+Command,+Arg,+Arg...)	*/
   
   PICKUP_CURRENT_GOAL(icgoal,cgoal,gem) ;

   if(!isGaeaTerm(cgoal) || _TermArity(cgoal) < 2) return bipNoMatch(gem) ;

   InstGaeaData strvar = _fetchedArgv(icgoal,0) ;

   gostrstream ostr ; 
   for(Int i = 1 ; i < _TermArity(cgoal) ; i++) {
      if(i > 0) ostr << " " ;
      GaeaData comname = fetchedArgv(icgoal,i) ;
      if(isGaeaString(comname)) 
	 ostr << _StrVal(comname) ;
      else
	 ostr << comname ;
   }
   ostr << ends ;
   char* comstr =  ostr.str() ;

   FILE* fp = popen(comstr,"w") ;
   if(isNullPtr(fp)) {
      *(gem->estrm) << comstr << endl ;
      return bipError(gem,"Abnormal end of popen call") ;
   } else {
      gostream* pstrm = new gofstream(fileno(fp)) ;
      GaeaData strm = GaeaExtObject(EOT_IOStream,pstrm) ;
      if(unifyWith(strm,strvar,gem)) {
	 return bipSuccess(gem) ;
      } else {
	 return bipCutFail(gem) ;
      }
   }

}

//======================================================================
// Load
//

BipStatus bip_load(Gem* gem) {
   /*	$load(+Filename)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1);

   GaeaData fname = fetchedArgv(icgoal,0) ;
   if(isGaeaString(fname)) {
      gem->popAndState() ;
      gem->load(_StrVal(fname)) ;
      return bipSuccess0(gem) ;
   } else {
      return bipError(gem,"Illegal load parameter") ;
   }
}

BipStatus bip_clearloadpath(Gem* gem) {
   /*	$clearloadpath	*/

   PICKUP_CURRENT_GOAL_CHECK_NOARITY(icgoal,cgoal,gem);

   LoadPath.clear() ;

   return bipSuccess(gem) ;
}

BipStatus bip_addloadpath(Gem* gem) {
   /*	$addloadpath(+Path)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1);

   ARGVAL_CHARPTR(path,gem,icgoal,0) ;

   LoadPath.push(path) ;

   return bipSuccess(gem) ;
}

BipStatus bip_delloadpath(Gem* gem) {
   /*	$delloadpath(-Path)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1);

   ARGVAL_VAR(path,gem,icgoal,0) ;

   String p = LoadPath.pop() ;
   if(unifyWith(GaeaString(p),path,gem)) {
      return bipSuccess(gem) ;
   } else {
      return bipCutFail(gem) ;
   }
}

Bool unifyLoadPathList(InstGaeaData& var, Gem* gem) {
   BindingState bstate(var.binding,&(gem->uiheap)) ;
   Bool r = True ;
   LoadPath.lock() ;
   InstGaeaData v = var ;
   for(UInt i = 0 ; i < LoadPath._toppos() ; i++) {
      GaeaData path = GaeaString(LoadPath._bottomN(i)) ;
      Binding* newbinding = gem->uiheap.newBinding() ;
      if(!unify(v,InstGaeaData(ConsTerm,newbinding),&gem->uiheap)) {
	 r = False ; break ; }
      if(!unifyWith(path,InstGaeaData(ConsCarVar,newbinding),gem)) {
	 r = False ; break ; } 
      v = InstGaeaData(ConsCdrVar,newbinding) ;
   }
   if(!unifyWith(NullSymbol,v,gem)) r = False ;
   LoadPath.unlock() ;
   if(r) {
      return True ;
   } else {
      rewindBindingState(bstate) ;
      return False ;
   } 
}

BipStatus bip_loadpath(Gem* gem) {
   /*	$loadpath(-Path)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1);

   ARGVAL_VAR(var,gem,icgoal,0) ;

   if(unifyLoadPathList(var,gem))
      return bipSuccess(gem) ;
   else
      return bipCutFail(gem) ;
}

//======================================================================
// load dynamic library
//

BipStatus bip_dlload(Gem* gem) {
   /*	$dlload(-Handle,+Filename[,+InitFunc])	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY2(icgoal,cgoal,gem,2,3) ;
   
   InstGaeaData var = _fetchedArgv(icgoal,0) ;
   GaeaData handle = genGaeaSymbol(DLHandleSymbol) ;
   if(!unifyWith(handle,var,gem)) return bipCutFail(gem) ;

   GaeaData fname = fetchedArgv(icgoal,1) ;
   if(!isGaeaString(fname)) return bipError(gem,"Illegal filename") ;

   char* initfunc = NULLPTR ;
   if(chkArity(cgoal,3)) {
      GaeaData ifunc = fetchedArgv(icgoal,2) ;
      if(!isGaeaString(ifunc)) return bipError(gem,"Illegal function name") ;
      initfunc = _StrVal(ifunc) ;
   } 

   if(!DLoad(_StrVal(fname),initfunc,handle,False,
	     (gem->verbosep > VBL_silent)))
      return bipError(gem,"Fail dlopen") ;

   return bipSuccess(gem) ;
}

BipStatus bip_dlload_sure(Gem* gem) {
   /*	$dlload_sure(-Handle,+Filename,+InitFunc)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,3) ;
   
   InstGaeaData var = _fetchedArgv(icgoal,0) ;
   GaeaData handle = genGaeaSymbol(DLHandleSymbol) ;
   if(!unifyWith(handle,var,gem)) return bipCutFail(gem) ;

   GaeaData fname = fetchedArgv(icgoal,1) ;
   if(!isGaeaString(fname)) return bipError(gem,"Illegal filename") ;

   char* initfunc = NULLPTR ;
   GaeaData ifunc = fetchedArgv(icgoal,2) ;
   if(!isGaeaString(ifunc)) return bipError(gem,"Illegal function name") ;
   initfunc = _StrVal(ifunc) ;
      
   if(!DLoad(_StrVal(fname),initfunc,handle,True,
	     (gem->verbosep > VBL_silent)))
      return bipError(gem,"Fail dlopen") ;

   return bipSuccess(gem) ;
}

BipStatus bip_find_function(Gem* gem) {
   /*	$find_function(-FuncObj,+Handle,+FuncName)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,3) ;
   
   InstGaeaData var = _fetchedArgv(icgoal,0) ;

   GaeaData handle = fetchedArgv(icgoal,1) ;
   if(!isGaeaSymbol(handle)) return bipError(gem,"Illegal Symbol") ;
   if(!hasDLHandle(handle))
      return bipError(gem,"Symbol has not dlhandle") ;

   GaeaData funcname = fetchedArgv(icgoal,2) ;
   if(!isGaeaString(funcname)) return bipError(gem,"Illegal Function Name");
   
   void* fp = dlsym(_SymVal(handle)->dlhandle,_StrVal(funcname)) ;

   if(isNullPtr(fp)) {
      *(gem->estrm) << dlerror() << endl ;
      return bipError(gem,"Can't find function") ;
   }

   if(!unifyWith(GaeaPointer(fp),var,gem)) 
      return bipCutFail(gem) ;
   
   return bipSuccess(gem) ;
}

BipStatus bip_symbol_function(Gem* gem) {
   /*	$symbol_function(+Symbol,-FuncObj)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

   GaeaData sym = fetchedArgv(icgoal,0) ;
   if(!isGaeaSymbol(sym)) return bipError(gem,"Illegal Symbol") ;

   InstGaeaData var = _fetchedArgv(icgoal,1) ;

   if(!_SymVal(sym)->hasFunc()) return bipCutFail(gem) ;

   GaeaData funcobj = GaeaPointer((void*)(SymVal(sym)->func)) ;

   if(!unifyWith(funcobj,var,gem)) return bipCutFail(gem) ;

   return bipSuccess(gem) ;
}

BipStatus bip_set_symbol_function(Gem* gem) {
   /*	$set_symbol_function(+Symbol,+FuncObj)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

   GaeaData sym = fetchedArgv(icgoal,0) ;
   if(!isGaeaSymbol(sym)) return bipError(gem,"Illegal Symbol") ;

   GaeaData funcobj = fetchedArgv(icgoal,1) ;
   if(!isGaeaPointer(funcobj)) return bipError(gem,"Illegal function pointer") ;

   _SymVal(sym)->setfunc((BipStatus (*)(Gem*))_PtrVal(funcobj)) ;

   return bipSuccess(gem) ;
}

BipStatus bip_clear_symbol_function(Gem* gem) {
   /*	$clear_symbol_function(+Symbol)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;
   ARGVAL_SYM(sym,gem,icgoal,0) ;

   if(sym->hasFunc()) return bipCutFail(gem) ;
   sym->clearfunc() ;

   return bipSuccess(gem) ;
}

//======================================================================
// io for parser
//

//--------------------------------------------------
// make buffer 

BipStatus bip_make_buffer(Gem* gem) {
   /*	$make_buffer(-Buffer,Size)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;
   ARGVAL_VAR(vbuf,gem,icgoal,0) ;
   ARGVAL_INT(size,gem,icgoal,1) ;

   GaeaBuffer *buf = new GaeaBuffer(size) ;
   if(unifyWith(GaeaExtObject(EOT_Buffer,buf),vbuf,gem)) 
      return bipSuccess(gem) ;
   else
      return bipCutFail(gem) ;
}

//--------------------------------------------------
// get string from buffer

BipStatus bip_get_string_from_buffer(Gem* gem) {
   /*	$get_string_from_buffer(-Str,+Buffer)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;
   ARGVAL_VAR(vstr,gem,icgoal,0) ;
   ARGVAL(vbuf,gem,icgoal,1,isBuffer,"Illegal Buffer Type") ;

   GaeaBuffer* buf = (GaeaBuffer*)_ExObjVal(vbuf) ;

   String* s = new String() ;
   s->data = new char[buf->bodyLength()+1] ;
   s->length = buf->bodyLength() ;
   s->originalp = True ;

   char *dd = s->data ;
   for(char* d = buf->readptr ; d < buf->writeptr ; d++) *(dd++) = *d ;
   *dd = NULLCHAR ;

   if(unifyWith(GaeaString(*s),vstr,gem)) 
      return bipSuccess(gem) ;
   else
      return bipCutFail(gem) ;
}

//--------------------------------------------------
// get string from buffer

BipStatus bip_read_to_buffer(Gem* gem) {
   /*	$read_to_buffer(+Buffer [,+Stream])	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY2(icgoal,cgoal,gem,1,2) ;

   ARGVAL(vbuf,gem,icgoal,0,isBuffer,"Illegal Buffer Type") ;
   GaeaBuffer* buf = (GaeaBuffer*)_ExObjVal(vbuf) ;

   gistream *istr = gem->istrm ;
   if(chkArity(cgoal,2)) {
      GaeaData strm = fetchedArgv(icgoal,1) ;
      if(!isIOStream(strm))
	 return bipError(gem,"Illegal IOStream object") ;
      istr = (gistream*)_ExObjVal(strm) ;
   }

   buf->rewind() ; 
   Bool r = readGeneralSentense(*istr,*buf) ;

   if(r)
      return bipSuccess(gem) ;
   else
      return bipCutFail(gem) ;
}

//--------------------------------------------------
// get a token from buffer

BipStatus bip_get_token(Gem* gem) {
   /*	$get_token(-Token,+Buffer)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

   ARGVAL_VAR(vtoken,gem,icgoal,0) ;
   ARGVAL(vbuf,gem,icgoal,1,isBuffer,"Illegal Buffer Type") ;
   GaeaBuffer* buf = (GaeaBuffer*)_ExObjVal(vbuf) ;

   char* head ; char* tail ;
   TokenType type ;
   Bool r = getToken(*buf,head,tail,type) ;
   if(!r) return bipError(gem,"Can not get a token.") ;

   GaeaData dt ;

   switch(type) {
      case TT_none : return bipCutFail(gem) ;
      case TT_int_bin : { dt = GaeaInteger(strtol(head,NULL,2)) ; break ; }
      case TT_int_oct : { dt = GaeaInteger(strtol(head,NULL,8)) ; break ; }
      case TT_int_dig : { dt = GaeaInteger(strtol(head,NULL,10)) ; break ; }
      case TT_int_hex : { dt = GaeaInteger(strtol(head,NULL,16)) ; break ; }
      case TT_float : 
      case TT_float_e : { dt = GaeaFloat(strtod(head,NULL)) ; break ; }
      case TT_symbol :
      case TT_quoted_symbol : {
	 dt = GaeaSymbol(FindSymbol(head,tail)) ; break ; }
      case TT_paren : {
	 dt = GaeaSymbol(FindSymbol(head,tail)) ; break ; }
      case TT_op : {
	 dt = GaeaSymbol(FindSymbol(head,tail)) ; break ; }
      case TT_variable : {
	 GaeaData v = GaeaSymbol(FindSymbol(head,tail)) ; 
	 Binding* b = gem->uiheap.newBinding() ;
	 if(!unify(vtoken,InstGaeaData(ReaderVarPackTerm,b),&(gem->uiheap)))
	    return bipCutFail(gem) ;
	 if(!unifyWith(v,InstGaeaData(ReaderVarPackArg,b),gem))
	    return bipCutFail(gem) ;
	 return bipSuccess(gem) ;
      }
      default : return bipError(gem,"Illegal token.") ;
   }

   if(unifyWith(dt,vtoken,gem))
      return bipSuccess(gem) ;
   else
      return bipCutFail(gem) ;
}

//======================================================================
// Initialize (external)
//

static Bool bipInit() {

   _SymVal(GaeaSymbol("$write"))->setfunc(bip_write) ;
   _SymVal(GaeaSymbol("$nl"))->setfunc(bip_nl) ;
   _SymVal(GaeaSymbol("$tab"))->setfunc(bip_tab) ;
   _SymVal(GaeaSymbol("$write_special_char"))->setfunc(bip_write_special_char) ;
   _SymVal(GaeaSymbol("$write_ctime"))->setfunc(bip_write_ctime) ;

   _SymVal(GaeaSymbol("$read"))->setfunc(bip_read) ;
   _SymVal(GaeaSymbol("$readline"))->setfunc(bip_readline) ;

   _SymVal(GaeaSymbol("$stdin"))->setfunc(bip_stdin) ;
   _SymVal(GaeaSymbol("$stdout"))->setfunc(bip_stdout) ;
   _SymVal(GaeaSymbol("$stderr"))->setfunc(bip_stderr) ;
   _SymVal(GaeaSymbol("$fopen"))->setfunc(bip_fopen) ;
   _SymVal(GaeaSymbol("$fclose"))->setfunc(bip_fclose) ;
   _SymVal(GaeaSymbol("$listen"))->setfunc(bip_listen) ;
   _SymVal(GaeaSymbol("$listened"))->setfunc(bip_listened) ;
   _SymVal(GaeaSymbol("$tell"))->setfunc(bip_tell) ;
   _SymVal(GaeaSymbol("$told"))->setfunc(bip_told) ;

   _SymVal(GaeaSymbol("$popen"))->setfunc(bip_popen) ;

   _SymVal(GaeaSymbol("$load"))->setfunc(bip_load) ;
   _SymVal(GaeaSymbol("$loadpath"))->setfunc(bip_loadpath) ;
   _SymVal(GaeaSymbol("$addloadpath"))->setfunc(bip_addloadpath) ;
   _SymVal(GaeaSymbol("$delloadpath"))->setfunc(bip_delloadpath) ;
   _SymVal(GaeaSymbol("$clearloadpath"))->setfunc(bip_clearloadpath) ;

   _SymVal(GaeaSymbol("$dlload"))->setfunc(bip_dlload) ;
   _SymVal(GaeaSymbol("$dlload_sure"))->setfunc(bip_dlload_sure) ;

   _SymVal(GaeaSymbol("$find_function"))->setfunc(bip_find_function) ;
   _SymVal(GaeaSymbol("$symbol_function"))->setfunc(bip_symbol_function) ;
   _SymVal(GaeaSymbol("$set_symbol_function"))->setfunc(bip_set_symbol_function) ;
   _SymVal(GaeaSymbol("$clear_symbol_function"))->setfunc(bip_clear_symbol_function) ;

   _SymVal(GaeaSymbol("$make_buffer"))->setfunc(bip_make_buffer) ;
   _SymVal(GaeaSymbol("$get_string_from_buffer"))->setfunc(bip_get_string_from_buffer) ;
   _SymVal(GaeaSymbol("$read_to_buffer"))->setfunc(bip_read_to_buffer) ;
   _SymVal(GaeaSymbol("$get_token"))->setfunc(bip_get_token) ;
   
} ;





