// -*- Mode:C++ -*-
//Header:
//File: bipred_misc.cc
//Title: built-in predicate 
//Author: NODA Itsuki
//Date: 1999/5/3
//EndHeader:

//ModifyHistory:
// 1999/5/3: Create this file
//EndModifyHistory:

//Memo:
//EndMemo:

#include "bipred.h"

//======================================================================
// Initialize (external)
//

static Bool bipInit() ;

extern "C" {
   Bool bipredInit_misc() {
      static Bool initializedp = False ;
      if(!initializedp) {
	 bipInit() ;
	 initializedp = True ;
	 return True ;
      } else {
	 return False ;
      }
   } ;
} ;

//======================================================================
// Halt
//

BipStatus bip_halt(Gem* gem) {
   /*	$halt	*/

   exit(0) ;
} ;

//======================================================================
// Trace/Notrace
//

BipStatus bip_trace(Gem* gem) {
   /*	$trace[(+PID)]	*/

   PICKUP_CURRENT_GOAL(icgoal,cgoal,gem) ;

   if(isGaeaSymbol(cgoal)) {
      gem->traceon(Gem::TRC_normal) ;
      return bipSuccess(gem) ;
   } else if (chkArity(cgoal,1)) {
      GaeaData g = fetchedArgv(icgoal,0) ;
      if(hasGem(g)) {
	 _SymGem(g)->traceon(Gem::TRC_normal) ;
	 return bipSuccess(gem) ;
      } else {
	 return bipError(gem,"Illegal GEM") ;
      }
   } else {
      return bipNoMatch(gem) ;
   } 
}

BipStatus bip_systrace(Gem* gem) {
   /*	$systrace[(+PID)]	*/

   PICKUP_CURRENT_GOAL(icgoal,cgoal,gem);

   if(isGaeaSymbol(cgoal)) {
      gem->traceon(Gem::TRC_kernel) ;
      return bipSuccess(gem) ;
   } else if (chkArity(cgoal,1)) {
      GaeaData g = fetchedArgv(icgoal,0) ;
      if(hasGem(g)) {
	 _SymGem(g)->traceon(Gem::TRC_kernel) ;
	 return bipSuccess(gem) ;
      } else {
	 return bipError(gem,"Illegal GEM") ;
      }
   } else {
      return bipNoMatch(gem) ;
   } 
}

BipStatus bip_notrace(Gem* gem) {
   /*	$notrace[(+PID)]	*/

   PICKUP_CURRENT_GOAL(icgoal,cgoal,gem);

   if(isGaeaSymbol(cgoal)) {
      gem->notrace() ;
      return bipSuccess(gem) ;
   } else if (chkArity(cgoal,1)) {
      GaeaData g = fetchedArgv(icgoal,0) ;
      if(hasGem(g)) {
	 _SymGem(g)->notrace() ;
	 return bipSuccess(gem) ;
      } else {
	 return bipError(gem,"Illegal GEM") ;
      }
   } else {
      return bipNoMatch(gem) ;
   } 

}

//======================================================================
// showState
//

BipStatus bip_showstate(Gem* gem) {
   /*	$showstate[(+PID)] */

   PICKUP_CURRENT_GOAL(icgoal,cgoal,gem);

   if(isGaeaSymbol(cgoal)) {
      gem->describe(*(gem->ostrm),True) ;
      return bipSuccess(gem) ;
   } else if (chkArity(cgoal,1)) {
      GaeaData g = fetchedArgv(icgoal,0) ;
      if(hasGem(g)) {
	 _SymGem(g)->describe(*(gem->ostrm),True) ;
	 return bipSuccess(gem) ;
      } else {
	 return bipError(gem,"Illegal GEM") ;
      }
   } else {
      return bipNoMatch(gem) ;
   } 
}

//======================================================================
// Show GaeaCell Stack
//

BipStatus bip_showenv(Gem* gem) {
   /* $showenv[(+Cell)] */

   PICKUP_CURRENT_GOAL(icgoal,cgoal,gem);

   if(isGaeaSymbol(cgoal)) {
      gem->gcell->showGaeaCellStack(*(gem->ostrm),True) ;
      return bipSuccess(gem) ;
   } else if (chkArity(cgoal,1)) {
      GaeaData g = fetchedArgv(icgoal,0) ;
      if(hasGaeaCell(g)) {
	 _SymGaeaCell(g)->showGaeaCellStack(*(gem->ostrm),True) ;
	 return bipSuccess(gem) ;
      } else {
	 return bipError(gem,"Illegal GaeaCell") ;
      }
   } else {
      return bipNoMatch(gem) ;
   } 
}



//======================================================================
// Chdir
//

BipStatus bip_getcwd(Gem* gem) {
   /*	$getcwd(-Dir)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1);

   InstGaeaData var = _fetchedArgv(icgoal,0);

   char buf[256] ;
   char* dirname = getcwd(buf,256) ;
   
   if(isNullPtr(dirname)) 
      return bipError(gem,"Can't get current working directory") ;

   if(unifyWith(GaeaString(dirname),var,gem)) 
      return bipSuccess(gem) ;
   else
      return bipCutFail(gem) ;
}

BipStatus bip_cd(Gem* gem) {
   /*	$cd(+Dir)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1);

   GaeaData dirname = fetchedArgv(icgoal,0) ;
   if(isGaeaString(dirname)) {
      if(chdir(_StrVal(dirname))!=0) 
	 return bipError(gem,"Directory not found") ;
      return bipSuccess(gem) ;
   }

   return bipError(gem,"Illegal cd parameter") ;
}

BipStatus bip_system(Gem* gem) {
   /*	$system(+Command,+Arg,+Arg...)	*/

   PICKUP_CURRENT_GOAL(icgoal,cgoal,gem) ;

   if(!isGaeaTerm(cgoal) || _TermArity(cgoal) < 1) return bipNoMatch(gem) ;

   gostrstream ostr ; 
   for(Int i = 0 ; i < _TermArity(cgoal) ; i++) {
      if(i > 0) ostr << " " ;
      GaeaData comname = fetchedArgv(icgoal,i) ;
      if(isGaeaString(comname)) 
	 ostr << _StrVal(comname) ;
      else
	 ostr << comname ;
   }
   ostr << ends ;
   char* comstr =  ostr.str() ;
   if(system(comstr)!=0) {
      *(gem->estrm) << comstr << endl ;
      return bipError(gem,"Abnormal end of system call") ;
   } else {
      return bipSuccess(gem) ;
   }
}

//======================================================================
// get System PID
//

BipStatus bip_system_pid(Gem* gem) {
   /*	$system_pid(-PID)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;

   InstGaeaData var = _fetchedArgv(icgoal,0);
   pid_t pid = getpid() ;
   if(unifyWith(GaeaInteger(pid),var,gem)) {
      return bipSuccess(gem) ;
   } else {
      return bipCutFail(gem) ;
   }
}



//======================================================================
// Initialize (external)
//

static Bool bipInit() {

   _SymVal(GaeaSymbol("$halt"))->setfunc(bip_halt) ;

   _SymVal(GaeaSymbol("$trace"))->setfunc(bip_trace) ;
   _SymVal(GaeaSymbol("$systrace"))->setfunc(bip_systrace) ;
   _SymVal(GaeaSymbol("$notrace"))->setfunc(bip_notrace) ;
   _SymVal(GaeaSymbol("$showstate"))->setfunc(bip_showstate) ;
   _SymVal(GaeaSymbol("$showenv"))->setfunc(bip_showenv) ;

   _SymVal(GaeaSymbol("$cd"))->setfunc(bip_cd) ;
   _SymVal(GaeaSymbol("$getcwd"))->setfunc(bip_getcwd) ;
   _SymVal(GaeaSymbol("$system"))->setfunc(bip_system) ;
   _SymVal(GaeaSymbol("$system_pid"))->setfunc(bip_system_pid) ;
   
} ;





