// -*- Mode:C++ -*-
//Header:
//File: bipred_proc.cc
//Title: built-in predicate 
//Author: NODA Itsuki
//Date: 1999/5/3
//EndHeader:

//ModifyHistory:
// 1999/5/3: Create this file
//EndModifyHistory:

//Memo:
//EndMemo:

#include "bipred.h"

//======================================================================
// Initialize (external)
//

static Bool bipInit() ;

extern "C" {
   Bool bipredInit_proc() {
      static Bool initializedp = False ;
      if(!initializedp) {
	 bipInit() ;
	 initializedp = True ;
	 return True ;
      } else {
	 return False ;
      }
   } ;
} ;

//======================================================================
// Fork
//

BipStatus bip_fork(Gem* gem) {
   /*	$fork(+Goal,-PID [,+PNameSymbol]	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY2(icgoal,cgoal,gem,2,3) ;

   GaeaData newgoal = fetchedValueArgv(icgoal,0,gem) ;
   InstGaeaData pid = _fetchedArgv(icgoal,1) ;
   Symbol* pname = NULLPTR ;
   if(chkArity(cgoal,3)) {
      GaeaData pnamesym = fetchedArgv(icgoal,2) ;
      if(isGaeaSymbol(pnamesym)) {
	 pname = _SymVal(pnamesym) ;
      } else {
	 return bipError(gem,"Illegal Process Name") ;
      }
   }
#ifdef THREADED
//   Gem* newgem = new Gem(pname) ;
   Gem* newgem ;
   if(isNullPtr(pname)) {
      newgem = NewGem(pname,false) ;
   } else {
      newgem = NewGem(pname,true) ;
   }

   /*  OLD version
   newgem->gcell->copyGaeaCellStack(gem->gcell) ;
   newgem->settopgoal(newgoal) ;
   newgem->pid.create(NULL,UnfoldLoop,newgem) ;
   */
   newgem->start(newgoal,gem) ;

   unifyWith(GaeaSymbol(newgem->name),pid,gem) ;
   return bipSuccess(gem) ;
#else
   return bipCutFail(gem) ;
#endif ;
}

//======================================================================
// PID
//

BipStatus bip_pid(Gem* gem) {
   /*	$pid(-PID)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;

   InstGaeaData var = _fetchedArgv(icgoal,0);
   if(unifyWith(GaeaSymbol(gem->name),var,gem)) {
      return bipSuccess(gem) ;
   } else {
      return bipCutFail(gem) ;
   }
}
      
//======================================================================
// yield
//

BipStatus bip_yield(Gem* gem) {
   /*	$yield	*/

   PICKUP_CURRENT_GOAL_CHECK_NOARITY(icgoal,cgoal,gem) ;

   Yield() ;
   return bipSuccess(gem) ;
} ;
      
//======================================================================
// die
//

BipStatus bip_die(Gem* gem) {
   /*	$die	*/

   PICKUP_CURRENT_GOAL_CHECK_NOARITY(icgoal,cgoal,gem) ;

   gem->orStateStack.reset() ;
   return bipCutFail(gem) ;
} ;
      
//======================================================================
// show process status
//

BipStatus bip_showstatus(Gem* gem) {
   /* $showstatus(PID) */

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1);

   GaeaData gd = fetchedArgv(icgoal,0) ;
   if(hasGem(gd)) {
      _SymGem(gd)->showstatus(*(gem->ostrm)) ;
      return bipSuccess(gem) ;
   } else {
      return bipError(gem,"Illegal GEM") ;
   } 
}

//======================================================================
// Timer & counter
//

BipStatus bip_ctime(Gem* gem) {
   /*	$ctime(-Sec,-USec)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2);

   InstGaeaData var0 = _fetchedArgv(icgoal,0);
   InstGaeaData var1 = _fetchedArgv(icgoal,1);
   TimeVal ctime ;
   ctime.update() ;

   if(unifyWith(GaeaInteger(ctime.sec()),var0,gem) &&
      unifyWith(GaeaInteger(ctime.usec()),var1,gem)) {
      return bipSuccess(gem) ;
   } else {
      return bipCutFail(gem) ;
   }
}

BipStatus bip_birthtime(Gem* gem) {
   /*	$birthtime(-Sec,-USec)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2);

   InstGaeaData var0 = _fetchedArgv(icgoal,0);
   InstGaeaData var1 = _fetchedArgv(icgoal,1);

   if(unifyWith(GaeaInteger(gem->birthTime.sec()),var0,gem) &&
      unifyWith(GaeaInteger(gem->birthTime.usec()),var1,gem)) {
      return bipSuccess(gem) ;
   } else {
      return bipCutFail(gem) ;
   }
}

BipStatus bip_stepcount(Gem* gem) {
   /*	$stepcount(-Count)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1);

   InstGaeaData var0 = _fetchedArgv(icgoal,0);

   if(unifyWith(GaeaInteger(gem->stepcount),var0,gem)) {
      return bipSuccess(gem) ;
   } else {
      return bipCutFail(gem) ;
   }
}

BipStatus bip_utimedif(Gem* gem) {
   /*	$utimedif(+Sec1,+USec1,+Sec2,+USec2,-Dif)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,5);

   GaeaData stsec = fetchedArgv(icgoal,0);
   GaeaData stusec = fetchedArgv(icgoal,1);
   GaeaData edsec = fetchedArgv(icgoal,2);
   GaeaData edusec = fetchedArgv(icgoal,3);
   InstGaeaData var = _fetchedArgv(icgoal,4);

   Int dsec = IntVal(edsec) - IntVal(stsec) ;
   Int dusec = dsec * 1000000 + IntVal(edusec) - IntVal(stusec) ;
//      GaeaData dif = GaeaFloat((Float)dusec/1000000.0) ;
   GaeaData dif = GaeaInteger(dusec) ;

   if(unifyWith(dif,var,gem)) {
      return bipSuccess(gem) ;
   } else {
      return bipCutFail(gem) ;
   }
}

//======================================================================
// timer from start gaea
//

TimeVal GaeaStartTime ;

BipStatus bip_rtime_msec(Gem* gem) {
   /*	$rtime_msec(-MSec)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1);

   InstGaeaData var0 = _fetchedArgv(icgoal,0);

   TimeVal ctime ;
   ctime.update() ;

   UInt dmsec = GaeaStartTime.diffMSec(ctime) ;

   if(unifyWith(GaeaInteger(dmsec),var0,gem)) {
      return bipSuccess(gem) ;
   } else {
      return bipCutFail(gem) ;
   }
}

//======================================================================
// Sleep
//

BipStatus bip_sleep(Gem* gem) {
   /*	$sleep(+Sec)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1);

   GaeaData sec = fetchedArgv(icgoal,0);
   if(isGaeaInteger(sec)) {
      Sleep(IntVal(sec)) ;
      return bipSuccess(gem) ;
   } else {
      return bipCutFail(gem) ;
   }
}

BipStatus bip_msleep(Gem* gem) {
   /*	$msleep([+Sec,] +MSec)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY2(icgoal,cgoal,gem,1,2);

   if (chkArity(cgoal,1)) {
      GaeaData msec = fetchedArgv(icgoal,0);
      if(isGaeaInteger(msec)) {
	 MSleep(IntVal(msec)) ;
	 return bipSuccess(gem) ;
      } else {
	 return bipError(gem,"Illegal Argument") ;
      }
   } else if (chkArity(cgoal,2)) {
      GaeaData sec = fetchedArgv(icgoal,0) ;
      GaeaData msec = fetchedArgv(icgoal,1) ;
      if(isGaeaInteger(sec) && isGaeaInteger(msec)) {
	 MSleep(IntVal(sec),IntVal(msec)) ;
	 return bipSuccess(gem) ;
      } else {
	 return bipError(gem,"Illegal Argument") ;
      }
   } else {
      return bipNoMatch(gem) ;
   }
}

BipStatus bip_usleep(Gem* gem) {
   /*	$usleep([+Sec,] +USec)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY2(icgoal,cgoal,gem,1,2);
   
   if (chkArity(cgoal,1)) {
      GaeaData usec = fetchedArgv(icgoal,0);
      if(isGaeaInteger(usec)) {
	 USleep(IntVal(usec)) ;
	 return bipSuccess(gem) ;
      } else {
	 return bipError(gem,"Illegal Argument") ;
      }
   } else if (chkArity(cgoal,2)) {
      GaeaData sec = fetchedArgv(icgoal,0) ;
      GaeaData usec = fetchedArgv(icgoal,1) ;
      if(isGaeaInteger(sec) && isGaeaInteger(usec)) {
	 USleep(IntVal(sec),IntVal(usec)) ;
	 return bipSuccess(gem) ;
      } else {
	 return bipError(gem,"Illegal Argument") ;
      }
   } else {
      return bipNoMatch(gem) ;
   }
}

BipStatus bip_sleep_forever(Gem* gem) {
   /*	$sleep_forever	*/

   PICKUP_CURRENT_GOAL_CHECK_NOARITY(icgoal,cgoal,gem) ;
   
   SleepForever() ;
   return bipSuccess(gem) ;

}

//======================================================================
// suspend
//

BipStatus bip_suspend(Gem* gem) {
   /*	$suspend	*/

   PICKUP_CURRENT_GOAL_CHECK_NOARITY(icgoal,cgoal,gem) ;

   gem->suspend() ;
   return bipSuccess(gem) ;
}

BipStatus bip_timedsuspend(Gem* gem) {
   /*	$timedsuspend(+Sec [, +MSec [,+USec]])	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY2(icgoal,cgoal,gem,1,3) ;

   Int sec = 0 ; Int msec = 0 ; Int usec = 0 ; 

   ARGVAL0_INT(sec,gem,icgoal,0) ;
   if(chkArityGE(cgoal,2)) {
      ARGVAL0_INT(msec,gem,icgoal,1) ;
      if(chkArityGE(cgoal,3)) {
	 ARGVAL0_INT(usec,gem,icgoal,2) ;
      }
   }

   if(gem->timedsuspend(sec,msec,usec))
      return bipSuccess(gem) ;
   else
      return bipCutFail(gem) ;

}

BipStatus bip_resume(Gem* gem) {
   /*	$resume(+PID)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;

   GaeaData gd = fetchedArgv(icgoal,0) ;
   if(hasGem(gd)) {
      _SymGem(gd)->resume() ;
      return bipSuccess(gem) ;
   } else {
      return bipError(gem,"Illegal GEM") ;
   } 
}

BipStatus bip_suspending(Gem* gem) {
   /*	$suspending(+PID)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;

   GaeaData gd = fetchedArgv(icgoal,0) ;
   if(hasGem(gd)) {
      if(_SymGem(gd)->suspendp()) {
	 return bipSuccess(gem) ;
      } else {
	 return bipCutFail(gem) ;
      }
   } else {
      return bipError(gem,"Illegal GEM") ;
   } 
}

//======================================================================
// signal
//

static Bool analizeSignalList(InstGaeaData& sym, Symbol** symlist, UInt& n) {
   if(isGaeaSymbol(sym.data)) {
      symlist[n] = SymVal(sym.data) ; n = 1 ;
   } else {
      InstGaeaData slist ;
      for(slist = sym ; isInstConsTerm(slist) ; 
	  slist = InstConsCdrVal(slist)) {
	 InstGaeaData s = InstConsCarVal(slist) ;
	 if(isGaeaSymbol(s.data)) {
	    symlist[n] = SymVal(s.data) ;
	    n++ ;
	 } else {
	    return False ;
	 }
      }
      if(!isNullSymbol(slist.data)) {
	 return False ;
      }
   }
   return True ;
}
   

BipStatus bip_wait_signal(Gem* gem) {
   /*	$wait_signal(-Signal,+SignalSymbol | +SignalSymbolList) */

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

   ARGVAL_VAR(sig,gem,icgoal,0) ;
   ARGVAL_VAR(sym,gem,icgoal,1) ;

   Symbol* symlist[GEMSIGNALENTRY_MAX] ;
   UInt n = 0 ;

   if(!analizeSignalList(sym,symlist,n)) {
      return bipError(gem,
		      "Signal should be a symbol or a list of symbols.",
		      icgoal) ;
   } 

   Symbol* sg = gem->waitSignal(symlist,n) ;
   if(unifyWith(GaeaSymbol(sg),sig,gem)) {
      return bipSuccess(gem) ;
   } else {
      return bipCutFail(gem) ;
   }
}

BipStatus bip_timedwait_signal(Gem* gem) {
   /*	$timedwait_signal(-Signal,+SignalSymbol | +SignalSymbolList, +Sec [, +MSec [,+USec]]) */

   PICKUP_CURRENT_GOAL_CHECK_ARITY2(icgoal,cgoal,gem,3,5) ;

   ARGVAL_VAR(sig,gem,icgoal,0) ;
   ARGVAL_VAR(sym,gem,icgoal,1) ;

   Int sec = 0, msec = 0, usec = 0 ;
   ARGVAL0_INT(sec,gem,icgoal,2) ;
   if(chkArityGE(cgoal,4)) {
      ARGVAL0_INT(msec,gem,icgoal,3) ;
      if(chkArityGE(cgoal,5)) {
	 ARGVAL0_INT(usec,gem,icgoal,4) ;
      }
   }

   Symbol* symlist[GEMSIGNALENTRY_MAX] ;
   UInt n = 0 ;

   if(!analizeSignalList(sym,symlist,n)) {
      return bipError(gem,
		      "Signal should be a symbol or a list of symbols.",
		      icgoal) ;
   } 

   Symbol* sg = gem->timedwaitSignal(symlist,n,sec,msec,usec) ;
   if(unifyWith(GaeaSymbol(sg),sig,gem)) {
      return bipSuccess(gem) ;
   } else {
      return bipCutFail(gem) ;
   }
}

BipStatus bip_signal(Gem* gem) {
   /*	$signal(+Signal) */

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;

   ARGVAL_SYM(sig,gem,icgoal,0) ;

   broadcastSignal(sig) ;

   return bipSuccess(gem) ;
}

BipStatus bip_signal_wait_list(Gem* gem) {
   /*	$signal_wait_list(+Signal,-List) */

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

   ARGVAL_SYM(sig,gem,icgoal,0) ;
   ARGVAL_VAR(list,gem,icgoal,1) ;

   InstGaeaData l = list ;

   for(GaeaSignalEntry* sigentry = sig->signalWaitList ;
       !isNullPtr(sigentry) ;
       sigentry = sigentry->next) {
      if(!unifyWithListShift(l,GaeaSymbol(sigentry->gem->name),gem))
	 return bipCutFail(gem) ;
   }
   if(!unifyWith(NullSymbol,l,gem)) 
      return bipCutFail(gem) ;

   return bipSuccess(gem) ;
}

BipStatus bip_waiting_signal(Gem* gem) {
   /*	$waiting_signal(+Gem,-SignalList) */
   
   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

   ARGVALCONV(Gem*,g,gem,icgoal,0,hasGem,_SymGem,"Illegal GEM") ;
   ARGVAL_VAR(list,gem,icgoal,1) ;

   InstGaeaData l = list ;
   for(UInt i = 0 ; i < g->signalEntryN ; i++) {
      if(!unifyWithListShift(l,GaeaSymbol(g->signalEntry[i].symbol),gem))
	 return bipCutFail(gem) ;
   }
   if(!unifyWith(NullSymbol,l,gem)) 
      return bipCutFail(gem) ;

   return bipSuccess(gem) ;
}

BipStatus bip_last_signal_time(Gem* gem) {
   /*	$last_signal_time(+Symbol,-Sec,-USec) */
   
   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,3) ;

   ARGVAL_SYM(sig,gem,icgoal,0) ;
   ARGVAL_VAR(secv,gem,icgoal,1) ;
   ARGVAL_VAR(usecv,gem,icgoal,2) ;
  
   Int sec,msec,usec ;
   sig->mutex.lock() ;
   sig->lastSignalTime.extract(sec,msec,usec) ;
   sig->mutex.unlock() ;

   if(unifyWith(GaeaInteger(sec),secv,gem) &&
      unifyWith(GaeaInteger(usec+msec*1000),usecv,gem)) 
      return bipSuccess(gem) ;
   else
      return bipCutFail(gem) ;
}

BipStatus bip_signal_is_newer(Gem* gem) {
   /*	$signal_is_newer(+Symbol,+Sec,+USec) */
   
   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,3) ;

   ARGVAL_SYM(sig,gem,icgoal,0) ;
   ARGVAL_INT(rsec,gem,icgoal,1) ;
   ARGVAL_INT(rusec,gem,icgoal,2) ;
  
   Int sec,msec,usec ;
   sig->mutex.lock() ;
   sig->lastSignalTime.extract(sec,msec,usec) ;
   sig->mutex.unlock() ;

   if(sec > rsec || (sec == rsec && (usec+msec*1000 > rusec)))
      return bipSuccess(gem) ;
   else
      return bipCutFail(gem) ;
}



//======================================================================
// CondBarrier
//

BipStatus bip_newbarrier(Gem* gem) {
   /*	$newbarrier(-NewCondBarrier)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;

   InstGaeaData var = _fetchedArgv(icgoal,0) ;
   GaeaMutexCondVar* bar = new GaeaMutexCondVar() ;
   if(unifyWith(GaeaExtObject(EOT_Barrier,bar),var,gem)) {
      return bipSuccess(gem) ;
   } else {
      return bipCutFail(gem) ;
   }
}

BipStatus bip_barrier_wait(Gem* gem) {
   /*	$barrier_wait(+CondBarrier)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;

   GaeaData bar = fetchedArgv(icgoal,0) ;
   if(isBarrier(bar)) {
      ((GaeaMutexCondVar*)_ExObjVal(bar))->lockwait() ;
      return bipSuccess(gem) ;
   } else {
      return bipCutFail(gem) ;
   }
}

BipStatus bip_barrier_timedwait(Gem* gem) {
   /*	$barrier_timedwait(+CondBarrier, Sec [, MSec [, USec ]])	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY2(icgoal,cgoal,gem,2,4) ;

   GaeaData bar = fetchedArgv(icgoal,0) ;

   Int sec = 0 ; Int msec = 0 ; Int usec = 0 ; 

   ARGVAL0_INT(sec,gem,icgoal,1) ;
   if(chkArityGE(cgoal,3)) {
      ARGVAL0_INT(msec,gem,icgoal,2) ;
      if(chkArityGE(cgoal,4)) {
	 ARGVAL0_INT(usec,gem,icgoal,3) ;
      }
   }
   
   if(isBarrier(bar)) {
      if(((GaeaMutexCondVar*)_ExObjVal(bar))->timedlockwait(sec,msec,usec))
	 return bipSuccess(gem) ;
      else
	 return bipCutFail(gem) ;
   } else {
      return bipCutFail(gem) ;
   }
}

BipStatus bip_barrier_signal(Gem* gem) {
   /*	$barrier_signal(+CondBarrier)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;

   GaeaData bar = fetchedArgv(icgoal,0) ;
   if(isBarrier(bar)) {
      ((GaeaMutexCondVar*)_ExObjVal(bar))->signal() ;
      return bipSuccess(gem) ;
   } else {
      return bipCutFail(gem) ;
   }
}
      
BipStatus bip_barrier_broadcast(Gem* gem) {
   /*	$barrier_broadcast(+CondBarrier)	*/

   PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;

   GaeaData bar = fetchedArgv(icgoal,0) ;
   if(isBarrier(bar)) {
      ((GaeaMutexCondVar*)_ExObjVal(bar))->broadcast() ;
      return bipSuccess(gem) ;
   } else {
      return bipCutFail(gem) ;
   }
}
      
//======================================================================
// Initialize (local)
//

static Bool bipInit() {

   _SymVal(GaeaSymbol("$fork"))->setfunc(bip_fork) ;
   _SymVal(GaeaSymbol("$pid"))->setfunc(bip_pid) ;
   _SymVal(GaeaSymbol("$yield"))->setfunc(bip_yield) ;
   _SymVal(GaeaSymbol("$die"))->setfunc(bip_die) ;

   _SymVal(GaeaSymbol("$showstatus"))->setfunc(bip_showstatus) ;

   _SymVal(GaeaSymbol("$ctime"))->setfunc(bip_ctime) ;
   _SymVal(GaeaSymbol("$birthtime"))->setfunc(bip_birthtime) ;
   _SymVal(GaeaSymbol("$stepcount"))->setfunc(bip_stepcount) ;

   _SymVal(GaeaSymbol("$utimedif"))->setfunc(bip_utimedif) ;

   _SymVal(GaeaSymbol("$rtime_msec"))->setfunc(bip_rtime_msec) ;

   _SymVal(GaeaSymbol("$sleep"))->setfunc(bip_sleep) ;
   _SymVal(GaeaSymbol("$msleep"))->setfunc(bip_msleep) ;
   _SymVal(GaeaSymbol("$usleep"))->setfunc(bip_usleep) ;
   _SymVal(GaeaSymbol("$sleep_forever"))->setfunc(bip_sleep_forever) ;

   _SymVal(GaeaSymbol("$suspend"))->setfunc(bip_suspend) ;
   _SymVal(GaeaSymbol("$timedsuspend"))->setfunc(bip_timedsuspend) ;
   _SymVal(GaeaSymbol("$resume"))->setfunc(bip_resume) ;
   _SymVal(GaeaSymbol("$suspending"))->setfunc(bip_suspending) ;

   _SymVal(GaeaSymbol("$wait_signal"))->setfunc(bip_wait_signal) ;
   _SymVal(GaeaSymbol("$timedwait_signal"))->setfunc(bip_timedwait_signal) ;
   _SymVal(GaeaSymbol("$signal"))->setfunc(bip_signal) ;
   _SymVal(GaeaSymbol("$signal_wait_list"))->setfunc(bip_signal_wait_list) ;
   _SymVal(GaeaSymbol("$waiting_signal"))->setfunc(bip_waiting_signal) ;
   _SymVal(GaeaSymbol("$last_signal_time"))->setfunc(bip_last_signal_time) ;
   _SymVal(GaeaSymbol("$signal_is_newer"))->setfunc(bip_signal_is_newer) ;

   _SymVal(GaeaSymbol("$newbarrier"))->setfunc(bip_newbarrier) ;
   _SymVal(GaeaSymbol("$barrier_wait"))->setfunc(bip_barrier_wait) ;
   _SymVal(GaeaSymbol("$barrier_timedwait"))->setfunc(bip_barrier_timedwait) ;
   _SymVal(GaeaSymbol("$barrier_signal"))->setfunc(bip_barrier_signal) ;
   _SymVal(GaeaSymbol("$barrier_broadcast"))->setfunc(bip_barrier_broadcast) ;

   GaeaStartTime.update() ;
   
} ;





