// -*- Mode:C++ -*-
//Header:
//File: cgaea.cc
//Title: gaea (C++ version)
//Author: NODA Itsuki
//Date: 1998/6/20
//EndHeader:

//ModifyHistory:
// 1998/6/20: Create this file
// 1999/3/18: Add -rest option
//EndModifyHistory:

//Memo:
//EndMemo:

//======================================================================
// Includes
//

#include <strings.h>
#include <netdb.h>

#include "gem.h"

#include "bipred.h"

#include "version.h"

//======================================================================
// Parameters
//

#define _MAKESTRING_SUB_(x) #x
#define _MAKESTRING_(x) _MAKESTRING_SUB_(x)

#ifdef CGAEA_LIBDIR
#define _CGAEA_LIBDIR _MAKESTRING_(CGAEA_LIBDIR)
#else
#define _CGAEA_LIBDIR "."
#endif

char* InitFile = "init.g" ;

#define ExecFileMax 64
UInt ExecFilesN = 0 ;
char* ExecFiles[ExecFileMax] ;

#define CallStrMax 64
UInt CallStrN = 0 ;
char* CallStr[CallStrMax] ;

char* MainGemName = "gem" ;

//======================================================================
// GetOptions
//

Int GetOption(int argc, char** argv) {
//   Int c = 0 ;
   for(Int i = 1 ; i < argc ; i++) {
      if        (!strcmp(argv[i],"-e")) {
	 i++ ;
	 ExecFiles[ExecFilesN] = argv[i] ;
	 ExecFilesN++ ;
      } else if (!strcmp(argv[i],"-init")) {
	 i++ ;
	 InitFile = argv[i] ;
      } else if (!strcmp(argv[i],"-noinit")) {
	 InitFile = NULLPTR ;
      } else if (!strcmp(argv[i],"-call")) {
	 i++ ;
	 CallStr[CallStrN] = argv[i] ;
	 CallStrN++ ;
      } else if (!strcmp(argv[i],"-path")) {
	 i++ ;
	 LoadPath.push(argv[i]) ;
#ifdef THREADED
      } else if (!strcmp(argv[i],"-gemN")) {
	 i++ ;
	 GemHeapTabSize = atol(argv[i]) ;
#endif
      } else if (!strcmp(argv[i],"-silent")) {
	 DefaultVerboseLevel = VBL_silent ;
      } else if (!strcmp(argv[i],"-verbose")) {
	 DefaultVerboseLevel = VBL_verbose ;
      } else if (!strcmp(argv[i],"-version")) {
	 gout << CGaeaVersion << "  [" << CGaeaFeatures << "]" << endl
	      << CGaeaCopyright << endl ;
	 exit(0) ;
      } else if (!strcmp(argv[i],"-rest")) {
	 Argc = argc - i ;
	 Argv = &(argv[i]) ;
	 return 0 ;
      } else if (!strcmp(argv[i],"-help")) {
	 gout << CGaeaVersion << "  [" << CGaeaFeatures << "]" << endl
	      << CGaeaCopyright << endl ;
	 gout << "\t% " << argv[0] << " [options]" << endl ;
	 gout << "\t   Option:" << endl ;
	 gout << "\t\t-e FILENAME : file to load before prompt." << endl ;
	 gout << "\t\t-init FILENAME : initfile. [default:init.g]" << endl ;
	 gout << "\t\t-noinit : do not load initfile." << endl ;
	 gout << "\t\t-call FORM : form to call before prompt." << endl ;
	 gout << "\t\t-path DIR : add DIR to search path of load." << endl ;
#ifdef THREADED
	 gout << "\t\t-gemN N : initial size of GEM table." << endl ;
#endif
	 gout << "\t\t-silent : silent mode." << endl ;
	 gout << "\t\t-verbose : verbose mode." << endl ;
	 gout << "\t\t-version : show version." << endl ;
	 gout << "\t\t-rest ... : skip scanning rest of options." << endl ;
	 gout << "\t\t          | The options are passed to additional modules. " << endl ;
	 gout << "\t\t-help : show this message." << endl ;
	 exit(0) ;
      } else {
	 gerr << "Unknown Option: " << argv[i] << endl ;
	 exit(1) ;
      }
   }
   return 0 ;
}


//======================================================================
// Main
//

int main(int argc, char** argv) {
   GetOption(argc,argv) ;

#ifdef Linux
//	This is required for dynamic loading of network facility.
//	I think there are better way.
   gethostbyname("localhost") ;
#endif

   LoadPath.push(_CGAEA_LIBDIR) ;
   LoadPath.push(".") ;

   gglobalInit() ;
   bipredInit() ;
   
//   Gem gem(MainGemName,False) ;
   Gem* _gem = NewGem(MainGemName,False) ;
   Gem& gem = *_gem ;

   gem.override_by_self_id() ;

   if(gem.verbosep > VBL_silent) {
      gout << CGaeaVersion << "  [" << CGaeaFeatures << "]" << endl
	   << CGaeaCopyright << endl ;
#ifdef THREADED
      gout << "GEM table size=" << GemHeapTabSize << endl ;
#endif
   }

   if(!isNullPtr(InitFile)) {
      gem.load(InitFile,BasicCell,True) ;
   }
   for(UInt i = 0 ; i < ExecFilesN ; i++) 
      gem.load(ExecFiles[i],NULLPTR,True) ;

   for(UInt i = 0 ; i < CallStrN ; i++) {
      gistrstream istr(CallStr[i]) ;
      gem.loadloop(istr,NULLPTR,True) ;
   }

   gem.reploop0(*(gem.istrm),*(gem.ostrm)) ;

   exit(0) ;
}




