//   -*- Mode: C++ -*-
//Header:
//EndHeader:

//Memo:
//  Number:
//    Integer:
//	Hex:  0xNNNN   N = [0-9a-fA-F]
//            0XNNNN
//      Dec:  sMNNN    S = [-|+|] M = [1-9], N = [0-9]
//	Oct:  0NNNN    N = [0-8]
//    Float:  sNNN.NNN[e|E]sNNN  (do not permit ".1")
//  Symbol:
//    Symbol = [a-z$][a-zA-Z$_]* | 'any string' | Operator
//    Operator = any symbol charactor except $ and _.
//  Variable:
//      
//EndMemo:

//======================================================================
// Include Files
//----------------------------------------------------------------------
#include "gbio.h"

//======================================================================
// Local Buffer
//----------------------------------------------------------------------

Bool LocalBuffer::putChar(char c) {
   if(max > len) {
      buffer[len] = c ; 
      len++ ;
      return True ;
   } else {
      if(extendp) {
	 extendBuffer() ;
	 return putChar(c) ;
      } else {
	 return False ;
      }
   }
} ;

char* LocalBuffer::duplicateBuffer() const {
   Int l = strlen(buffer) ;
   char* r = new char[l+1] ;
   strcpy(r,buffer) ;
   return r ;
} ;


//======================================================================
// Utility Functions
//----------------------------------------------------------------------
Int skipSpace(gistream& istr, ReadStatus& rstatus) {
   Int i=0 ;
   for(char c = istr.peek() ; isspace(c) ; c = istr.peek()) {
      getChar(istr,rstatus) ;
      i++ ;
   }
   return i ;
}

Bool skipComment(gistream& istr, ReadStatus& rstatus) {
   if(istr.peek() == COMMENTSTARTCHAR) {
      char c ;
      while((c = getChar(istr,rstatus)) != COMMENTENDCHAR) {}
      return True ;
   } else {
      return False ;
   }
}

void skipSpaceComment(gistream& istr, ReadStatus& rstatus) {
   skipSpace(istr,rstatus) ;
   if(skipComment(istr,rstatus)) 
      skipSpaceComment(istr,rstatus) ;
}

Bool readUntilChar(gistream& istr, LocalBuffer& buf, 
		   char endchar, char escapechar,
		   Bool discardlastp, ReadStatus& rstatus) 
{
   char c ;
   for(c = istr.peek() ; c != endchar ; c = istr.peek()) {
      if(c == EOF) {
	 rstatus.errorID = REI_IllegalEOF ;
	 return False ;
      } 
      c = getChar(istr,rstatus) ;
      if(c == escapechar) c = getChar(istr,rstatus) ;
      buf.putChar(c) ;
   }
   if(discardlastp) getChar(istr,rstatus) ;
   buf.putChar(EOSCHAR) ;
   return True ;
}

Bool skipLine(gistream& istr, ReadStatus& rstatus) {
   while(True) {
      if(istr.peek() == EOF) return False ;
      if(istr.peek() == NULLCHAR) return False ;
      if(CRCHAR == getChar(istr,rstatus)) return True ;
   }
}


//======================================================================
// readGaeaData
//----------------------------------------------------------------------

GaeaData readSentence(gistream& istr, ReadStatus& rstatus) {
   rstatus.initNewSentence() ;
   GaeaData r = readSentenceSub(istr, rstatus, 0) ;
   return r ;
}

GaeaData readSentenceSub(gistream& istr, ReadStatus& rstatus, UInt n) {
   GaeaData dt = readGaeaData(istr, rstatus) ;
   if(rstatus.errorID == REI_None) {
      skipSpaceComment(istr,rstatus) ;
      char nextc = getChar(istr,rstatus) ;
      if(nextc == CLAUSEEND) {
	 if(n == 0) {
	    return dt ;
	 } else {
	    GaeaData andterm = GaeaTerm(AndSymbol,n+1) ;
	    _TermArgv(andterm,n) = dt ;
	    return andterm ;
	 }
      } else if (nextc == CLAUSECONJ) {
	 GaeaData andterm = readSentenceSub(istr, rstatus, n+1) ;
	 _TermArgv(andterm,n) = dt ;
	 return andterm ;
      } else {
	 rstatus.errorID = REI_IllegalSentenceFormat ;
	 return ErrorSymbol ;
      } ;
   } else {
      return dt ;
   }
}

GaeaData readSentence0(gistream& istr, ReadStatus& rstatus) {
   rstatus.errorID = REI_None ;
   GaeaData dt = readGaeaData(istr, rstatus) ;
   if(rstatus.errorID == REI_None) {
      skipSpaceComment(istr,rstatus) ;
      char nextc = istr.peek() ;
      if(nextc == CLAUSEEND) getChar(istr,rstatus) ;
   }
   if(eq(dt,ErrorSymbol)) skipLine(istr,rstatus) ; 
   return dt ;
}
   

GaeaData readGaeaData(gistream& istr, ReadStatus& rstatus) {
   skipSpaceComment(istr,rstatus) ;
   char nextc = istr.peek() ;
   GaeaData dt ;
   switch(nextc) {
      case CLAUSEEND :
	 rstatus.errorID = REI_NilData ;
	 return ErrorSymbol;
      case QUOTESYMBOLBEGIN :
	 dt = readQuotedGaeaSymbol(istr,rstatus) ;
	 break ;
      case STRINGBEGIN :
	 dt = readGaeaString(istr,rstatus) ;
	 break ;
      case LISTBEGINCHAR :
	 dt = readList(istr,rstatus) ;
	 break ;
      default:
	 dt = readGNumberGaeaSymbol(istr,rstatus) ;
	 break ;
   }
   if(isGaeaSymbol(dt)) 
      dt = checkArgs(dt,istr,rstatus) ;

   return dt ;
}

GaeaData readQuotedGaeaSymbol(gistream& istr, ReadStatus& rstatus) {
   if(istr.peek() != QUOTESYMBOLBEGIN) {
      rstatus.errorID = REI_IllegalSymbolFormat ;
      return ErrorSymbol ;
   }
   getChar(istr,rstatus) ;

   LocalBuffer buf(BUFSIZE) ;

   readUntilChar(istr,buf,QUOTESYMBOLEND,ESCAPECHAR,True,rstatus) ;

   return GaeaSymbol(buf.buffer) ;
}

GaeaData readGaeaString(gistream& istr, ReadStatus& rstatus) {
   if(istr.peek() != STRINGBEGIN) {
      rstatus.errorID = REI_IllegalStringFormat ;
      return ErrorSymbol ;
   }
   getChar(istr,rstatus) ;

   LocalBuffer buf(BUFSIZE) ;

   readUntilChar(istr,buf,STRINGEND,ESCAPECHAR,True,rstatus) ;

   return GaeaString(buf.buffer) ;
}

GaeaData readList(gistream& istr, ReadStatus& rstatus, const Int listDepth) {
   if(listDepth == 0) getChar(istr,rstatus) ; // skip '['
   
   skipSpaceComment(istr,rstatus) ;
   char nextc = istr.peek() ;

   if(isListEndChar(nextc)) {
      getChar(istr,rstatus) ;
      return NullSymbol ;
   }

   GaeaData cons = GaeaTerm(_SymVal(ConsSymbol),2) ;

   GaeaData car = readGaeaData(istr,rstatus) ;
   GaeaData cdr ;
   cons.val.term->argv(0) = car ;

   skipSpaceComment(istr,rstatus) ;
   nextc = istr.peek() ;
   
   if(isListArgSepChar(nextc)) {
      getChar(istr,rstatus) ;
      cdr = readList(istr,rstatus,listDepth+1) ;
   } else if(isListDotChar(nextc)) {
      getChar(istr,rstatus) ;
      cdr = readGaeaData(istr,rstatus) ;
      if(!isListEndChar(istr.peek())) {
	 rstatus.errorID = REI_IllegalListFormat ;
	 return ErrorSymbol ;
      }
      getChar(istr,rstatus) ;
   } else if(isListEndChar(nextc)) {
      getChar(istr,rstatus) ;
      cdr = NullSymbol ;
   } else {
      rstatus.errorID = REI_IllegalListFormat ;
      return ErrorSymbol ;
   }
   if(eq(cdr,ErrorSymbol)) {
      return ErrorSymbol ;
   }

   cons.val.term->argv(1) = cdr ;
   return cons ;
}

GaeaData readGNumberGaeaSymbol(gistream& istr, ReadStatus& rstatus) {
   LocalBuffer buf(BUFSIZE) ;

   char c = istr.peek() ;
   if(c == EOF) { 
      rstatus.errorID = REI_IllegalEOF ;
      return EOFSymbol ;
   }
   c = getChar(istr,rstatus) ;

   buf.putChar(c) ;
   if(isSignChar(c)) {				/* "+" or "-" */
      char nc = istr.peek() ;
      if(isdigit(nc)) {				/* "+[0-9]" */
	 return readGIntFloatRest(istr,buf,rstatus) ;	/* -> DecInt/Float */
      } else {
	 buf.finish() ;
	 return GaeaSymbol(buf.buffer) ;			/* -> '+' or '-' */
      }
   } else if(isZeroChar(c)) {			/* "0" */
      char nc = istr.peek() ;
      if(isDotChar(nc)) {				/* "0." */
	 return readGIntFloatRest(istr,buf,rstatus) ;	/* -> Float */
      } else if(isHexPreChar(nc)) {		/* "0x" or "0X" */
	 return readGHexIntRest(istr,buf,rstatus) ;	/* -> HexInt */
      } else {					/* "0" */
	 return readGOctIntRest(istr,buf,rstatus) ;	/* ->  OctInt */
      }
   } else if(isdigit(c)) {			/* "[1-9]" */
      return readGIntFloatRest(istr,buf,rstatus) ;	/* -> DecInt/Float */
   } else if(isSymHeadChar(c)) {			/* "[a-z$]" */
      return readGaeaSymbolRest(istr,buf,rstatus) ;		/* -> Symbol */
   } else if(isVarHeadChar(c)) {			/* "[A-Z_]" */
      return readGaeaVariableRest(istr,buf,rstatus);	      /* -> Variable */
   } else {					/* else */
      buf.finish() ;
      return GaeaSymbol(buf.buffer) ;			/* -> Operator */
   }
}

GaeaData readGIntFloatRest(gistream& istr, LocalBuffer& buf, ReadStatus& rstatus){
   enum { IFM_beforePeriod, 
	  IFM_afterPeriod, 
	  IFM_afterExponent 
   } mode ;

   mode = IFM_beforePeriod ;
   while(True) {
      char c = istr.peek() ; 
      if (isdigit(c)) {
	 buf.putChar(getChar(istr,rstatus)) ;
      } else if (isDotChar(c)) {
	 if(mode == IFM_beforePeriod) {
	    buf.putChar(getChar(istr,rstatus)) ;
	    mode = IFM_afterPeriod ;
	 } else {
	    break ;
	 }
      } else if (isExpSepChar(c)) {
	 if(mode == IFM_afterExponent) {
	    break;
	 } else {
	    buf.putChar(getChar(istr,rstatus)) ;
	    char nc = istr.peek() ;
	    if(isSignChar(nc)) {
	       buf.putChar(getChar(istr,rstatus)) ;
	       nc = istr.peek() ;
	    }
	    if(!isdigit(nc)) {
	       rstatus.errorID = REI_IllegalNumberFormat ;
	       return ErrorSymbol ;
	    } 
	 }
      } else {
	 break ;
      }
   }
   buf.finish() ;
  
   if(mode == IFM_beforePeriod) {
      Int ival = strtol(buf.buffer,NULL,10) ;
      return GaeaInteger(ival) ;
   } else {
      Float fval = strtod(buf.buffer,NULL) ;
      return GaeaFloat(fval) ;
   } 
}

GaeaData readGOctIntRest(gistream& istr, LocalBuffer& buf, ReadStatus& rstatus){
   while(True) {
      char c = istr.peek() ;
      if (isOctIntChar(c)) {
	 buf.putChar(getChar(istr,rstatus)) ;
      } else {
	 break ;
      }
   }
   buf.finish() ;
  
   Int ival = strtol(buf.buffer,NULL,8) ;
   return GaeaInteger(ival) ;
}

GaeaData readGHexIntRest(gistream& istr, LocalBuffer& buf, ReadStatus& rstatus){
   buf.putChar(getChar(istr,rstatus)) ;
   while(True) {
      char c = istr.peek() ;
      if(isxdigit(c)) {
	 buf.putChar(getChar(istr,rstatus)) ;
      } else {
	 break ;
      }
   }
   buf.finish() ;
  
   Int ival = strtol(buf.buffer,NULL,16) ;
   return GaeaInteger(ival) ;
}

GaeaData readGaeaSymbolRest(gistream& istr, LocalBuffer& buf, ReadStatus& rstatus){
   while(True) {
      char c = istr.peek() ;
      if (isSymBodyChar(c)){
	 buf.putChar(getChar(istr,rstatus)) ;
      } else {
	 break ;
      }
   }
   buf.finish() ;
  
   return GaeaSymbol(buf.buffer) ;
}

GaeaData readGaeaVariableRest(gistream& istr, LocalBuffer& buf, ReadStatus& rstatus){
   while(True) {
      char c = istr.peek() ;
      if(isSymBodyChar(c)) {
	 buf.putChar(getChar(istr,rstatus)) ;
      } else {
	 break ;
      }
   }
   buf.finish() ;
   
   if(!strcmp(buf.buffer,ANONYMOUSVARSTR)) {
      return rstatus.newAnonymousVar() ;
   } else {
      return GaeaVariable(buf.buffer) ;
   }
}

GaeaData checkArgs(const GaeaData& head, gistream& istr, ReadStatus& rstatus) {
   skipSpaceComment(istr,rstatus) ;
   char nextc = istr.peek() ;
   if(isArgBeginChar(nextc)) {
      getChar(istr,rstatus) ;
      return readArgs(head,istr,rstatus,0) ;
   } else {
      return head ;
   }
}

GaeaData readArgs(const GaeaData& head, gistream& istr, ReadStatus& rstatus,
	       Int argn) 
{
   skipSpaceComment(istr,rstatus) ;
   char nextc = istr.peek() ;
   if(isArgEndChar(nextc) || isGArgSepChar(nextc)) {
      getChar(istr,rstatus) ;
      if(argn < 0)  {
	 rstatus.errorID = REI_IllegalTermFormat ;
	 return ErrorSymbol ;
      } else {
	 GaeaData r = GaeaTerm(head.val.symbol, argn) ;
	 if(isGArgSepChar(nextc)) {
	    _TermGArgP(r) = True ;
	    if(readGArgs(r,istr,rstatus)) {
	       return r ;
	    } else {
	       return ErrorSymbol ;
	    }
	 } else {
	    return r ; 
	 }
      }
   } else if(isArgSepChar(nextc) && argn < 1) {
      rstatus.errorID = REI_IllegalTermFormat ;
      return ErrorSymbol ;
   } else if( (! isArgSepChar(nextc)) && argn > 0) {
      rstatus.errorID = REI_IllegalTermFormat ;
      return ErrorSymbol ;
   } else {
      if(isArgSepChar(nextc)) getChar(istr,rstatus) ;
      GaeaData arg = readGaeaData(istr,rstatus) ;
      GaeaData r = readArgs(head,istr,rstatus,argn + 1) ;
      if(rstatus.errorID != REI_None) return r ;
      r.val.term->argv(argn) = arg ;
      return r ;
   } ;
}

Bool readGArgs(TermGArg*& garg, gistream& istr, ReadStatus& rstatus) {
   skipSpaceComment(istr,rstatus) ;
   char nextc = istr.peek() ;
   if(isArgEndChar(nextc)) {
      getChar(istr,rstatus) ;
      garg = NULLPTR ;
      return True ;
   } else if(isGArgSepChar(nextc)) {
      rstatus.errorID = REI_IllegalTermFormat ;
      return False ;
   } else if (isArgSepChar(nextc)) {
      getChar(istr,rstatus) ;   /* skip "," */
   }

   GaeaData value = readGaeaData(istr,rstatus) ;
   skipSpaceComment(istr,rstatus) ;
   nextc = istr.peek() ;

   if(!isGArgPairChar(nextc)) {
      rstatus.errorID = REI_IllegalTermFormat ;
      return False ;
   }

   getChar(istr,rstatus) ;
   skipSpaceComment(istr,rstatus) ;
   GaeaData role = readGaeaData(istr,rstatus) ;

   if(!isGaeaSymbol(role)) {
      rstatus.errorID = REI_IllegalTermFormat ;
      return False ;
   }

   garg = NewTermGArg(SymVal(role),value) ;

   return readGArgs(garg->next,istr,rstatus) ;
}
      

Bool GaeaSymbol::shouldquotep() const {
   Int i = 0;
   for(char* c = val.symbol->name.data ; *c != NULLCHAR ; c++) {
      if(i == 0 && !isSymHeadChar(*c)) return True ;
      if(!isSymBodyChar(*c)) return True ;
      i++ ;
   }
   return False ;
}
    
//======================================================================
// read for general purpose
//----------------------------------------------------------------------

//----------------------------------------------------------------------
// control char set

GaeaReadControlCharSet DefaultReadControlCharSet = 
	DEFAULT_READ_CONTROL_CHAR_SET ;

//----------------------------------------------------------------------
// read general sentense

Bool readGeneralSentense(gistream& istr, GaeaBuffer& buffer,
			 const GaeaReadControlCharSet& ctrSet) {
   ReadStatus rstatus ;
   enum local_mode { LM_normal, LM_string, LM_quote, LM_comment, 
		     LM_number } ;
   local_mode mode = LM_normal ;

   while(True) {
      char c = getChar(istr,rstatus) ;
      if(isNullChar(c)) {
	 return True ;
      } else if(mode == LM_comment) {
	 if(c == ctrSet.commentEnd) {
	    mode = LM_normal ;
	 }
	 continue ;
      } else if(mode == LM_string) {
	 if(c == ctrSet.stringEnd) mode = LM_normal ;
      } else if(mode == LM_quote) {
	 if(c == ctrSet.quoteEnd) mode = LM_normal ;
      } else if(mode == LM_number) {
	 if(!isdigit(c)) mode = LM_normal ;
      } else { // normal mode
	 if(c == ctrSet.terminate) {
	    buffer.pushChar(c) ;
	    return True ;
	 } else if(c == ctrSet.commentBegin) {
	    mode = LM_comment ;
	    continue ;
	 } else if(c == ctrSet.stringBegin) {
	    mode = LM_string ;
	 } else if(c == ctrSet.quoteBegin) {
	    mode = LM_quote ;
	 } else if (isdigit(c)) {
	    mode = LM_number ;
	 }
      }
      if (c == ctrSet.escape) c = getChar(istr,rstatus) ;
      if(buffer.isFull()) { return False ; } 
      buffer.pushChar(c) ;
   }
}

Bool getToken(GaeaBuffer& buffer, char*& head, char*& tail, TokenType& type) {
   while(isspace(buffer.peekChar())) buffer.getChar() ;
   head = buffer.readptr ;
   char c = buffer.getChar() ;
   if(c == '0') {
      c = buffer.getChar() ;
      if     (c == 'B' || c == 'b') { 
	 type = TT_int_bin ; head = buffer.readptr ;
      } else if(c == 'O' || c == 'o') {
	 type = TT_int_oct ; head = buffer.readptr ;
      } else if(c == 'X' || c == 'x') {
	 type = TT_int_hex ; head = buffer.readptr ;
      } else if('0' <= c && c <= '7') { 
	 type = TT_int_oct ; 
      } else if('0' == '.') type = TT_float ;
      else { type = TT_error ; return False ; }
   } else if(isdigit(c)) type = TT_int_dig ;
   else if (isupper(c) || c == '_') type = TT_variable ;
   else if (islower(c) || c == '$') type = TT_symbol ;
   else if (c == '\'') { 
      type = TT_quoted_symbol ; head = buffer.readptr ;
   } else if (c == '"') {
      type = TT_string ; head = buffer.readptr ;
   } else if (c == '(' || c == ')' || c == '{' || c == '}' || 
	    c == '[' || c == ']') {
      type = TT_paren ; 
      tail = buffer.readptr ;
      return True ;
   } else if (ispunct(c)) {
      type = TT_op ;
   } else {
      type = TT_none ;
      return True ;
   }

   while(True) {
      c = buffer.getChar() ;
      switch(type) {
	 case TT_int_bin: { 
	    if(c == '0' || c == '1') break ;
	    else {
	       buffer.ungetChar() ;
	       tail = buffer.readptr ;
	       return True ; 
	    } 
	 }
	 case TT_int_oct : {
	    if('0' <= c && c <= '7') break ;
	    else {
	       buffer.ungetChar() ;
	       tail = buffer.readptr ;
	       return True ; 
	    } 
	 }
	 case TT_int_dig : {
	    if(isdigit(c)) break ;
	    else if(c == '.') { type = TT_float ; break ; }
	    else {
	       buffer.ungetChar() ;
	       tail = buffer.readptr ;
	       return True ; 
	    } 
	 } 
	 case TT_int_hex : {
	    if(isxdigit(c)) break ;
	    else {
	       buffer.ungetChar() ;
	       tail = buffer.readptr ;
	       return True ; 
	    } 
	 }
	 case TT_float : {
	    if(isdigit(c)) break ;
	    else if (c == 'e' || c == 'E') { type = TT_float_e ; break ; } 
	    else {
	       buffer.ungetChar() ;
	       tail = buffer.readptr ;
	       return True ; 
	    } 
	 } 
	 case TT_float_e : {
	    if(isdigit(c)) break ;
	    else {
	       buffer.ungetChar() ;
	       tail = buffer.readptr ;
	       return True ; 
	    } 
	 } 
	 case TT_string : {
	    if(c == '"') {
	       tail = buffer.readptr ; tail-- ;
	       return True ;
	    } else if (c == '\\') {
	       buffer.getChar() ;
	       break ;
	    } else {
	       break ;
	    }
	 }
	 case TT_quoted_symbol : {
	    if(c == '\'') {
	       tail = buffer.readptr ; tail-- ;
	       return True ;
	    } else if (c == '\\') {
	       buffer.getChar() ;
	       break ;
	    } else {
	       break ;
	    }
	 }
	 case TT_symbol : 
	 case TT_variable : {
	    if(isalnum(c) || c == '_' || c == '$') break ;
	    else {
	       buffer.ungetChar() ;
	       tail = buffer.readptr ;
	       return True ; 
	    } 
	 } 
	 case TT_op : {
	    if(ispunct(c) && c != '_' && c != '$' && 
	       c != '(' && c != ')' && c != '{' && c != '}' && 
	       c != '[' && c != ']') break ;
	    else {
	       if(!isNullChar(c)) buffer.ungetChar() ;
	       tail = buffer.readptr ;
	       return True ; 
	    } 
	 } 
	 default : return False ;
      }
   }
}


//======================================================================
// output for debug
//----------------------------------------------------------------------
void dbp(GaeaData dt) { gerr << dt ; }
void dbd(GaeaData dt) { dt.describe(gerr) ; } 
