//   -*- Mode: C++ -*-
//Header:
//EndHeader:

#ifndef _gbio_h_
#define _gbio_h_
////////////////////////////////////////////////////////////////////////

//======================================================================
// Include Files
extern "C" {
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
} ;

#include "gdata.h"

//======================================================================
// check CharSets

inline Bool isSignChar(char c) { return (c == PLUSCHAR || c == MINUSCHAR) ; }
inline Bool isZeroChar(char c) { return (c == ZEROCHAR) ; }
inline Bool isDotChar(char c) { return (c == DOTCHAR) ; }
inline Bool isHexPreChar(char c) { return (c == 'x' || c == 'X') ; }
inline Bool isOctIntChar(char c) { return ('0' <= c && c <= '7');} 
inline Bool isExpSepChar(char c) { return (c == 'e' || c == 'E') ; }

inline Bool isSymHeadChar(char c) { return (islower(c) || c == XTRSYMCHAR);}
inline Bool isVarHeadChar(char c) { return (isupper(c) || c == VARPRECHAR);}
inline Bool isSymBodyChar(char c) { 
   return (isalnum(c) || c == XTRSYMCHAR || c == VARPRECHAR) ;
} 

inline Bool isArgBeginChar(char c) { return (c == ARGBEGIN) ; }
inline Bool isArgEndChar(char c) { return (c == ARGEND) ; }
inline Bool isArgSepChar(char c) { return (c == ARGSEP) ; }

inline Bool isGArgSepChar(char c) { return (c == GARGSEP) ; }
inline Bool isGArgPairChar(char c) { return (c == GARGPAIR) ; }

inline Bool isListArgSepChar(char c) { return (c == LISTARGSEP) ; }
inline Bool isListBeginChar(char c) { return (c == LISTBEGINCHAR) ; }
inline Bool isListEndChar(char c) { return (c == LISTENDCHAR) ; }
inline Bool isListDotChar(char c) { return (c == LISTDOTCHAR) ; }

//======================================================================
// OTHER CONSTANTS

#define BUFSIZE 1024 
#define EXTENDMAGNIFY 2

//======================================================================
// class Local Buffer
//----------------------------------------------------------------------
class LocalBuffer {
   public:
      char *buffer ;
      Int len ;
      Int max ;
      Bool extendp ;
      Bool originalp ;

      inline Bool assignBuffer(Int maxlen) {
	 buffer = new char[maxlen] ;
	 originalp = True ;
	 max = maxlen ;
	 return True ;
      } ;

      inline LocalBuffer(Int maxlen = 0,
			 Bool exp = False) {
			    originalp = False ;
			    len = 0 ;
			    max = maxlen ;
			    buffer = NULLPTR ;
			    if(max > 0) assignBuffer(max) ;
			    extendp = exp ;
      } ;

      inline LocalBuffer(char* buf,
			 Int maxlen,
			 Bool exp = False,
			 Bool copyp = False) {
			    originalp = False ;
			    len = 0 ;
			    max = maxlen ;
			    if(copyp) {
			       assignBuffer(max) ;
			       strncpy(buffer,buf,max) ;
			    } else {
			       buffer = buf ;
			    }
			    extendp = exp ;
      } ;

      inline LocalBuffer(LocalBuffer& orgbuf,
			 Int offset = 0,
			 Int l = 0,
			 Bool copyp = False) {
			    originalp = False ;
			    len = l ;
			    max = orgbuf.max - offset ;
			    if(copyp) {
			       assignBuffer(max) ;
			       strncpy(buffer,&(orgbuf.buffer[offset]),max) ;
			    } else {
			       buffer = &(orgbuf.buffer[offset]) ;
			    }
			    extendp = orgbuf.extendp ;
      } ;

      inline ~LocalBuffer() {
	 if(originalp) delete buffer ;
      } ;
      inline Bool deleteBuffer() {
	 if(originalp) delete buffer ;
	 return originalp ;
      } ;

      inline Bool extendBuffer() {
	 char* oldbuf = buffer ;
	 Bool origp = originalp ;
	 Bool r = assignBuffer(max * EXTENDMAGNIFY) ;
	 strncpy(buffer,oldbuf,len) ;
	 if(origp) deleteBuffer() ;

	 return r ;
      } ;

      Bool putChar(char c) ;
      inline Bool finish() { return putChar(EOSCHAR) ; } ;

      char* duplicateBuffer() const ;
} ;

//======================================================================
// class ReadStatus
//----------------------------------------------------------------------
enum ReadErrorID {
   REI_None,
   REI_NilData,
   REI_IllegalEOF,
   REI_IllegalNumberFormat,
   REI_IllegalSymbolFormat,
   REI_IllegalStringFormat,
   REI_IllegalTermFormat,
   REI_IllegalSentenceFormat,
   REI_IllegalListFormat,
   REI_Max
} ;

class ReadStatus {
   public:
      ReadErrorID errorID ;
      Int charCount ;
      Int lineCount ;
      Int anonymousVarCount ;
  
      inline ReadStatus() {
	 errorID = REI_None ;
	 charCount = 0  ;
	 lineCount = 0 ;
	 anonymousVarCount = 0 ;
      } ;

      inline void incChar(char c) {
	 charCount++ ;
	 if(c == CRCHAR) lineCount++ ;
      } ;

      inline void initNewSentence() {
	 errorID = REI_None ;
	 anonymousVarCount = 0 ;
      } ;
      inline GaeaVariable newAnonymousVar() {
	 ostrstream buf ;
	 buf << ANONYMOUSVARSTR << anonymousVarCount << ends ;
	 anonymousVarCount++ ;
	 return GaeaVariable(buf.str()) ;
      } ;
} ;

//======================================================================
// Utility
//
extern Int skipSpace(gistream& istr, ReadStatus& rstatus) ;
extern Bool skipComment(gistream& istr, ReadStatus& rstatus) ;
extern void skipSpaceComment(gistream& istr, ReadStatus& rstatus) ;
extern Bool readUntilChar(gistream& istr, LocalBuffer& buf, 
			  char endchar, char escapechar,
			  Bool discardlastp, ReadStatus& rstatus) ;
extern Bool skipLine(gistream& istr, ReadStatus& rstatus) ;

inline char getChar(gistream& istr, ReadStatus& rstatus) {
   char c ;
/*     This part is for problem of old gistream in multithread */
#ifdef xxxTHREADEDxxx
/*	gistream bug?? 	
	Following code can not wait input.
	if((c = istr.peek()) != EOF) {
	istr.get(c) ;
	}
	*/
   do {
      c = NULLCHAR ;
      istr.get(c) ;
      if (c == NULLCHAR) { 
	 istr.clear();
	 MSleep(10) ;
      }
   } while (c == NULLCHAR) ;
/*     gistream in the new version is MT safe, so, istr.get is fine. */ 
#else
   istr.get(c) ;
#endif
   rstatus.incChar(c) ;
   return c ;
} ;

//======================================================================
// readGaeaData
//----------------------------------------------------------------------
extern GaeaData readSentence(gistream& istr, ReadStatus& rstatus) ;
inline GaeaData readSentence(gistream& istr) {
   ReadStatus rstatus ;
   return readSentence(istr,rstatus) ;
}
inline GaeaData readSentence(char* str) {
   gistrstream istr(str) ;
   return readSentence(istr) ;
}
extern GaeaData readSentenceSub(gistream& istr, ReadStatus& rstatus, UInt n) ;

extern GaeaData readSentence0(gistream& istr, ReadStatus& rstatus) ;
inline GaeaData readSentence0(gistream& istr) {
   ReadStatus rstatus ;
   return readSentence0(istr,rstatus) ;
}
inline GaeaData readSentence0(char* str) {
   gistrstream istr(str) ;
   return readSentence(istr) ;
}
   

extern GaeaData readGaeaData(gistream& istr, ReadStatus& rstatus) ;
inline GaeaData readGaeaData(gistream& istr) {
   ReadStatus rstatus ;
   return readGaeaData(istr,rstatus) ;
}
inline GaeaData readGaeaData(char* str) {
   gistrstream istr(str) ;
   return readGaeaData(istr) ;
}

extern GaeaData readQuotedGaeaSymbol(gistream& istr, ReadStatus& rstatus) ;
extern GaeaData readGaeaString(gistream& istr, ReadStatus& rstatus) ;
extern GaeaData readList(gistream& istr, ReadStatus& rstatus, 
		      const Int listDepth = 0) ;
extern GaeaData readGNumberGaeaSymbol(gistream& istr, ReadStatus& rstatus) ;
extern GaeaData readGIntFloatRest(gistream& istr, LocalBuffer& buf, 
			       ReadStatus& rstatus);
extern GaeaData readGOctIntRest(gistream& istr, LocalBuffer& buf, 
			     ReadStatus& rstatus);
extern GaeaData readGHexIntRest(gistream& istr, LocalBuffer& buf, 
			     ReadStatus& rstatus);
extern GaeaData readGaeaSymbolRest(gistream& istr, LocalBuffer& buf, 
			     ReadStatus& rstatus);
extern GaeaData readGaeaVariableRest(gistream& istr, LocalBuffer& buf, 
			       ReadStatus& rstatus);
extern GaeaData checkArgs(const GaeaData& head, gistream& istr, ReadStatus& rstatus);
extern GaeaData readArgs(const GaeaData& head, gistream& istr, ReadStatus& rstatus,
		      Int argn) ;

extern Bool readGArgs(TermGArg*& garg, gistream& istr, ReadStatus& rstatus) ;
inline Bool readGArgs(GaeaData& term, gistream& istr, ReadStatus& rstatus) {
   return readGArgs(term.val.term->_garg,istr,rstatus) ;
} ;


//======================================================================
// read for general purpose
//----------------------------------------------------------------------

//----------------------------------------------------------------------
// GaeaBuffer class

class GaeaBuffer {
   public:
      char* data ;
      UInt size ;
      char* readptr ;
      char* writeptr ;

      //--------------------------------------------------
      // constructor

      GaeaBuffer() { 
	 data = NULLPTR ; size = 0 ; 
	 rewind() ;
      } ;
      GaeaBuffer(UInt s) {
	 size = s ;
	 if(size == 0) { data = NULLPTR ; }
	 else { data = new char[size] ; }
	 rewind() ;
      } ;

      //--------------------------------------------------
      // destructor

      ~GaeaBuffer() { if(size>0) delete data ; } ;

      //--------------------------------------------------
      // rewind

      void rewindRead(Int offset = 0) { 
	 readptr = &(data[offset]) ; } ;
      void rewindWrite(Int offset = 0) { 
	 writeptr = &(data[offset]) ; } ;

      void rewind() {
	 rewindRead() ; rewindWrite() ;
      } ;

      //--------------------------------------------------
      // getchar

      char peekChar() { return *readptr ; } ;

      char getChar() {
	 if(!isEmpty())
	    return *(readptr++) ;
	 else
	    return NULLCHAR ;
      } ;
      Bool ungetChar() { readptr-- ; return True ; } ;
      Bool ungetChar(const char c) { 
	 readptr-- ; *readptr = c ; return True ; } ;

      //--------------------------------------------------
      // pushchar

      Bool pushChar(char c) {
	 *(writeptr++) = c ;
	 return True ;
      } ;

      //--------------------------------------------------
      // full/empty check

      Bool isFull() const {
	 if(writeptr >= &(data[size])) return True ;
	 else return False ;
      } ;

      Bool isEmpty() const {
	 if(readptr >= writeptr) return True ;
	 else return False ;
      } ;

      //--------------------------------------------------
      // length

      UInt bodyLength() const { 
	 return (((UInt)writeptr) - ((UInt)readptr)) ;
      } ;

      //--------------------------------------------------
      // describe

      void describe(ostream& ostr, Bool detailp = True, 
		    Bool newlinep = True) const {
	 ostr << "#Buffer[" << size 
	      << ":" << (void*)data 
	      << "/" << (void*)readptr 
	      << "/" << (void*)writeptr 
	      << "(l=" << bodyLength() << ")]" ;
	 if(detailp) {
	    ostr << endl << "\tdata="; 
	    for(UInt i = 0 ; i < size ; i++) ostr << data[i] ;
	    ostr << endl << "\tread-write=" ;
	    for(char* d = readptr ; d < writeptr ; d++) 
	       ostr << *d ;
	 }
	 if(newlinep) ostr << endl ;
      } ;

      friend ostream& operator << (ostream& ostr, const GaeaBuffer& buf) {
	 buf.describe(ostr,False,False) ;
	 return ostr ;
      } ;

} ;

//----------------------------------------------------------------------
// Control Char Set

class GaeaReadControlCharSet {
   public:
      char terminate ;
      char escape ;
      char stringBegin ;
      char stringEnd ;
      char quoteBegin ;
      char quoteEnd ;
      char commentBegin ;
      char commentEnd ;
} ;

#define DEFAULT_READ_CONTROL_CHAR_SET {\
   CLAUSEEND, ESCAPECHAR,\
   STRINGBEGIN, STRINGEND,\
   QUOTESYMBOLBEGIN, QUOTESYMBOLEND,\
   COMMENTSTARTCHAR, COMMENTENDCHAR}
extern GaeaReadControlCharSet DefaultReadControlCharSet ;

//----------------------------------------------------------------------
// Read General Sentense

extern Bool readGeneralSentense(gistream& istr, GaeaBuffer& buf,
				const GaeaReadControlCharSet& ctrSet = 
				DefaultReadControlCharSet) ;

//----------------------------------------------------------------------
// get token

enum TokenType {
   TT_none,
   TT_symbol,
   TT_quoted_symbol,
   TT_variable,
   TT_int_bin,
   TT_int_dig,
   TT_int_oct,
   TT_int_hex,
   TT_float,
   TT_float_e,
   TT_string,
   TT_paren,
   TT_op,	// operator 
   TT_end,
   TT_error
} ;

Bool getToken(GaeaBuffer& buffer, char*& head, char*& tail, TokenType& type) ;

//======================================================================
// output for debug
//----------------------------------------------------------------------
extern void dbp(GaeaData dt) ;
extern void dbd(GaeaData dt) ;

//======================================================================
// Other Util
//----------------------------------------------------------------------
  





////////////////////////////////////////////////////////////////////////
#endif
