// -*- Mode:C++ -*-
//Header:
//File: gc.h
//Author: NODA Itsuki
//Date: 1998/1/22
//EndHeader:

//ModifyHistory:
// 1998/1/22: Create this file
//EndModifyHistory:

//Memo:
//  GCNode: node of link from GCTopNode to GCEntity.
//	GCNode may not be actual entity.
//  GCEntity: actual entity of heap memory.
//  GCTopNode: top node where markup process starts.
//EndMemo:

#ifndef _gc_h_
#define _gc_h_
////////////////////////////////////////////////////////////////////////
extern "C" {
#include <stdlib.h>
} ;

#include "gbtype.h"
#include "gthread.h"
#include "gstream.h"
#include "utility.h"

//======================================================================
// GCNode
class GCNode {
   public:
      Bool markup() ;
} ;

//======================================================================
// GCEntity
class GCEntity : public GCNode {
   public:
      Bool	used ;
      int	instack ;

      inline GCEntity() { used = False ; instack = 0 ; } ;

      inline stackIn() { ++instack ; } ;
      inline stackOut() { --instack ; } ;

      Bool markup() ;
} ;

//======================================================================
// GCTopNode
class GCTopNode : public GCNode {
   public:

} ;

//======================================================================
// GCEntityTable
class GCEntityTable {
   public:
      GaeaMutex mutex ;

      inline GCEntityTable() {
      } ;

      inline void Lock() { 
	 mutex.lock() ;
      } ;
      inline void UnLock() { 
	 mutex.unlock() ;
      } ;

      Bool sweep(Bool flag) ;
} ;

//======================================================================
// GCTopNodeTable
class GCTopNodeTable : public GCTopNode {
   public:
      Int n ;
      Int max ;
      GCTopNode **node ;
#ifdef THREADED
      pthread_mutex_t mutex ;
#endif

      inline GCTopNodeTable(Int size = 0) {
	 max = size ;
	 n = 0 ;
	 if(size > 0) node = new (GCTopNode *)[size] ;
	 else node = NULLPTR ;
#ifdef THREADED
	 pthread_mutex_init(&mutex,NULLPTR) ;
#endif
      } ;

      inline Bool pushTopNode(GCTopNode* nd) {
#ifdef THREADED
	 pthread_mutex_lock(&mutex) ;
#endif
	 Bool r ;
	 if(n < max) {
	    node[n] = nd ;
	    n++ ;
	    r = True ;
	 } else {
	    r = False ;
	 }
#ifdef THREADED
	 pthread_mutex_unlock(&mutex) ;
#endif
	 return r ;
      } ;
     
      Bool markup() ;
} ;





////////////////////////////////////////////////////////////////////////
#endif
