// -*- Mode:C++ -*-
//Header:
//File: gcell.cc
//Author: NODA Itsuki
//Date: 1998/6/5
//EndHeader:

//ModifyHistory:
// 1998/6/5: Create this file
//EndModifyHistory:

#include "gcell.h"

//======================================================================
// GaeaClause Table

//--------------------------------------------------
// New Ring

GaeaClauseRing* GaeaClauseTable::newRing() {
   GaeaClauseRing* r ;
   mutex.lock() ;
   if(trash->isSingleRing()) {
      r = new GaeaClauseRing() ;
   } else {
      r = trash->next ;
      r->removeSelf() ;
   }
   mutex.unlock() ;
   return r ;
}

//--------------------------------------------------
// asserta

GaeaClauseRing* GaeaClauseTable::asserta(const GaeaClause& gclause) {
   GaeaClauseRing* r = newRing() ;
   r->set(gclause) ;
   mutex.lock() ;
   root->addNext(r) ;
   mutex.unlock() ;
   return r ;
} 

//--------------------------------------------------
// assertz
	    
GaeaClauseRing* GaeaClauseTable::assertz(const GaeaClause& gclause) {
   GaeaClauseRing* r = newRing() ;
   r->set(gclause) ;
   mutex.lock() ;
   root->addPrevious(r) ;
   mutex.unlock() ;
   return r ;
}

//--------------------------------------------------
// retract
	    
GaeaClauseRing* GaeaClauseTable::retract(GaeaClauseRing* gclring) {
   mutex.lock() ;
   gclring->removeSelf() ;
   trash->addPrevious(gclring) ;
   mutex.unlock() ;
   return gclring ;
}

//--------------------------------------------------
// clear
	    
void GaeaClauseTable::clear() {
   mutex.lock() ;
   for(GaeaClauseRing* gclring = root->next ; 
       gclring != root ;
       gclring = gclring->next) {
      gclring->removeSelf() ;
      trash->addPrevious(gclring) ;
   }
   mutex.unlock() ;
}

//--------------------------------------------------
// describe

void GaeaClauseTable::describe(gostream& o, Bool detailp = True) const {
   if(detailp) {
      o << "#GaeaClauseTable[" << this << "]" << endl;
      o << "  table=" ;
      root->describe(o,True) ;
      o << "  # of garbage=" << trash->clength() << endl ;
      o << "----- end GaeaClauseTable" << endl ;
   } else {
      o << "#GaeaClauseTable[" << this 
	<< ":# of defs=" << root->clength() 
	<< ", # of garbage=" << trash->clength() 
	<< "]" ;
   } 
} 

//======================================================================
// CellVar

//--------------------------------------------------
// listing

void GaeaCellVarTable::listing(gostream& ostr) {
   mutex.lock() ;
   for(GaeaCellVarRing* c = next ; c != this ; c = c->next) { 
      ostr << "\t" << *(c->entity.key) << " = " << c->entity.value << endl ;
   }
   mutex.unlock() ;
} ;

//======================================================================
// GaeaCell

//--------------------------------------------------
// copy gcell stack
UInt GaeaCell::copyGaeaCellStack(GaeaCell* src) {
   gcs_mutex.lock() ;
   gcellStack.reset() ;
   UInt n = src->gcellStack.toppos() ;
   for(UInt i = n ; i > 0  ; i--) {
      gcellStack.push(src->gcellStack.topN(i-1)) ;
   }
   gcs_mutex.unlock() ;
   return n ;
}

//--------------------------------------------------
// check gcell is in gcell stack
Bool GaeaCell::isInGaeaCellStack(const GaeaCell* target) const {
   UInt n = gcellStack.toppos() ;
   for(UInt i = 0 ; i < n ; i++) {
      if(gcellStack.topN(i) == target) 
	 return True ;
   }
   return False ;
}

Bool GaeaCell::isInGaeaCellStackR(const GaeaCell* target) const {
   if(target == this) return True ;

   UInt n = gcellStack.toppos() ;
   for(UInt i = 0 ; i < n ; i++) {
      if(gcellStack.topN(i)->isInGaeaCellStackR(target))
	 return True ;
   }
   return False ;
}

//--------------------------------------------------
// copy gcell stack

void GaeaCell::abolishR() {
   abolish() ;
   
   UInt n = gcellStack.toppos() ;
   for(UInt i = 0 ; i < n ; i++) {
      gcellStack.topN(i)->abolishR() ;
   }
}

//--------------------------------------------------
// find GaeaCellVar recursive

GaeaDataBox* GaeaCell::findGaeaCellVarR(Symbol* key) {
   GaeaDataBox* gdb = findGaeaCellVar(key) ;
   if(!isNullPtr(gdb)) return gdb ;

   UInt n = gcellStack.toppos() ;
   for(UInt i = 0 ; i < n ; i++) {
      gdb = gcellStack.topN(i)->findGaeaCellVarR(key) ;
      if(!isNullPtr(gdb)) return gdb ;
   }
   return NULLPTR ;
}


//--------------------------------------------------
// get/put GaeaCellVar recursive

Bool GaeaCell::getGaeaCellVarValueR(GaeaData& value, Symbol* key, 
				    Bool retractp, Bool waitp) {
   GaeaDataBox* gdb = findGaeaCellVar(key) ;
   if(!isNullPtr(gdb)) {
      if(gdb->get(value,retractp,waitp)) {
	 return True ;
      }
   }

   UInt n = gcellStack.toppos() ;
   for(UInt i = 0 ; i < n ; i++) {
      if(gcellStack.topN(i)->getGaeaCellVarValueR(value,key,retractp,waitp))
	 return True ;
   }
   return False ;
}

Bool GaeaCell::putGaeaCellVarValueR(const GaeaData& value, Symbol* key, 
				    GaeaDataBox::PutType puttype) {
   GaeaDataBox* gdb = findGaeaCellVar(key) ;
   if(!isNullPtr(gdb)) {
      if(gdb->put(value,puttype)) {
	 return True ;
      }
   }

   UInt n = gcellStack.toppos() ;
   for(UInt i = 0 ; i < n ; i++) {
      if(gcellStack.topN(i)->putGaeaCellVarValueR(value,key,puttype)) 
	 return True ;
   }
   return False ;
}

//--------------------------------------------------
// describe
void GaeaCell::describe(gostream& o, Bool detailp = True) const {
   o << "#GaeaCell[" ;

   if(isNullPtr(name)) o << this ;
   else o << *name ;

   o << "]" ;

   if(detailp) {
      o << endl ;
      o << "    Definitions:" ;
      gclauseTable.describe(o,True) ;
      o << "    CellStack:" << gcellStack << endl ;
      for(Int i = 0 ; i < gcellStack.toppos() ; i++) {
	 o << "\t" << *(gcellStack.topN(i)) << endl ;
      }
   } 
}

//--------------------------------------------------
// show GaeaCell Stack

void GaeaCell::showGaeaCellStack(gostream& o, Bool newlinep) const {
   if(isNullPtr(name)) o << this ;
   else o << *name ;
   if(!gcellStack.bottomp()) {
      o << "(" ;
      Bool firstp = True ;
      for(UInt i = 0 ; i < gcellStack.toppos() ; i++) {
	 if(firstp) firstp = False ; 
	 else o << "," ;

	 gcellStack.topN(i)->showGaeaCellStack(o,False) ;
      }
      o << ")" ;
   }
   if(newlinep) o << endl ;
}

//--------------------------------------------------
// listing

void GaeaCell::listing(gostream& o, Bool recursivep, Bool showHiddenp,
		    Bool detailp) const 
{
   if((!showHiddenp) && hiddenInListingp) return ;

   o << "In " << *this << "." << endl ;

//   o << "  " ; 
//   gclauseTable.root->describe(o,True) ;

   GaeaClauseRing* tail = gclauseTable.root ;
   for(GaeaClauseRing* r = gclauseTable.root->next ;
	r != tail ;
	r = r->next) {
      if(detailp) {
	 o << "  " << r << "\t" << r->entity << endl ;
      } else {
	 o << "  " << r->entity << endl ; ;
      }
    }

   if(recursivep) {
      for(UInt i = 0 ; i < gcellStack.toppos() ; i++) {
	 gcellStack.topN(i)->listing(o,recursivep,showHiddenp,detailp) ;
      }
   }
} 
      

//======================================================================
// GaeaCellPointer
   
//--------------------------------------------------
// Output

void GaeaCellPointer::describe(gostream& o, Bool detailp) const {
   o << "#GaeaCellPointer[" << gcellstack << "," << n << "]" ;
   if(detailp) {
      o << endl ;
      o << "  GaeaCellStack:" << endl ;
      gcellstack->describe(o, True) ;
   } ;
}

//======================================================================
// GaeaClausePointer

//--------------------------------------------------
// push/pop

void GaeaClausePointer::pushgcp(GaeaCell* tcell) {
   if(0 < tcell->gcellStack.toppos()) {
//      gcpstack.push(GaeaCellPointer(&(tcell->gcellStack),0)) ; // XXX
      gcpstack.push(GaeaCellPointer(&(tcell->gcellStack))) ;      
   } 
} 

GaeaCell* GaeaClausePointer::popgcell() {
  popgcell_top:
   
   if(gcpstack.toppos() <= 0) return NULLPTR ;

   GaeaCellPointer gcp = gcpstack.pop() ;
//   if(gcp.gcellstack->toppos() <= gcp.n) // XXX
   if(gcp.n <= 0) goto popgcell_top ;
   if(gcp.gcellstack->toppos() < gcp.n) gcp.n = gcp.gcellstack->toppos() ;
	 
   gcp.n-- ;
   GaeaCell* r = gcp.gcellstack->bottomN(gcp.n) ;
   gcpstack.push(gcp) ;
   
   return r ;
} ;

//--------------------------------------------------
// find GaeaClause in Ring

GaeaClauseRing* GaeaClausePointer::findGaeaClauseInRing(const GaeaTermHeadArity& gtermha) {
   GaeaClauseRing* c ;
   for(c = top->next ; c != topcell->gclauseTable.root ; c = c->next ) {
      if(c->entity.head_arity.eq(gtermha)) {
	 top = c ;
	 return c ;
      }
   }
   return NULLPTR ;
}

//--------------------------------------------------
// find GaeaClause

GaeaClauseRing* GaeaClausePointer::findGaeaClause(const GaeaTermHeadArity& gtermha) {
  findGaeaClause_top:

   GaeaClauseRing* r = findGaeaClauseInRing(gtermha) ;
   if(!isNullPtr(r)) return r ;

   pushgcp(topcell) ;
   GaeaCell* newtopcell = popgcell() ;
   if(isNullPtr(newtopcell)) return NULLPTR ;
	 
   settopcell(newtopcell) ;
   goto findGaeaClause_top ;
}

//--------------------------------------------------
// Output

void GaeaClausePointer::describe(gostream& o, Bool detailp) const {
   o << "#GaeaClausePointer[" << this << ":" ;
   if(isNullPtr(topcell)) o << "null" ;
   else o << *topcell ;
   o << "]" ;
   if(detailp && !isNullPtr(this)) {
      o << endl ; 
      o << "   top=" << *top << endl ;
      o << "   topcell=" << *topcell << endl ;
      o << "   gcellpointerstack:" << endl; 
      for(Int i = 0 ; i < gcpstack.toppos() ; i++) {
	 (gcpstack.topN(i)).describe(o,True) ;
      }
   } ;
}

#ifdef _NEVER_UNDEFINED_      

//======================================================================
// GaeaCellVarPointer

//--------------------------------------------------
// push/pop

void GaeaCellVarPointer::pushgcp(GaeaCell* tcell) {
   if(0 < tcell->gcellStack.toppos()) {
//      gcpstack.push(GaeaCellPointer(&(tcell->gcellStack),0)) ; // XXX
      gcpstack.push(GaeaCellPointer(&(tcell->gcellStack))) ;      
   } 
} 

GaeaCell* GaeaCellVarPointer::popgcell() {
  popgcell_top:
   
   if(gcpstack.toppos() <= 0) return NULLPTR ;

   GaeaCellPointer gcp = gcpstack.pop() ;
//   if(gcp.gcellstack->toppos() <= gcp.n) // XXX
   if(gcp.n <= 0) goto popgcell_top ;
   if(gcp.gcellstack->toppos() < gcp.n) gcp.n = gcp.gcellstack->toppos() ;
	 
   gcp.n-- ;
   GaeaCell* r = gcp.gcellstack->bottomN(gcp.n) ;
   gcpstack.push(gcp) ;
   
   return r ;
} ;

//--------------------------------------------------
// find GaeaCellVar in environment

GaeaDataBox* GaeaCellVarPointer::findGaeaCellVar(Symbol* name, 
						 const Bool firstp) {
  findGaeaCellVar_top:

   if(!firstp) {
      pushgcp(topcell) ;
      GaeaCell* newtopcell = popgcell() ;
      if(isNullPtr(newtopcell)) return NULLPTR ;
      settopcell(newtopcell) ;
   }

   GaeaDataBox* gdb = topcell->findGaeaCellVar(name) ;
   if(!isNullPtr(gdb)) return gdb ;

   goto findGaeaCellVar_top ;
}

//--------------------------------------------------
// Output

void GaeaCellVarPointer::describe(gostream& o, Bool detailp) const {
   o << "#GaeaCellVar[" << this << ":" ;
   if(isNullPtr(topcell)) o << "null" ;
   else o << *topcell ;
   o << "]" ;
   if(detailp && !isNullPtr(this)) {
      o << endl ; 
      o << "   topcell=" << *topcell << endl ;
      o << "   gcellpointerstack:" << endl; 
      for(Int i = 0 ; i < gcpstack.toppos() ; i++) {
	 (gcpstack.topN(i)).describe(o,True) ;
      }
   } ;
}

#endif

