// -*- Mode:C++ -*-
//Header:
//File: gcell.h
//Author: NODA Itsuki
//Date: 1998/2/8
//EndHeader:

//ModifyHistory:
// 1998/2/8: Create this file
//EndModifyHistory:

#ifndef _gcell_h_
#define _gcell_h_
////////////////////////////////////////////////////////////////////////
#ifdef NULL
#undef NULL
#endif
#define NULL 0

extern "C" { 
#include <string.h>
}

#include "gthread.h"
#include "mmtoolT.h"
#include "ringT.h"
#include "gdata.h"

//======================================================================
// GaeaTermHeadArity
class GaeaTermHeadArity {
   public:
      Symbol *head ;
      Int arity ;

      //--------------------------------------------------
      // Creater

      inline GaeaTermHeadArity(Symbol *hd = NULLPTR, const Int a = 0) {
	 head = hd ;
	 arity = a ;
      } ;
      inline GaeaTermHeadArity(const Term& term) {
	 head = term.head ;
	 arity = term.arity ;
      } ;
      inline GaeaTermHeadArity(const Term* term) {
	 head = term->head ;
	 arity = term->arity ;
      } ;
      inline GaeaTermHeadArity(const GaeaData& gd) {
	 if(isGaeaTerm(gd)) {
	    head = gd.val.term->head ;
	    arity = gd.val.term->arity ;
	 } else if(isGaeaSymbol(gd)) {
	    head = gd.val.symbol ;
	    arity = 0 ;
	 } else {
	    head = NULLPTR ;
	    arity = 0 ;
	 }
      } ;
      inline GaeaTermHeadArity(const GaeaTermHeadArity& src) {
	 head = src.head ;
	 arity = src.arity ;
      } ; 

      //--------------------------------------------------
      // equal and let
      inline Bool eq(const GaeaTermHeadArity& target) const {
	 return (head == target.head)&&(arity == target.arity) ;
      } ;

      inline GaeaTermHeadArity operator=(const GaeaTermHeadArity& src) {
	 head = src.head ; arity = src.arity ;
	 return *this ;
      } ;

      //--------------------------------------------------
      // output
      inline friend gostream& operator<< (gostream& o, const GaeaTermHeadArity& ha){
	 o << ha.head << "/" << ha.arity ;
	 return o ;
      } ;
} ;

//======================================================================
// GaeaClause

inline Bool isFactTerm(const GaeaData &def) {
   return (isGaeaTerm(def) 
	   && 
	   (_TermHead(def) == _SymVal(FactOpSymbol))
	   &&
	   (_TermArity(def) == 1)) ;
} ;
inline Bool isRuleTerm(const GaeaData &def) {
   return (isGaeaTerm(def) 
	   && 
	   (_TermHead(def) == _SymVal(RuleOpSymbol))
	   &&
	   (_TermArity(def) == 2)) ;
} ;

class GaeaClause {
   public:
      GaeaTermHeadArity head_arity ;
      GaeaData head ;
      GaeaData body ;
     
      //--------------------------------------------------
      // Creater

      inline GaeaClause() {} ;

      inline GaeaClause(const GaeaData &def) {
	 if (isFactTerm(def)) {
	    init(_TermArgv(def,0),TrueSymbol) ;
	 } else if (isRuleTerm(def)) {
	    init(_TermArgv(def,0),_TermArgv(def,1)) ;
	 } else {
	    gerr << "Error: Illegal rule/fact: " << def << endl ;
	 } 
      } ;

      inline GaeaClause(const GaeaData &h, const GaeaData &b) {
	 init(h,b) ;
      } ;

      inline Bool init(const GaeaData &h, const GaeaData &b) {
	 head_arity = GaeaTermHeadArity(h) ;
	 head = h ;
	 body = b ;
      } ;
	 

      //--------------------------------------------------
      // Printer

      inline friend gostream& operator<< (gostream& o, const GaeaClause& gc) {
	 o << gc.head << " :- " << gc.body << "." ;
	 return o ;
      } ;

} ;

//======================================================================
// GaeaClause Table

typedef Ring<GaeaClause> GaeaClauseRing ;
const GaeaClause NullGaeaClause ; 

class GaeaClauseTable {
   public:
      GaeaClauseRing* root ;
      GaeaClauseRing* trash ;
      GaeaMutex mutex ;
      
      //--------------------------------------------------
      // Creater

      inline GaeaClauseTable() {
	 root = new GaeaClauseRing(NullGaeaClause,True) ;
	 trash = new GaeaClauseRing(NullGaeaClause,True) ;
      } ;

      //--------------------------------------------------
      // New Ring

      GaeaClauseRing* newRing() ;
	 
      //--------------------------------------------------
      // asserta

      GaeaClauseRing* asserta(const GaeaClause& gclause) ;
	    
      //--------------------------------------------------
      // assertz
	    
      GaeaClauseRing* assertz(const GaeaClause& gclause) ;

      //--------------------------------------------------
      // retract
	    
      GaeaClauseRing* retract(GaeaClauseRing* gclring) ;

      //--------------------------------------------------
      // retract
	    
      void clear() ;

      //--------------------------------------------------
      // describe

      void describe(gostream& o, Bool detailp = True) const ;

      //--------------------------------------------------
      // Output
      inline friend gostream& operator<< (gostream& o, 
					 const GaeaClauseTable& gcltable) {
	 gcltable.describe(o,False) ;
	 return o ;
      } ;
} ;

//======================================================================
// GaeaCellVar Table

class GaeaCellVar {
   public:
      Symbol* key ;
      GaeaDataBox value ;

      GaeaCellVar() { key = NULLPTR ; } ;
      GaeaCellVar(Symbol* k) { key = k ; } ;

      inline GaeaCellVar operator=(const GaeaCellVar& src) {
	 key = src.key ;
	 value = src.value ;
	 return *this ;
      } ;
} ;

typedef Ring<GaeaCellVar> GaeaCellVarRing ;

class GaeaCellVarTable : public GaeaCellVarRing {
   public:

      GaeaMutex mutex ;

      //--------------------------------------------------
      // constructor

      GaeaCellVarTable() {
	 clear();
      } ;

      //--------------------------------------------------
      // clear

      inline void clear() {
	 entity.key = _NullSymbol ;
	 previous = this ;
	 next = this ; 
      } ;

      //--------------------------------------------------
      // new/find

      inline GaeaCellVarRing* _newvar(Symbol* key) {
	 mutex.lock() ;
	 GaeaCellVarRing* r = insertLast(GaeaCellVar(key)) ;
	 mutex.unlock() ;
	 return r ;
      } ;

      inline GaeaCellVarRing* _find(Symbol* key) {
	 mutex.lock() ;
	 GaeaCellVarRing* r = NULLPTR ;
	 for(GaeaCellVarRing* c = next ; c != this ; c = c->next) {
	    if(c->entity.key == key) {
	       r = c ;
	       break ;
	    }
	 }
	 mutex.unlock() ;
	 return r ;
      } ;

      inline GaeaCellVarRing* newvar(Symbol* key) {
	 GaeaCellVarRing* e = _find(key) ;
	 if(isNullPtr(e)) 
	    e = _newvar(key) ;
	 return e ;
      } ;

      inline GaeaDataBox* find(Symbol* key, Bool forcep = False) {
	 GaeaCellVarRing* e = _find(key) ;
	 if(isNullPtr(e)) {
	    if(!forcep)
	       return NULLPTR ;
	    else
	       e = _newvar(key) ;
	 }
	 return &(e->entity.value) ;
      } ;

      //--------------------------------------------------
      // listing

      void listing(gostream& ostr) ;
	 
} ;

//======================================================================
// GaeaCell
class GaeaCell ;
typedef AStack<GaeaCell*> GaeaCellStack ;

#define GCELLSTACKSIZE	4

class GaeaCell {
   public:
      Symbol* name ;
      GaeaData gsymbol ;
      GaeaClauseTable gclauseTable ;
      GaeaCellVarTable gcellVarTable ;
      GaeaCellStack gcellStack ;
      GaeaMutex gcs_mutex ;
      Bool hiddenInListingp ;

      //--------------------------------------------------
      // Creater

      inline GaeaCell() {
	 init(_SymVal(NullSymbol)) ;
      } ;
      inline GaeaCell(Symbol* nm) {
	 init(nm) ;
      } ;
      inline GaeaCell(const GaeaData& gd) {
	 init(gd) ;
      } ;

      inline void init(const GaeaData gd) {
	 if(isGaeaSymbol(gd)) {
	    init(_SymVal(gd)) ;
	 } else {
	    init(_SymVal(NullSymbol)) ;
	 }
      } ;
      inline void init(Symbol* nm) {
	 if(!(isNullPtr(nm))) {
	    name = nm ;
	    gsymbol = GaeaSymbol(nm) ;
	    nm->gcell = this ;
	    gcellStack.init(nm->name,GCELLSTACKSIZE) ;
	    hiddenInListingp = False ;
	 }
      } ;

      //--------------------------------------------------
      // assert/retract

      inline GaeaClauseRing* asserta(const GaeaClause& gclause) {
	 return gclauseTable.asserta(gclause) ;
      } ;
      inline GaeaClauseRing* assertz(const GaeaClause& gclause) {
	 return gclauseTable.assertz(gclause) ;
      } ;

      inline GaeaClauseRing* retract(GaeaClauseRing* gclring) {
	 return gclauseTable.retract(gclring) ;
      } ;

      inline void abolish() { gclauseTable.clear() ; } ;
      void abolishR() ;

      //--------------------------------------------------
      // GaeaCellVar

      inline GaeaDataBox* findGaeaCellVar(Symbol* key, 
					  Bool forcep = False) {
	 return gcellVarTable.find(key,forcep) ;
      } ;
      
      GaeaDataBox* findGaeaCellVarR(Symbol* key) ;

      Bool getGaeaCellVarValueR(GaeaData& value, Symbol* key, 
				Bool retractp, Bool waitp) ;
      Bool putGaeaCellVarValueR(const GaeaData& value, Symbol* key, 
				GaeaDataBox::PutType puttype) ;
      
      inline Bool newGaeaCellVar(Symbol* key) {
	 GaeaCellVarRing* r = gcellVarTable._find(key) ;
	 if(!isNullPtr(r)) return False ;
	 gcellVarTable._newvar(key) ;
	 return True ;
      } ;

      //--------------------------------------------------
      // push/pop gcell

      inline GaeaCellStack* pushGaeaCell(GaeaCell* gcell) {
	 gcs_mutex.lock() ;
	 GaeaCellStack* r = gcellStack.push(gcell) ;
	 gcs_mutex.unlock() ;
	 return r ;
      } ;
      
      inline GaeaCell* popGaeaCell() {
	 gcs_mutex.lock() ;
	 GaeaCell* r = gcellStack.pop() ;
	 gcs_mutex.unlock() ;
	 return r ;
      } ;

      inline Int removeGaeaCell(GaeaCell* gcell) {
	 gcs_mutex.lock() ;
	 Int r = 0 ;
	 for(UInt i = 0 ; i < gcellStack.toppos() ; i++) {
	    if(gcellStack.topN(i) == gcell) {
	       gcellStack.removeN(i) ;
	       i-- ;
	       r++ ;
	    } 
	 }
	 gcs_mutex.unlock() ;
	 return r ;
      } ;

      inline Bool remove1GaeaCell(GaeaCell* gcell) {
	 gcs_mutex.lock() ;
	 Bool r = False ;
	 for(UInt i = 0 ; i < gcellStack.toppos() ; i++) {
	    if(gcellStack.topN(i) == gcell) {
	       gcellStack.removeN(i) ;
	       r = True ;
	       break ;
	    } 
	 }
	 gcs_mutex.unlock() ;
	 return r ;
      } ;

      //--------------------------------------------------
      // clear gcell

      inline void clear() {
	 gcs_mutex.lock();
	 gcellStack.reset() ;
	 gcs_mutex.unlock();
	 gclauseTable.clear() ;
	 gcellVarTable.clear() ;
      } ;

      //--------------------------------------------------
      // copy cell stack
      UInt copyGaeaCellStack(GaeaCell* src) ;

      //--------------------------------------------------
      // check gcell is in gcell stack
      Bool isInGaeaCellStack(const GaeaCell* target) const ;

      //--------------------------------------------------
      // check gcell is in gcell stack (recursive)
      Bool isInGaeaCellStackR(const GaeaCell* target) const ;

      //--------------------------------------------------
      // describe

      void describe(gostream& o, Bool detailp = True) const ;

      void showGaeaCellStack(gostream& o, Bool newlinep = False) const ;

      void listing(gostream& o, Bool recursive = False, 
		   Bool showHiddenp = False,
		   Bool detailp = False) const ;

      //--------------------------------------------------
      // Output
      inline friend gostream& operator<< (gostream& o, const GaeaCell& gcell) {
	 gcell.describe(o,False) ;
	 return o ;
      } ;

} ;

//======================================================================
// GaeaCellPointer

class GaeaCellPointer {
   public:
      GaeaCellStack* gcellstack ;
      Int n ;

      //--------------------------------------------------
      // Creator

      inline GaeaCellPointer() {gcellstack = NULLPTR ; n = 0 ; } ;
      inline GaeaCellPointer(GaeaCellStack* gcstk) {
	 gcellstack = gcstk ; n = gcellstack->toppos() ;
      } ;
      inline GaeaCellPointer(GaeaCellStack* gcstk, Int i) {
	 gcellstack = gcstk ; n = i ;
      } ;
      inline GaeaCellPointer(const GaeaCellPointer& gcp) {
	 gcellstack = gcp.gcellstack ; n = gcp.n ;
      } ;

      //--------------------------------------------------
      // Output

      void describe(gostream& o, Bool detailp = True) const ;

      inline friend gostream& operator<< (gostream& o, 
					 const GaeaCellPointer& gcp) {
	 gcp.describe(o,False) ;
	 return o ;
      } ;

} ;

//======================================================================
// GaeaClausePointer

typedef AStack<GaeaCellPointer> GaeaCellPointerStack ;

class GaeaClausePointer {
   public:
      GaeaClauseRing* top ;
      GaeaCell* topcell ;
      GaeaCellPointerStack gcpstack ;

      //--------------------------------------------------
      // Creator

      inline GaeaClausePointer() { init0() ; } ;
      inline GaeaClausePointer(GaeaCell* tcell) { init0() ; init(tcell) ; } ;

      inline void init0() {
	 top = NULLPTR ;
	 topcell = NULLPTR ;
	 gcpstack.init("GaeaClausePointer",GCELLSTACKSIZE) ;
      } ;

      inline void init() {
	 top = NULLPTR ;
	 topcell = NULLPTR ;
      } ;

      inline void init(GaeaCell* tcell) {
	 settopcell(tcell) ;
	 gcpstack.reset() ;
      } ;

      //--------------------------------------------------
      // copy

      inline void copy(const GaeaClausePointer& src) {
	 top = src.top ;
	 topcell = src.topcell ;
	 gcpstack.copy(src.gcpstack) ;
      } ;

      //--------------------------------------------------
      // set top cell

      inline void settopcell(GaeaCell* tcell) {
	 topcell = tcell ;
	 top = tcell->gclauseTable.root ;
      } ;

      //--------------------------------------------------
      // push/pop

      void pushgcp(GaeaCell* tcell) ;

      GaeaCell* popgcell() ;

      //--------------------------------------------------
      // find GaeaClause in Ring

      GaeaClauseRing* findGaeaClauseInRing(const GaeaTermHeadArity& gtermha) ;

      //--------------------------------------------------
      // find GaeaClause

      GaeaClauseRing* findGaeaClause(const GaeaTermHeadArity& gtermha) ;

      //--------------------------------------------------
      // Output

      void describe(gostream& o, Bool detailp = True) const ;

      inline friend gostream& operator<< (gostream& o, 
					  const GaeaClausePointer& gcp) {
	 gcp.describe(o,False) ;
	 return o ;
      } ;

} ;

#ifdef _NEVER_UNDEFINED_      
// This may be useful when enabling redo of unification of @-infon
//======================================================================
// GaeaCellVarPointer

class GaeaCellVarPointer {
   public:
      GaeaCell* topcell ;
      GaeaCellPointerStack gcpstack ;

      //--------------------------------------------------
      // Creator

      inline GaeaCellVarPointer() { init0() ; } ;
      inline GaeaCellVarPointer(GaeaCell* tcell) { init0() ; init(tcell) ; } ;

      inline void init0() {
	 topcell = NULLPTR ;
	 gcpstack.init("GaeaCellVarPointer",GCELLSTACKSIZE) ;
      } ;

      inline void init() {
	 topcell = NULLPTR ;
      } ;

      inline void init(GaeaCell* tcell) {
	 settopcell(tcell) ;
	 gcpstack.reset() ;
      } ;

      //--------------------------------------------------
      // copy

      inline void copy(const GaeaCellVarPointer& src) {
	 topcell = src.topcell ;
	 gcpstack.copy(src.gcpstack) ;
      } ;

      //--------------------------------------------------
      // set top cell

      inline void settopcell(GaeaCell* tcell) {
	 topcell = tcell ;
      } ;

      //--------------------------------------------------
      // push/pop

      void pushgcp(GaeaCell* tcell) ;

      GaeaCell* popgcell() ;

      //--------------------------------------------------
      // find GaeaCellVar in environment

      GaeaDataBox* findGaeaCellVar(Symbol* name,
				   const Bool firstp) ;

      //--------------------------------------------------
      // Output

      void describe(gostream& o, Bool detailp = True) const ;

      inline friend gostream& operator<< (gostream& o, 
					  const GaeaCellVarPointer& gcp) {
	 gcp.describe(o,False) ;
	 return o ;
      } ;

} ;
      
#endif

////////////////////////////////////////////////////////////////////////
#endif

