// -*- Mode:C++ -*-
//Header:
//File: gdata.cc
//Author: NODA Itsuki
//Date: 1997/9/12
//EndHeader:

//ModifyHistory:
// 1997/9/12: Create this file
//EndModifyHistory:

//======================================================================
// Include Files
//
#include "gdata.h"

//======================================================================
// Symbol Table & Term Table

char * GaeaDTypeName[] = GaeaDTypeNameString ;

ExtObjTypeTable EObjTable ;

SymbolTable SymTab ;

Symbol * SymbolTable::newSymbol(const String &nm) {
   if(n < size) {
      Lock() ;
      Symbol *s = entry(n) ; n++ ;
      s->name = nm ;
      s->id = idmax ; idmax++ ;  
      UnLock() ;
      return s ;
   } else if (expand()) {
      return newSymbol(nm) ;
   } else {
      gerr << "Warning: no more symbol (in generating '" << nm << "'.)" 
	   << endl ;
   } ;
   return NULL ;
} ;

//--------------------------------------------------
// gensym

Symbol * SymbolTable::genSymbol(const char* seed) {
   UInt len = strlen(seed) + 64 ;
   char* name = new char[len] ;
   for(UInt i = 0 ; True ; i++) {
      ostrstream ostr(name,len) ;
      ostr << seed << VARCONCATCHAR << i << ends ;
      Symbol* symp = _findSymbol(name) ;
      if(symp == NULLPTR) break ;
   }
   char* newname = strdup(name) ;
   delete name ;
   return findSymbol(newname) ;
} ;

//--------------------------------------------------
// Expand Table

//------------------------------
// Symbol Table

Bool SymbolTable::expand(UInt addsize) {
   Symbol ** oldentry = _entry ;
   UInt oldsize = size ;
   size = size + addsize ;
  
#ifdef DEBUG0
   gerr << "Status: expanding Symbol Table:" << oldsize << "->" << size 
	<< endl ;
#endif

   _entry = new (Symbol*)[size] ;
   if(_entry == NULL) return False ;
  
   for(UInt i = 0 ; i < oldsize ; i++) { _entry[i] = oldentry[i] ; } ;
   if(oldsize > 0) delete oldentry ;
  
   Symbol * newentry = new Symbol[addsize] ;
   if(newentry == NULL) return False ; 
    
   for(UInt i = oldsize ; i < size ; i++) {
      _entry[i] = &(newentry[i - oldsize]) ; 
   } ;
   return True ;
} ;

//------------------------------
// Term Argv Table

#ifndef NOT_USE_TERM_ARGV

TermArgvTable TermArgvTab ;

Bool TermArgvTable::expand(UInt addsize) {
   TermArgv ** oldentry = _entry ;
   UInt oldsize = size ;
   size = size + addsize ;
 
   _entry = new (TermArgv *)[size] ;
   if(_entry == NULL) return False ;

   for(UInt i = 0 ; i < oldsize ; i++) { _entry[i] = oldentry[i] ; } ;
   if(oldsize > 0) delete oldentry ;
    
   TermArgv * newentry = new TermArgv[addsize] ;
   if(newentry == NULL) return False ;
    
   for(UInt i = oldsize ; i < size ; i++) {
      _entry[i] = &(newentry[i - oldsize]) ;
   } ;
   return True ;
} ;

#endif

//------------------------------
// Term G Argv Table

TermGArgTable TermGArgTab ;

Bool TermGArgTable::expand(UInt addsize) {
   TermGArg ** oldentry = _entry ;
   UInt oldsize = size ;
   size = size + addsize ;
 
   _entry = new (TermGArg *)[size] ;
   if(_entry == NULL) return False ;

   for(UInt i = 0 ; i < oldsize ; i++) { _entry[i] = oldentry[i] ; } ;
   if(oldsize > 0) delete oldentry ;
    
   TermGArg * newentry = new TermGArg[addsize] ;
   if(newentry == NULL) return False ;
    
   for(UInt i = oldsize ; i < size ; i++) {
      _entry[i] = &(newentry[i - oldsize]) ;
   } ;
   return True ;
} ;


//------------------------------
// Term Table

TermTable TermTab ;

Bool TermTable::expand(UInt addsize) {
   Term ** oldentry = _entry ;
   UInt oldsize = size ;
   size = size + addsize ;
 
#ifdef DEBUG0
   gerr << "Status: expanding Term Table:" << oldsize << "->" << size 
	<< endl ;
#endif

   _entry = new (Term *)[size] ;
   if(_entry == NULL) return False ;

   for(UInt i = 0 ; i < oldsize ; i++) { _entry[i] = oldentry[i] ; } ;
   if(oldsize > 0) delete oldentry ;
    
   Term * newentry = new Term[addsize] ;
   if(newentry == NULL) return False ;
    
   for(UInt i = oldsize ; i < size ; i++) {
      _entry[i] = &(newentry[i - oldsize]) ;
   } ;
   return True ;
} ;

//--------------------------------------------------
// describe Table

void SymbolTable::describe(gostream& o, Bool detailp) const {
   o << "#SymbolTable[" << n << "/" << size << "(" << idmax << ")]" ;
   if(detailp) {
      o << endl ;
      for(UInt i = 0 ; i < n ; i++) {
	 o << "\t" << *entry(i) << "\t" ;
	 if(entry(i)->funcp) o << " func" ;
	 if(!isNullPtr(entry(i)->gcell)) o << " gcell" ;
	 if(!isNullPtr(entry(i)->gem)) o << " gem" ;
	 if(!isNullPtr(entry(i)->gdbox)) o << " gdbox" ;
	 if(!isNullPtr(entry(i)->dlhandle)) o << " dlhandle" ;
	 o << endl ;
      }
   }
}

void showSymTab(gostream& ostr) { SymTab.describe(ostr,True) ; } ; 

//======================================================================
// Describe Data

void Symbol::_describe(gostream& o, const Bool newlinep, const Bool detailp,
		       const Bool quotep) 
const 
{
   if(detailp) o << "#Sym[" << id << ":" ;

   if(this == _NullSymbol && !detailp) {
      o << LISTBEGINCHAR << LISTENDCHAR ;
   } else {
      if(quotep)  o << "'" << name << "'" ;
      else        o << name ;
   }

   if(detailp) {
      if(funcp) o << ", function=" << (void*)func ;
      if(!isNullPtr(gcell)) o << ", gcell=" << *gcell ;
      if(detailp) o << "]" ;
   }

   if(newlinep) o << endl ;
} ;
     
void GaeaData::_describe(gostream& o, const Bool newlinep, const Bool detailp,
		      const Int listformp) 
const 
{
   switch(type) {
      case GDT_Variable: 
	 ((GaeaVariable*)this)->_describe(o,newlinep,detailp); break;
      case GDT_Integer:  
	 ((GaeaInteger*)this)->_describe(o,newlinep,detailp); break;
      case GDT_Float: 
	 ((GaeaFloat*)this)->_describe(o,newlinep,detailp); break;
      case GDT_Symbol: 
	 ((GaeaSymbol*)this)->_describe(o,newlinep,detailp); break;
      case GDT_String: 
	 ((GaeaString*)this)->_describe(o,newlinep,detailp); break;
      case GDT_Pointer: 
	 ((GaeaPointer*)this)->_describe(o,newlinep,detailp); break;
//      case GDT_TimeVal: 
//	 ((GaeaTimeVal*)this)->_describe(o,newlinep,detailp); break;
      case GDT_Term: 
	 ((GaeaTerm*)this)->_describe(o,newlinep,detailp,listformp); break;
     default:
	 if(isExtObjectType(type)) {
	    ((GaeaExtObject*)this)->_describe(o,newlinep,detailp); break;
	 }
	 if(detailp)
	    o << "#DT[type=" << GaeaDTypeName[type] << "]" ;
	 else
	    o << "#DT[" << this << "]" ;
	 if(newlinep) o << endl ;
   }
} ;

void TermGArg::describe(gostream& o,const Bool newlinep,const Bool detailp) 
   const {
   if(detailp) {
      o << "#TermGArg[" ;
   }
   o << value << GARGPAIR << *role ;
   if(detailp) {
      o << "]" ;
   }
   if(newlinep) o << endl ;
}

void Term::_describe(gostream& o, const Bool newlinep, const Bool detailp,
		     const Int listformp) const 
{
   if(detailp) {
      o << "#Term[" << head->name << "/" << arity << ARGBEGIN ;
   } else if(listformp > 0 && isConsTerm(this)) {
      /* Special for list (== $/2) term */
      return _describe_list(o,newlinep,detailp,listformp) ;
   } else {
      o << head->name << ARGBEGIN ;
   }

   for(Int i = 0 ; i < arity ; i++) {
      if(i>0) o << ARGSEP ;
      (argv(i))._describe(o, False, detailp,listformp) ;
   }

   if(gargp) {
      o << GARGSEP ;
      Bool firstp = True ;
      for(TermGArg* garg = _garg ; !isNullPtr(garg) ; garg = garg->next) {
	 if(!firstp) o << ARGSEP ;
	 o << *garg ;
	 firstp = False ;
      }
   } 

   o << ARGEND ;

   if(detailp)  o << "]" ;

   if(newlinep) o << endl ;
} ;

void Term::_describe_list(gostream& o, const Bool newlinep, const Bool detailp,
			  const Int listformp) const 
{
   GaeaData car = argv(0) ;
   GaeaData cdr = argv(1) ;

   if(listformp == GaeaData::ShowList_On) {
      o << LISTBEGINCHAR ;
   } else {
      o << LISTARGSEP ;
   }
   car._describe(o, False, detailp, GaeaData::ShowList_On) ;

   if(isConsTerm(cdr)) {
      cdr._describe(o,False,detailp,listformp + 1) ;
   } else if(eq(cdr,NullSymbol)) {
      o << LISTENDCHAR ;
   } else {
      o << LISTDOTCHAR ;
      cdr._describe(o,False,detailp, GaeaData::ShowList_On) ;
      o << LISTENDCHAR ;
   }
   if(newlinep) o << endl ;
}

void GaeaInteger::_describe(gostream& o, const Bool newlinep, const Bool detailp) 
const 
{
   if(detailp) 
      o << "#Int[" << val.inum << "]" ;
   else
      o << val.inum ;
   if(newlinep) o << endl ;
} ;

void GaeaFloat::_describe(gostream& o, const Bool newlinep, const Bool detailp) 
const 
{
   if(detailp)
      o << "#Flt[" << val.fnum << "]" ;
   else
      o << val.fnum ;
   if(newlinep) o << endl ;
} ;

void GaeaSymbol::_describe(gostream& o, const Bool newlinep, const Bool detailp) 
const 
{
   (*(val.symbol))._describe(o,newlinep,detailp,shouldquotep()) ;
} ;

void GaeaVariable::_describe(gostream& o, const Bool newlinep, const Bool detailp) 
const 
{
   if(detailp)
      o << "#Var" ;
  
   (*(val.symbol))._describe(o,newlinep,detailp,False) ;
} ;

void GaeaString::_describe(gostream& o, const Bool newlinep, const Bool detailp) 
const 
{
   if(detailp)
      o << "#Str[" << val.string << ":\"" << *val.string << "\"]" ;
   else
      o << "\"" << *val.string << "\"" ;
   if(newlinep) o << endl ;
} ;

void GaeaPointer::_describe(gostream& o, const Bool newlinep, const Bool detailp) 
const 
{
   if(detailp)
      o << "#Ptr[" << val.pointer << "]" ;
   else
      o << "#Ptr[" << val.pointer << "]" ;
   if(newlinep) o << endl ;
} ;

void GaeaExtObject::_describe(gostream& o, const Bool newlinep, const Bool detailp)
const 
{
   if(detailp)
      o << "#ExObj[" << ExtObjTypeName(GaeaDType2ExtObjTypeID(type)) 
	<< ":" << val.pointer << "]" ;
   else
      o << "#ExObj[" << ExtObjTypeName(GaeaDType2ExtObjTypeID(type)) 
	<< ":" << val.pointer << "]" ;
   if(newlinep) o << endl ;
} ;

/*
void GaeaTimeVal::_describe(gostream& o, const Bool newlinep, const Bool detailp) 
const 
{
   o << val.time ;
   if(newlinep) o << endl ;
} ;
*/

void GaeaTerm::_describe(gostream& o, const Bool newlinep, const Bool detailp,
		      const Int listformp) const 
{
   (*(val.term))._describe(o,newlinep,detailp,listformp) ;
} ;

//======================================================================
// Equal Data

#ifndef USE_INLINE_EQ
Bool GaeaData::eq(const GaeaData & gd) const 
{
   if(type == gd.type) {
      if        (type == GDT_Integer) {
	 return val.inum == gd.val.inum ;
      } else if (type == GDT_Float) {
	 return val.fnum == gd.val.fnum ;
//      } else if (type == GDT_TimeVal) {
//	 return val.time == gd.val.time ;
      } else {
	 return val.pointer == gd.val.pointer ;
      }
   } else {
      return False ;
   } 
} 
#endif

//======================================================================
// Generate GaeaSymbol and GaeaVariable 

GaeaData genGaeaSymbol(const GaeaData &seed) {
   Symbol* sym ;
   if(isGaeaSymbol(seed)){
      sym = GenSymbol(seed.val.symbol) ;
   } else {
      sym = GenSymbol(NullSymbol.val.symbol) ;
   }
   return GaeaSymbol(sym) ;
}

GaeaData genGaeaVariable(const GaeaData &seed) {
   Symbol* sym ;
   if(isGaeaVariable(seed)) {
      sym = GenSymbol(seed.val.symbol) ;
   } else {
      sym = GenSymbol(AnonymousVar.val.symbol) ;
   }
   return GaeaVariable(sym) ;
} ;

//======================================================================
// Assign GaeaData Box to Symbol

Bool Symbol::assignGDBox() {
   if(hasGDBox()) return False ;
   gdbox = new GaeaDataBox() ;
   return True ;
}

Bool GaeaDataBox::put(const GaeaData& v, GaeaDataBox::PutType puttype) {
   Bool r = True ;
   mutex.lock() ;
   switch(puttype) {
      case PT_force: 
	 break ;
      case PT_wait: 
	 while(assigned) 
	    mutex.wait() ;
	 break ;
      case PT_nowait: 
	 if(assigned) r = False ;
	 break ;
      default: 
	 r = False ;
	 break ;
   }
   if(r) {
      val = v ;
      assigned = True ;
      mutex.broadcast() ;
   }
   mutex.unlock() ;
   return r ;
}

Bool GaeaDataBox::timedwait_put(const GaeaData& v, UInt sec, UInt msec, UInt usec) {
   Bool contp = True ;
   mutex.lock() ;
   while(assigned && contp) 
      contp = mutex.timedwait(sec,msec,usec) ;
   if(contp) {
      val = v ;
      assigned = True ;
      mutex.broadcast() ;
   }
   mutex.unlock() ;
   return contp ;
}

Bool GaeaDataBox::get(GaeaData& v, Bool retractp, Bool waitp) {
   Bool r = True ;
	 
   mutex.lock() ;
   if(waitp) {
      while(!assigned) 
	 mutex.wait() ;
   } else {
      if(!assigned) r = False ;
   }
   if(r) {
      v = val ;
      if(retractp) {
	 assigned = False ;
	 mutex.broadcast() ;
      }
   }
   mutex.unlock() ;
   return r ;
}

Bool GaeaDataBox::timedwait_get(GaeaData& v, Bool retractp, 
			     UInt sec, UInt msec, UInt usec) 
{
   Bool contp = True ;
   mutex.lock() ;
   while(!assigned && contp)
      contp = mutex.timedwait(sec,msec,usec) ;
   if(contp) {
      v = val ;
      if(retractp) {
	 assigned = False ;
	 mutex.broadcast() ;
      }
   } 
   mutex.unlock() ;
   return contp ;
}





