// -*- Mode:C++ -*-
//Header:
//File: gdata.h
//Author: NODA Itsuki
//Date: 1997/9/9
//EndHeader:

//ModifyHistory:
// 1997/9/9: Create this file
//EndModifyHistory:

#ifndef _gdata_h_
#define _gdata_h_
////////////////////////////////////////////////////////////////////////

//======================================================================
// Include Files
//

extern "C" {
#include <strings.h>
#include <string.h>
#include <stdio.h>
#include <dlfcn.h>
} ;

#include "gbtype.h"
#include "gstream.h"
#include "utility.h"
#include "gc.h"
#include "astableT.h"

//======================================================================
// Some switches
//

#define USE_INLINE_EQ 1

//======================================================================
// Table Default Size
//

#define SymbolTableDefaultSize 1024
#define TermTableDefaultSize 1024

//======================================================================
// Value Type
//
enum GaeaDType {
   GDT_None,
   GDT_Variable,
   GDT_Integer,
   GDT_Float,
   GDT_Symbol,
   GDT_Term,
   GDT_String,
   GDT_Pointer,
   GDT_ExtObject,
//   GDT_TimeVal,
   GDT_Max
} ;

#define GaeaDTypeNameString {"none",\
   "variable",\
   "integer",\
   "float",\
   "symbol",\
   "term",\
   "string",\
   "pointer",\
   "ext_object",\
/*   "timeval", */\
   "unknown" } 

extern char * GaeaDTypeName[] ;

class Symbol ;
class Term ;
class GaeaData ;

union GaeaDValue {
      Int	inum ;
      Float 	fnum ;
      Symbol *	symbol ;
      Term *	term ;
      String *	string ;
      void *	pointer ;
//      TimeVal	time ;
} ;

inline Bool operator==(const GaeaDValue& a,  const GaeaDValue& b) {
   return !strncmp((char *)&a, (char *)&b, sizeof(GaeaDValue)) ;
} ;

//======================================================================
// extended object
//
class ExtObjTypeEntry {
   public:
      char* name ;

      inline ExtObjTypeEntry() { name = NULLPTR ; } ;
      inline ExtObjTypeEntry(char* nm) { name = nm ; } ;

      inline void init() { name = NULLPTR ; } ;

      inline friend gostream& operator<< (gostream& o, const ExtObjTypeEntry& e){
	 o << e.name ; return o ;
      } ;
} ;
      

typedef UInt ExtObjTypeID ;
class ExtObjTypeTable : public AHeap<ExtObjTypeEntry> {
   public:
      GaeaMutex mutex ;

      typedef char* Name_t ;

      inline ExtObjTypeEntry& entry(ExtObjTypeID i) { return *(table[i]) ; } ;
      inline ExtObjTypeID defType(Name_t name) {
	 mutex.lock() ;
	 ExtObjTypeID id = _top ;
	 ExtObjTypeEntry* entry = get() ;
	 entry->name = name ;
	 mutex.unlock() ;
	 return id ;
      } ;
} ;
extern ExtObjTypeTable EObjTable ;

inline ExtObjTypeID DefExtObjType(char* name) {
   return EObjTable.defType(name) ;
} ;
inline char* ExtObjTypeName(ExtObjTypeID id) {
   return EObjTable.entry(id).name ;
} ;
inline bool isExtObjectType(GaeaDType type) { return type >= GDT_ExtObject ; } ;
inline ExtObjTypeID GaeaDType2ExtObjTypeID(GaeaDType type) { 
   return type - (ExtObjTypeID)GDT_ExtObject ;
} ;
inline GaeaDType ExtObjTypeID2GaeaDType(ExtObjTypeID id) {
   return (GaeaDType)(id + (ExtObjTypeID)GDT_ExtObject) ; 
} ;

//======================================================================
// Return Status of Built-In-Predicates
//

enum BipStatus {
   BIP_NoMatch,
   BIP_CutFail,
   BIP_Error,
   BIP_Success,
   BIP_max
} ;


//======================================================================
// Symbol
//
class GaeaCell ;
class Gem ;
class GaeaDataBox ;
class GaeaSignalEntry ;

class Symbol : public GCEntity {
   public:
      String name ;
      ID id ;
      Bool funcp ;
      BipStatus (*func)(Gem*) ;

      GaeaCell* gcell ;
      Gem* gem ;
      GaeaMutex mutex ; 
      GaeaDataBox* gdbox ;
      GaeaSignalEntry* signalWaitList ;
      TimeVal lastSignalTime ;

      void* dlhandle ; // handle for dynamic loading 

      //--------------------------------------------------
      // Creater

      inline Symbol() {
	 funcp = False ; gcell = NULLPTR ; 
	 gem = NULLPTR ; gdbox = NULLPTR ; 
	 signalWaitList = NULLPTR; lastSignalTime.update() ;
	 dlhandle = NULLPTR ;
      } ;
      inline Symbol(const String &nm, const ID i = 0) { 
	 name = nm ; id = i ; funcp = False ; gcell = NULLPTR ;
	 gem = NULLPTR ; gdbox = NULLPTR ; 
	 signalWaitList = NULLPTR; lastSignalTime.update() ;
	 dlhandle = NULLPTR ;
      } ;
      inline Symbol(const char* nm, const ID i = 0) { 
	 name = nm ; id = i ; funcp = False ; gcell = NULLPTR ;
	 gem = NULLPTR ; gdbox = NULLPTR ; 
	 signalWaitList = NULLPTR; lastSignalTime.update() ;
	 dlhandle = NULLPTR ;
      } ;
  
      //--------------------------------------------------
      // Equal

      inline Bool eq(Symbol* that) { return this==that ; } ; 

      //--------------------------------------------------
      // set func
      
      inline void setfunc(BipStatus (*f)(Gem*)) {
	 funcp = True ;
	 func = f ;
      } ;
      inline void clearfunc() {
	 funcp = False ;
	 func = NULLPTR ;
      } ;

      //--------------------------------------------------
      // assign gdata box

      Bool assignGDBox() ;
      
      //--------------------------------------------------
      // check type

      inline Bool hasFunc() const { return funcp ; } ;
      inline Bool hasGaeaCell() const { return !isNullPtr(gcell) ; } ;
      inline Bool hasGem() const { return !isNullPtr(gcell) ; } ;
      inline Bool hasGDBox() const { return !isNullPtr(gdbox) ; } ;
      inline Bool hasDLHandle() const { return !isNullPtr(dlhandle) ; } ;

      //--------------------------------------------------
      // Printer
  
      inline void describe(gostream& o = gout, 
			   const Bool newlinep = True, 
			   const Bool detailp = True,
			   const Bool quotep = False) const {
			      _describe(o,newlinep,detailp,quotep) ;} ;
      void _describe(gostream& o, const Bool newlinep, const Bool detailp,
		     const Bool quotep ) const ;

      inline friend gostream& operator<< (gostream& o, const Symbol& s) {
	 s.describe(o,False,False) ;
	 return o ;
      } ;
} ;

//======================================================================
// Symbol Table
//

class SymbolTable : public GCEntityTable {
   public:
      UInt size ;
      UInt n ;
      UInt idmax ;
      Symbol ** _entry;
  
      //--------------------------------------------------
      // Creater

      inline SymbolTable(const UInt sz = SymbolTableDefaultSize) {
	 size = 0 ;
	 n = 0 ;
	 idmax = 0 ;
	 _entry = NULLPTR ;
	 Lock() ;
	 expand(sz) ;
	 UnLock() ;
      } ;

      //--------------------------------------------------
      // Get Nth Entry

      inline Symbol * entry(UInt i) const { return _entry[i] ; } ;

      //--------------------------------------------------
      // New Symbol

      inline Symbol * newSymbol(const char * nm) { 
	 return newSymbol(String(nm)) ; } ;
      Symbol * newSymbol(const String &nm) ;
      inline Symbol * newSymbol(char* head, char* tail) {
	 char c = *tail ;
	 *tail = NULLCHAR ;
	 char * nm = strdup(head) ;
	 *tail = c ;
	 return  newSymbol(nm) ;
      } ;

      //--------------------------------------------------
      // Find Symbol

      inline Symbol * _findSymbol(const char* sym) const {
	 return _findSymbol(String(sym)) ;
      } ;

      inline Symbol * _findSymbol(const String & sym) const {
	 for(UInt i = 0 ; i < n ; i++) {
	    Symbol *e = entry(i) ;
	    if(e->name == sym) return e ;
	 }
	 return (Symbol *)NULLPTR ;
      } ;

      inline Symbol * findSymbol(const char * nm) { 
	 return findSymbol(String(nm)) ; } ;
      inline Symbol * findSymbol(const String & nm) {
//	 Lock() ; // this Lock is nonsense
	 Symbol * s = _findSymbol(nm) ;
	 if(s == NULLPTR) { s = newSymbol(nm) ; } 
//	 UnLock() ; // this UnLock is nonsense
	 return s ;
      } ;

      //------------------------------
      // find symbol by substring

      inline Symbol * _findSymbol(char* head, char* tail) const {
	 char c = *tail ;
	 *tail = NULLCHAR ;
	 Symbol* r = _findSymbol(head) ;
	 *tail = c ;
	 return r ;
      } ;
      inline Symbol * findSymbol(char * head, char* tail) { 
//	 Lock() ; // this Lock is nonsense
	 Symbol * s = _findSymbol(head,tail) ;
	 if(s == NULLPTR) { s = newSymbol(head,tail) ; } 
//	 UnLock() ; // this UnLock is nonsense
	 return s ;
      } ;

      //--------------------------------------------------
      // Generate New Symbol

      Symbol * genSymbol(const char* seed) ;
      inline Symbol * genSymbol(const String& seed) {
//       return genSymbol((char*)seed) ;
	 return genSymbol(seed.data) ;
      } ;
      inline Symbol * genSymbol(const Symbol* seed) {
	 return genSymbol(seed->name) ;
      } ;
       

      //--------------------------------------------------
      // Expand Table

      Bool expand(UInt addsize = SymbolTableDefaultSize) ;
      inline Bool Expand(UInt addsize = SymbolTableDefaultSize) {
	 Lock() ;
	 Bool r = expand(addsize) ;
	 UnLock() ;
	 return r ;
      }

      //--------------------------------------------------
      // describe/output

      void describe(gostream& o, Bool detailp = True) const ;

      inline friend gostream& operator<< (gostream& o, const SymbolTable& st) {
	 st.describe(o,False) ;
	 return o ;
      } ;

} ;

extern SymbolTable SymTab ;

inline Bool ExistSymbol(const char * nm) {
   return !isNullPtr(SymTab._findSymbol(nm)) ;
} ;
inline Bool ExistSymbol(const String & nm) {
   return !isNullPtr(SymTab._findSymbol(nm)) ;
} ;

inline Symbol * FindSymbol(const char * nm) { 
   return SymTab.findSymbol(String(nm)) ; 
} ;
inline Symbol * FindSymbol(const String & nm) {
   return SymTab.findSymbol(nm) ; 
} ;
inline Symbol * FindSymbol(char* head, char* tail){
   return SymTab.findSymbol(head,tail) ; 
} ;


inline Symbol * GenSymbol(const char* seed) {
   return SymTab.genSymbol(seed) ;
} ;
inline Symbol * GenSymbol(const String& seed) {
   return SymTab.genSymbol(seed) ;
} ;
inline Symbol * GenSymbol(const Symbol* seed) {
   return SymTab.genSymbol(seed) ;
} ;

extern void showSymTab(gostream& ostr = gout) ;

//======================================================================
// Data
//

class GaeaData : public GCNode {
   public:
      GaeaDType type ;
      GaeaDValue val ;

      static const Int ShowList_Off = 0 ;
      static const Int ShowList_On = 1 ;

      inline void describe(gostream& o = gout, 
			   const Bool newlinep = True,
			   const Bool detailp = True,
			   const Int listformp = ShowList_Off) const 
      {
			      _describe(o,newlinep,detailp,listformp) ;
      } ;

      void _describe(gostream& o,const Bool newlinep,const Bool detailp,
		     const Int listformp) const ;

      /* followings do not work well */
//  virtual void __describe(gostream& o,const Bool newlinep,const Bool detailp) 
//    const ;
  
      inline friend gostream& operator<< (gostream& o, const GaeaData& dt) {
	 dt.describe(o,False,False,ShowList_On) ;
	 return o ;
      } ;

#ifdef USE_INLINE_EQ
      Bool eq(const GaeaData &gd) const {
	 if(type == gd.type) {
	    if        (type == GDT_Integer) {
	       return val.inum == gd.val.inum ;
	    } else if (type == GDT_Float) {
	       return val.fnum == gd.val.fnum ;
//	    } else if (type == GDT_TimeVal) {
//	       return val.time == gd.val.time ;
	    } else {
	       return val.pointer == gd.val.pointer ;
	    }
	 } else {
	    return False ;
	 } 
      } ;
#else
      Bool eq(const GaeaData &gd) const ;
#endif

	 
      /* maybe bug 
	 inline GaeaData operator=(const GaeaData& dt) { 
	 _copy_data(dt) ; 
	 return *this ; 
	 } ;
	 */
      inline void _copy_data(const GaeaData& dt) { 
	 memcpy(this,&dt,sizeof(*this)) ;
      };

      inline void init() ;
      inline void setNoData() { type = GDT_None ; } ;
      inline Bool isNoData() { return type == GDT_None ; } ;
} ;

inline Bool isNoData(GaeaData& dt) { return dt.isNoData() ; } ; 


extern GaeaData NullSymbol ;

//======================================================================
// Term Argv
//

const Int TermArityStep = 3 ;

#ifndef NOT_USE_TERM_ARGV
class TermArgv : public GCEntity {
   public:
      GaeaData   _argv[TermArityStep] ;
      TermArgv * next ;
      
      //--------------------------------------------------
      // constructor
      
      inline TermArgv() {} ;

      //--------------------------------------------------
      // access

      inline GaeaData& argv(Int idx) const {
	 Int n = idx/TermArityStep ;
	 TermArgv *targv = (TermArgv *)this ;
	 for(Int i = 0 ; i < n ; i++) { targv = targv->next ; } 
	 return targv->_argv[idx % TermArityStep] ;
      } ;
} ;


//======================================================================
// TermArgvTable
//

class TermArgvTable : public GCEntityTable {
   public:
      UInt size ;
      UInt n ;
      TermArgv ** _entry ;

      //--------------------------------------------------
      // Creater

      inline TermArgvTable(const UInt sz = TermTableDefaultSize) {
	 size = 0 ;
	 n = 0 ;
	 Lock() ;
	 expand(sz) ;
	 UnLock() ;
      } ;

      //--------------------------------------------------
      // Get Nth Entry

      inline TermArgv * entry(UInt i) const { return _entry[i] ; } ;

      //--------------------------------------------------
      // New Term

      inline TermArgv * _newTermArgv() {
	 if(n < size) { return entry(n++) ; } 
	 else if(expand()) { return _newTermArgv() ; }
	 else { 
	    gerr << "Warning: no more term argv." << endl ;
	 } ;
	 return NULLPTR ;
      } ;
      inline TermArgv * newTermArgv() {
	 Lock() ;
	 TermArgv * r = _newTermArgv() ;
	 UnLock() ;
	 return r ;
      } ;

      inline TermArgv * newTermArgv(const UInt arity) {
	 Int l = arity / TermArityStep ;
	 Int r = arity % TermArityStep ;
	 if (r) l++ ;
	 TermArgv * hd = newTermArgv() ;
	 hd->next = NULLPTR ;
	 TermArgv * md = hd ;
	 for(l-- ; l > 0 ; l--) {
	    TermArgv * tl = newTermArgv() ;
	    md->next = tl ;
	    tl->next = NULLPTR ;
	    md = tl ;
	 } 
	 return hd ;
      } ;

      //--------------------------------------------------
      // Expand Table

      Bool expand(UInt addsize = TermTableDefaultSize) ;
} ;

extern TermArgvTable TermArgvTab ;

inline TermArgv * NewTermArgv(const UInt arity) { 
   return TermArgvTab.newTermArgv(arity) ; 
} ;

#endif


//======================================================================
// Term G Arg
//

class TermGArg : public GCEntity {
   public:
      Symbol* role ;
      GaeaData value ;
      TermGArg* next ;

      //--------------------------------------------------
      // constructor

      inline TermGArg(Symbol* r = NULLPTR, const GaeaData& v = NullSymbol, 
		      TermGArg* n = NULLPTR) {
	 init(r,v,n) ;
      } ;

      //--------------------------------------------------
      // init

      inline void init(Symbol* r = NULLPTR, const GaeaData& v = NullSymbol, 
		       TermGArg* n = NULLPTR) {
	 role = r ; value = v ; next = n ;
      } ;

      //--------------------------------------------------
      // findRole

      inline TermGArg* findRole(const Symbol* r) const {
	 for(TermGArg* tga = (TermGArg *)this ; !isNullPtr(tga) ; 
	     tga = tga->next) {
	    if(r == tga->role) return tga ;
	 }
	 return (TermGArg*)NULLPTR ;
      } ;

      //--------------------------------------------------
      // printer

      void describe(gostream& o = gout,
		    const Bool newlinep = True,
		    const Bool detailp = True) const ;

      inline friend gostream& operator<< (gostream& o, const TermGArg& tga) {
	 tga.describe(o,False,False) ;
	 return o ;
      } ;
} ;


//======================================================================
// Term G Arg Table
//

class TermGArgTable : public GCEntityTable {
   public:
      UInt size ;
      UInt n ;
      TermGArg ** _entry ;

      //--------------------------------------------------
      // Creater

      inline TermGArgTable(const UInt sz = TermTableDefaultSize) {
	 size = 0 ;
	 n = 0 ;
	 Lock() ;
	 expand(sz) ;
	 UnLock() ;
      } ;

      //--------------------------------------------------
      // Get Nth Entry

      inline TermGArg * entry(UInt i) const { return _entry[i] ; } ;

      //--------------------------------------------------
      // New Term

      inline TermGArg * _newTermGArg() {
	 if(n < size) { return entry(n++) ; } 
	 else if(expand()) { return _newTermGArg() ; }
	 else { 
	    gerr << "Warning: no more term argv." << endl ;
	 } ;
	 return NULLPTR ;
      } ;
      inline TermGArg * newTermGArg() {
	 Lock() ;
	 TermGArg * r = _newTermGArg() ;
	 UnLock() ;
	 return r ;
      } ;

      //--------------------------------------------------
      // Expand Table

      Bool expand(UInt addsize = TermTableDefaultSize) ;
} ;

extern TermGArgTable TermGArgTab ;

inline TermGArg * NewTermGArg() {
   return TermGArgTab.newTermGArg() ;
} ;

inline TermGArg * NewTermGArg(Symbol* role, const GaeaData& value, 
			      TermGArg* next = NULLPTR) {
   TermGArg* tga = NewTermGArg() ;
   tga->init(role,value,next) ;
   return tga ;
} ;

//======================================================================
// Term
//

class Term : public GCEntity {
   public:
      Symbol * 	head ;
      Int	arity ;
      Bool	gargp ;
#ifdef NOT_USE_TERM_ARGV
      GaeaData	_argv[TermArityStep] ;
      Term *	next ;
#else
      TermArgv _argv ;
#endif
      TermGArg * _garg ;
      
      //--------------------------------------------------
      // Constructor

      inline Term() {} ;

#ifdef _NOT_USED_

      inline Term(Symbol* hd, const GaeaData& arg1) { 
	 head = hd ; arity = 1 ; next = NULLPTR ;
	 argv(0) = arg1 ; } ;
      inline Term(Symbol* hd, 
		  const GaeaData& arg1, const GaeaData& arg2) {
		     head = hd ; arity = 2 ; next = NULLPTR ;
		     argv(0) = arg1 ; argv(1) = arg2 ; } ;
      inline Term(Symbol* hd, 
		  const GaeaData& arg1, const GaeaData& arg2, const GaeaData& arg3) {
		     head = hd ; arity = 3 ; next = NULLPTR ;
		     argv(0) = arg1 ; argv(1) = arg2 ; argv(2) = arg3 ; } ;
      inline Term(Symbol* hd, 
		  const GaeaData& arg1, const GaeaData& arg2, const GaeaData& arg3,
		  const GaeaData& arg4) {
		     head = hd ; arity = 4 ; next = NULLPTR ;
		     argv(0) = arg1 ; argv(1) = arg2 ; argv(2) = arg3 ; 
		     argv(3) = arg4 ; } ; 

#endif


      //--------------------------------------------------
      // Printer

      inline void describe(gostream& o = gout, 
			   const Bool newlinep = True,
			   const Bool detailp = True,
			   const Int listformp = GaeaData::ShowList_Off) const 
      {
			      _describe(o,newlinep,detailp,listformp) ;
      } ;			 
			 
      void _describe(gostream& o,const Bool newlinep,const Bool detailp,
		     const Int listformp) const ;
      void _describe_list(gostream& o,const Bool newlinep,const Bool detailp,
			  const Int listformp) const ;
      inline friend gostream& operator<< (gostream& o, Term& tm) {
	 tm.describe(o,False,False,GaeaData::ShowList_On) ;
	 return o ;
      } ;
  
      //--------------------------------------------------
      // Get Arg Entry

#ifdef NOT_USE_TERM_ARGV
      inline GaeaData& argv(Int idx) const {
	 Int n = idx/TermArityStep ;
	 Term *tm = (Term *)this ;
	 for(Int i = 0 ; i < n ; i++) { tm = tm->next ; }
	 return tm->_argv[idx % TermArityStep] ;
      } ;
#else
      inline GaeaData& argv(Int idx) const {
	 return _argv.argv(idx) ;
      } ;
#endif

      //--------------------------------------------------
      // Assign Argv

#ifdef NOT_USE_TERM_ARGV
      inline Bool allocArgv(UInt new_arity, const Bool gap = False) {
	 arity = new_arity ;
	 gargp = gap ;
	 _garg = NULLPTR ;
	 return True ;
      } ;
#else
      inline Bool allocArgv(UInt new_arity, const Bool gap = False) {
	 arity = new_arity ;
	 gargp = gap ;
	 _garg = NULLPTR ;
	 if(new_arity <= TermArityStep) {
	    _argv.next = NULLPTR ;
	    return True ;
	 } else {
	    UInt rest_arity = new_arity - TermArityStep ;
	    TermArgv* next = NewTermArgv(rest_arity) ;
	    _argv.next = next ;
	    if(isNullPtr(next)) return False ;
	    else                return True ;
	 }
      } ;
#endif

      //--------------------------------------------------
      // G Arg

      inline TermGArg* _findRole(const Symbol* r) const {
	 return _garg->findRole(r) ;
      } ;

      //--------------------------------------------------
      // G Arg tail

      inline TermGArg** gargTail() const {
	 TermGArg** tail ;
	 for(tail = (TermGArg **)&_garg ; !isNullPtr(*tail) ; 
	     tail = &((*tail)->next)) ;
	 return tail ;
      } ;

      inline void addGArg(TermGArg* garg) {
	 TermGArg** tail = gargTail();
	 *tail = garg ;
      } ;

      inline void addGArg(Symbol* role, const GaeaData& value) {
	 TermGArg* garg = NewTermGArg(role,value) ;
	 addGArg(garg) ;
      } ;

} ;

//======================================================================
// Term Table
//

class TermTable : public GCEntityTable {
   public:
      UInt size ;
      UInt n ;
      Term ** _entry ;

      //--------------------------------------------------
      // Creater

      inline TermTable(const UInt sz = TermTableDefaultSize) {
	 size = 0 ;
	 n = 0 ;
	 Lock() ;
	 expand(sz) ;
	 UnLock() ;
      } ;

      //--------------------------------------------------
      // Get Nth Entry

      inline Term * entry(UInt i) const { return _entry[i] ; } ;

      //--------------------------------------------------
      // New Term

      inline Term * _newTerm() {
	 if(n < size) { return entry(n++) ; } 
	 else if(expand()) { return _newTerm() ; }
	 else { 
	    gerr << "Warning: no more term." << endl ;
	 } ;
	 return NULLPTR ;
      } ;
      inline Term * newTerm() {
	 Lock() ;
	 Term * r = _newTerm() ;
	 UnLock() ;
	 return r ;
      } ;

#ifdef NOT_USE_TERM_ARGV
      inline Term * newTerm(const UInt arity, const Bool gargp = False) {
	 UInt l = arity / TermArityStep ;
	 UInt r = arity % TermArityStep ;
	 if (r) l++ ;
	 Term * hd = newTerm() ;
	 hd->arity = arity ;
	 hd->gargp = gargp ;
	 hd->next = NULLPTR ;
	 Term * md = hd ;
	 for(l-- ; l > 0 ; l--) {
	    Term * tl = newTerm() ;
	    md->next = tl ;
	    tl->next = NULLPTR ;
	    md = tl ;
	 } 
	 return hd ;
      } ;
#else
      inline Term * newTerm(const UInt arity, const Bool gargp = False) {
	 Term *hd = newTerm() ;
	 hd->allocArgv(arity, gargp) ;
	 return hd ;
      } ;
#endif

      inline Term * newTerm(Symbol * head, const UInt arity, 
			    const Bool gargp = False ) {
	 Term * headterm = newTerm(arity, gargp) ;
	 headterm->head = head ;
	 return headterm ;
      } ;

      //--------------------------------------------------
      // Expand Table

      Bool expand(UInt addsize = TermTableDefaultSize) ;
} ;

extern TermTable TermTab ;

inline Term * NewTerm(const UInt arity, const Bool gargp = False ) { 
   return TermTab.newTerm(arity,gargp) ; } ;
inline Term * NewTerm(Symbol * head, const UInt arity, 
		      const Bool gargp = False) {
   return TermTab.newTerm(head, arity, gargp) ;
} ;
inline Term * NewTerm(Symbol *head, const GaeaData & arg0) {
   Term * term = NewTerm(head,1) ;
   term->argv(0) = arg0 ;
   return term ;
} ;
inline Term * NewTerm(Symbol *head, const GaeaData & arg0, const GaeaData & arg1) {
   Term * term = NewTerm(head,2) ;
   term->argv(0) = arg0 ; term->argv(1) = arg1 ;
   return term ;
} ;
inline Term * NewTerm(Symbol *head, const GaeaData & arg0, const GaeaData & arg1, 
		      const GaeaData & arg2) {
   Term * term = NewTerm(head,3) ;
   term->argv(0) = arg0 ; term->argv(1) = arg1 ; term->argv(2) = arg2 ;
   return term ;
} ;
inline Term * NewTerm(Symbol *head, const GaeaData & arg0, const GaeaData & arg1, 
		      const GaeaData & arg2, const GaeaData & arg3) {
			 Term * term = NewTerm(head,4) ;
			 term->argv(0) = arg0 ; term->argv(1) = arg1 ; 
			 term->argv(2) = arg2 ; term->argv(3) = arg3 ;
			 return term ;
} ;
inline Term * NewTerm(Symbol *head, const GaeaData & arg0, const GaeaData & arg1, 
		      const GaeaData & arg2, const GaeaData & arg3, 
		      const GaeaData & arg4) {
			 Term * term = NewTerm(head,5) ;
			 term->argv(0) = arg0 ; term->argv(1) = arg1 ; 
			 term->argv(2) = arg2 ; term->argv(3) = arg3 ; 
			 term->argv(4) = arg4 ;
			 return term ;
} ;
inline Term * NewTerm(Symbol *head, const GaeaData & arg0, const GaeaData & arg1, 
		      const GaeaData & arg2, const GaeaData & arg3, 
		      const GaeaData & arg4, const GaeaData & arg5) {
			 Term * term = NewTerm(head,6) ;
			 term->argv(0) = arg0 ; term->argv(1) = arg1 ; 
			 term->argv(2) = arg2 ; term->argv(3) = arg3 ; 
			 term->argv(4) = arg4 ; term->argv(5) = arg5 ;
			 return term ;
} ;
inline Term * NewTerm(Symbol *head, const GaeaData & arg0, const GaeaData & arg1, 
		      const GaeaData & arg2, const GaeaData & arg3, 
		      const GaeaData & arg4, const GaeaData & arg5, 
		      const GaeaData & arg6) {
			 Term * term = NewTerm(head,7) ;
			 term->argv(0) = arg0 ; term->argv(1) = arg1 ; 
			 term->argv(2) = arg2 ; term->argv(3) = arg3 ; 
			 term->argv(4) = arg4 ; term->argv(5) = arg5 ;
			 term->argv(6) = arg6 ;
			 return term ;
} ;
inline Term * NewTerm(Symbol *head, const GaeaData & arg0, const GaeaData & arg1,
		      const GaeaData & arg2, const GaeaData & arg3, 
		      const GaeaData & arg4, const GaeaData & arg5, 
		      const GaeaData & arg6, const GaeaData & arg7) {
			 Term * term = NewTerm(head,8) ;
			 term->argv(0) = arg0 ; term->argv(1) = arg1 ; 
			 term->argv(2) = arg2 ; term->argv(3) = arg3 ; 
			 term->argv(4) = arg4 ; term->argv(5) = arg5 ;
			 term->argv(6) = arg6 ; term->argv(7) = arg7 ;
			 return term ;
} ;


//======================================================================
// GaeaInteger
//
class GaeaInteger : public GaeaData {
   public:
      inline GaeaInteger(const Int i) {
	 type = GDT_Integer ;
	 val.inum = i ;
   } ;
  
      void _describe(gostream& o,const Bool newlinep,const Bool detailp) 
   const ;
//  virtual void __describe(gostream& o,const Bool newlinep,const Bool detailp) 
//    const ;
} ;


//======================================================================
// GaeaFloat
//
class GaeaFloat : public GaeaData {
   public:
      inline GaeaFloat(const Float & f) {
	 type = GDT_Float ;
	 val.fnum = f ;
   } ;
      void _describe(gostream& o,const Bool newlinep,const Bool detailp) 
   const ;
//  virtual void __describe(gostream& o,const Bool newlinep,const Bool detailp) 
//    const ;
} ;

//======================================================================
// GaeaSymbol
//

class GaeaSymbol : public GaeaData {
   public:
      inline GaeaSymbol(const char * nm) {
	 type = GDT_Symbol ;
	 val.symbol = FindSymbol(nm) ;
   } ;
      inline GaeaSymbol(const String & nm) {
	 type = GDT_Symbol ;
	 val.symbol = FindSymbol(nm) ;
      } ;
      inline GaeaSymbol(Symbol * sym = NULLPTR) {
	 type = GDT_Symbol ;
	 val.symbol = sym ;
      } ;
    
      void _describe(gostream& o,const Bool newlinep,const Bool detailp) 
   const ;
//  virtual void __describe(gostream& o,const Bool newlinep,const Bool detailp) 
//    const ;

      Bool shouldquotep() const ;

} ;

//======================================================================
// GaeaVariable
//
class GaeaVariable : public GaeaSymbol {
   public:
      inline GaeaVariable(const char * nm) : GaeaSymbol(nm) {
	 type = GDT_Variable ; 
      } ;
      inline GaeaVariable(const String & nm) : GaeaSymbol(nm) { 
	 type = GDT_Variable ; 
      } ;
      inline GaeaVariable(Symbol * sym = NULLPTR) : GaeaSymbol(sym) { 
	 type = GDT_Variable ; 
      } ; 

      void _describe(gostream& o,const Bool newlinep,const Bool detailp) 
   const ;
//  virtual void __describe(gostream& o,const Bool newlinep,const Bool detailp) 
//    const ;
} ;

//======================================================================
// GaeaString
//
class GaeaString : public GaeaData {
   public:
      inline GaeaString() {
	 type = GDT_String ;
	 val.string = new String() ;
      } ;
      inline GaeaString(const char *s) {
	 type = GDT_String ;
	 val.string = new String(s) ;
      } ;
      inline GaeaString(const String& s) {
	 type = GDT_String ;
	 val.string = new String(s) ;
      } ;

      inline void _let(char *s, Bool deletep = False) {
	 if(deletep && !isNullPtr(val.string)) delete val.string ;
	 val.string = new String() ;
	 val.string->_let(s) ;
      } ;
      inline void _let(String* s) {
	 delete val.string ;
	 val.string = s ;
      } ;

      void _describe(gostream& o, const Bool newlinep, const Bool detailp) const ;
} ;

inline GaeaString TmpGaeaString(char *str) {
   GaeaString r ;
   r._let(str,True) ;
   return r ;
} ;

//======================================================================
// GaeaPointer
//
class GaeaPointer : public GaeaData {
   public:
      inline GaeaPointer() {
	 type = GDT_Pointer ;
	 val.pointer = NULLPTR ;
      } ;
      inline GaeaPointer(void* ptr) {
	 type = GDT_Pointer ;
	 val.pointer = ptr ;
      } ;

      inline void _let(void* ptr) {
	 val.pointer = ptr ;
      } ;
      void _describe(gostream& o, const Bool newlinep, const Bool detailp) const ;
} ;


//======================================================================
// GaeaExtObject
//
class GaeaExtObject : public GaeaData {
   public:
      inline GaeaExtObject() {
	 type = GDT_ExtObject ;
	 val.pointer = NULLPTR ;
      } ;
      inline GaeaExtObject(ExtObjTypeID id, void* ptr) {
	 type = (GaeaDType)(ExtObjTypeID2GaeaDType(id)) ;
	 val.pointer = ptr ;
      } ;

      inline void _let(ExtObjTypeID id, void* ptr) {
	 type = (GaeaDType)(id+GDT_ExtObject) ;
	 val.pointer = ptr ;
      } ;
      void _describe(gostream& o, const Bool newlinep, const Bool detailp) const ;
} ;


//======================================================================
// GaeaInteger
//
/*
class GaeaTimeVal : public GaeaData {
   public:
      inline GaeaTimeVal(const Int sec = 0, 
			 const Int msec = 0,
			 const Int usec = 0) {
	 init(TimeVal(sec,msec,usec)) ;
      } ;
      inline GaeaTimeVal(const TimeVal& tval) {
	 init(tval) ; 
      } ;

      inline init(const TimeVal& tval) {
	 type = GDT_TimeVal ;
	 val.time = tval ;
      } ;
  
      void _describe(gostream& o,const Bool newlinep,const Bool detailp) 
   const ;
//  virtual void __describe(gostream& o,const Bool newlinep,const Bool detailp) 
//    const ;
} ;
*/

//======================================================================
// GaeaTerm
//
class GaeaTerm : public GaeaData {
   public:
      //------------------------------
      // constructor by symbol head
      inline GaeaTerm() {} ;
      inline GaeaTerm(Term * term) { type = GDT_Term ; val.term = term ; } ;
      inline GaeaTerm(Symbol * sym, const UInt arity, 
		      const Bool gargp = False) {
	 type = GDT_Term ;
	 val.term = NewTerm(sym, arity, gargp) ;
      } ;
      inline GaeaTerm(Symbol * sym, const GaeaData & arg0) {
	 type = GDT_Term ;
	 val.term = NewTerm(sym, arg0) ;
      } ;
      inline GaeaTerm(Symbol * sym, const GaeaData & arg0, const GaeaData & arg1) {
	 type = GDT_Term ;
	 val.term = NewTerm(sym, arg0, arg1) ;
      } ;
      inline GaeaTerm(Symbol * sym, const GaeaData & arg0, const GaeaData & arg1, 
		   const GaeaData & arg2) {
	 type = GDT_Term ;
	 val.term = NewTerm(sym, arg0, arg1, arg2) ;
      } ;
      inline GaeaTerm(Symbol * sym, const GaeaData & arg0, const GaeaData & arg1, 
		   const GaeaData & arg2, const GaeaData & arg3) {
		      type = GDT_Term ;
		      val.term = NewTerm(sym, arg0, arg1, arg2, arg3) ;
      } ;
      inline GaeaTerm(Symbol * sym, const GaeaData & arg0, const GaeaData & arg1, 
		   const GaeaData & arg2, const GaeaData & arg3, const GaeaData & arg4){
		      type = GDT_Term ;
		      val.term = NewTerm(sym, arg0, arg1, arg2, arg3, arg4) ;
      } ;
      inline GaeaTerm(Symbol * sym, const GaeaData & arg0, const GaeaData & arg1, 
		   const GaeaData & arg2, const GaeaData & arg3, const GaeaData & arg4,
		   const GaeaData & arg5) {
		      type = GDT_Term ;
		      val.term = NewTerm(sym, arg0, arg1, arg2, arg3, arg4, 
					 arg5) ;
      } ;
      inline GaeaTerm(Symbol * sym, const GaeaData & arg0, const GaeaData & arg1, 
		   const GaeaData & arg2,const GaeaData & arg3, const GaeaData & arg4, 
		   const GaeaData & arg5, const GaeaData & arg6) {
		      type = GDT_Term ;
		      val.term = NewTerm(sym, arg0, arg1, arg2, arg3, arg4, 
					 arg5, arg6) ;
      } ;
      inline GaeaTerm(Symbol * sym, const GaeaData & arg0, const GaeaData & arg1, 
		   const GaeaData & arg2, const GaeaData & arg3, const GaeaData & arg4,
		   const GaeaData & arg5, const GaeaData & arg6, const GaeaData & arg7){
		      type = GDT_Term ;
		      val.term = NewTerm(sym, arg0, arg1, arg2, arg3, arg4, 
					 arg5, arg6, arg7) ;
      } ;

      //------------------------------
      // constructor by gdata head

      inline Symbol* symval(const GaeaData& sym) {
	 if(sym.type == GDT_Symbol) return sym.val.symbol ;
	 else return NullSymbol.val.symbol ;
      } ;

      inline GaeaTerm(const GaeaData& sym, const UInt arity) {
	 type = GDT_Term ;
	 val.term = NewTerm(symval(sym), arity) ;
      } ;
      inline GaeaTerm(const GaeaData& sym, const GaeaData & arg0) {
	 type = GDT_Term ;
	 val.term = NewTerm(symval(sym), arg0) ;
      } ;
      inline GaeaTerm(const GaeaData& sym, const GaeaData & arg0, const GaeaData & arg1) {
	 type = GDT_Term ;
	 val.term = NewTerm(symval(sym), arg0, arg1) ;
      } ;
      inline GaeaTerm(const GaeaData& sym, const GaeaData & arg0, const GaeaData & arg1, 
		   const GaeaData & arg2) {
	 type = GDT_Term ;
	 val.term = NewTerm(symval(sym), arg0, arg1, arg2) ;
      } ;
      inline GaeaTerm(const GaeaData& sym, const GaeaData & arg0, const GaeaData & arg1, 
		   const GaeaData & arg2, const GaeaData & arg3) {
		      type = GDT_Term ;
		      val.term = NewTerm(symval(sym), arg0, arg1, arg2, arg3) ;
      } ;
      inline GaeaTerm(const GaeaData& sym, const GaeaData & arg0, const GaeaData & arg1, 
		   const GaeaData & arg2, const GaeaData & arg3, const GaeaData & arg4){
		      type = GDT_Term ;
		      val.term = NewTerm(symval(sym), arg0, arg1, arg2, arg3, 
					 arg4) ;
      } ;
      inline GaeaTerm(const GaeaData& sym, const GaeaData & arg0, const GaeaData & arg1, 
		   const GaeaData & arg2, const GaeaData & arg3, const GaeaData & arg4,
		   const GaeaData & arg5) {
		      type = GDT_Term ;
		      val.term = NewTerm(symval(sym), arg0, arg1, arg2, arg3, 
					 arg4, arg5) ;
      } ;
      inline GaeaTerm(const GaeaData& sym, const GaeaData & arg0, const GaeaData & arg1, 
		   const GaeaData & arg2,const GaeaData & arg3, const GaeaData & arg4, 
		   const GaeaData & arg5, const GaeaData & arg6) {
		      type = GDT_Term ;
		      val.term = NewTerm(symval(sym), arg0, arg1, arg2, arg3, 
					 arg4, arg5, arg6) ;
      } ;
      inline GaeaTerm(const GaeaData& sym, const GaeaData & arg0, const GaeaData & arg1, 
		   const GaeaData & arg2, const GaeaData & arg3, const GaeaData & arg4,
		   const GaeaData & arg5, const GaeaData & arg6, const GaeaData & arg7){
		      type = GDT_Term ;
		      val.term = NewTerm(symval(sym), arg0, arg1, arg2, arg3, 
					 arg4, arg5, arg6, arg7) ;
      } ;

      //--------------------------------------------------
      // describe

      void _describe(gostream& o,const Bool newlinep,const Bool detailp,
		     const Int listformp) const ;
//  virtual void __describe(gostream& o,const Bool newlinep,const Bool detailp) 
//    const ;
} ;

//======================================================================
// check data type
//
inline Bool isGaeaInteger(const GaeaData &dt) { return dt.type == GDT_Integer ; } ;
inline Bool isGaeaFloat(const GaeaData &dt) { return dt.type == GDT_Float ; } ;
inline Bool isGaeaSymbol(const GaeaData &dt) { return dt.type == GDT_Symbol ; } ;
inline Bool isGaeaVariable(const GaeaData &dt) { return dt.type == GDT_Variable ; } ;
inline Bool isGaeaString(const GaeaData &dt) { return dt.type == GDT_String ; } ;
inline Bool isGaeaPointer(const GaeaData &dt) { return dt.type == GDT_Pointer ; } ;
//inline Bool isGaeaTimeVal(const GaeaData &dt) { return dt.type == GDT_TimeVal ; } ;
inline Bool isGaeaTerm(const GaeaData &dt) { return dt.type == GDT_Term ; } ;
inline Bool isGaeaExtObject(const ExtObjTypeID tid, const GaeaData &dt) {
   return dt.type == ExtObjTypeID2GaeaDType(tid) ;
} ;
inline Bool isGaeaExtObject(const GaeaData &dt) {
   return isExtObjectType(dt.type) ;
} ;

inline Bool isGNumber(const GaeaData &dt) { 
   return isGaeaInteger(dt)||isGaeaFloat(dt) ; 
} ;

inline Bool isTGaeaString(const GaeaData &dt) {
   return isGaeaSymbol(dt) || isGaeaString(dt) ;
} ;

inline Bool isGAtom(const GaeaData &dt) {
   return  isGaeaSymbol(dt) || isGNumber(dt) || isGaeaString(dt) 
//      || isGaeaTimeVal(dt)
      || isGaeaPointer(dt) || isGaeaExtObject(dt) ;
} ;

inline Bool eq(const GaeaData& x, const GaeaData& y) { return x.eq(y) ; };
inline Bool eqAsPointer(const GaeaData& x, const GaeaData& y) { 
   return x.val.pointer == y.val.pointer ; };

inline Int IntVal(const GaeaData &dt) {
   if(isGaeaInteger(dt)) return dt.val.inum ;
   if(isGaeaFloat(dt)) return (Int)(dt.val.fnum) ;
   return 0 ;
} ;
inline Float FltVal(const GaeaData &dt) {
   if(isGaeaFloat(dt)) return dt.val.fnum ;
   if(isGaeaInteger(dt)) return (Float)(dt.val.inum) ;
   return 0.0 ;
} ;
//inline TimeVal _TimeVal(const GaeaData& dt) {
//   if(isGaeaTimeVal(dt)) return dt.val.time ;
//   return TimeVal(0,0,0) ;
//} ;

inline Symbol* _SymVal(const GaeaData& dt) { return dt.val.symbol ; } ;
inline Symbol* _VarVal(const GaeaData& dt) { return dt.val.symbol ; } ;
inline Symbol* SymVal(const GaeaData& dt) {
   if(isGaeaSymbol(dt)) return _SymVal(dt) ;
   else return _SymVal(NullSymbol) ;
} ;

inline GaeaCell* _SymGaeaCell(const GaeaData& dt) { return _SymVal(dt)->gcell ; };
inline Gem* _SymGem(const GaeaData& dt) { return _SymVal(dt)->gem ; };
inline GaeaDataBox* _SymGDBox(const GaeaData& dt) { return _SymVal(dt)->gdbox ; };
inline GaeaMutex& _SymMutex(const GaeaData& dt) { return _SymVal(dt)->mutex ; };
inline void* _SymDLHandle(const GaeaData& dt) { return _SymVal(dt)->dlhandle ; }; 

inline Bool hasGaeaCell(const GaeaData& dt) { 
   return isGaeaSymbol(dt) && _SymVal(dt)->hasGaeaCell() ; 
}
inline Bool hasGem(const GaeaData& dt) { 
   return isGaeaSymbol(dt) && _SymVal(dt)->hasGem() ; 
}
inline Bool hasGDBox(const GaeaData& dt) { 
   return isGaeaSymbol(dt) && _SymVal(dt)->hasGDBox() ; 
}
inline Bool hasDLHandle(const GaeaData& dt) { 
   return isGaeaSymbol(dt) && _SymVal(dt)->hasDLHandle() ; 
}

inline Term* _TermVal(const GaeaData& dt) { return dt.val.term ; } ;
inline Symbol* _TermHead(const GaeaData& dt) { return dt.val.term->head ; } ;
inline Int _TermArity(const GaeaData& dt) { return dt.val.term->arity ; } ;
inline GaeaData& _TermArgv(const GaeaData& dt, Int n) { 
   return dt.val.term->argv(n) ; } ;
inline Bool& _TermGArgP(const GaeaData& dt) { return _TermVal(dt)->gargp ; } ;
inline TermGArg* _TermGArgTop(const GaeaData& dt) { 
   return _TermVal(dt)->_garg ; } ;
inline TermGArg* _TermGArg(const GaeaData& dt, const Symbol* role) { 
   return _TermVal(dt)->_garg->findRole(role) ; } ;

inline String* _StrPtrVal(const GaeaData& dt) { return dt.val.string ; } ;
inline String _StrVal(const GaeaData& dt) { return *_StrPtrVal(dt) ; } ;
inline String _TStrVal(const GaeaData& dt) { 
   if(isGaeaSymbol(dt)) return _SymVal(dt)->name ;
   else *_StrPtrVal(dt) ; } ;
inline void* _PtrVal(const GaeaData& dt) { return dt.val.pointer ; } ;
inline void* _ExObjVal(const GaeaData& dt) { return dt.val.pointer ; } ;

//======================================================================
// GaeaData Box
//

class GaeaDataBox {
   public:
      GaeaData val ;
      Bool assigned ;
      GaeaMutexCondVar mutex ;

      //--------------------------------------------------
      // Creater

      inline GaeaDataBox() {
	 val = NullSymbol ;
	 assigned = False ;
      } ;

      inline GaeaDataBox(const GaeaDataBox& src) {
	 val = src.val ;
	 assigned = src.assigned ;
      } ;

      inline GaeaDataBox(const GaeaData& v) {
	 val = v ;
	 assigned = True ;
      } ;

      //--------------------------------------------------
      // let

      inline GaeaDataBox operator=(const GaeaDataBox& src) {
	 val = src.val ;
	 assigned = src.assigned ;
      } ;

      //--------------------------------------------------
      // put

      enum PutType {
	 PT_nowait,
	 PT_wait,
	 PT_force,
	 PT_max
      } ;

      Bool put(const GaeaData& v, PutType puttype) ;
      Bool timedwait_put(const GaeaData& v, 
			 UInt sec, UInt msec = 0, UInt usec = 0) ;

      //--------------------------------------------------
      // get

      Bool get(GaeaData& v, Bool retractp, Bool waitp) ;
      Bool timedwait_get(GaeaData& v, Bool retractp,
			 UInt sec, UInt msec = 0, UInt usec = 0) ;

      //--------------------------------------------------
      // Printer
  
      inline void describe(gostream& o, Bool detailp = True) {
	 o << "#GaeaDataBox[" ;
	 if(assigned) o << "*:" << val ;
	 else         o << "_" ;
	 o << "]" ;
	 if(detailp) o << endl ;
      } ;
      inline friend gostream& operator<< (gostream& o, GaeaDataBox& gdbox) {
	 gdbox.describe(o,False) ;
	 return o ; 
      } ;
      
} ;

//======================================================================
// GaeaDataStack
//

class GaeaDataStack : public  AStack<GaeaData> {
   public:
      GaeaMutex mutex ;
      
      GaeaDataStack(char *nm, UInt sz = DEFAULT_ARRAY_SIZE) {
	 astack_t::init(nm,sz) ;
      } ;

      inline void Lock() { mutex.lock() ; } ;
      inline void UnLock() { mutex.unlock() ; } ;

      GaeaDataStack* push(const GaeaData& val) {
	 Lock() ; astack_t::push(val) ; UnLock() ;
	 return this ;
      } ;
      
      GaeaData pop() {
	 Lock() ; GaeaData v = astack_t::pop() ; UnLock() ;
	 return v ;
      } ;

      GaeaData top() { 
	 Lock() ; GaeaData v = astack_t::top() ; UnLock() ;
	 return v ;
      } ;

      GaeaData topN(UInt c) { 
	 Lock() ; GaeaData v = astack_t::topN(c) ; UnLock() ;
	 return v ;
      } ;

      GaeaData bottomN(UInt c) { 
	 Lock() ; GaeaData v = astack_t::bottomN(c) ; UnLock() ;
	 return v ;
      } ;

      UInt setpos(UInt newtop) { 
	 Lock() ; UInt c = astack_t::setpos(newtop) ; UnLock() ;
	 return c ;
      } ;

      UInt rewind(UInt newtop) { 
	 Lock() ; UInt c = astack_t::rewind(newtop) ; UnLock() ;
	 return c ;
      } ;

      UInt toppos() { 
	 Lock() ; UInt c = astack_t::toppos() ; UnLock() ;
	 return c ;
      } ;

      void fill(const GaeaData& val, UInt from, UInt to) { 
	 Lock() ; astack_t::fill(val,from,to) ; UnLock() ;
      } ;
} ;


//======================================================================
// Special Data
//

#include "gglobal.h"

#define VARCONCATCHAR "_"

inline Bool isNullSymbol(const GaeaData& dt) { return eq(dt,NullSymbol) ; } ;
inline Bool isConsTerm(const Term& tm) {
   return tm.head == _SymVal(ConsSymbol) && tm.arity == 2 && !(tm.gargp) ; } ;
inline Bool isConsTerm(const Term* tm) { return isConsTerm(*tm) ; } ;
inline Bool isConsTerm(const GaeaData& dt) { 
   return (isGaeaTerm(dt) && isConsTerm(_TermVal(dt))) ; } ;
inline GaeaData _ConsCar(const GaeaData& dt) { return _TermArgv(dt,0) ; } ;
inline GaeaData _ConsCdr(const GaeaData& dt) { return _TermArgv(dt,1) ; } ;

inline void GaeaData::init() { *this = NullSymbol ; } ;

GaeaData genGaeaSymbol(const GaeaData &seed) ;
inline GaeaData genGaeaSymbol() { return genGaeaSymbol(NullSymbol) ; } ;

GaeaData genGaeaVariable(const GaeaData &seed) ;
inline GaeaData genGaeaVariable() { return genGaeaVariable(AnonymousVar) ; } ;

//======================================================================
#include "gcell.h"

////////////////////////////////////////////////////////////////////////
#endif


