// -*- Mode:C++ -*-
//Header:
//File: gem.cc
//Title: gaea elementary machine
//Author: NODA Itsuki
//Date: 1998/6/09
//EndHeader:

//ModifyHistory:
// 1998/6/09: Create this file
//EndModifyHistory:

//Memo:
//EndMemo:

#include "gem.h"

//======================================================================
// Global Ver
//

VerboseLevel DefaultVerboseLevel = VBL_normal ;
DirSearchPath LoadPath ;

//======================================================================
// Load Path
//

String DirSearchPath::findFullPath(const String filename) {
   char* fname = (char*)((String)filename) ;
   if(isNullPtr(fname)) return filename ;
   if(strchr(fname,DIRSEPARATOR)) return filename ;

   Bool foundp = False ;
   char r[256]; 
   lock() ;
   for(Int i = 0 ; i < _toppos() ; i++) {
      strcpy(r,(char*)(_bottomN(i))) ;
      Int l = strlen(r) ;
      if(r[l-1] != DIRSEPARATOR) { r[l] = DIRSEPARATOR ; r[l+1] = NULLCHAR ; } 
      strcat(r,fname) ;
      if(!access(r,R_OK)) {
	 foundp = True ;
	 break ;
      }
   }
   unlock() ;
   if(foundp) 
      return String(r) ;
   else
      return String() ;
}

//======================================================================
// OrState
//

//--------------------------------------------------
// output

void OrState::describe(gostream& o, Bool detailp) const {
   o << "#OrState[" << this << ":asstack=" << andStateStackPointer << ",bstate=" 
     << bindingState << "]" ;
   if(detailp) {
      o << endl ;
      o << "   AndStateStackPointer=" << andStateStackPointer << endl ;
      o << "   BindingState=" << bindingState << endl ;
      o << endl ;
   }
} ;

//======================================================================
// Signal Entry
//

GaeaMutex GaeaSignalMutex ;

void GaeaSignalEntry::pushSignal(Gem* g, Symbol* sym) {
   sym->mutex.lock();
   GaeaSignalEntry* wlist = sym->signalWaitList;
   gem = g ;
   symbol = sym ;
   prev = NULLPTR;
   next = wlist ;
   if(!isNullPtr(wlist)) wlist->prev = this ;
   sym->signalWaitList = this ;
   sym->mutex.unlock();
} ;

void GaeaSignalEntry::popSignal() {
   symbol->mutex.lock();
   if(isNullPtr(prev)) {
      symbol->signalWaitList = next ;
   } else {
      prev->next = next ;
   }
   if(!isNullPtr(next)) {
      next->prev = prev ;
   }
   symbol->mutex.unlock();
} ;

void GaeaSignalEntry::broadcast() {
   symbol->mutex.lock();
   broadcast0();
   symbol->mutex.unlock();
} ;

void GaeaSignalEntry::broadcast0() {
   if(!isNullPtr(next)) next->broadcast0() ;
//   gem->runlevel_lock.lock() ;
   if(gem->runlevel == Gem::RL_suspend) {
      gem->runlevel = Gem::RL_wakeup ;
      gem->signalSymbol = symbol ;
      gem->runlevel_lock.signal() ;
   }
//   gem->runlevel_lock.unlock() ;
} ;

void broadcastSignal(Symbol* symbol) {
   GaeaSignalMutex.lock() ;
   symbol->mutex.lock();
   symbol->lastSignalTime.update() ;
   if(!isNullPtr(symbol->signalWaitList)) {
      symbol->signalWaitList->broadcast0() ;
   }
   symbol->mutex.unlock();
   GaeaSignalMutex.unlock() ;
} 
   
//======================================================================
// GEM Gaea Elementary Machine

//--------------------------------------------------
// wait signal

Symbol* Gem::waitSignal(Symbol** symlist, UInt n) {
   runlevel_lock.lock() ;
   for(UInt i = 0 ; i < n ; i++) {
      signalEntry[i].pushSignal(this,symlist[i]) ;
   }
   signalEntryN = n ;
   runlevel = RL_suspend ;
   while(runlevel == RL_suspend) {
      runlevel_lock.wait() ;
   }
   Symbol* sigsym = signalSymbol ;
   for(UInt i = 0 ; i < signalEntryN ; i++) {
      signalEntry[i].popSignal() ;
   }
   signalEntryN = 0 ;
   runlevel = RL_run ;
   runlevel_lock.unlock() ;
   
   return sigsym ;
} ;
	 
Symbol* Gem::timedwaitSignal(Symbol** symlist, UInt n, 
			  UInt sec, UInt msec, UInt usec = 0) {
   runlevel_lock.lock() ;
   for(UInt i = 0 ; i < n ; i++) {
      signalEntry[i].pushSignal(this,symlist[i]) ;
   }
   signalEntryN = n ;
   runlevel = RL_suspend ;
   runlevel_lock.timedwait(sec,msec,usec) ;
   Symbol* sigsym ;
   if(runlevel == RL_wakeup) {
      sigsym = signalSymbol ;
   } else {
      sigsym = _SymVal(TimeOutSymbol) ;
   }
   for(UInt i = 0 ; i < signalEntryN ; i++) {
      signalEntry[i].popSignal() ;
   }
   signalEntryN = 0 ;
   runlevel = RL_run ;
   runlevel_lock.unlock() ;

   return sigsym ;
}

//--------------------------------------------------
// read/exec/print loop

void Gem::reploop0(gistream& istr, gostream& ostr, Bool toplevelp) {
   Bool endp = False ;
   ReadStatus rstatus ;
   while(!endp) {
      ostr << "?- " ; ostr.flush() ;
      GaeaData gdt = readSentence(istr,rstatus) ;
      if(eq(gdt,EOFSymbol)) { endp = True ; continue ; }
      if(eq(gdt,ErrorSymbol)) { 
	 ostr << "Error: Illegal form." << endl ;
	 skipLine(istr,rstatus) ;
	 continue ; 
      }
      if(isGaeaTerm(gdt)) {
	 if((_TermHead(gdt) == _SymVal(RuleOpSymbol) &&
	     _TermArity(gdt) == 2) ||
	    (_TermHead(gdt) == _SymVal(FactOpSymbol) &&
	     _TermArity(gdt) == 1)) {
	    gcell->assertz(GaeaClause(gdt)) ;
	    continue ;
	 }
      }
      skipLine(istr,rstatus) ;
      settopgoal(gdt,NULLPTR,toplevelp) ;
      Bool r = unfoldLoop();
      if(r) {
//	 ostr << fetchValue(topGoal,&uiheap) << endl ;
	 fetchValue(topGoal,&uiheap)._describe(ostr,True,False,1) ;
      } else {
	 ostr << "Fail!" << endl ;
      }
   }
}

//--------------------------------------------------
// load loop

void Gem::load(String filename, GaeaCell* dstcell, Bool toplevelp) {
   String fname = LoadPath.findFullPath(filename) ;
   if(isNullPtr((char*)fname)) {
      *estrm << "Error: can not find " << filename << " in the path:" ;
      LoadPath.describe(*estrm) ;
      return ;
   }
   gifstream istr((char*)fname) ;
   if(istr) { 
      if(verbosep > VBL_silent) {
	 *ostrm << "Loading " << filename << "..." ;
      }
      loadloop(istr, dstcell, toplevelp) ;
      if(verbosep > VBL_silent) {
	 *ostrm << "done." << endl ;
      }
   } else {
      *estrm << "Error: Can not open file \"" << filename << "\"" << endl ;
   }
//	??? is this safe?
//   istr.close() ;
}

void Gem::loadloop(gistream& istr, GaeaCell* dstcell, Bool toplevelp) {
   ReadStatus rstatus ;
   Bool endp = False ;
   while(!endp) {
      GaeaData gdt = readSentence(istr,rstatus) ;
      if(eq(gdt,EOFSymbol)) {
	 endp = True ;
	 continue ;
      }
      if(eq(gdt,ErrorSymbol)) {
	 *estrm << "Syntax error: erroID=" << rstatus.errorID 
		<< " char=" << rstatus.charCount
		<< " line=" << rstatus.lineCount 
		<< endl ;
	 endp = True ;
	 continue ;
      }
      if(isGaeaTerm(gdt)) {
	 if((_TermHead(gdt) == _SymVal(RuleOpSymbol) &&
	     _TermArity(gdt) == 2) ||
	    (_TermHead(gdt) == _SymVal(FactOpSymbol) &&
	     _TermArity(gdt) == 1)) {
	    if(isNullPtr(dstcell)) { 
	       agcell()->assertz(GaeaClause(gdt)) ;
	    } else {
	       dstcell->assertz(GaeaClause(gdt)) ;
	    }
	    continue ;
	 }
      }
      UInt orlevel = orStateStack.toppos() ;
      settopgoal(gdt,NULLPTR,toplevelp) ;
      Bool r = unfoldLoop();
      orStateStack.setpos(orlevel) ;
   }
}

//--------------------------------------------------
// settopgoal

void Gem::settopgoal(GaeaData goal, Binding* binding, Bool resetp) {
   if(resetp) {
      andStateStack.reset() ;
      orStateStack.reset() ;
      uiheap.reset() ;
   }
   if(isNullPtr(binding)) 
      binding = uiheap.newBinding() ;
   
   topGoal = InstGaeaData(goal,binding) ;
   andStateStack.push(tmpAndState(topGoal,orStateStack.toppos())) ;
} ;

//--------------------------------------------------
// over toploop for sub process

void Gem::overTopLoop() {
   while(True) {
      runlevel_lock.lock() ;
      while(runlevel == RL_discard) {
	 runlevel_lock.wait() ;
      }
      runlevel_lock.unlock() ;
      switch(runlevel) {
	 case RL_start : {
	    birthTime.update() ;
	    UnfoldLoop() ; 
	    DeleteGem(this) ; 
	    break ;
	 }
	 case RL_dead: {
	    runlevel_lock.signal() ;
	    return ;
	 }
	 default: {
	    cerr << "Warning: something wrong in run level." << endl ;
	    showstatus(gerr) ;
	    break ;
	 }
      }
   }
}
	 

void* GemOverTopLoop(void* g) {
   Gem* gem = ((Gem*) g) ;
   gem->overTopLoop() ;
   return NULLPTR;
}

void Gem::start(GaeaData goal, const Gem* mother_gem) {
   runlevel_lock.lock() ;
   gcell->copyGaeaCellStack(mother_gem->gcell) ;
   settopgoal(goal) ;
   runlevel = RL_start ;
   runlevel_lock.signal();
   runlevel_lock.unlock();
}

void Gem::override_by_self_id() {
   runlevel_lock.lock() ;
   runlevel = RL_dead ;
   runlevel_lock.signal();
   runlevel_lock.wait();
   runlevel_lock.unlock();
   pid.set_self_id() ;
}

//--------------------------------------------------
// Unforld Loop for sub process

/* OLD version 
void* UnfoldLoop(void* g) {
   Gem* gem = ((Gem*)g) ;
   if(gem->verbosep > VBL_silent) {
      *(gem->ostrm) << "Start process " << *(gem->name) << "." << endl ;
   } 
   Bool r = gem->unfoldLoop() ;
   if(gem->verbosep > VBL_silent) {
      *(gem->ostrm) << "Terminate process " << *(gem->name) << "." 
		    << "[final state = " ;
      if (r) *(gem->ostrm) << "success" ; 
      else   *(gem->ostrm) << "fail" ; 
      *(gem->ostrm) << "]" << endl ;
   } 
   DeleteGem(gem) ;
   return g ;
}
*/

void Gem::UnfoldLoop() {
   if(verbosep > VBL_silent) {
      *ostrm << "Start process " << *name << "." << endl ;
   } 
   Bool r = unfoldLoop() ;
   if(verbosep > VBL_silent) {
      *ostrm << "Terminate process " << *name << "." 
		    << "[final state = " ;
      if (r) *ostrm << "success" ; 
      else   *ostrm << "fail" ; 
      *ostrm << "]" << endl ;
   } 
}


Bool Gem::unfoldLoop() {
   runlevel_lock.lock() ;
   runlevel = RL_run ;
   runlevel_lock.unlock() ;

   UnfoldState r ;
   Bool successp ;

   while(True) {
      r = unfold() ;
      if(r == US_fail) {
	 if(orStateStack.bottomp()) {
	    successp = False ;
	    break ;
	 } else {
	    popOrState() ;
	 }
      } else if (r == US_succeed) {
	 successp = True ;
	 break ;
      }
      Yield() ;
   } ;

   runlevel_lock.lock() ;
   runlevel = RL_discard ;
   runlevel_lock.unlock() ;
   return successp ;
}

//--------------------------------------------------
// unfold

Gem::UnfoldState Gem::unfold() {
   if(andStateStack.bottomp())	return US_succeed ;
   else if(tryDef()) 		return US_continue ;
   else 			return US_fail ;
}

//--------------------------------------------------
// try system def

Bool Gem::tryDef() {
   stepcount++ ;
#ifdef xxDEBUG
   lastActiveTime.update() ;
#endif
   
   astate = andStateStack.toppointer() ;

   Bool r = False ;

   if(tracelevel > TRC_none) {
      if(!(isGaeaTerm(currentGoal().data) && 
	   _TermHead(currentGoal().data) == _SymVal(AndSymbol))) {
	 switch(tracelevel) {
	    case TRC_normal: {
	       *estrm << "Call:" 
		      << fetchValue(currentGoal(),(UnifyInfoHeap *)&uiheap) 
		      << endl ; 
	       break ;
	    }
	    case TRC_kernel: {
	       *estrm << endl << "<<<<<<<<<<Enter tryDef<<<<<<<<<<" << endl ;
	       showstatus(*estrm) ;
	       break ;
	    }
	 }
      }
   }

   workingGoal = fetch(currentGoal()) ;

   BipStatus sysresult = trySystemDef() ;

   if(sysresult != BIP_NoMatch) {
      if(sysresult == BIP_Success) {
	 r = True ;
      } else if (sysresult == BIP_CutFail) {
	 r = False ;
      } else if (sysresult == BIP_Error) {
	 r = False ;
      } else {
	 gerr << "Unknown BIP status!!" << endl ;
      }
   } else {
      if(astate->firsttryp()) {
	 astate->initGaeaClausePointer(gcell) ;
	 astate->trycounter = 0 ;
      } ;
      if(tryUserDef()) { r = True ;}
      else {r = False ;}
   }

   if(tracelevel > TRC_none) {
      if(!(isGaeaTerm(currentGoal().data) && 
	   _TermHead(currentGoal().data) == _SymVal(AndSymbol))) {
	 switch(tracelevel) {
	    case TRC_normal: {
	       if(r) *estrm << "Succ[" << astate->trycounter << "]:" ;
	       else  *estrm << "Fail[" << astate->trycounter << "]:" ;
	       *estrm << fetchValue(currentGoal(),(UnifyInfoHeap *)&uiheap) 
		      << endl ; 
	       break ;
	    }
	    case TRC_kernel: {
	       *estrm << endl << ">>>>>>>>>>Exit tryDef>>>>>>>>>>" ;
	       if(r) *estrm << "Succ[" << astate->trycounter << "]:" << endl ;
	       else  *estrm << "Fail[" << astate->trycounter << "]:" << endl ;
	       showstatus(*estrm) ;
	       break ;
	    }
	 }
      }
   }

   return r ;
}

//--------------------------------------------------
// try system def

BipStatus Gem::trySystemDef() {
   GaeaData& goal = workingGoal.data ;

   if(isGaeaSymbol(goal)
      && _SymVal(goal)->funcp) {
      return (*(_SymVal(goal)->func))(this) ;
   } 
   if(isGaeaTerm(goal)
      && _TermHead(goal)->funcp) {
      return (*(_TermHead(goal)->func))(this) ;
   } 
   return BIP_NoMatch ;
}

#ifndef ORIGINAL
//--------------------------------------------------
// empty role term operation

InstGaeaData Gem::makeRoleTerm(const InstGaeaData& term,
			       Symbol* role, InstGaeaData& value) {
   Binding* newbinding = uiheap.newBinding() ;
   unify(InstGaeaData(RoleRoleVar,newbinding),
	 InstGaeaData(GaeaSymbol(role),newbinding),
	 &uiheap) ;
   unify(InstGaeaData(RoleValueVar,newbinding),
	 value,
	 &uiheap) ;
   unify(InstGaeaData(RoleTermVar,newbinding),
	 term,
	 &uiheap) ;
   return InstGaeaData(RoleTerm3,newbinding) ;
}

void Gem::pushRoleTerms() {
   for(UInt i = 0 ; i < uiheap.gargstack.toppos() ; i++) {
      GaeaArgEntry& garg = uiheap.gargstack.topN(i) ;
      InstGaeaData goal = makeRoleTerm(garg.term,
				       garg.role,garg.value) ;
      pushAndState(goal,orStateStack.toppos()) ;
   }
}
#endif

//--------------------------------------------------
// try user def

Bool Gem::tryUserDef() {
   Binding* newbinding = uiheap.newBinding() ;
   OrState* ostate = pushOrState() ;
   BindingState* bstate = &(ostate->bindingState) ;

   GaeaClauseRing* gcr ;
   while(True) {
      gcr = gclausePointer()->findGaeaClause(GaeaTermHeadArity(workingGoal.data)) ;
      if(isNullPtr(gcr)) { 
	 popOrState() ;
	 return False ;
      } 

      astate->trycounter++ ; 

      newbinding->init() ;
#ifndef ORIGINAL
      uiheap.resetGArgStack();
#endif
      if(unify(workingGoal,InstGaeaData(gcr->entity.head,newbinding),
	       &uiheap)) {
	 popAndState() ;
	 if(!eq(gcr->entity.body,TrueSymbol)) {
	    pushAndState(InstGaeaData(gcr->entity.body,newbinding),
			 orStateStack.toppos()) ;
	 }
#ifndef ORIGINAL
	 if(uiheap.gargstack.toppos()>0) {
	    // push operation for empty garg
	    pushRoleTerms() ;
	 }
#endif
	 return True ;
      } 
      rewindBindingState(bstate) ;
   }
}

//--------------------------------------------------
// Describe
      
void Gem::describe(gostream& o, Bool detailp = True) const {
   o << "#GEM[" << *name << ":SC=" << stepcount << ":RL=" ;
   switch(runlevel){ 
      case RL_dead: o << "Z" ; break ;
      case RL_discard: o << "D" ; break ;
      case RL_start: o << "s" ; break ;
      case RL_suspend: o << "S" ; break ;
      case RL_wakeup: o << "W" ; break ;
      case RL_run: o << "R" ; break ;
      default : o << (int)runlevel ; break ;
   }
   o << "]" ;

   if(detailp) {
      o << endl ;
      o << "  pid=" << pid << endl ;
      o << "  env=" ; 
      gcell->describe(o,True) ;
      o << "  top goal=" << topGoal << endl ;

      if(isNullPtr(astate)) { 
	 o << "  current goal= unbounded" << endl ;
	 o << "  gclausePointer= unbounded" << endl ;
      } else {
	 o << "  current goal=" ;
	 o << fetchValue(currentGoal(),(UnifyInfoHeap *)&uiheap) << endl ; 
	 o << "  gclausePointer=" ;
	 o << *gclausePointer() << endl; 
      }
      
      o << "  and state stack=" ;
      andStateStack.describe(o,True) ;
      o << "  or state stack=" ;
      orStateStack.describe(o,True) ;
      o << "  uiheap=" ;
      uiheap.describe(o,True) ;
      o << endl ;
   }
}

//--------------------------------------------------
// show gem status

void Gem::showstatus(gostream& o) const {
   o << "#GEM[" << *name << "]" ;
   o << " runlevel = " ;
   switch(runlevel){ 
      case RL_dead: o << "dead" ; break ;
      case RL_discard: o << "discarded" ; break ;
      case RL_start: o << "start" ; break ;
      case RL_suspend: o << "suspended" ; break ;
      case RL_wakeup: o << "wakeup" ; break ;
      case RL_run: o << "running" ; break ;
      default : o << (int)runlevel ; break ;
   }
   o << endl ;

   o << "  Env=" ;
   gcell->showGaeaCellStack(o,True) ;
   
   o << "  TGoal=" << topGoal << endl ;
   
   if(!isNullPtr(astate)) {
      o << "  CGoal=" ;
      o << fetchValue(currentGoal(),(UnifyInfoHeap *)&uiheap) ;
      o << "  [" << currentGoal().data << "|" << *(currentGoal().binding) 
	<< "]" << endl ;
      o << "  GCP=" << *gclausePointer() << endl ;
   } ;
   
   o << "  AndStack=" ; 
   andStateStack.describe(o,True) ;
   o << endl ;

   o << "  OrStack=" ;
   orStateStack.describe(o,True) ;
   o << "============================================================" << endl;
}


//======================================================================
// Gem Heap
//

GemHeap* GemHeapTab ;
UInt GemHeapTabSize = GEMHEAP_SIZE ;
UInt GemHeapTabFactor = GEMHEAP_FACTOR ;

void showGemTable(gostream& ostr) {
   GemHeapTab->describe(ostr,True) ;
}

//--------------------------------------------------
// show gem status

void showGemList(gostream& ostr) {
   GemHeapTab->mutex.lock() ;
   for(UInt i = 0 ; i < SymTab.n ; i++) {
      if(!isNullPtr(SymTab.entry(i)->gem)) {
	 ostr << *(SymTab.entry(i)) << ":" << *(SymTab.entry(i)->gem) << endl ;
      }
   }
   GemHeapTab->mutex.unlock() ;
}

//======================================================================
// Dynamic Loader
//

Bool DLoad(char* filename, char* initfuncname, GaeaData entrysymbol, 
	   Bool sureinitcallp, Bool verbosep) 
{
   if(verbosep) 
      gout << "Dynamic-linking " << filename << "..." ;
   void *handle ;
   if(isNullPtr(filename)) {
      handle=dlopen((char*)NULLPTR,RTLD_GLOBAL|RTLD_LAZY) ; 
   } else {
      String fname = LoadPath.findFullPath(String(filename,False)) ;
      if(isNullPtr((char*)fname)) {
	 if(verbosep)
	    gout << "warning: file not found." << endl ;
	 return False ;
      } else {
	 handle=dlopen((char*)fname,RTLD_GLOBAL|RTLD_LAZY) ; 
      }
   }

   if(isNullPtr(handle)) {
      gerr << dlerror() << endl ;
      return False ;
   } ;
   
   if(isGaeaSymbol(entrysymbol))  {
      _SymVal(entrysymbol)->dlhandle = handle ;
      if(verbosep)
	 gout << "[handle = " << entrysymbol << "]" ;
   }


   if(isNullPtr(initfuncname)) {
      if(verbosep) gout << "done." << endl ;
      return True ;
   } ;

   void* fp = dlsym(handle,initfuncname) ;
   if(isNullPtr(fp)) {
      if(sureinitcallp) {
	 gerr << dlerror() << endl ;
	 return False ;
      } else {
	 if(verbosep) gout << "[no initialize] done." << endl ;
	 return True ;
      }
   }

   Bool (*initfunc)() ;
   initfunc = (Bool (*)())fp ;

   Bool r = (*initfunc)() ;

   if(verbosep && r) 
      gout << "[call " << initfuncname << "()]...done." << endl ;
   
   return True ;

}

//======================================================================
// for debug
      
void dbdGem(Gem* gem) { 
   gem->describe(*(gem->estrm), True) ; 
   for(UInt i = 0 ; i < gem->andStateStack._top ; i++) {
      *(gem->estrm) << i << ":" ;
      dbp((gem->andStateStack[i])->entity.igdata.data) ;
      *(gem->estrm) << endl ;
   }
} 







