// -*- Mode:C++ -*-
//Header:
//File: gem.h
//Title: gaea elementary machine
//Author: NODA Itsuki
//Date: 1998/6/09
//EndHeader:

//ModifyHistory:
// 1998/6/09: Create this file
//EndModifyHistory:

//Memo:
//EndMemo:

#ifndef _gem_h_
#define _gem_h_
////////////////////////////////////////////////////////////////////////

extern "C" {
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
}

#include "gstream.h"
#include "gdata.h"
#include "gbio.h"
#include "unify.h"
#include "gcell.h"

//======================================================================
// Load Path
//

#define DIRSEARCHPATHNAME	"Path"
#define DIRSEPARATOR		'/'

typedef AStack<String> StringStack ;

class DirSearchPath : public StringStack {
   public:
      GaeaMutex mutex ;

      inline DirSearchPath() : StringStack(DIRSEARCHPATHNAME) {} ;

      inline void lock() { mutex.lock() ; } ;
      inline void unlock() { mutex.unlock() ; } ;

      inline void push(const String path) {
	 lock() ;
	 StringStack::push(path) ;
	 unlock() ;
      } ;

      inline String pop() {
	 lock() ;
	 String r = StringStack::pop() ;
	 unlock() ;
	 return r ;
      } ;

      inline void clear() {
	 lock() ;
	 StringStack::reset() ;
	 unlock() ;
      } ;
      
      inline Int _toppos() { return StringStack::toppos() ; } ;
      inline Int toppos() {
	 lock() ;
	 Int r = _toppos() ;
	 unlock() ;
	 return r ;
      } ;

      inline String& _topN(UInt n) { return StringStack::topN(n) ; } ;
      inline String& topN(UInt n) {
	 lock() ;
	 String& r = _topN(n) ;
	 unlock() ;
	 return r ;
      } ;

      inline String& _bottomN(UInt n) { return StringStack::bottomN(n) ; } ;
      inline String& bottomN(UInt n) {
	 lock() ;
	 String& r = _bottomN(n) ;
	 unlock() ;
	 return r ;
      } ;

      String findFullPath(const String filename) ;
} ;	 

extern DirSearchPath LoadPath ;

//======================================================================
// AndState
//

class AndState {
   public:
      InstGaeaData igdata ;
      UInt cutOrLevel ;
      GaeaClausePointer gclpointer ;
      GaeaData workdata ;
      UInt trycounter ;
      
      inline AndState() {
	 init() ;
      } ;
      inline AndState(const AndState& src) {
	 igdata = src.igdata ;
	 cutOrLevel= src.cutOrLevel ;
	 gclpointer = src.gclpointer ;
	 workdata = src.workdata ;
      } ;
      inline AndState(const GaeaData& goal, Binding *binding, UInt cutlevel = 0) {
	 init(goal,binding,cutlevel) ;
      } ;
      inline AndState(const InstGaeaData& goal, UInt cutlevel = 0) {
	 init(goal,cutlevel) ;
      } ; 

      inline AndState& operator=(const AndState& src) {
	 igdata = src.igdata ;
	 cutOrLevel= src.cutOrLevel ;
//	 gclpointer = src.gclpointer ;
	 gclpointer.copy(src.gclpointer) ;
	 workdata = src.workdata ;
	 return *this ;
      } ;

      inline void init() {
	 igdata = InstGaeaData() ;
	 cutOrLevel = 0 ;
	 gclpointer.init() ;
	 workdata.setNoData() ;
      } ;

      inline void init(GaeaData goal, Binding* binding, UInt cutlevel = 0) {
	 init(InstGaeaData(goal,binding), cutlevel) ;
      } ;
      inline void init(const InstGaeaData& goal, UInt cutlevel = 0, 
		       Bool init_gclp_p = True) {
	 igdata = goal ;
	 cutOrLevel = cutlevel ;
	 if(init_gclp_p) gclpointer.init() ;
	 workdata.setNoData() ;
      } ;

      inline void initGaeaClausePointer(GaeaCell* gcell) {
	 gclpointer.init(gcell) ;
      } ;

      inline void copyGaeaClausePointer(const AndState& src) {
	 gclpointer.copy(src.gclpointer) ;
      } ;
      inline void copyGaeaClausePointer(const GaeaClausePointer& gcp) {
	 gclpointer.copy(gcp) ;
      } ;

      inline Bool firsttryp() {
	 return isNullPtr(gclpointer.top) ;
      } ;
	 
      inline void describe(gostream& o, Bool detailp = True) const {
	 o << "#AndState[" << igdata ;
	 o << " ,cutLevel=" << cutOrLevel ;
	 o << " ,gcp=" << gclpointer ;
	 o << " ,work=" << workdata ;
	 o << " ,try=" << trycounter ;
	 o << "]" ;
	 if(detailp) { o << endl ; } ;
      } ;
      inline friend gostream& operator<<(gostream& o, const AndState& astate) {
	 astate.describe(o,False) ;
	 return o ;
      } ;
} ;

typedef SpStack<AndState> AndStateStack ;
typedef SpStackEntry<AndState> AndStateStackEntry ;

//======================================================================
// OrState
//

class OrState {
   public:
      AndStateStack::rewindPos andStateStackPointer ;
      BindingState bindingState ;
      UInt gclausePointerRewindPos ;

      //--------------------------------------------------
      // Constructor

      inline OrState() {
	 andStateStackPointer = AndStateStack::rewindPos(NULLPTR,0) ;
	 gclausePointerRewindPos = 0 ;
      } ;

      inline OrState(const OrState& src) {
	 andStateStackPointer = src.andStateStackPointer ;
	 bindingState = src.bindingState ;
      } ;
      inline OrState(AndStateStack::rewindPos assp,
		     const BindingState& bstate)
      {
	 andStateStackPointer = assp ;
	 bindingState = bstate ;
      } ;
      inline OrState(AndStateStack::rewindPos assp,
		     Binding* binding,
		     UnifyInfoHeap* uiheap) {
	 andStateStackPointer = assp ;
	 bindingState.memory(binding,uiheap) ;
      } ;

      //--------------------------------------------------
      // Let

      inline OrState& operator=(const OrState& src) {
	 andStateStackPointer = src.andStateStackPointer ;
	 bindingState = src.bindingState ;
	 return *this ;
      } ;

      inline OrState& let(AndStateStack::rewindPos assp,
			  Binding* binding,
			  UnifyInfoHeap* uiheap) {
	 andStateStackPointer = assp ;
	 bindingState.memory(binding,uiheap) ;
	 return *this ;
      } ;

      //--------------------------------------------------
      // output

      void describe(gostream& o, Bool detailp = True) const ;

      inline friend gostream& operator<< (gostream& o, const OrState& os) {
	 os.describe(o,False) ;
	 return o ;
      } ;

} ;

typedef AStack<OrState> OrStateStack ;

//======================================================================
// Gaea Signal Entry
//

extern GaeaMutex GaeaSignalMutex ;

#define GEMSIGNALENTRY_MAX 16

class GaeaSignalEntry {
   public:
      Gem* gem ;
      Symbol* symbol ;
      GaeaSignalEntry* prev ;
      GaeaSignalEntry* next ;

      //--------------------------------------------------
      // constructor

      inline GaeaSignalEntry() { init() ; } ;

      //--------------------------------------------------
      // initialize & clear

      inline void init() { clear() ; } ;

      inline void clear() {
	 gem = NULLPTR ;
	 symbol = NULLPTR;
	 prev = NULLPTR ;
	 next = NULLPTR ;
      } ;

      //--------------------------------------------------
      // push signal entry

      void pushSignal(Gem* g, Symbol* sym) ;

      void popSignal() ;

      void broadcast() ;
	 
      void broadcast0() ;
} ;

extern void broadcastSignal(Symbol* symbol) ;

//======================================================================
// GEM Gaea Elementary Machine
//

typedef AHeap<GaeaClausePointer> GaeaClausePointerHeap ;

#define DEFAULTGEMPREFIX	"gem"
#define ANDSTATESTACKINITSIZE	256
#define ORSTATESTACKINITSIZE	64
#define UIHEAPINITSIZE		256
#define GCPHEAPINITSIZE		512

enum VerboseLevel { VBL_silent, VBL_normal, VBL_verbose, VBL_max } ;
extern VerboseLevel DefaultVerboseLevel ;

extern void* GemOverTopLoop(void *g) ;

class Gem {
   public:
      GaeaThread pid ;
      Symbol* name ;
      GaeaCell* gcell ;

      inline GaeaCell* agcell() {
	 if(gcell->gcellStack.bottomp()) return gcell ;
	 else return gcell->gcellStack.top() ;
      } ;

      InstGaeaData topGoal ;

      AndState* astate ;

      InstGaeaData workingGoal ;

      inline InstGaeaData& currentGoal() const {
	 return astate->igdata ;
      } ;
      inline GaeaClausePointer* gclausePointer() const {
	 return &(astate->gclpointer) ;
      } ;

      enum TryStatus { TS_none, TS_call, TS_success, TS_fail, TS_match } ;
      TryStatus tstate ;

      AndStateStack andStateStack ;
      OrStateStack orStateStack ;
      
      UnifyInfoHeap uiheap ;
      
      enum traceLevel { TRC_none, TRC_normal, TRC_kernel, TRC_max } ;
      traceLevel tracelevel ;

      inline void traceon(traceLevel tl) { tracelevel = tl ; } ;
      inline void notrace() { tracelevel = TRC_none ; } ;

      VerboseLevel verbosep ;

      gistream* istrm ;
      gostream* ostrm ;
      gostream* estrm ;

      TimeVal birthTime ;
      TimeVal lastActiveTime ;
      UInt stepcount ;

      enum runLevel { RL_dead, RL_discard, 
		      RL_start, 
		      RL_suspend, RL_wakeup, 
		      RL_run } ;
      runLevel runlevel ;
//      GaeaMutexCondVar runlevel_lock ;
      GaeaSharedMutexCondVar runlevel_lock ;

      GaeaSignalEntry signalEntry[GEMSIGNALENTRY_MAX] ;
      UInt signalEntryN ;
      Symbol* signalSymbol ;

      AndState workAndState ;
      AndState& tmpAndState(const InstGaeaData& goal, UInt cutlevel = 0) {
	 workAndState.init(goal,cutlevel,False) ;
	 return workAndState ;
      } ;


      //--------------------------------------------------
      // Constructor

      inline Gem() { 
	 runlevel = RL_dead ;
	 init() ; } ;
      inline Gem(Symbol* nm, Bool newsymbolp = True) {
	 runlevel = RL_dead ;
	 init(nm, newsymbolp) ;
      } ;
      inline Gem(const char* nm, Bool newsymbolp = True) {
	 runlevel = RL_dead ;
	 init(nm,newsymbolp) ;
      } ;

      //--------------------------------------------------
      // initializer

      inline void init(Symbol* nm = NULLPTR, Bool newsymbolp = True) {
	 if(runlevel == RL_dead) clear(True,nm,newsymbolp) ;
	 else                    clear(False,nm,False) ;
      } ;

      inline void init(const char* nm, Bool newsymbolp = True) {
	 init(FindSymbol(nm),newsymbolp) ;
      } ;

      //--------------------------------------------------
      // clean up

      inline void clear(Bool initializep = False, 
			Symbol* nm = NULLPTR,
			Bool newsymbolp = True) {
	 if(!initializep) {
	    name->gem = NULLPTR ;
	 }

	 if(initializep) name = NULLPTR ;
	 if(!isNullPtr(nm)) name = nm ;
	 if(newsymbolp) {
	    if(!isNullPtr(name)) {
	       name = GenSymbol(name)  ;
	    } else {
	       name = GenSymbol(DEFAULTGEMPREFIX) ;
	    }
	 }
				    
	 if(initializep) {
	    gcell = new GaeaCell(name) ;
	 } else {
	    gcell->clear() ;
	 }

	 gcell->pushGaeaCell(BasicCell) ;
	 gcell->pushGaeaCell(MainCell) ;

	 topGoal = InstGaeaData() ;

	 astate = NULLPTR ;

	 if(initializep) {
	    andStateStack.expand(ANDSTATESTACKINITSIZE,False) ;
	    orStateStack.expand(ORSTATESTACKINITSIZE,False) ;
	    uiheap.init(UIHEAPINITSIZE,False) ;
	 }

	 tracelevel = TRC_none ;
	 verbosep = DefaultVerboseLevel ;

	 istrm = &gin ;
	 ostrm = &gout ;
	 estrm = &gerr ;

	 stepcount = 0 ;
	 birthTime.update() ;
	 lastActiveTime.update() ;

	 if(initializep) runlevel_lock.init(&GaeaSignalMutex) ;
	 runlevel_lock.lock() ;
	 runlevel = RL_discard ;
	 runlevel_lock.unlock() ;
	 for(UInt i = 0 ; i < GEMSIGNALENTRY_MAX ; i++) {
	    signalEntry[i].clear() ;
	 }
	 signalEntryN = 0 ;
	 signalSymbol = SymVal(NullSymbol) ;

//	 pid.set_self_id() ;
	 if(initializep) {
	    pid.create(GemOverTopLoop,this) ;
	 }
	 
	 name->gem = this ;

      } ;

      //--------------------------------------------------
      // And State
      
      inline AndState* pushAndState(const InstGaeaData& newgoal,
				    UInt cutlevel,
				    AndStateStackEntry* next = NULLPTR) 
      {
//	 return andStateStack.push(AndState(newgoal,cutlevel)) ;
	 AndState* as ;
	 if(isNullPtr(next)) {
	    as = andStateStack.pushget() ;
	 } else {
	    as = andStateStack.pushget(next) ;
	 }
	 as->init(newgoal,cutlevel) ;
	 return as ;
      } ;
      inline AndState* pushAndState(const InstGaeaData& newgoal) {
	 return pushAndState(newgoal,0) ;
      } ;
      inline AndState* pushAndState(const GaeaData dt, 
				    const Binding* binding) 
      {
	 return pushAndState(InstGaeaData(dt,binding)) ;
      } ;
      inline AndState* pushAndState(const GaeaData dt, 
				    const Binding* binding,
				    UInt cutlevel)
      {
	 return pushAndState(InstGaeaData(dt,binding),cutlevel) ;
      } ;
      
      inline AndState& popAndState() {
	 AndState* as = &(andStateStack.pop()) ;
	 if(!andStateStack.bottomp()) 
	    andStateStack.top().gclpointer.init() ;
	 return *as ;
      } ;

      inline UInt topposAndStateStack() {
	 return andStateStack.toppos() ;
      } ;

      inline UInt rewindAndStateStack(UInt pos) {
	 return andStateStack._top = pos ;
      } ;

      //--------------------------------------------------
      // Or State

      inline OrState* pushOrState() {
	 /* BindinState becomes too big, so...
	 orStateStack.push(OrState(andStateStack.rewindpos(),
				   memoryBindingState(currentGoal().binding,
						      &uiheap))) ;
	  */ 
	 /* This may still be slow...
	 orStateStack.push(OrState(andStateStack.rewindpos(),
				   currentGoal().binding,
				   &uiheap)) ;
	  */ 
	 OrState* ostate = orStateStack.pushGetTopPtr() ;
	 ostate->let(andStateStack.rewindpos(),
		     currentGoal().binding, &uiheap) ;
		     
	 return &(orStateStack.top()) ;
      } ;
      inline void popOrState() {
	 OrState *ostate = &(orStateStack.pop()) ;
	 andStateStack.rewind(ostate->andStateStackPointer) ;
	 rewindBindingState(ostate->bindingState) ;
      } ;

      //--------------------------------------------------
      // overtoploop

      void overTopLoop() ;
      void start(GaeaData goal, const Gem* mother_gem) ;
      void override_by_self_id() ;

      //--------------------------------------------------
      // Unfold

      enum UnfoldState {
	 US_fail,
	 US_continue,
	 US_succeed,
	 US_max
      } ;

      void load(String filename, GaeaCell* dstcell = NULLPTR, 
		Bool toplevelp = False) ;
      inline void load(char* filename, GaeaCell* dstcell = NULLPTR, 
		       Bool toplevelp = False) { 
	 load(String(filename), dstcell, toplevelp) ; 
      } ;
      void loadloop(gistream& istr, GaeaCell* dstcell = NULLPTR, 
		    Bool toplevelp = False) ;

      void reploop0(gistream& istr = gin, gostream& ostr = gout, 
		    Bool toplevelp = True) ;
      void settopgoal(GaeaData goal, Binding* binding = NULLPTR, 
		      Bool resetp = True) ;
      void UnfoldLoop() ;
      Bool unfoldLoop() ;
      UnfoldState unfold() ;
      Bool tryDef() ;
      BipStatus trySystemDef() ;
      Bool tryUserDef() ;

#ifndef ORIGINAL
      //--------------------------------------------------
      // make role term

      InstGaeaData makeRoleTerm(const InstGaeaData& term,
				Symbol* role, InstGaeaData& value) ;
      void pushRoleTerms() ;
#endif

      //--------------------------------------------------
      // Cell Operation

      inline Bool pushGaeaCell(GaeaCell* gc, GaeaCell* dst) { 
	 dst->pushGaeaCell(gc) ; 
	 return True ;
      } ;
      inline Bool pushGaeaCell(GaeaCell* gc) { return pushGaeaCell(gc, gcell) ;} ;
      inline Bool pushGaeaCell(GaeaData gc, GaeaData src) { 
	 if(isGaeaSymbol(gc) && isGaeaSymbol(src))
	    return pushGaeaCell(_SymVal(gc)->gcell,_SymVal(src)->gcell) ;
	 return False ;
      } ;
      inline Bool pushGaeaCell(GaeaData gc) {
	 if(isGaeaSymbol(gc))
	    return pushGaeaCell(_SymVal(gc)->gcell) ;
	 return False ;
      } ;

      //--------------------------------------------------
      // Suspend
      
      inline void suspend() {
	 runlevel_lock.lock() ;
	 runlevel = RL_suspend ;
	 while(runlevel == RL_suspend) {
	    runlevel_lock.wait() ;
	 } 
	 runlevel_lock.unlock() ;
      } ;
      
      inline Bool timedsuspend(UInt sec, UInt msec = 0, UInt usec = 0) {
	 Bool r = False ;
	 runlevel_lock.lock() ;
	 runlevel = RL_suspend ;
	 runlevel_lock.timedwait(sec,msec,usec) ;
	 if(runlevel != RL_suspend) r = True ;
	 runlevel_lock.unlock() ;
	 return r ;
      } ;
      
      inline Bool resume() {
	 Bool r = False ;
	 runlevel_lock.lock() ;
	 if(runlevel == RL_suspend) {
	    runlevel = RL_run ;
	    runlevel_lock.signal() ;
	    r = True ;
	 }
	 runlevel_lock.unlock() ;
	 return r ;
      } ;

      inline Bool suspendp() {
	 Bool r = False ;
	 runlevel_lock.lock() ;
	 if(runlevel == RL_suspend) r = True ;
	 runlevel_lock.unlock() ;
	 return r ;
      } ;

      //--------------------------------------------------
      // wait signal

      Symbol* waitSignal(Symbol** symlist, UInt n) ;
      Symbol* timedwaitSignal(Symbol** symlist, UInt n,
			       UInt sec = 0, UInt msec = 0, UInt usec = 0) ;

      //--------------------------------------------------
      // Output
      
      void describe(gostream& o, Bool detailp = True) const ;
      void showstatus(gostream& o) const ;

      inline friend gostream& operator<<(gostream& o, const Gem& gem) {
	 gem.describe(o,False) ;
	 return o ;
      } ;

} ;

extern void dbdGem(Gem* gem) ;
//extern void* UnfoldLoop(void* gem) ;
extern void showGemList(gostream& ostr = gout) ;

//======================================================================
// Gem Heap
//

#define GEMHEAP_NAME	"GemHeap"
#ifdef THREADED
#define GEMHEAP_SIZE	8
#define GEMHEAP_FACTOR	2
#else
#define GEMHEAP_SIZE	1
#define GEMHEAP_FACTOR	1
#endif

typedef PRHeap<Gem> GemHeap ;
extern GemHeap* GemHeapTab ;
extern UInt GemHeapTabSize ;
extern UInt GemHeapTabFactor ;


inline Gem* NewGem(Symbol* nm = NULLPTR, Bool newsymbolp = True) {
   Gem* gem = GemHeapTab->get() ;
   gem->clear(False,nm,newsymbolp) ;
   return gem ;
} ;
inline Gem* NewGem(char* nm, Bool newsymbolp = True) {
   return NewGem(FindSymbol(nm),newsymbolp) ;
} ;


inline void DeleteGem(Gem* gem) {
   GemHeapTab->put(gem) ;
} ;

extern void showGemTable(gostream& ostr = gout) ;

//======================================================================
// Dynamic Loader
//

extern Bool DLoad(char* filename, char* initfuncname, GaeaData entrysymbol,
		  Bool sureinitcallp = True, Bool verbosep = False) ;


//======================================================================
// for degub
//

extern void myBreakPoint() ;


////////////////////////////////////////////////////////////////////////
#endif
