// -*- Mode:C++ -*-
//Header:
//File: geometric.h
//Title: Geometric Class
//Author: NODA Itsuki
//Date: 1998/9/11
//EndHeader:

//ModifyHistory:
// 1998/9/11: Create this file
//EndModifyHistory:

//Memo:
//EndMemo:

#ifndef _geometric_h_
#define _geometric_h_
////////////////////////////////////////////////////////////////////////

//======================================================================
// Include Files
//

#include <iostream.h>
extern "C" {
#include <math.h>
} ;
#include "base.h"

//======================================================================
// Utils
//

#ifndef M_PI
#define M_PI        3.14159265358979323846	/* pi */
#endif
#define DEG2RAD	(M_PI/180.0)
#define RAD2DEG	(180.0/M_PI)

inline Float sq(Float v) { return v * v ; } ;

inline Float Sin(Float deg) { return sin(deg * DEG2RAD) ; };
inline Float Cos(Float deg) { return cos(deg * DEG2RAD) ; };
inline Float Atan2(Float y, Float x) { return RAD2DEG * atan2(y,x) ; } ;
inline Float ACos(Float cosval) { return acos(cosval) * RAD2DEG ; } ;

inline Float reducedDegAngle(Float angle) {
   while(True) {
      if(angle > 180.0) 
	 angle -= 360.0 ;
      else if (angle < -180.0)
	 angle += 360.0 ;
      else
	 return angle ;
   }
}

inline Float reducedRadAngle(Float angle) {
   while(True) {
      if(angle > M_PI) 
	 angle -= 2 * M_PI ;
      else if (angle < -M_PI)
	 angle += 2 * M_PI ;
      else
	 return angle ;
   }
}

inline Float flipDegAngle(Float angle, Float base) {
   Float dif = angle - base ;
   Float newdir = base - dif ;
   return reducedDegAngle(newdir) ;
} 
inline Float flipRadAngle(Float angle, Float base) {
   Float dif = angle - base ;
   Float newdir = base - dif ;
   return reducedRadAngle(newdir) ;
} 

//======================================================================
// 2D Coordinate
//

class PPrPos ;
class PPdPos ;

class PPos {
   public:
      Float x ;
      Float y ;

      //--------------------------------------------------
      // constructor
      PPos(Float vx = 0.0, Float vy = 0.0) { x = vx ; y = vy ; } ;
      PPos(const PPos& v) { x = v.x ; y = v.y ; } ;
      
      //--------------------------------------------------
      // set
      inline PPos& set(Float vx, Float vy) { 
	 x = vx ; y = vy ; return *this ;} ;
      inline PPos& set(const PPos& v) { x = v.x ; y = v.y ; return *this ; } ;
      inline PPos& operator=(const PPos& v) { return set(v) ; } ;

      //--------------------------------------------------
      // set rand
      
      inline PPos& setRand(Float lower, Float upper) {
	 x = FltRand(lower,upper) ; y = FltRand(lower,upper) ;
	 return *this ; 
      } ;

      //--------------------------------------------------
      // equal
      inline Bool operator==(const PPos& v) const {
	 return x == v.x && y == v.y ; } ;

      //--------------------------------------------------
      // polar

      inline PPos& setpolarRad(Float radius, Float angle) {
	 x = radius * cos(angle) ;
	 y = radius * sin(angle) ;
	 return *this ;
      } ;
      inline PPos& setpolarDeg(Float radius, Float angle) {
	 x = radius * Cos(angle) ;
	 y = radius * Sin(angle) ;
	 return *this ;
      } ;
      inline Float radAngle() { return atan2(y,x) ; } ;
      inline Float degAngle() { return Atan2(y,x) ; } ;
      inline Float radius() { return sqrt(sq(x)+sq(y)) ; } ;
      inline Float norm() { return radius() ; } ;

      //--------------------------------------------------
      // calc

      inline PPos& operator+=(const PPos& v) { 
	 x += v.x ; y += v.y ; return *this ; } ;
      inline PPos& operator-=(const PPos& v) {
	 x -= v.x ; y -= v.y ; return *this ; } ;

      inline PPos& operator*=(const Float r) {
	 x *= r ; y *= r ; return *this ; } ;

      inline PPos operator+ (const PPos& v) const {
	 return PPos(x + v.x, y + v.y) ; } ;
      inline PPos operator- (const PPos& v) const {
	 return PPos(x - v.x, y - v.y) ; } ;

      inline PPos operator* (const Float& r) const {
	 return PPos(x * r, y * r) ; } ;

      inline PPos operator/ (const Float& r) const {
	 return PPos(x / r, y / r) ; } ;

      inline Float innerproduct(const PPos& v) const {
	 return (x * v.x + y * v.y) ; } ;

      //--------------------------------------------------
      // output

      inline friend ostream& operator<< (ostream& o, const PPos& p){
	 o << "(" << p.x << "," << p.y << ")" ;
	 return o ;
      } ;
} ;

inline PPos operator* (const Float& r, const PPos& p) {
   return PPos(r * p.x, r * p.y) ; } ;
inline PPos operator/ (const Float& r, const PPos& p) {
   return PPos(r / p.x, r / p.y) ; } ;

//======================================================================
// Polar Coordinate
//

class PPrPos { // Polar in Degree
   public:
      Float r ;
      Float a ;
} ;

class PPdPos { // Polar in Degree
   public:
      Float r ;
      Float a ;

      //--------------------------------------------------
      // constructor
      PPdPos(Float radius = 0.0, Float angle = 0.0) { r = radius ; a = angle ; } ;
      PPdPos(const PPdPos& v) { r = v.r ; a = v.a ; } ;
//      PPdPos(const PPos& v) ;
//      PPdPos(const PPrPos& v) ;
      PPdPos(const PPos& target, const PPos& origin, const Float basedir = 0.0) {
	 PPos dpos = target - origin ;
	 r = dpos.radius() ;
	 a = reducedDegAngle(dpos.degAngle() - basedir) ;
      } ;
      //--------------------------------------------------
      // set
      inline PPdPos& set(Float vr, Float va) { 
	 r = vr ; a = va ; return *this ;} ;
      inline PPdPos& set(const PPdPos& v) { r = v.r ; a = v.a ; return *this ; } ;
      inline PPdPos& operator=(const PPdPos& v) { return set(v) ; } ;

      //--------------------------------------------------
      // equal
      inline Bool operator==(const PPdPos& v) const {
	 return r == v.r && a == v.a ; } ;

      //--------------------------------------------------
      // Converter
      inline PPos ppos() { return PPos(r * Cos(a),r * Sin(a)) ; } ;

      //--------------------------------------------------
      // output

      inline friend ostream& operator<< (ostream& o, const PPdPos& p){
	 o << "(" << p.r << "," << p.a << "deg)" ;
	 return o ;
      } ;

} ;      

//======================================================================
// 3D Coordinate
//

class SPos {
   public:
      Float x ;
      Float y ;
      Float z ;

      //--------------------------------------------------
      // constructor

      SPos(Float vx = 0.0, Float vy = 0.0, Float vz = 0.0 ) { 
	 x = vx ; y = vy ; z = vz ; } ;
      SPos(const SPos& v) {
	 x = v.x ; y = v.y ; z = v.z ; } ;
      
      //--------------------------------------------------
      // set

      inline SPos& set(Float vx, Float vy, Float vz) { 
	 x = vx ; y = vy ; z = vz ; return *this ; } ;

      //--------------------------------------------------
      // output

      inline friend ostream& operator<< (ostream& o, const SPos& p){
	 o << "(" << p.x << "," << p.y << "," << p.z << ")" ;
	 return o ;
      } ;
} ;

////////////////////////////////////////////////////////////////////////
#endif




