#ifndef _GSTREAM_H
#define _GSTREAM_H
extern "C" {
#include <strings.h>
#include <unistd.h>
}

#include <iostream.h>
#include <strstream.h>
#include <fstream.h>
#include <iomanip.h>

#include "gthread.h"

typedef ostream &(* _gomanip) (ostream &);
class gtstream;
struct _gomanip2 {
  void (* func) (gtstream &, char *);
  char *arg;
  _gomanip2 (void (* _func) (gtstream &, char *), char *_arg)
    : func (_func), arg (_arg) {}
};

class gstream
{
 private:
  /* Copy constructor disabled.  */
  gstream (const gstream &);
  /* `=' operator disabled.  */
  gstream& operator= (gstream &);

  GaeaMutex mutex;

 protected:
  void lock () { mutex.lock (); }
  void unlock () { mutex.unlock (); }

 public:
  gstream () { mutex.init() ; } 
  virtual ~gstream () { mutex.destroy() ; }
};

class gistream : public virtual gstream
{
 private:
  typedef int iostate;

  // Manipulator handlers.
  void idispatch (void (istream::*) ());
  void idispatch (istream& (istream::*) ());
  int idispatch (int (istream::*) ());
  int idispatch (unsigned int (istream::*) ());

 protected:
  istream *is;

 public:
  gistream (istream *istrm) : is (istrm) {}
  virtual ~gistream () {}

  virtual gistream& get (char *ptr, int len, char delim = '\n')
    {
      lock (); is->get (ptr, len, delim); unlock ();
      return *this;
    }
  virtual gistream& get (unsigned char* ptr, int len, char delim = '\n')
    { return get ((char *)ptr, len, delim); }
  virtual gistream& get (char& c)
    { lock (); is->get (c); unlock (); return *this; }
  virtual gistream& get (unsigned char& c)
    { return get ((char &) c); }
  virtual gistream& getline (char* ptr, int len, char delim = '\n')
    { lock (); is->getline (ptr, len, delim); unlock (); return *this; }
  virtual gistream& getline (unsigned char *ptr, int len, char delim = '\n')
    { return getline ((char *) ptr, len, delim); }
  virtual gistream& read (char *ptr, int n)
    { lock (); is->read (ptr, n); unlock (); return *this; }
  virtual gistream& read (unsigned char *ptr, int n)
    { return read ((char *) ptr, n); }
  virtual int ipfx (int need = 0)
    { lock (); int ret = is->ipfx (need); unlock (); return ret; }
  virtual int ipfx0 () { return idispatch (&istream::ipfx0); }
  virtual int ipfx1 () { return idispatch (&istream::ipfx1); }
  virtual void isfx() { idispatch (&istream::isfx); }
  virtual int get() { return idispatch (&istream::get); }
  virtual int peek() { return idispatch (&istream::peek); }
  virtual int gcount () { return idispatch (&istream::gcount); }
  virtual gistream& ignore (int n=1, int delim = EOF)
    { lock (); is->ignore (n, delim); unlock (); return *this; }
  virtual int sync () { return idispatch (&istream::sync); }
  virtual gistream& seekg (streampos pos)
    { lock (); is->seekg (pos); unlock (); return *this; }
  virtual streampos tellg ()
    { lock (); streampos pos = is->tellg (); unlock (); return pos; }
  virtual gistream& putback (char ch)
    { lock (); is->putback (ch); unlock (); return *this; }
  virtual gistream& operator>> (char *p)
    { lock (); *is >> p; unlock (); return *this; }
  virtual gistream& operator>> (unsigned char *p)
    { return operator>> ((char *) p); }
  virtual gistream& operator>> (char &c)
    { lock (); *is >> c; unlock (); return *this; }
  virtual gistream& operator>> (unsigned char &c)
    { return operator>> ((char &) c); }
  virtual gistream& operator>> (int &i)
    { lock (); *is >> i; unlock (); return *this; }
  virtual gistream& operator>> (long &l)
    { lock (); *is >> l; unlock (); return *this; }
  virtual gistream& operator>> (float &f)
    { lock (); *is >> f; unlock (); return *this; }
  virtual gistream& operator>> (double &d)
    { lock (); *is >> d; unlock (); return *this; }
  // Simulate member functions of ios.
  virtual void clear (iostate state = 0)
    { lock (); is->clear (state); unlock (); }
  virtual int good ()
    { lock (); int ret = is->good (); unlock (); return ret; }
  virtual int eof ()
    { lock (); int ret = is->eof (); unlock (); return ret; }
  virtual int fail ()
    { lock (); int ret = is->fail (); unlock (); return ret; }
  virtual int bad ()
    { lock (); int ret = is->bad (); unlock (); return ret; }
  virtual iostate rdstate ()
    { lock (); int ret = is->rdstate (); unlock (); return ret; }
  virtual operator void* ()
    { return fail () ? (void *) 0 : (void *) (-1); }
  virtual int operator! ()
    { return fail (); }
};

class gostream : public virtual gstream
{
 private:
  // Manipulator handlers.
  void odispatch (void (ostream::*) ());
  void odispatch (ostream& (ostream::*) ());
  int odispatch (int (ostream::*) ());

 protected:
  ostream *os;

 public:
  gostream (ostream *ostrm) : os (ostrm) {}
  virtual ~gostream () {}

  //virtual gostream& gflush () { return *this; }
  //virtual gostream& gendl () { return *this; }
  //virtual gostream& gends () { return *this; }

  // For output.
  virtual int opfx() { return odispatch (&ostream::opfx); }
  virtual void osfx() { odispatch (&ostream::osfx); }
  virtual gostream& flush () { odispatch (&ostream::flush); return *this; }
  virtual gostream& put (const char c)
    { lock (); os->put (c); unlock (); return *this; }
  virtual gostream& write (const char *s, int n)
    { lock (); os->write (s, n); unlock (); return *this; }
  virtual gostream& write (const unsigned char *s, int n)
    { return write ((const char*) s, n);}
  virtual gostream& seekp (streampos p)
    { lock (); os->seekp (p); unlock (); return *this; }
  virtual streampos tellp()
    { lock (); streampos pos = os->tellp (); unlock (); return pos; }
  virtual gostream& operator<< (const char c)
    { lock (); *os << c; unlock (); return *this; }
  virtual gostream& operator<< (const unsigned char c)
    { return (*this) << (char) c; }
  virtual gostream& operator<< (const char *s)
    { lock (); *os << s; unlock (); return *this; }
  virtual gostream& operator<< (const unsigned char *s)
    { return (*this) << (const char *) s; }
  virtual gostream& operator<< (const void *p)
    { lock (); *os << p; unlock (); return *this; }
  virtual gostream& operator<< (const int n)
    { lock (); *os << n; unlock (); return *this; }
  virtual gostream& operator<< (const unsigned int n)
    { lock (); *os << n; unlock (); return *this; }
  virtual gostream& operator<< (const long n)
    { lock (); *os << n; unlock (); return *this; }
  virtual gostream& operator<< (const unsigned long n)
    { lock (); *os << n; unlock (); return *this; }
  virtual gostream& operator<< (const double n)
    { lock (); *os << n; unlock (); return *this; }
  virtual gostream& operator<< (const float n)
    { return operator<< ((const double) n); }
  virtual gostream& operator<< (_gomanip func)
    { lock (); (*func) (*os); unlock (); return *this; }
  virtual gostream& operator<< (_gomanip2 func)
    { return *this; }
  virtual gostream& operator<< (const smanip<int>& dt)
    { lock (); *os << dt ; unlock (); return *this; }
  virtual int precision ()
    { lock (); int ret = os->precision (); unlock (); return ret; }
  virtual int precision (int newp)
    { lock (); int ret = os->precision (newp); unlock (); return ret; }
  virtual long setf (long val)
    { lock (); long ret = os->setf (val); unlock (); return ret; }
  virtual int width ()
    { lock (); int ret = os->width (); unlock (); return ret; }
  virtual int width (int val)
    { lock (); int ret = os->width (val); unlock (); return ret; }

#ifdef _STREAM_COMPAT
  virtual streambuf* ostreambuf () const
    { lock (); streambuf p = os->_strbuf; unlock (); return p; }
#endif
  // Gaea specific objects.
  virtual gostream& operator<< (const GaeaThread &gt)
    { return *this << ((int) gt.id); }
};

class giostream : public gistream, public gostream
{
 protected:
  iostream *ios;

 public:
  giostream (iostream *strm) : gistream (strm), gostream (strm), ios (strm) {}
  virtual ~giostream () {}
};

class gistrstream : public gistream
{
 public:
  gistrstream (const char *p, int n = 0) : gistream (new istrstream (p, n)) {}
  virtual ~gistrstream () { delete is; }
};

class gostrstream : public gostream
{
 public:
  gostrstream () : gostream (new ostrstream ()) {}
  gostrstream (char *cp, int n, int mode = ios::out)
    : gostream (new ostrstream (cp, n, mode)) {}
  virtual ~gostrstream () { delete os; }

  char *str ()
    { lock (); char *p = (((ostrstream *) os)->str) (); unlock (); return p; }
};

class gifstream : public gistream
{
 public:
  gifstream() : gistream (new ifstream ()) {}
  gifstream (int fd) : gistream (new ifstream (fd)) {}
  gifstream (int fd, char *p, int l) : gistream (new ifstream (fd, p, l)) {}
  gifstream (const char *name, int mode = ios::in, int prot = 0664)
    : gistream (new ifstream (name, mode, prot)) {}
  virtual ~gifstream () { delete is; }

  void open (const char *name, int mode = ios::in, int prot = 0664)
    { lock (); ((ifstream *) is)->open (name, mode, prot); unlock (); }
};

class gofstream : public gostream
{
 public:
  gofstream() : gostream (new ofstream ()) {}
  gofstream (int fd) : gostream (new ofstream (fd)) {}
  gofstream (int fd, char *p, int l) : gostream (new ofstream (fd, p, l)) {}
  gofstream (const char *name, int mode = ios::out, int prot = 0664)
    : gostream (new ofstream (name, mode, prot)) {}
  virtual ~gofstream () { delete os; }

  void open (const char *name, int mode = ios::out, int prot = 0664)
    { lock (); ((ofstream *) os)->open (name, mode, prot); unlock (); }
};

class gfstream : public giostream
{
 public:
  gfstream() : giostream (new fstream ()) {}
  gfstream (int fd) : giostream (new fstream (fd)) {}
  gfstream (const char *name, int mode, int prot = 0664)
    : giostream (new fstream (name, mode, prot)) {}
  gfstream (int fd, char *p, int l) : giostream (new fstream (fd, p, l)) {}
  ~gfstream () { delete ios; }

  void open (const char *name, int mode, int prot = 0664)
    { lock (); ((fstream *) ios)->open (name, mode, prot); unlock (); }
};

// Predefined streams.
extern gifstream gin;
extern gofstream gout;
extern gofstream gerr;

#endif  /* _GSTREAM_H */
