%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% builtin-functions
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%======================================================================
% control
%

%--------------------------------------------------
% eq (buggy, please use $unify and $gunify)

$rule(eqcut(X,X),$cut).
$rule(eq(X,X),$true).

%--------------------------------------------------
% not

$rule(not(Goal),$and(Goal,!,fail)).
$fact(not(_)).

%--------------------------------------------------
% true

$fact(true).
$fact(true(true)).

%--------------------------------------------------
% repeat

$rule(repeat,$repeat).

$rule(repeat(Goal),$and(repeat,once(Goal),fail)).

%--------------------------------------------------
% dotimes

$rule(dotimes(I,M),dotimes(I,0,M)).
$rule(dotimes(I,C,M),
	$and($comp_ge(C,M),$cut,$fail)).
$fact(dotimes(C,C,M)).
$rule(dotimes(I,P,M),
	$and($inc(P,N),$cut,dotimes(I,N,M))).

%--------------------------------------------------
% success

$rule(succeed(Goal),Goal).
$rule(succeed(_),$true).

%--------------------------------------------------
% once

$rule(once(Goal),
	$and(Goal,!)).
$rule(once(Goal),
	$cut).	

%--------------------------------------------------
% if

$rule(if(Cond,Then),$and(Cond,!,Then)).
$rule(if(Cond,Then,Else),$and(Cond,!,Then)).
$rule(if(Cond,Then,Else),$and(!,Else)).

%======================================================================
% environment
%======================================================================

%--------------------------------------------------
% environment

$rule(environment(Env),$environment(Env)).

%--------------------------------------------------
% listing

$rule(listing,$and($cut,$listing)).
$rule(listing(Cell),$and($cut,$listing(Cell))).

%--------------------------------------------------
% push,pop
% push/pop cause serious change to gclausepointer, 
% so it should be built-in pred

$rule(push(Cell),$and($cut,$push(Cell))).
$rule(push(Cell,Dst),$and($cut,$push(Cell,Dst))).
$rule(pop(Cell),$pop(Cell)).
$rule(pop(Cell,Src),$pop(Cell,Src)).

$rule(repush(Cell,Dst),
	$and($in_env(Cell,Dst),$remove_cell(Cell,Dst),$push(Cell,Dst),$cut)).
$rule(repush(Cell,Dst),
	$push(Cell,Dst)).
$rule(needs(Cell,Dst),
	$and($in_env(Cell,Dst),$cut)).
$rule(needs(Cell,Dst),
	$push(Cell,Dst)).
$rule(remove(Cell,Dst),
	$and($cut,$remove_cell(Cell,Dst))).
$rule(remove_a_cell(Cell,Dst),
	$and($cut,$remove_a_cell(Cell,Dst))).
$rule(fremove(Cell,Dst),
	$and($in_env(Cell,Dst),$remove_cell(Cell,Dst),$cut)).
$rule(fremove(Cell,Dst),
	$cut).
$rule(fremove_a_cell(Cell,Dst),
	$and($in_env(Cell,Dst),$remove_a_cell(Cell,Dst),$cut)).
$rule(fremove_a_cell(Cell,Dst),
	$cut).
$rule(pushsecond(NewCell),
	$and($cut,pop(Top),push(NewCell),push(Top))).
$rule(pushsecond(NewCell,Dst),
	$and($cut,pop(Top,Dst),push(NewCell,Dst),push(Top,Dst))).


%--------------------------------------------------
% new cell / gencell

$rule(newcell(Cell),$newcell(Cell)).

$rule(gencell(NewCell,Seed),
	$and($cut,$gensym(NewCell,Seed),newcell(NewCell))).

%--------------------------------------------------
% assert

$rule(assert(Fact),asserta(Fact)).
$rule(assert_in(Fact,Dst),asserta_in(Fact,Dst)).
$rule(as(Fact),asserta(Fact)).
$rule(aa(Fact),asserta(Fact)).
$rule(az(Fact),assertz(Fact)).

$rule(asserta(Fact),$asserta_fact(Fact)).
$rule(asserta_in(Fact,Dst),$asserta_fact(Fact,Dst)).

$rule(assertz(Fact),$assertz_fact(Fact)).
$rule(assertz_in(Fact,Dst),$assertz_fact(Fact,Dst)).

%--------------------------------------------------
% with

$rule(with(Cell,Goal),
	$and(!,
	     $environment1(Env1),
	     $or($push(Cell),$and($set_environment1(Env1),fail)),
	     Goal,
	     $environment1(Env2),
	     $or($set_environment1(Env1),$and($set_environment1(Env2),fail)))).

%======================================================================
% misc
%======================================================================

%--------------------------------------------------
% halt

$rule(halt,$halt).

%--------------------------------------------------
% loading

$rule(load(Filename),$load(Filename)).

%--------------------------------------------------
% is form

$rule(is(Var,Form),$is(Var,Form)).

%--------------------------------------------------
% timer

$rule(ctime(time(Sec,USec)),
	$ctime(Sec,USec)).
$rule(timedif(time(BSec,BUSec),time(ESec,EUSec),DUSec),
	$is(DUSec,add(sub(ESec,BSec),div(sub(EUSec,BUSec),1000000.0)))).

%======================================================================
% process
%======================================================================

%--------------------------------------------------
% fork/pid

$rule(fork(Goal,PID),$fork(Goal,PID)).

$rule(pid(PID),$pid(PID)).

%--------------------------------------------------
% sleep

$rule(sleep(forever),$and($cut,$sleep_forever)).
$rule(sleep(Sec),$sleep(Sec)).

%======================================================================
% list operations
%======================================================================

$fact(member(Item,[Item|_])).
$rule(member(Item,[_|Rest]),
	member(Item,Rest)).

$fact(append([],Tail,Tail)).
$rule(append([Head|Body],Tail,[Head|Rest]),
	append(Body,Tail,Rest)).

$rule(reverse(List,Rev),reverse(List,[],Rev)).
$fact(reverse([],Rev,Rev)).
$rule(reverse([Head|Rest],Stack,Rev),
	reverse(Rest,[Head|Stack],Rev)).

$fact(length([],0)).
$rule(length([_|Rest],Len),
	$and(!,length(Rest,Len1),$inc(Len1,Len))).


%======================================================================
% dynamic loader
%======================================================================

$newcell($dynamic_loader_db).
$hide_cell($dynamic_loader_db).

$rule(binload(FileName),
	binload(FileName,"gdlinit")).

$rule(binload(FileName,InitFunc),
	$and(	$dlload(Handle,FileName,InitFunc),
		$assertz_fact(module(Handle,FileName,InitFunc,unsure),
			      $dynamic_loader_db))).

$rule(binload_sure(FileName,InitFunc),
	$and(	$dlload_sure(Handle,FileName,InitFunc),
		$assertz_fact(module(Handle,FileName,InitFunc,sure),
			      $dynamic_loader_db))).

%======================================================================
% fill empty role
%======================================================================

$rule($role(Role,Value,_),
	$and($cv_get_r(Role,Value,no,no),!)).
$rule($role(Role,Value,_),!).

%% $role/2
% $rule($role(Role,Value),
% 	$and($cv_get_r(Role,Value,no,no),!)).
% $rule($role(Role,Value),!).

%% $role/4
% $rule($role(Head,Arity,Role,Value),
% 	$and($cv_get_r(Role,Value,no,no),!)).
% $rule($role(Head,Arity,Role,Value),!).

$write('init.g is loaded').
