// -*- Mode:C++ -*-
//Header:
//File: unify.cc
//Author: NODA Itsuki
//Date: 1998/3/28
//EndHeader:

//ModifyHistory:
// 1998/3/28: Create this file
//EndModifyHistory:

//======================================================================
// Include Files
//
#include "unify.h"

//======================================================================
// InstGaeaData
//
gostream& operator<< (gostream& o, const InstGaeaData& dt) {
   o << "#InstGaeaData[" << dt.data ;
   o << "|" << dt.binding ;
   o << "]" ;
   return o ;
} ;

//======================================================================
// Bind
//
gostream& operator<< (gostream& o, const Bind& bind) {
   o << "#Bind[" << bind.variable ;
   o << " = " << bind.value ;
   o << "]" ;
   return o ;
} ;

//======================================================================
// New Var Entry
//
gostream& operator<< (gostream& o, const NewVarEntry& nventry) {
   o << "#NewVar[" << nventry.variable ;
   o << " -> " << nventry.newvariable ;
   o << "]" ;
   return o ;
} ;

GaeaData newVariable(const InstGaeaData& var, 
		  NewVarList* nvlist,
		  UnifyInfoHeap* uiheap) 
{
   for(NewVarList* nv = nvlist->next ; !(nv->endp()) ; 
					nv = nv->next ) {
      if((eq(nv->element.variable.data, var.data))&&
	 (nv->element.variable.binding == var.binding)) {
	 return nv->element.newvariable ;
      }
   }
   GaeaData newvar = genGaeaVariable(var.data) ;	
   NewVarList* nv = 
      uiheap->pushNewVarList(NewVarEntry(var,newvar),nvlist->next) ;
   nvlist->next = nv ;
   return newvar ;
} ;

//======================================================================
// Binding
//
Bind* Binding::findBind(const GaeaData var) {
   BindList* blist ;
   for(blist = bindlist ; !(blist->endp()) ; blist = blist->next ) {
//      if(eqAsPointer(blist->element.variable, var)) { /* skip tag check */
      /* but, usual eq is faster (checked by gprof) */
      if(eq(blist->element.variable, var)) {
	 return &(blist->element) ;
      }
   }
   return NULLPTR ;
} ;

InstGaeaData Binding::findBindValue(const GaeaData var) {
   Bind* r = findBind(var) ;
   if(r == NULLPTR)
      return InstGaeaData() ;
   else
      return r->value ;
} ;

BindList* Binding::pushBindList(GaeaData& var, InstGaeaData& val, 
				UnifyInfoHeap* uiheap) 
{
   return bindlist = 
      uiheap->pushBindList(Bind(var,val), bindlist) ;
} ;

GBind* Binding::_findGBind(const GaeaData gterm) {
   GBindList* gblist ;
   for(gblist = gbindlist ; !(gblist->endp()) ; gblist = gblist->next ) {
      if(gterm.eq(gblist->element->entity.data)) {
	 return gblist->element ;
      }
   }
   return NULLPTR;
} ;

GBind* Binding::findGBind(const GaeaData gterm, UnifyInfoHeap* uiheap) {
   GBind* r = _findGBind(gterm) ;
   if(isNullPtr(r)) {
      r = uiheap->newGBind() ;
      r->init(gterm,this) ;
      pushGBindList(r,uiheap) ;
   }
   return r ;
} ;

GBindList* Binding::pushGBindList(GBind* gbind,
				  UnifyInfoHeap* uiheap) {
   return gbindlist = 
      uiheap->pushGBindList(gbind, gbindlist) ;
} ;

GBindList* Binding::pushGBindList(const GaeaData& gterm, 
				  UnifyInfoHeap* uiheap) {
   return gbindlist = 
      uiheap->pushGBindList(InstGArgTerm(gterm,this), gbindlist) ;
} ;

gostream& operator<< (gostream& o, const Binding& binding) {
   o << "#Binding[" << (void*)&binding ;
   o << ":" << binding.bindlist->length() ;
   o << ":" << binding.bindlist ;
//   o << "," << binding.newvartable ;
   o << "]" ;
   return o ;
} ;
gostream& operator<< (gostream& o, const Binding* binding) {
   o << "#Binding[" << (void *)binding ;
   if(!isNullPtr(binding)) {
      o << ":" << binding->bindlist->length() ;
   }
   o << "]" ;
   return o ;
} ;

//======================================================================
// Binding
//
void UnifyInfoHeap::describe(gostream& o, Bool detailp = True) const {
   o << "#UIHeap[" ;
//   o << bindListHeap ;
//   o << "," << bindingHeap ;
//   o << "," << newvarListHeap ;
   o << "]" ;
} ;


//======================================================================
// isAssigned
//

Bind* isAssigned(GaeaData var, Binding* binding) { 
   return binding->findBind(var) ;
} ;

InstGaeaData isAssignedIGD(GaeaData var, Binding* binding) {
   return binding->findBindValue(var) ;
} ;

Bool isGround(GaeaData var, Binding* binding) {
   if (isGaeaVariable(var)) {
      Bind* bind = isAssigned(var, binding) ;
      if(!isNullPtr(bind)) 
	 return isGround(bind->value.data, bind->value.binding) ;
      else 
	 return False ;
   } else {
      return True ;
   }
} ; 

Bool isObjective(const GaeaData dt) {
   if(isGAtom(dt)) return True ;
   if(!isGaeaTerm(dt)) return False ;
   for(UInt i = 0 ; i < _TermArity(dt) ; i++) {
      if(!isObjective(_TermArgv(dt,i))) return False ;
   }
   if(_TermGArgP(dt)) {
      for(TermGArg* garg = _TermGArgTop(dt) ; !isNullPtr(garg) ;
	  garg = garg->next) {
	 if(!isObjective(garg->value)) return False ;
      }
   }
   return True ;
}

Bool isObjective(const InstGaeaData& idt) {
   if(!isObjective(idt.data)) return False ;

   if(isGaeaTerm(idt.data) && _TermGArgP(idt.data)) {
      GBind* gb = idt.binding->_findGBind(idt.data) ;
      if(!isNullPtr(gb) && !gb->isSingleRing()) 
	 return False ;
   }

   return True ;
} ;
   

//======================================================================
// Safty Unify (rewind if fail)
//

Bool Unify(InstGaeaData x, InstGaeaData y, UnifyInfoHeap* uiheap) {
   BindingState2 bstate(x.binding,y.binding,uiheap) ;
   Bool r = unify(x,y,uiheap) ;
   if(r) {
      return True ;
   } else {
      rewindBindingState2(bstate) ;
      return False ;
   }
}

//======================================================================
// Unify
//

Bool unify(InstGaeaData x, InstGaeaData y, UnifyInfoHeap* uiheap) {
  _unify_:
   if       (eq(x.data, y.data) && (x.binding == y.binding)) {
      return True ;
   } else if(isGaeaVariable(x.data)) {
      Bind* as = isAssigned(x) ;
      if(!isNullPtr(as)) {
	 x = fetch(as->value) ;
	 goto _unify_ ;
      } else {
	 linkBind(x,y,uiheap) ;
	 return True ;
      }
   } else if(isGaeaVariable(y.data)) {
      Bind* as = isAssigned(y) ;
      if(!isNullPtr(as)) {
	 y = fetch(as->value) ;
	 goto _unify_ ;
      } else {
	 linkBind(y,x,uiheap) ;
	 return True ;
      }
   } else if(!isGaeaTerm(x.data) || !isGaeaTerm(y.data)) {
      return eq(x.data, y.data) ;
   } else {
      return unifyGaeaTerm(x,y,uiheap) ;
   }
} ;

Bool unifyGaeaTerm(const InstGaeaData& x, const InstGaeaData& y, 
		   UnifyInfoHeap* uiheap) {
   Term* xterm = _TermVal(x.data) ;
   Term* yterm = _TermVal(y.data) ;
   if(xterm->head != yterm->head ||
      xterm->arity != yterm->arity)
      return False ;
//!!!!!!!!!!
// 	create data in argument is faster?
//
   Binding* xbinding = x.binding ;
   Binding* ybinding = y.binding ;
   for(Int i = 0 ; i < xterm->arity ; i++) {
      if(!unify(InstGaeaData(xterm->argv(i),xbinding),
		InstGaeaData(yterm->argv(i),ybinding),
		uiheap))
	 return False ;
   }
//!!!!!!!!!!
#ifdef __this_is_slower__
   InstGaeaData xdt ; xdt.binding = x.binding ;
   InstGaeaData ydt ; ydt.binding = y.binding ;
   for(Int i = 0 ; i < xterm->arity ; i++) {
      xdt.data = xterm->argv(i) ;
      ydt.data = yterm->argv(i) ;
      if(!unify(xdt,ydt,uiheap,gcell))
	 return False ;
   }
#endif
//!!!!!!!!!!
   if(xterm->gargp != yterm->gargp) return False ;
   if(xterm->gargp) {
      if(!unifyGArgs(x,
		     xbinding->findGBind(x.data,uiheap),
		     ybinding->findGBind(y.data,uiheap),
		     uiheap))
	 return False ;
   }
   return True ;
}

Bool unifyGArgsOneWay(const InstGaeaData& term,
		      GBind* xgbind, GBind* ygbind, UnifyInfoHeap* uiheap) {
   GBind * xgb = xgbind ;
   while(True) {
      for(TermGArg* xtgarg = _TermGArgTop(xgb->entity.data) ;
	  !isNullPtr(xtgarg) ;
	  xtgarg = xtgarg->next) {
	 Bool empty_rolep = True ;
	 GBind * ygb = ygbind ;
	 while(True) {
	    TermGArg* ytgarg = _TermGArg(ygb->entity.data,xtgarg->role) ;
	    if(!isNullPtr(ytgarg)) {
	       empty_rolep = False ;
	       if(!unify(InstGaeaData(xtgarg->value,xgb->entity.binding),
			 InstGaeaData(ytgarg->value,ygb->entity.binding),
			 uiheap))
		  return False ;
	    }	/* this will be wasted process.  Once the role found,
		   then it needs not continue to check. */
	    ygb = (GBind*)ygb->next ;
	    if(ygbind->endp(ygb)) break ;
	 }
	 if(empty_rolep) {
	    //  operation for retrieve infon from environment
#ifndef ORIGINAL
	    uiheap->pushGArgIfNeed(term, xtgarg->role,
				   InstGaeaData(xtgarg->value,
						xgb->entity.binding)) ;
#else
	    GaeaData value ;
	    if(gcell->getGaeaCellVarValueR(value,xtgarg->role,False,False)) {
	       // retrieveing infon from environment never fail.
	       Unify(InstGaeaData(xtgarg->value,xgb->entity.binding),
		     InstGaeaData(value,uiheap->newBinding()),
		     uiheap, gcell) ;
	    }
#endif
	 }
      }
      xgb = (GBind*)xgb->next ;
      if(xgbind->endp(xgb)) break ;
   }
   return True ;
}

Bool unifyGArgs(const InstGaeaData& term, GBind* xgbind, GBind* ygbind, 
		UnifyInfoHeap* uiheap) {
   if(unifyGArgsOneWay(term, xgbind,ygbind,uiheap) &&
      unifyGArgsOneWay(term, ygbind,xgbind,uiheap)) {
      linkGBind(xgbind,ygbind,uiheap) ;
      return True ;
   } else {
      return False ;
   }
}

void linkBind(InstGaeaData var, InstGaeaData val, UnifyInfoHeap* uiheap) {
   uiheap->saveLinkBindHistory(var.binding) ;
   var.binding->pushBindList(var.data, val, uiheap) ;
} ;

void linkGBind(GBind *xgbind, GBind *ygbind, UnifyInfoHeap* uiheap) {
   uiheap->saveLinkGBindHistory(xgbind,ygbind) ;
   xgbind->joinPrevious(ygbind) ;
} ;

//======================================================================
// fetch
//

const InstGaeaData& fetch(const InstGaeaData& idt) {
   const GaeaData& dt = idt.data ;
   Binding* binding = idt.binding ;
     
   if(isGaeaVariable(dt)) {
      Bind* bind = binding->findBind(dt) ;
      if(bind == NULLPTR) {
	 return idt ;
      } else {
	 return fetch(bind->value) ;
      }
   } else {
      return idt ;
   }
}

//======================================================================
// fetch-value
//

GaeaData fetchValue(const InstGaeaData& idt, UnifyInfoHeap* uiheap) {
   uiheap->newvarListHeap.reset() ;
   NewVarList* nvlist = uiheap->pushNewVarList(NewVarEntry(), NULLPTR) ;
   return fetchValueBody(idt, uiheap, nvlist) ;
} ;

GaeaData fetchValueBody(const InstGaeaData& idt, UnifyInfoHeap* uiheap, 
		     NewVarList* nvlist) 
{
   GaeaData dt = idt.data ;
   Binding* binding = idt.binding ;
   if        (isGaeaVariable(dt)) {
      Bind* bind = isAssigned(dt,binding) ;
      if(isNullPtr(bind)) {
	 return newVariable(idt,nvlist,uiheap) ;
      } else {
	 return fetchValueBody(bind->value, uiheap, nvlist) ;
      }
   } else if (isObjective(idt)) {
      return dt ;
   } else { /* GaeaTerm with Var */
      return fetchValueGaeaTerm(idt, uiheap, nvlist) ;   }
} ;

GaeaData fetchValueGaeaTerm(const InstGaeaData& idt, UnifyInfoHeap* uiheap, 
			    NewVarList* nvlist) 
{
   GaeaData dt = idt.data ;
   Binding* binding = idt.binding ;

   GaeaData newdt = GaeaTerm(_TermHead(dt),_TermArity(dt),_TermGArgP(dt)) ;
   for(UInt i = 0 ; i < dt.val.term->arity ; i++) {
      newdt.val.term->argv(i) = 
	 fetchValueBody(InstGaeaData(dt.val.term->argv(i),binding),
			uiheap, nvlist) ;
   }
   if(_TermGArgP(dt)) {
      GBind* gbind = binding->findGBind(dt,uiheap) ;
      GBind* gb = gbind ;
      while(True) {
	 for(TermGArg* garg = _TermGArgTop(gb->entity.data) ;
	     !isNullPtr(garg) ;
	     garg = garg->next) {
	    Symbol* role = garg->role ;
	    if(isNullPtr(_TermVal(newdt)->_findRole(role))) {
	       GaeaData value = 
		  fetchValueBody(InstGaeaData(garg->value,gb->entity.binding),
				 uiheap, nvlist) ;
	       _TermVal(newdt)->addGArg(role,value) ;
	    }
	 }
	 gb = (GBind*)gb->next ;
	 if(gbind->endp(gb)) break ;
      }
   }

   return newdt ;
}
			      
//======================================================================
// fetch garg
//

InstGaeaData fetchGArg(const InstGaeaData& idt, Symbol* role, 
		       UnifyInfoHeap* uiheap) {
   const GaeaData& term = idt.data ;
   if(isGaeaTerm(term) && _TermGArgP(term)) {
      GBind* gbind = idt.binding->findGBind(term,uiheap) ;
      GBind* gb = gbind ;
      while(True) {
	 for(TermGArg* garg = _TermGArgTop(gb->entity.data) ;
	     !isNullPtr(garg) ;
	     garg = garg->next) {
	    if(garg->role == role)
	       return InstGaeaData(garg->value, gb->entity.binding) ;
	 }
	 gb = (GBind*)gb->next ;
	 if(gbind->endp(gb)) break ;
      }
   }
   // come here if idt is not term, not have garg or not have role

   return InstGaeaData(UserNeverUseSymbol,idt.binding) ;
}

   
   
   

