// -*- Mode:C++ -*-
//Header:
//File: unify.h
//Author: NODA Itsuki
//Date: 1998/3/27
//EndHeader:

//ModifyHistory:
// 1998/3/27: Create this file
//EndModifyHistory:

//Memo:
//EndMemo:

#ifndef _unify_h_
#define _unify_h_
////////////////////////////////////////////////////////////////////////
extern "C" {
#include <string.h>
#include <stdlib.h>
} ;

#include "gdata.h"
#include "gcell.h"

//======================================================================
// bind
//

class UnifyInfoHeap ;
class Binding ;

class InstGaeaData {
   public: 
      GaeaData data ;
      Binding* binding ;

      inline InstGaeaData(const GaeaData d = NullSymbol, 
			  const Binding* b = NULLPTR) 
      {
	 data = d ; binding = (Binding*)b ;
      } ;

      inline void init() {
	 data.init() ;	
	 binding = NULLPTR ;
      } ;

      inline InstGaeaData& operator=(const InstGaeaData& src) {
	 data = src.data ;
	 binding = src.binding ;
	 return *this ;
      } ;

      inline Bool operator==(const InstGaeaData& ref) const {
	 return (data.eq(ref.data)) && (binding == ref.binding) ;
      } ;
     
      friend gostream& operator<< (gostream& o, const InstGaeaData& dt) ;
} ; 

class Bind {
   public:
      GaeaData variable ;
      InstGaeaData value ;

      inline Bind(const GaeaData& var, const InstGaeaData& val) {
	 variable = var ; 
	 value = val ;
      } ;
      inline Bind() {} ;
     
      inline void init() { 
	 variable.init() ;
	 value.init() ;
      } ;

      friend gostream& operator<< (gostream& o, const Bind& bind) ;

} ;

//======================================================================
// G bind
//

typedef InstGaeaData InstGArgTerm ;

typedef Ring<InstGArgTerm> InstGArgTermRing ;

class GBind : public InstGArgTermRing {
   public:

      //--------------------------------------------------
      // constructor

      GBind(const GaeaData& gargterm = NullSymbol, 
	    Binding* binding = NULLPTR) {
	 init(gargterm,binding) ;
      } ;

      GBind(const InstGArgTerm& org) { init(org) ; } ;

      //--------------------------------------------------
      // init

      void init(const GaeaData gargterm = NullSymbol, 
		Binding* binding = NULLPTR) {
	 entity.data = gargterm ;
	 entity.binding = binding ;
	 beSingle() ;
      } ;

      void init(const InstGArgTerm & org) {
	 init(org.data,org.binding) ;
      } ;

} ;

      
//======================================================================
// newvartable
//
class NewVarEntry {
   public:
      InstGaeaData variable ;
      GaeaData newvariable ;

      inline NewVarEntry() {} ;
      inline NewVarEntry(InstGaeaData var, GaeaData newvar) {
	 variable = var ;
	 newvariable = newvar ;
      } ;
      inline NewVarEntry(GaeaData var, Binding* binding, GaeaData newvar) {
	 NewVarEntry(InstGaeaData(var,binding),newvar) ;
      } ;

      inline void init() {
	 variable.init() ;
	 newvariable.init() ;
      } ;

      friend gostream& operator<< (gostream& o, const NewVarEntry& nventry) ;

} ;

typedef SList<NewVarEntry> NewVarList ;

extern GaeaData newVariable(const InstGaeaData& var, 
			 NewVarList* nvlist,
			 UnifyInfoHeap* uiheap) ;


		    

//======================================================================
// binding
//
typedef SList<Bind> BindList ;
typedef SList<GBind *> GBindList ;

class Binding {
   public:
      BindList* bindlist ;
      GBindList* gbindlist ;

      inline Binding() { init() ; } ;

      inline void init() { 
	 bindlist = NULLPTR ; 
	 gbindlist = NULLPTR ;
      } ;

      Bind* findBind(const GaeaData var) ;
      InstGaeaData findBindValue(const GaeaData var) ;

      BindList* pushBindList(GaeaData& var, InstGaeaData& val, 
			     UnifyInfoHeap* uiheap) ;

      GBind* _findGBind(const GaeaData gterm) ;
      GBind* findGBind(const GaeaData gterm, UnifyInfoHeap* uiheap) ;

      GBindList* pushGBindList(GBind* gbind,
			       UnifyInfoHeap* uiheap) ;
      GBindList* pushGBindList(const GaeaData& gterm,
			       UnifyInfoHeap* uiheap) ;

      GaeaData newVariable(GaeaData var, UnifyInfoHeap* uiheap) ;

      friend gostream& operator<< (gostream& o, const Binding& binding) ;
      friend gostream& operator<< (gostream& o, const Binding* binding) ;
} ;

#ifndef ORIGINAL
//======================================================================
// open garg info
//

class GaeaArgEntry {
   public:
      InstGaeaData term ;
      Symbol* role ;
      InstGaeaData value ;

      GaeaArgEntry() {} ;
      GaeaArgEntry(const InstGaeaData& t, Symbol* r, 
		   const InstGaeaData& v) {
	 term = t ;
	 role = r ; value = v ;
      } ;

      void describe(gostream& ostr, Bool detailp = True, 
		    Bool newlinep = True) const {
	 ostr << "#GaeaArgEntry[" << ":" << *role << "=" << value 
	      << "|" << term << "]" ;
	 if(newlinep) ostr << endl ;
      } ;
      friend gostream& operator<< (gostream& ostr, const GaeaArgEntry& dt) {
	 dt.describe(ostr,False,False) ;
	 return ostr ;
      } ;
} ;

typedef AStack<GaeaArgEntry> GaeaArgEntryStack ;

#endif

//======================================================================
// Unify Info Heap
//
typedef SListHeap<Bind> BindListHeap ;

typedef AHeap<GBind> GBindHeap ;
typedef SListHeap<GBind*> GBindListHeap ;

typedef AHeap<Binding> BindingHeap ;

typedef SListHeap<NewVarEntry> NewVarListHeap ;

class LinkBindHistoryEntry {
   public:
      Binding* binding ;
      BindList* bindlist ;

      inline LinkBindHistoryEntry() { 
	 binding = NULLPTR ; 
	 bindlist = NULLPTR ;
      } ;

      inline LinkBindHistoryEntry(const LinkBindHistoryEntry& lb) {
	 binding = lb.binding ;
	 bindlist = lb.bindlist ;
      } ;

      inline LinkBindHistoryEntry(Binding* b, BindList* bl) {
	 binding = b ;
	 bindlist = bl ;
      } ;

      inline friend gostream& operator<< (gostream& o, 
					 const LinkBindHistoryEntry& lb) 
      {
	 o << "#LinkBind[binding=" << lb.binding 
	   << ",bindlist=" << lb.bindlist << "]" ;
	 return o ;
      } ;
} ;
typedef AStack<LinkBindHistoryEntry> LinkBindHistory ;

class LinkGBindHistoryEntry {
   public:
      GBind *xgbind ;
      GBind *ygbind ;

      LinkGBindHistoryEntry(GBind *x = NULLPTR, GBind *y = NULLPTR) {
	 init(x,y) ;
      } ;

      LinkGBindHistoryEntry(const LinkGBindHistoryEntry& lgb) {
	 init(lgb.xgbind, lgb.ygbind) ;
      }

      void init(GBind *x = NULLPTR, GBind *y = NULLPTR) {
	 xgbind = x ; ygbind = y ;
      } ;

      void describe(gostream& ostr, 
		    Bool detailp = True, Bool newlinep = True) const {
	 ostr << "#LinkGBind[x=" << xgbind << ":y=" << ygbind << "]" ;
	 if(newlinep) ostr << endl ;
      } ;

      friend gostream& operator<< (gostream& o, 
				   const LinkGBindHistoryEntry& lgb) {
	 lgb.describe(o,False,False) ;
      } ;

} ;

typedef AStack<LinkGBindHistoryEntry> LinkGBindHistory ;

class UnifyInfoHeap {
   public:
      BindListHeap bindListHeap ;
      GBindHeap gbindHeap ;
      GBindListHeap gbindListHeap ;
      BindingHeap bindingHeap ;
      NewVarListHeap newvarListHeap ;
      LinkBindHistory linkBindHistory;
      LinkGBindHistory linkGBindHistory;
#ifndef ORIGINAL
      GaeaArgEntryStack gargstack ;
#endif

      //--------------------------------------------------
      // init

      inline void init(UInt size = 0, Bool verbosep = False) {
	 if(size>0) {
	    bindListHeap.expand(size,verbosep) ;
	    gbindHeap.expand(size,verbosep) ;
	    gbindListHeap.expand(size,verbosep) ;
	    bindingHeap.expand(size,verbosep) ;
	    newvarListHeap.expand(size,verbosep) ;
	    linkBindHistory.expand(size,verbosep) ;
	    linkGBindHistory.expand(size,verbosep) ;
	 }
      } ;

      //--------------------------------------------------
      // reset

      inline void reset() {
	 bindListHeap.reset() ;
	 gbindHeap.reset() ;
	 gbindListHeap.reset() ;
	 bindingHeap.reset() ;
	 newvarListHeap.reset() ;
	 linkBindHistory.reset() ;
	 linkGBindHistory.reset() ;
      } ;

      //--------------------------------------------------
      // push BindList/GBindList

      inline BindList* pushBindList(const Bind& bind, BindList* blist) {
	 return bindListHeap.push(bind,blist) ;
      } ;

      inline GBindList* pushGBindList(GBind* gbind, GBindList* gblist) {
	 return gbindListHeap.push(gbind,gblist) ;
      } ;

      inline GBindList* pushGBindList(const InstGArgTerm& igterm, 
				      GBindList* gblist) {
	 GBind* gbind = gbindHeap.get() ;
	 gbind->init(igterm) ;
	 return pushGBindList(gbind,gblist) ;
      } ;

      //--------------------------------------------------
      // get new GBind/BInding

      inline GBind* newGBind() {
	 return gbindHeap.get() ;
      } ;

      inline Binding* newBinding() {
	 return bindingHeap.get() ;
      } ;

      //--------------------------------------------------
      // NewVarList

      inline NewVarList* pushNewVarList(NewVarEntry nventry, 
					NewVarList* nvlist) 
      {
	 return this->newvarListHeap.push(nventry,nvlist) ;
      } ;

      //--------------------------------------------------
      // Rewind Pointer

      class rewindPointer {
	 public:
	    UInt bindListHeapPointer ;
	    UInt gbindHeapPointer ;
	    UInt gbindListHeapPointer ;
	    UInt bindingHeapPointer ;
	    UInt linkBindHistoryPointer ;
	    UInt linkGBindHistoryPointer ;

	    //--------------------------------------------------
	    // constructor

	    inline rewindPointer() {
	       bindListHeapPointer = 0 ;
	       gbindHeapPointer = 0 ;
	       gbindListHeapPointer = 0 ;
	       bindingHeapPointer = 0 ;
	       linkBindHistoryPointer = 0 ;
	       linkGBindHistoryPointer = 0 ;
	    } ;
	    inline rewindPointer(const rewindPointer& up) {
	       bindListHeapPointer = up.bindListHeapPointer ;
	       gbindHeapPointer = up.bindListHeapPointer ;
	       gbindListHeapPointer = up.bindListHeapPointer ;
	       bindingHeapPointer = up.bindingHeapPointer ;
	       linkBindHistoryPointer = up.linkBindHistoryPointer ;
	       linkGBindHistoryPointer = up.linkGBindHistoryPointer ;
	    } ;
	    inline rewindPointer(const UInt blisthp, 
				 const UInt gbindhp,
				 const UInt gblisthp,
				 const UInt bindinghp,
				 const UInt lbhistp,
				 const UInt lgbhistp) 
	    {
	       bindListHeapPointer = blisthp ;
	       gbindHeapPointer = gbindhp ;
	       gbindListHeapPointer = gblisthp ;
	       bindingHeapPointer = bindinghp ;
	       linkBindHistoryPointer = lbhistp ;
	       linkGBindHistoryPointer = lgbhistp ;
	    } ;
	    
	    //--------------------------------------------------
	    // operator=

	    inline rewindPointer operator= (const rewindPointer& up) {
	       bindListHeapPointer = up.bindListHeapPointer ;
	       gbindHeapPointer = up.gbindHeapPointer ;
	       gbindListHeapPointer = up.gbindListHeapPointer ;
	       bindingHeapPointer = up.bindingHeapPointer ;
	       linkBindHistoryPointer = up.linkBindHistoryPointer ;
	       linkGBindHistoryPointer = up.linkGBindHistoryPointer ;
	       return *this ;
	    } ;

      } ;

      //--------------------------------------------------
      // rewind

      inline rewindPointer rewindPoint() {
	 return rewindPointer(bindListHeap.toppos(),
			      gbindHeap.toppos(),
			      gbindListHeap.toppos(),
			      bindingHeap.toppos(),
			      linkBindHistory.toppos(),
			      linkGBindHistory.toppos()) ;
      } ;
      inline Bool rewind(const rewindPointer& rp) {
	 return rewind(rp.bindListHeapPointer,
		       rp.gbindHeapPointer,
		       rp.gbindListHeapPointer,
		       rp.bindingHeapPointer,
		       rp.linkBindHistoryPointer,
		       rp.linkGBindHistoryPointer) ; 
      };

      Bool rewind(UInt bindListHeapPointer,
		  UInt gbindHeapPointer,
		  UInt gbindListHeapPointer,
		  UInt bindingHeapPointer,
		  UInt linkBindHistoryPointer,
		  UInt linkGBindHistoryPointer) 
      {
	 bindListHeap.rewind(bindListHeapPointer) ;
	 gbindHeap.rewind(gbindHeapPointer) ;
	 gbindListHeap.rewind(gbindListHeapPointer) ;
	 bindingHeap.rewind(bindingHeapPointer) ;

	 for(Int i = (Int)(linkBindHistory.toppos()) - 1 ;
	     i >= (Int)linkBindHistoryPointer ;
	     i--) {
	    linkBindHistory.table[i].binding->bindlist = 
	       linkBindHistory.table[i].bindlist ;
	 }
	 linkBindHistory.rewind(linkBindHistoryPointer) ;

	 for(Int i = (Int)(linkGBindHistory.toppos()) - 1 ;
	     i >= (Int)linkGBindHistoryPointer ;
	     i--) {
	    LinkGBindHistoryEntry& e = linkGBindHistory.table[i] ;
	    e.xgbind->split(e.ygbind) ;
	 }
	 linkGBindHistory.rewind(linkGBindHistoryPointer) ;
	 return True ;
      } ; 

      //--------------------------------------------------
      // save 

      void saveLinkBindHistory(Binding* b) {
	 linkBindHistory.push(LinkBindHistoryEntry(b,b->bindlist)) ;
      } ;

      void saveLinkGBindHistory(GBind* xgbind, GBind* ygbind) {
	 linkGBindHistory.push(LinkGBindHistoryEntry(xgbind,ygbind)) ;
      } ;


      //--------------------------------------------------
      // garg stack operation

      void resetGArgStack() { gargstack.reset() ; }
      void pushGArgIfNeed(const InstGaeaData& term, Symbol* role, 
			  const InstGaeaData& value) {
	 for(UInt i = 0 ; i < gargstack.toppos() ; i++) {
	    GaeaArgEntry& garg = gargstack.topN(i) ;
	    if((garg.role == role) && (garg.value == value)) 
	       return ;
	 }
	 gargstack.push(GaeaArgEntry(term,role,value)) ;
      }

      //--------------------------------------------------
      // describe

      void describe(gostream& o, Bool detailp = True) const ;

      inline friend gostream& operator<<(gostream& o, 
					const UnifyInfoHeap& uiheap) 
      {
	 uiheap.describe(o,False) ;
	 return o ;
      } ;
} ;	
     
//======================================================================
// Binding State
//
class BindingState {
   public:
      Binding* binding ;
      UnifyInfoHeap* uiheap ;
      BindList* bindlist ;
      GBindList* gbindlist ;
      UInt bindListHeapPtr ;
      UInt gbindHeapPtr ;
      UInt gbindListHeapPtr ;
      UInt bindingHeapPtr ;
      UInt linkBindHistoryPtr ; 
      UInt linkGBindHistoryPtr ; 

      inline BindingState() { 
	 binding = NULLPTR ;
	 uiheap = NULLPTR ;
	 bindlist = NULLPTR ; 
	 gbindlist = NULLPTR ; 
	 bindListHeapPtr = 0 ; 
	 gbindHeapPtr = 0 ; 
	 gbindListHeapPtr = 0 ; 
	 bindingHeapPtr = 0 ;
	 linkBindHistoryPtr = 0 ; 
	 linkGBindHistoryPtr = 0 ; 
      } ;
      inline BindingState(Binding* b, UnifyInfoHeap* uih) {
	 memory(b, uih) ;
      } ;
      inline BindingState(const BindingState& src) {
	 binding = src.binding ;
	 uiheap = src.uiheap ;
	 bindlist = src.bindlist ;
	 gbindlist = src.gbindlist ;
	 bindListHeapPtr = src.bindListHeapPtr ;
	 gbindHeapPtr = src.gbindHeapPtr ;
	 gbindListHeapPtr = src.gbindListHeapPtr ;
	 bindingHeapPtr = src.bindingHeapPtr ;
	 linkBindHistoryPtr = src.linkBindHistoryPtr ; 
	 linkGBindHistoryPtr = src.linkGBindHistoryPtr ; 
      } ;

      inline BindingState operator=(const BindingState& src) {
	 binding = src.binding ;
	 uiheap = src.uiheap ;
	 bindlist = src.bindlist ;
	 gbindlist = src.gbindlist ;
	 bindListHeapPtr = src.bindListHeapPtr ;
	 gbindHeapPtr = src.gbindHeapPtr ;
	 gbindListHeapPtr = src.gbindListHeapPtr ;
	 bindingHeapPtr = src.bindingHeapPtr ;
	 linkBindHistoryPtr = src.linkBindHistoryPtr ; 
	 linkGBindHistoryPtr = src.linkGBindHistoryPtr ; 
	 return *this ;
      } ;

      inline Bool memory(Binding* b, UnifyInfoHeap* uih) {
	 binding = b ;
	 uiheap = uih ;
	 bindlist = binding->bindlist ;
	 gbindlist = binding->gbindlist ;
	 bindListHeapPtr = uiheap->bindListHeap.toppos() ;
	 gbindHeapPtr = uiheap->gbindHeap.toppos() ;
	 gbindListHeapPtr = uiheap->gbindListHeap.toppos() ;
	 bindingHeapPtr = uiheap->bindingHeap.toppos() ;
	 linkBindHistoryPtr = uiheap->linkBindHistory.toppos() ;
	 linkGBindHistoryPtr = uiheap->linkGBindHistory.toppos() ;
	 return True ;
      } ;

      inline Bool rewind() {
	 binding->bindlist = bindlist ;
	 binding->gbindlist = gbindlist ;
	 uiheap->rewind(bindListHeapPtr,gbindHeapPtr,gbindListHeapPtr,
			bindingHeapPtr,linkBindHistoryPtr,
			linkGBindHistoryPtr) ;
	 return True ;
      } ;

      inline void describe(gostream& o, Bool detailp = True) const {
	 o << "#BindingState[" << bindlist << "," << gbindlist 
	   << ":" << bindListHeapPtr 
	   << "," << gbindHeapPtr << "," << gbindListHeapPtr 
	   << "," << bindingHeapPtr << "]" ;
	 if(detailp) {
	    o << endl << "   BindList=" << *bindlist << endl;
	    o << endl << "   GBindList=" << *gbindlist << endl;
	 }
      } ;

      inline friend gostream& operator<< (gostream& o, const BindingState& bs) {
	 bs.describe(o,False) ;
	 return o ;
      } ;
	    
} ;

inline BindingState memoryBindingState(Binding* binding, 
				       UnifyInfoHeap* uiheap) 
{
   return BindingState(binding,uiheap) ;
} ;

inline Bool memoryBindingState(BindingState& bstate, 
			       Binding* binding,
			       UnifyInfoHeap* uiheap) 
{
   return bstate.memory(binding,uiheap) ;
} ;


inline Bool rewindBindingState(BindingState& bstate) 
{
   return bstate.rewind() ;
} ;

inline Bool rewindBindingState(BindingState* bstate) 
{
   return bstate->rewind() ;
} ;
			       

     

//==================================================
// Binding State 2 
//

class BindingState2 {
   public:
      Binding* xbinding ;
      Binding* ybinding ;
      UnifyInfoHeap* uiheap ;
      BindList* xbindlist ;
      BindList* ybindlist ;
      GBindList* xgbindlist ;
      GBindList* ygbindlist ;
      UInt bindListHeapPtr ;
      UInt gbindHeapPtr ;
      UInt gbindListHeapPtr ;
      UInt bindingHeapPtr ;

      inline BindingState2() { 
	 xbinding = NULLPTR ;
	 ybinding = NULLPTR ;
	 uiheap = NULLPTR ;
	 xbindlist = NULLPTR ; 
	 ybindlist = NULLPTR ; 
	 xgbindlist = NULLPTR ; 
	 ygbindlist = NULLPTR ; 
	 bindListHeapPtr = 0 ; 
	 gbindHeapPtr = 0 ; 
	 gbindListHeapPtr = 0 ; 
	 bindingHeapPtr = 0 ;
      } ;
      inline BindingState2(Binding* xb, Binding* yb, UnifyInfoHeap* uih) {
	 memory(xb, yb, uih) ;
      } ;
      inline BindingState2(const BindingState2& src) {
	 xbinding = src.xbinding ;
	 ybinding = src.ybinding ;
	 uiheap = src.uiheap ;
	 xbindlist = src.xbindlist ;
	 ybindlist = src.ybindlist ;
	 xgbindlist = src.xgbindlist ;
	 ygbindlist = src.ygbindlist ;
	 bindListHeapPtr = src.bindListHeapPtr ;
	 gbindHeapPtr = src.gbindHeapPtr ;
	 gbindListHeapPtr = src.gbindListHeapPtr ;
	 bindingHeapPtr = src.bindingHeapPtr ;
      } ;

      inline BindingState2 operator=(const BindingState2& src) {
	 xbinding = src.xbinding ;
	 ybinding = src.ybinding ;
	 uiheap = src.uiheap ;
	 xbindlist = src.xbindlist ;
	 ybindlist = src.ybindlist ;
	 xgbindlist = src.xgbindlist ;
	 ygbindlist = src.ygbindlist ;
	 bindListHeapPtr = src.bindListHeapPtr ;
	 gbindHeapPtr = src.gbindHeapPtr ;
	 gbindListHeapPtr = src.gbindListHeapPtr ;
	 bindingHeapPtr = src.bindingHeapPtr ;
	 return *this ;
      } ;

      inline Bool memory(Binding* xb, Binding* yb, UnifyInfoHeap* uih) {
	 xbinding = xb ;
	 ybinding = yb ;
	 uiheap = uih ;
	 xbindlist = xbinding->bindlist ;
	 ybindlist = ybinding->bindlist ;
	 xgbindlist = xbinding->gbindlist ;
	 ygbindlist = ybinding->gbindlist ;
	 bindListHeapPtr = uiheap->bindListHeap.toppos() ;
	 gbindHeapPtr = uiheap->gbindHeap.toppos() ;
	 gbindListHeapPtr = uiheap->gbindListHeap.toppos() ;
	 bindingHeapPtr = uiheap->bindingHeap.toppos() ;
	 return True ;
      } ;

      inline Bool rewind() {
	 xbinding->bindlist = xbindlist ;
	 ybinding->bindlist = ybindlist ;
	 xbinding->gbindlist = xgbindlist ;
	 ybinding->gbindlist = ygbindlist ;
	 uiheap->bindListHeap.rewind(bindListHeapPtr) ;
	 uiheap->gbindHeap.rewind(gbindHeapPtr) ;
	 uiheap->gbindListHeap.rewind(gbindListHeapPtr) ;
	 uiheap->bindingHeap.rewind(bindingHeapPtr) ;
      } ;

      inline void describe(gostream& o, Bool detailp = True) const {
	 o << "#BindingState2[" << xbindlist << ":" << ybindlist << ":" 
	   << xgbindlist << ":" << ygbindlist << ":"
	   << bindListHeapPtr << "," 
	   << gbindHeapPtr << "," << bindListHeapPtr << "," 
	   << bindingHeapPtr << "]" ;
	 if(detailp) {
	    o << endl << "   XBindList=" << *xbindlist << endl;
	    o << endl << "   YBindList=" << *ybindlist << endl;
	    o << endl << "   XGBindList=" << *xgbindlist << endl;
	    o << endl << "   YGBindList=" << *ygbindlist << endl;
	 }
      } ;

      inline friend gostream& operator<< (gostream& o, const BindingState2& bs) {
	 bs.describe(o,False) ;
	 return o ;
      } ;
	    
} ;

inline BindingState2 memoryBindingState2(Binding* xbinding, 
					 Binding* ybinding, 
					 UnifyInfoHeap* uiheap) 
{
   return BindingState2(xbinding,ybinding,uiheap) ;
} ;

inline Bool memoryBindingState2(BindingState2& bstate, 
				Binding* xbinding,
				Binding* ybinding,
				UnifyInfoHeap* uiheap) 
{
   return bstate.memory(xbinding,ybinding,uiheap) ;
} ;


inline Bool rewindBindingState2(BindingState2& bstate) 
{
   return bstate.rewind() ;
} ;

inline Bool rewindBindingState2(BindingState2* bstate) 
{
   return bstate->rewind() ;
} ;
			       

     

//======================================================================
// isAssigned
//

extern Bind* isAssigned(GaeaData var, Binding* binding) ;
inline Bind* isAssigned(InstGaeaData dt) {
   return isAssigned(dt.data, dt.binding) ; 
};

extern InstGaeaData isAssignedIGD(GaeaData var, Binding* binding) ;
inline InstGaeaData isAssignedIGD(InstGaeaData dt) {
   return isAssignedIGD(dt.data, dt.binding) ; 
};

extern Bool isGround(GaeaData var, Binding* binding) ;
inline Bool isGround(InstGaeaData dt) {
   return isGround(dt.data, dt.binding) ;
};

extern Bool isObjective(const GaeaData dt) ;
extern Bool isObjective(const InstGaeaData& idt) ;

//======================================================================
// unify
//

extern Bool Unify(InstGaeaData x, InstGaeaData y, UnifyInfoHeap* uiheap) ;
extern Bool unify(InstGaeaData x, InstGaeaData y, UnifyInfoHeap* uiheap) ;
extern Bool unifyGaeaTerm(const InstGaeaData& x, const InstGaeaData& y, 
			  UnifyInfoHeap* uiheap) ;
extern Bool unifyGArgsOneWay(const InstGaeaData& term, 
			     GBind* xgbind, GBind* ygbind, 
			     UnifyInfoHeap* uiheap) ;
extern Bool unifyGArgs(const InstGaeaData& term, GBind* xgbind, GBind* ygbind, 
		       UnifyInfoHeap* uiheap) ;
extern void linkBind(InstGaeaData var, InstGaeaData val, 
		     UnifyInfoHeap* uiheap);
extern void linkGBind(GBind* xgbind, GBind* ygbind, UnifyInfoHeap* uiheap) ;

//======================================================================
// fetch
//

extern const InstGaeaData& fetch(const InstGaeaData& idt) ;
inline InstGaeaData fetch(const GaeaData dt, const Binding* binding) {
   return fetch(InstGaeaData(dt,binding)) ;
} 

//======================================================================
// fetch-value
//

extern GaeaData fetchValue(const InstGaeaData& idt, UnifyInfoHeap* uiheap) ;
extern GaeaData fetchValueBody(const InstGaeaData& idt, UnifyInfoHeap* uiheap,
			    NewVarList* nvlist) ;
extern GaeaData fetchValueGaeaTerm(const InstGaeaData& idt, UnifyInfoHeap* uiheap, 
			     NewVarList* nvlist) ;


//======================================================================
// fetch garg
//

extern InstGaeaData fetchGArg(const InstGaeaData& idt, Symbol* role,
			       UnifyInfoHeap* uiheap);




////////////////////////////////////////////////////////////////////////
#endif
