// -*- Mode:C++ -*-
//Header:
//File: thread_util.cc
//Author: NODA Itsuki
//Date: 1998/2/6
//EndHeader:

#include "utility.h"

//======================================================================
// for debug

void nop() { return ; } 

//======================================================================
// SLEEP
//

//==================================================
// USleep(sec,usec)

GaeaMutexCondVar _NeverAwakeMTXCV ;

#if defined(SELECTUSLEEP) || !defined(THREADED)
//--------------------------------------------------
// 		usleep using select
void USleep(const UInt sec, const UInt usec, GaeaMutexCondVar& mcv) {
				// mcv is dummy
   struct timeval time_out ;

   time_out.tv_sec = sec ;
   time_out.tv_usec = usec ;

   select(0,(fd_set *)NULL, (fd_set *)NULL, (fd_set *)NULL, &time_out) ;
}
#else
//--------------------------------------------------
// 		usleep using cond_timedwait
void USleep(const UInt sec, const UInt usec, GaeaMutexCondVar& mcv) {
   mcv.timedlockwait(sec,0,usec) ;
}
#endif

//==================================================
// usleep(usec)

void USleep(const UInt usec, GaeaMutexCondVar& mcv) {
   UInt sec = usec / 1000000 ;
   UInt ausec = usec % 1000000 ;
  
   USleep(sec,ausec,mcv) ;
}

//==================================================
// MSleep(sec,msec)

void MSleep(const UInt sec, const UInt msec, GaeaMutexCondVar& mcv) {
   UInt asec = sec + msec/1000 ;
   UInt ausec = (msec%1000) * 1000 ;

   USleep(asec,ausec,mcv) ;
}

//==================================================
// MSleep(msec)

void MSleep(const UInt msec, GaeaMutexCondVar& mcv) {
   UInt sec = msec/1000 ;
   UInt usec = (msec%1000) * 1000 ;

   USleep(sec,usec,mcv) ;
}

//==================================================
// Sleep(sec)

void Sleep(const UInt sec, GaeaMutexCondVar& mcv) {
   USleep(sec,0,mcv) ;
}

void FSleep(const Float sec, GaeaMutexCondVar& mcv) {
   UInt isec = (UInt)sec ;
   UInt iusec = (UInt)(sec * 1000000) ;

   USleep(isec,iusec,mcv) ;
}

//==================================================
// sleep forever

void SleepForever(GaeaMutexCondVar& mcv) {
#if defined(SELECTUSLEEP) || !defined(THREADED)
   select(0,(fd_set *)NULL, (fd_set *)NULL, (fd_set *)NULL, NULL) ;
#else
   mcv.lockwait() ;
#endif
}


//======================================================================
// File Test
//
// this function alloc new memory in some case

char* fullPathName(const char* filename, const char* path) {
   struct stat s ;
   if(isNullPtr(path) || index(path,DIRSEPCHAR)) {
      if(stat(filename, &s)) {
	 return NULLPTR ;
      } else {
	 return strdup(filename) ;
      }
   } else {
      char buf[FULLPATHNAMESIZE] ;
      const char* dir ; char* token ;
      for(dir = path ; !isNullPtr(dir) && !isNullChar(*dir) ; dir = token) {
	 while(*dir == PATHSEPCHAR) dir++ ;
	 token = index(dir,PATHSEPCHAR) ;
	 Int ld = token - dir ;
	 Int lf = strlen(filename) ;
	 memcpy(buf,dir,ld) ; buf[ld] = DIRSEPCHAR ;
	 memcpy(&(buf[ld+1]),filename,lf) ;
	 buf[ld+1+lf] = NULLCHAR ;
	 if(!stat(buf, &s)) {
	    char* fullpath = new char[ld+1+lf+1] ;
	    memcpy(fullpath,buf,ld+1+lf+1) ;
	    return fullpath ;
	 }
      }
      return NULLPTR ;
   }
}


	 
	 
	 
	 


	 

