// -*- Mode:C++ -*-
//Header:
//File: utility.h
//Author: NODA Itsuki
//Date: 1998/2/6
//EndHeader:

#ifndef _utility_h_
#define _utility_h_

////////////////////////////////////////////////////////////////////////

#define UseMyStringClass

extern "C" {
#include <strings.h>

#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
}

#include "gbtype.h"
#include "gthread.h"
#include "gstream.h"

//======================================================================
// for debug
//

extern void nop() ;

//======================================================================
// Min/Max & Sign
//

#define Min(x,y) ((x>y)?(y):(x))
#define Max(x,y) ((x>y)?(x):(y))
#define Min3(x,y,z) ((x>y)?((y>z)?(z):(y)):(x>z)?(z):(x))
#define Max3(x,y,z) ((x<y)?((y<z)?(z):(y)):(x<z)?(z):(x))
#define ISign(x) ((x>0)?  1 : ((x < 0)? -1 : 0))
#define FSign(x) ((x>0)?  1.0 : ((x < 0)? -1.0 : 0.0))


//======================================================================
// TimeStruct
//

class TimeVal {
   public:
      struct timeval tval ;

      //--------------------------------------------------
      // constructor

      inline TimeVal(Int sec = 0, Int msec = 0, Int usec = 0) {
	 Int us = msec * 1000 + usec ;
	 tval.tv_usec = us % 1000000 ;
	 tval.tv_sec = sec + (us / 1000000) ;
      } ;
      inline TimeVal(const TimeVal& src) {
	 tval = src.tval ;
      } ;

      //--------------------------------------
      // get value

      inline int sec() const {
	 return tval.tv_sec ;
      } ;
      inline int usec() const {
	 return tval.tv_usec ;
      } ;
      inline Int msec() const {
	 return tval.tv_usec / 1000 ;
      } ;
      inline Int rusec() const {
	 return tval.tv_usec % 1000 ;
      } ;
      inline Int MSec() const {
	 return msec() + (sec() * 1000) ;
      } ;
      inline Int USec() const {
	 return usec() + (sec() * 1000000) ;
      } ;

      //--------------------------------------
      // Extract 
      inline void extract(Int& s, Int& ms, Int& us) {
	 s = sec() ;
	 ms = msec() ;
	 us = rusec() ;
      } ;

      //--------------------------------------
      // comparison

      inline Bool operator==(const TimeVal& tv) const {
	 return (tval.tv_sec == tv.tval.tv_sec) 
	    &&  (tval.tv_usec == tv.tval.tv_usec) ;
      } ;

      inline Bool operator>(const TimeVal& tv) const {
	 return (tval.tv_sec > tv.tval.tv_sec) 
	    ||  ((tval.tv_sec == tv.tval.tv_sec) &&
		 (tval.tv_usec > tv.tval.tv_usec)) ;
      } ;

      inline Bool operator>=(const TimeVal& tv) const {
	 return (tval.tv_sec > tv.tval.tv_sec) 
	    ||  ((tval.tv_sec == tv.tval.tv_sec) &&
		 (tval.tv_usec >= tv.tval.tv_usec)) ;
      } ;

      inline Bool operator<(const TimeVal& tv) const {
	 return (tval.tv_sec < tv.tval.tv_sec) 
	    ||  ((tval.tv_sec == tv.tval.tv_sec) &&
		 (tval.tv_usec < tv.tval.tv_usec)) ;
      } ;

      inline Bool operator<=(const TimeVal& tv) const {
	 return (tval.tv_sec < tv.tval.tv_sec) 
	    ||  ((tval.tv_sec == tv.tval.tv_sec) &&
		 (tval.tv_usec <= tv.tval.tv_usec)) ;
      } ;

      //--------------------------------------
      // set value

      inline TimeVal& getTimeOfDay() {
	 gettimeofday(&tval,NULL) ;
	 return *this;
      } ;
      inline TimeVal& update() { return getTimeOfDay() ; } ;

      inline TimeVal& set(UInt sec = 0, UInt msec = 0, UInt usec = 0) {
	 UInt us = msec * 1000 + usec ;
	 tval.tv_usec = us % 1000000 ;
	 tval.tv_sec = sec + (us / 1000000) ;
	 return *this ;
      } ;

      inline TimeVal operator=(const TimeVal& tv) {
	 tval = tv.tval ;
	 return *this ;
      } ;

      //--------------------------------------
      // Inc / dec / diff

      inline TimeVal& inc(const Int sec, const Int msec = 0, 
			  const Int usec = 0) 
      {
	 UInt us = tval.tv_usec + (msec * 1000) + usec ;
	 tval.tv_usec = us % 1000000 ;
	 tval.tv_sec += sec + (us / 1000000) ;
	 return *this ;
      } ;
      inline TimeVal& inc(const TimeVal& x) {
	 return inc(x.sec(),0,x.usec()) ;
      } ;

      inline TimeVal& dec(const Int sec, const Int msec = 0, 
			  const Int usec = 0) 
      {
	 UInt d = (msec * 1000) + usec ;
	 UInt over=0 ;
	 if(d > tval.tv_usec) over++ ;
	 over += d/1000000 ;
	 tval.tv_usec += (over * 1000000 - d) ;
	 tval.tv_sec -= (over + sec) ;
	 return *this ;
      } ;
      inline TimeVal& dec(const TimeVal& x) {
	 return dec(x.sec(),0,x.usec()) ;
      } ;

      inline TimeVal diff(const TimeVal& x) const {
	 TimeVal v = x ;
	 return v.dec(*this) ;
      } ;

      inline UInt diffSec(const TimeVal& x) const {
	 TimeVal v = diff(x) ;
	 return v.sec() ;
      } ;
      inline UInt diffMSec(const TimeVal& x) const {
	 TimeVal v = diff(x) ;
	 return v.MSec() ;
      } ;
      inline UInt diffUSec(const TimeVal& x) const {
	 TimeVal v = diff(x) ;
	 return v.USec() ;
      } ;


      //--------------------------------------
      // output in readable format

      void output_normal_form(gostream& ostr) const {
	 UInt rsec = sec() % 60 ; UInt min = (sec() - rsec)/60 ;
	 UInt rmin = min % 60   ; UInt hour = (min - rmin)/60 ;
	 UInt rhour = hour % 24 ; UInt day = (hour - rhour)/24 ;
	 UInt rday = day % 365 ; 
	 ostr << setw(3) << setfill('0') << rday 
	      << "|" << setw(2) << setfill('0') << rhour 
	      << ":" << setw(2) << setfill('0') << rmin
	      << ":" << setw(2) << setfill('0') << rsec
	      << "." << setw(3) << setfill('0') << msec()
	      << "," << setw(3) << setfill('0') << rusec() ;
      } ;

      //--------------------------------------
      // output

      template <class _Stream_>
      inline void describe0(_Stream_ & ostr, Bool detailp) const {
	 ostr << "#TimeVal[" << sec() << "s " << msec() << "ms " 
	      << usec() << "]"  ;
      } ;

      inline void describe(gostream & ostr, Bool detailp) const {
	 describe0<gostream>(ostr,detailp) ;
      } ;
} ;

//--------------------------------------------------
// operator<<

inline gostream& operator<< (gostream& o, const TimeVal& tv) {
   tv.describe(o,False) ;
   return o ;
} ;
      
      
//======================================================================
// Sleep
//

extern GaeaMutexCondVar _NeverAwakeMTXCV ;

extern void USleep(const UInt sec, const UInt usec, 
		   GaeaMutexCondVar& mcv = _NeverAwakeMTXCV) ;
extern void USleep(const UInt usec, 
		   GaeaMutexCondVar& mcv = _NeverAwakeMTXCV) ;

extern void MSleep(const UInt sec, const UInt msec, 
		   GaeaMutexCondVar& mcv = _NeverAwakeMTXCV) ;
extern void MSleep(const UInt msec, 
		   GaeaMutexCondVar& mcv = _NeverAwakeMTXCV) ;

extern void Sleep(const UInt sec, 
		   GaeaMutexCondVar& mcv = _NeverAwakeMTXCV) ;
extern void FSleep(const Float sec, 
		   GaeaMutexCondVar& mcv = _NeverAwakeMTXCV) ;

extern void SleepForever(GaeaMutexCondVar& mcv = _NeverAwakeMTXCV) ;

//======================================================================
// Random Number
//

#ifdef Solaris
extern "C" {
long random();
int srandom(unsigned seed);
}
#endif

inline void randomize_by_ctime() {
   TimeVal ctime ;
   ctime.getTimeOfDay() ;
   UInt seed = ctime.usec() + ctime.sec() ;
   srandom(seed) ;
} ;

inline Int IntRand(Int max) { return (random() % max) ; } ;
inline Int IntRand(Int min, Int max) { return IntRand(max - min) + min ; } ;

#define FLOATRANDBASE (RAND_MAX/2) 
inline Float FltRand() { 
   return ((Float)IntRand(FLOATRANDBASE))/(Float)FLOATRANDBASE ;
} ;
inline Float FltRand(Float max) { return max * FltRand() ; } ;
inline Float FltRand(Float min, Float max) {
   return FltRand(max - min) + min ;
} ;

//======================================================================
// myString Class
//
#ifndef UseMyStringClass
#include <String.h>
#else
class myString {
   public:
      char* data ;
      UInt length ;
      Bool originalp ;

      inline Int _strlen(const char* s) const {
	 if(isNullPtr(s)) return 0 ;
	 else return strlen(s) ;
      }  ;

      inline myString& _let(char *s) {
	 data = s ;
	 length = _strlen(s) ;
	 originalp = False ;
	 return *this ;
      } ;
      inline myString& _let(myString s) { return _let(s.data) ; } ;

      inline myString& _copy(const char *s) {
	 length = _strlen(s) ;
	 data = new char[length+1] ;
	 if(length>0) strcpy(data,s) ;
	 else data[0] = NULLCHAR ;
	 originalp = True ;
	 return *this ;
      } ;
      inline myString& _copy(const myString s) { return _copy(s.data) ; } ;

      inline myString() { data = NULLPTR ; length = 0 ; originalp = False ; };
      inline myString(const char *dt) { this->_copy(dt) ; } ;
      inline myString(const char *dt, Bool copyp) {
	 if(copyp) {
	    _copy(dt) ;
	 } else {
	    _let(dt) ;
	 }
      } ;
	    
//  inline myString(const myString& dt) { this->_copy(dt) ; } ;
      inline myString(UInt size) {
	 length = size ;
	 data = new char[length] ;
	 originalp = True ;
      } ;
/*
      inline ~myString() {
	 if(originalp) delete data ;
      } ;
      */

      inline Bool operator== (const myString& s) {
	 return !strcmp(data,s.data) ;
      } ;
      inline Bool operator== (const char* s) {
	 return !strcmp(data,s) ;
      } ;
      inline myString& operator= (const myString& s) {
	 return this->_copy(s) ; 
      } ;
      inline myString& operator= (const char* s) {
	 return this->_copy(s) ;
      } ;
      inline operator char* () const { return data ; } ;

      inline myString operator+ (const myString& s) {
	 return *this + s.data ;
      } ;
      inline myString operator+ (const char* s) const {
	 Int dlen = _strlen(data) ;
	 Int slen = _strlen(s) ;
	 Int newlen = dlen + slen + 1 ;
	 myString newstr = myString(newlen) ;
	 strncpy(newstr.data,data,dlen) ;
	 strncpy(&(newstr.data[length]),s,slen) ;
	 newstr.data[newlen-1] = NULLCHAR ;
	 return newstr ;
      } ; 

      inline friend gostream& operator<< (gostream& o, const myString& s) {
	 return o << s.data ;
      } ;
};
typedef myString String ;
#endif

//======================================================================
// Tagged Basic Type
//

#define TBT_NONE_STRING		"(none)"
#define TBT_CHAR_BEGIN_CHAR	'\''
#define TBT_CHAR_END_CHAR	'\''
#define TBT_STR_BEGIN_CHAR	'\"'
#define TBT_STR_END_CHAR	'\"'
#define TBT_FLOAT_PRECISION	3
#define TBT_FLOAT_WIDTH		5


enum BasicTypeTag {
   BTT_none,
   BTT_char,
   BTT_Int,
   BTT_Float,
   BTT_charPtr,
   BTT_pointer,
   BTT_max
} ;

union TaggedBasicTypeVal {
      char	ch ;
      Int	inum ;
      Float	fnum ;
      char*	str ;
      void*	pointer ;
} ;

class TaggedBasicType {
   public:
      BasicTypeTag type ;
      TaggedBasicTypeVal val ;

      //--------------------------------------------------
      // constructor

      inline TaggedBasicType(const TaggedBasicType& src) {
	 set(src) ;
      } ;
      inline TaggedBasicType(void) { set(); } ;
      inline TaggedBasicType(const char ch) { set(ch) ; } ;
      inline TaggedBasicType(const int i) { set(i) ; } ;
      inline TaggedBasicType(const Int i) { set(i) ; } ;
      inline TaggedBasicType(const Float f) { set(f) ; } ;
      inline TaggedBasicType(const char* str) { set(str) ; } ;
      inline TaggedBasicType(const void* ptr) { set(ptr) ; } ;

      //--------------------------------------------------
      // set

      inline void set(const TaggedBasicType& src) {
	 type = src.type ;
	 memcpy(&val,&(src.val),sizeof(TaggedBasicTypeVal)) ;
      } ;
      
      inline void set(void) { 
	 type = BTT_none ; } ;
      inline void set(const char ch) { 
	 type = BTT_char ; val.ch = ch ; } ;
      inline void set(const int i) { set((Int)i) ; } ;
      inline void set(const Int i) { 
	 type = BTT_Int ; val.inum = i ; } ;
      inline void set(const Float f) { 
	 type = BTT_Float ; val.fnum = f ; } ;
      inline void set(const char* str) { 
	 type = BTT_charPtr ; val.str = (char*)str ; } ;
      inline void set(const void* ptr) { 
	 type = BTT_pointer ; val.pointer = (void*)ptr ; } ;

      //--------------------------------------------------
      // operator=

      inline TaggedBasicType& operator=(const TaggedBasicType& src) {
	 set(src) ; return *this ; } ;
      inline TaggedBasicType& operator=(const char ch) {
	 set(ch) ; return *this ; } ;
      inline TaggedBasicType& operator=(const Int i) {
	 set(i) ; return *this ; } ;
      inline TaggedBasicType& operator=(const Float f) {
	 set(f) ; return *this ; } ;
      inline TaggedBasicType& operator=(const char* str) {
	 set(str) ; return *this ; } ;
      inline TaggedBasicType& operator=(const void* ptr) {
	 set(ptr) ; return *this ; } ;

      //--------------------------------------------------
      // eq

      inline Bool eq() const {
	 return (type == BTT_none) ; } ;
      inline Bool eq(const char ch) const { 
	 return (type == BTT_char) && (ch == val.ch) ; } ;
      inline Bool eq(const Int i) const {
	 return (type == BTT_Int) && (i == val.inum) ; } ;
      inline Bool eq(const Float f) const {
	 return (type == BTT_Float) && (f == val.fnum) ; } ;
      inline Bool eq(const char* str) const {
	 return (type == BTT_charPtr) && !strcmp(str,val.str) ; } ;
      inline Bool eq(void * ptr) const {
	 return (type == BTT_pointer)  && (ptr = val.pointer) ; } ;

      inline Bool eq(const TaggedBasicType& v) const {
	 switch(v.type) {
	    case BTT_none : return eq() ;
	    case BTT_char: return eq(v.val.ch) ;
	    case BTT_Int: return eq(v.val.inum) ;
	    case BTT_Float: return eq(v.val.fnum) ;
	    case BTT_charPtr: return eq(v.val.str) ;
	    case BTT_pointer: return eq(v.val.pointer) ;
	    default: return False ;
	 }
      } ;

      //--------------------------------------------------
      // operator==

      inline Bool operator==(const TaggedBasicType& v) const {return eq(v);};
      inline Bool operator==(const char ch) const { return eq(ch) ; } ;
      inline Bool operator==(const Int i) const { return eq(i) ; } ;
      inline Bool operator==(const Float f) const { return eq(f) ; } ;
      inline Bool operator==(const char* str) const { return eq(str) ; } ;
      inline Bool operator==(const void* ptr) const { return eq(ptr) ; } ;

      //--------------------------------------------------
      // convert

      inline Bool convertAsChar(const char* str) {
	 Int len = strlen(str) ;
	 if(*str == TBT_CHAR_BEGIN_CHAR) {
	    str++ ; 
	    char v = *str ; str++ ; 
	    if((*str == TBT_CHAR_BEGIN_CHAR) && (len == 3)) {
	       set(v) ;
	       return True ;
	    } 
	 }
	 return False ;
      } ;

      inline Bool convertAsInt(const char* str) {
	 char *endp ;
	 Int len = strlen(str) ;
	 Int v = strtol(str,&endp,10) ;
	 if(endp == &(str[len])) {
	    set(v) ;
	    return True ;
	 } else {
	    return False ;
	 } 
      } ;
      
      inline Bool convertAsFloat(const char* str) {
	 char *endp ;
	 Int len = strlen(str) ;
	 Float v = strtod(str,&endp) ;
	 if(endp == &(str[len])) {
	    set(v) ;
	    return True ;
	 } else {
	    return False ;
	 } 
      } ;

      //--------------------------------------------------
      // output

      template <class _Stream_>
      inline void describe0(_Stream_ & o, Bool detailp) const {
	 switch(type) {
	    case BTT_none:
	       o << TBT_NONE_STRING ; break ;
	    case BTT_char:
	       o << TBT_CHAR_BEGIN_CHAR << val.ch << TBT_CHAR_END_CHAR ; 
	       break ;
	    case BTT_Int:
	       o << val.inum ; break ;
	    case BTT_Float: 
	       {
		  Int org_prec = o.precision(TBT_FLOAT_PRECISION) ;
		  o.setf(ios::fixed) ;
		  o.width(TBT_FLOAT_WIDTH) ;
		  o << val.fnum ; 
		  o.precision(org_prec) ;
	       }
	       break ;
	    case BTT_charPtr:
	       o << TBT_STR_BEGIN_CHAR << val.str << TBT_STR_END_CHAR ; 
	       break ;
	    case BTT_pointer :
	       o << val.pointer ; break ;
	    default:
	       break ;
	 } ;
      } ;
      inline void describe(gostream& o, Bool detailp) const {
	 describe0<gostream>(o,detailp) ;
      } ;
} ;

//--------------------------------------------------
// operator<<

inline gostream& operator<<(gostream& o, const TaggedBasicType& v) {
   v.describe(o,False) ;
   return o ; 
} 

//======================================================================
// File Test
//

#define FULLPATHNAMESIZE	1024
#define PATHSEPCHAR		':'
#define DIRSEPCHAR		'/'

extern char* fullPathName(const char* filename, const char* path) ;

////////////////////////////////////////////////////////////////////////
#endif
