$newcell(reader).
$push(reader).

%%======================================================================
%% read to token

$rule($read_to_token_list(List,Buffer),
	$and(	$read_to_buffer(Buffer),
		$extract_to_token_list(Buffer,List))).
$rule($extract_to_token_list(Buffer,[Token|Rest]),
	$and(	$get_token(Token,Buffer),!,
		$extract_to_token_list(Buffer,Rest))).
$rule($extract_to_token_list(_,[]),!).

%%======================================================================
%% basic rules

%%--------------------------------------------------
%% grammer

%%------------------------------
%% clause

$rule($parse(clause,Head,Tail,Clause,VarInfo),
	$and(	$parse(form,Head,Body,Clause,VarInfo),
		$unify(Body,['.' | Tail]))).

%%------------------------------
%% form

$rule($parse(form,Head,Tail,Form,VarInfo),
	$parse(variable,Head,Tail,Form,VarInfo)).
$rule($parse(form,Head,Tail,Form,VarInfo),
	$parse(term,Head,Tail,Form,VarInfo)).
$rule($parse(form,Head,Tail,Form,VarInfo),
	$parse(atom,Head,Tail,Form,VarInfo)).

%%------------------------------
%% atom

$rule($parse(atom,[Atom | Tail],Tail,Atom,VarInfo),
	$isatom(Atom)).

%%------------------------------
%% variable

$rule($parse(variable,[var(VarName) | Tail],Tail,Var,VarInfo),
	$varname(VarName,Var,VarInfo)).

$rule($varname(VarName,Var,[var(VarName,Var) | _ ]),!).
$rule($varname(VarName,Var,[_ | Rest]),
	$varname(VarName,Var,Rest)).

%%------------------------------
%% term

$rule($parse(term,Head,Tail,Term,VarInfo),
	$parse(term_normal,Head,Tail,Term,VarInfo)).

$rule($parse(term,Head,Tail,Term,VarInfo),
	$parse(op_prefix,Head,Tail,Term,VarInfo)).
$rule($parse(term,Head,Tail,Term,VarInfo),
	$parse(op_infix,Head,Tail,Term,VarInfo)).
$rule($parse(term,Head,Tail,Term,VarInfo),
	$parse(op_suffix,Head,Tail,Term,VarInfo)).
$rule($parse(term,Head,Tail,Term,VarInfo),
	$parse(list,Head,Tail,Term,VarInfo)).

%%------------------------------
%% normal term

$rule($parse(term_normal,Head,Tail,Term,VarInfo),
	$and(	$unify(Body1,['(' | Body2]),
		$unify(Body3,[')' | Tail]),
		$parse(term_head,Head,Body1,TermHead,VarInfo),
		$parse(term_args,Body2,Body3,ArgList,VarInfo),
		$maketerm(Term,TermHead,ArgList))).

$rule($parse(term_head,[Symbol | Tail],Tail,Symbol,VarInfo),
	$issymbol(Symbol)).

$rule($parse(term_args,Head,Head,[],VarInfo),true).
$rule($parse(term_args,Head,Tail,[Arg],VarInfo),
	$parse(form,Head,Tail,Arg,VarInfo)).
$rule($parse(term_args,Head,Tail,[Arg | Rest],VarInfo),
	$and(	$unify(Body,[',' | Body2]),
		$parse(form,Head,Body,Arg,VarInfo),
		$parse(term_args,Body2,Tail,Rest,VarInfo))).

$rule($maketerm(Term,Head,ArgList),
	$and(	length(ArgList,Len),
		$genterm(Head,Len,Term),
		$assign_termarg(Term,ArgList,0))).
$rule($assign_termarg(Term,[],_),true).
$rule($assign_termarg(Term,[Arg|Rest],N),
	$and(	$term_argv(Term,N,Arg),
		$inc(N,N1),
		$assign_termarg(Term,Rest,N1))).

%%------------------------------
%% prefix op

$rule($parse(op_prefix,[Op | Body], Tail, Term,VarInfo),
	$and(	$op(Op,fx,_),
		$parse(form,Body,Tail,Arg,VarInfo),
		$maketerm(Term,Op,[Arg]))).

%%------------------------------
%% infix op

$rule($parse(op_infix,Head,Tail,Term,VarInfo),
	$and(	$parse(form,Head,Body1,Arg1,VarInfo),
		$unify(Body1,[Op | Body2]),
		$op(Op,xfx,_),
		$parse(form,Body2,Tail,Arg2,VarInfo),
		$maketerm(Term,Op,[Arg1,Arg2]))).

%%------------------------------
%% infix op

$rule($parse(op_suffix,Head,Tail,Term,VarInfo),
	$and(	$parse(form,Head,Body,Arg,VarInfo),
		$unify(Body,[Op | Tail]),
		$op(Op,xf,_),
		$maketerm(Term,Op,[Arg]))).

%%------------------------------
%% operator

$fact($op('-',fx,100)).
$fact($op(':-',fx,100)).

$fact($op('-',xfx,100)).
$fact($op('+',xfx,100)).
$fact($op('*',xfx,100)).
$fact($op('/',xfx,100)).
$fact($op('=',xfx,100)).
$fact($op('&',xfx,100)).
$fact($op(';',xfx,100)).
$fact($op('<',xfx,100)).
$fact($op('>',xfx,100)).
$fact($op(':-',xfx,100)).





		


		

$pop(_).