// -*- Mode:C++ -*-
//Header:
//File: bipred_gsc.cc
//Author: NDOA, Itsuki
//Date: 1999/06/10
//

//ModifyHistory:
// 1999/06/10: Create this file
//EndModifyHistory:

#include <string.h>
#include "bip_util.h"
#include "GSPlayer.h"

//======================================================================
// Initialize (external)
//

static Bool bipInit() ;

extern "C" {
    Bool gdlinit() {
        static Bool initializedp = False ;
        if(!initializedp) {
            bipInit() ;
            initializedp = True ;
            return True ;
        } else {
            return False ;
        }
    } ;
} ;

//======================================================================
// create gsplayer
//

ExtObjTypeID EOT_GSPlayer ;
#define isGSPlayer(gdt) isGaeaExtObject(EOT_GSPlayer,gdt)

GaeaData GsSymError ;

//--------------------------------------------------
// create

BipStatus gs_create_player(Gem* gem) {
    /*   $gs$create_player(-Player,+TeamName,+SHost,+SPort)  */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,4) ;

    ARGVAL_VAR(vplayer,gem,icgoal,0) ;
    ARGVAL_CHARPTR(teamname,gem,icgoal,1) ;
    ARGVAL_CHARPTR(shost,gem,icgoal,2) ;
    ARGVAL_INT(sport,gem,icgoal,3) ;

    GSPlayer* player = new GSPlayer ;
    player->init(ScString(teamname),shost,sport,ScFalse) ;

    if(unifyWith(GaeaExtObject(EOT_GSPlayer,player),vplayer,gem))
        return bipSuccess(gem) ;
    else
        return bipCutFail(gem) ;
}

//--------------------------------------------------
// init

BipStatus gs_send_init(Gem* gem) {
    /*   $gs$send_init(+Player) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;

    ScBool r = player->sendInit(5.0,ScFalse) ;

    if(r)
        return bipSuccess(gem) ;
    else 
        return bipError(gem,"Can't send init command.") ;
}

//--------------------------------------------------
// init (goalie)

BipStatus gs_send_ginit(Gem* gem) {
    /*   $gs$send_ginit(+Player) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;

    ScBool r = player->sendInit(5.0,ScTrue) ;

    if(r)
        return bipSuccess(gem) ;
    else 
        return bipError(gem,"Can't send goalie init command.") ;
}


//--------------------------------------------------
// re-init (half-time)

BipStatus gs_half_time_reset(Gem* gem) {
    /*   $gs$half_time_reset(+Player) */
    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    player->halfTimeReset() ;
    return bipSuccess(gem);
}


//======================================================================
// sensor information
//

GaeaData GsSymSee ;
GaeaData GsSymHear ;
GaeaData GsSymSenseBody ;

BipStatus gs_recv_scan_sinfo(Gem* gem) {
    /*   $gs$recv_scan_sinfo(+Player,-Type) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_VAR(vtype,gem,icgoal,1) ;

    ScBool r = player->recvScanSInfo() ;

    GaeaData type = NullSymbol;
    if(r) {
        switch(player->sinfo.type) {
            case SC_SIT_see : { type = GsSymSee ; break ; } 
            case SC_SIT_hear : { type = GsSymHear ; break ; } 
            case SC_SIT_senseBody : { type = GsSymSenseBody ; break ; } 
        }
    }

    if(r && unifyWith(type,vtype,gem))
        return bipSuccess(gem) ;
    else 
        return bipError(gem,"Can't receive correct sensor information.") ;
}

//======================================================================
// basic command
//

//--------------------------------------------------
// move

BipStatus gs_move(Gem* gem) {
    /*   $gs$move(+Player,+X,+Y) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,3) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_FLT(x,gem,icgoal,1) ;
    ARGVAL_FLT(y,gem,icgoal,2) ;

    ScBool r = player->move(x,y) ;

    if(r)
        return bipSuccess(gem) ;
    else 
        return bipError(gem,"Can't send move command.") ;
}

//V.1.06
//--------------------------------------------------
// change view

GaeaData GsSymHigh ;
GaeaData GsSymLow ;
GaeaData GsSymWide ;
GaeaData GsSymNormal ;
GaeaData GsSymNarrow ;

BipStatus gs_change_view(Gem* gem) {
    /*   $gs$change_view(+Player,+Width,+Quality) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,3) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_SYM(width,gem,icgoal,1) ;
    ARGVAL_SYM(quality,gem,icgoal,2) ;

    ScViewWidth w ;
    if(width == SymVal(GsSymNarrow)) 
       w = SC_VW_narrow ;
    else if(width == SymVal(GsSymNormal)) 
       w = SC_VW_normal ;
    else if(width == SymVal(GsSymWide)) 
       w = SC_VW_wide ;
    else
       return bipError(gem,"Illegal view width mode") ;

    ScViewQuality q ;
    if(quality == SymVal(GsSymHigh)) 
       q = SC_VQ_high ;
    else if(quality == SymVal(GsSymLow)) 
       q = SC_VQ_low ;
    else
       return bipError(gem,"Illegal view quality mode") ;

    player->sendChangeView(w,q) ;

    return bipSuccess(gem) ;

}

//--------------------------------------------------
// turn

BipStatus gs_turn(Gem* gem) {
    /*   $gs$turn(+Player,+Moment) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_FLT(moment,gem,icgoal,1) ;

    ScBool r = player->turn(moment) ;

    if(r)
        return bipSuccess(gem) ;
    else 
        return bipError(gem,"Can't send turn command.") ;
}

//--------------------------------------------------
// dash

BipStatus gs_dash(Gem* gem) {
    /*   $gs$dash(+Player,+Power) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_FLT(power,gem,icgoal,1) ;

    ScBool r = player->dash(power) ;

    if(r)
        return bipSuccess(gem) ;
    else 
        return bipError(gem,"Can't send turn command.") ;
}

//--------------------------------------------------
// kick

BipStatus gs_kick(Gem* gem) {
    /*   $gs$kick(+Player,+Power,+Dir) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,3) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_FLT(power,gem,icgoal,1) ;
    ARGVAL_FLT(dir,gem,icgoal,2) ;

    ScBool r = player->kick(power,dir) ;

    if(r)
        return bipSuccess(gem) ;
    else 
        return bipError(gem,"Can't send kick command.") ;
}

//--------------------------------------------------
// catch

BipStatus gs_catch(Gem* gem) {
    /*   $gs$catch(+Player,+Dir) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_FLT(dir,gem,icgoal,1) ;

    ScBool r = player->catchBall(dir) ;

    if(r)
        return bipSuccess(gem) ;
    else 
        return bipError(gem,"Can't send catch command.") ;
}

// 1.04o
BipStatus gs_make_catch(Gem* gem) {
    //   $gs$make_catch(+Player)

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;
    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;

    player->commandLock();
    if (player->distanceToBall() > SC_CATCHABLE_AREA_L) {
        player->commandUnlock();
        return bipError(gem, "Not close enough to catch!");
    }
    ScFloat dir = player->ballAngleRelBodyDir();
    player->commandUnlock();
    
    if(player->catchBall(dir))
        return bipSuccess(gem) ;
    else 
        return bipError(gem,"Can't send catch command.") ;
}


//--------------------------------------------------
// turn neck

BipStatus gs_turn_neck(Gem* gem) {
    /*   $gs$turn_neck(+Player,+Power) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_FLT(power,gem,icgoal,1) ;

    ScBool r = player->turnNeck(power) ;

    if(r)
        return bipSuccess(gem) ;
    else 
        return bipError(gem,"Can't send turn_neck command.") ;
}

//--------------------------------------------------
// say

BipStatus gs_say(Gem* gem) {
    /*   $gs$say(+Player,+Message) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_CHARPTR(msg,gem,icgoal,1) ;

    ScBool r = player->say(msg) ;

    if(r)
        return bipSuccess(gem) ;
    else 
        return bipError(gem,"Can't send say command.") ;
}

//--------------------------------------------------
// say list

Bool GsListToBufferBody(Gem* gem, GSPlayer* player, 
                        InstGaeaData& idt,  ScSubString& buffer) {
   
    if(eq(idt.data,NullSymbol)) {
        buffer.terminate() ;
        return True ;
    } 

    if(!isConsTerm(idt.data)) return False ;

    InstGaeaData car = fetch(InstGaeaData(_ConsCar(idt.data),
                                          idt.binding)) ;
    int n ;

    if(isGaeaInteger(car.data)) {
        sprintf(buffer.tail,"i%ld%n",IntVal(car.data),&n) ;
    } else if(isGaeaFloat(car.data)) {
        sprintf(buffer.tail,"f%.2lf%n",FltVal(car.data),&n) ;
    } else if(isGaeaSymbol(car.data)) {
        sprintf(buffer.tail,"s%s%n",(char*)_SymVal(car.data)->name,&n) ;
    } else {
        *(gem->estrm) << "Error: unsuported data for say list:" 
                      << car.data << endl ;
        return False ;
    }
    sprintf(&(buffer.tail[n]),"%c",GS_SAY_SEPCHAR) ;
    buffer.tail = &(buffer.tail[n+1]) ;
   
    InstGaeaData cdr = fetch(InstGaeaData(_ConsCdr(idt.data),
                                          idt.binding)) ;

    return GsListToBufferBody(gem, player, cdr, buffer) ;
}

Bool GsListToBuffer(Gem* gem, GSPlayer* player, 
                    InstGaeaData& idt,  ScSubString& buffer) {
    int n ;
    sprintf(buffer.tail,"%s%c%n",(char*)(player->teamname),GS_SAY_SEPCHAR,&n) ;
    buffer.tail = &(buffer.tail[n]) ;
    return GsListToBufferBody(gem, player, idt, buffer) ;
} 

BipStatus gs_say_list(Gem* gem) {
    /*   $gs$say_list(+Player,+List) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_VAR(list,gem,icgoal,1) ;

    ScSubString buf ;
    buf.set(player->sayBuffer.head,player->sayBuffer.head) ;

    Bool r = GsListToBuffer(gem,player,list,buf) ;
    if(r) r = player->say(buf) ;
   
    if(r)
        return bipSuccess(gem) ;
    else 
        return bipError(gem,"Can't send say list command.") ;
}

//--------------------------------------------------
// scan teammate message

GaeaData GsSymOpponent ;
GaeaData GsSymTeammate ;

Bool GsScanBufferToListBody(Gem* gem, GSPlayer* player,
                            ScSubString& buffer, const InstGaeaData& list) {
    if(buffer.isNullStr()) {
        return unifyWith(NullSymbol,list,gem) ;
    } 

    Binding* newbinding = gem->uiheap.newBinding() ;
    if(!unify(list,InstGaeaData(ConsTerm,newbinding),&(gem->uiheap)))
        return False ;

    char c = buffer.popChar() ;
    GaeaData car ;
    switch(c) {
        case 'i' : {
            ScInt v ;
            if(!buffer.popIntVal(v)) return False ;
            car = GaeaInteger(v) ;
            break ;
        }
        case 'f' : {
            ScFloat v ;
            if(!buffer.popFloatVal(v)) return False ;
            car = GaeaFloat(v) ;
            break ;
        }
        case 's' : {
            ScSubString sym = buffer.popUntil(GS_SAY_SEPCHAR) ;
            char tailc = *(sym.tail) ;
            *(sym.tail) = SC_NULLCHR ;
 	    String tmpstr ; tmpstr._let(sym.head) ;
 	    car = GaeaSymbol(tmpstr) ;
            *(sym.tail) = tailc ;
            break ;
        }
        default : return False ;
    }
    buffer.popUntil(GS_SAY_SEPCHAR) ;
    buffer.popChar() ;   /* skip SEPCHAR */

    if(!unifyWith(car,InstGaeaData(ConsCarVar,newbinding),gem))
        return False ;

    return GsScanBufferToListBody(gem,player,buffer,
                                  InstGaeaData(ConsCdrVar,newbinding)) ;
   
} 

Bool GsScanBufferToList(Gem* gem, GSPlayer* player,
                        ScSubString& buffer, InstGaeaData& list) {
    if(!buffer.hasPrefix(player->teamname)) {
        return unifyWith(GsSymOpponent,list,gem) ;
    }
    ScSubString header = buffer.popUntil(GS_SAY_SEPCHAR) ;
    if(!(header == player->teamname)) {
        return unifyWith(GsSymOpponent,list,gem) ;
    }
    buffer.ignoreChar() ;
    return GsScanBufferToListBody(gem,player,buffer,list) ;
} 



BipStatus gs_scan_msg_list(Gem* gem) {
    /*   $gs$scan_msg_list(+Player,+Buffer,-List) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,3) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_VAR(vbuf,gem,icgoal,1) ;
    ARGVAL_VAR(vlist,gem,icgoal,2) ;

    if(!isGaeaPointer(vbuf.data)) 
        return bipError(gem,"Illegal buffer") ;

    ScSubString buf = *((ScSubString*)_PtrVal(vbuf.data)) ;

    Bool r = GsScanBufferToList(gem,player,buf,vlist) ;

    if(r)
        return bipSuccess(gem) ;
    else 
        return bipError(gem,"Can't scan message to list.") ;
}


//======================================================================
// get information
//

// v1.04o
GaeaData GsSymMe ;
GaeaData GsSymBall ;

//--------------------------------------------------
// side / unum

GaeaData GsSymLeft ;
GaeaData GsSymRight ;
GaeaData GsSymUnknown ;

// v1.04o
GaeaData GsSymRotLeft ;
GaeaData GsSymRotRight ;
GaeaData GsSymRotClosest ;
GaeaData GsSymRotFurthest ;
GaeaData GsSymRotAvoid ;


inline GaeaData& gsSideSymbol(const ScSide& side) {
    switch(side) {
        case SC_S_left : return GsSymLeft ;
        case SC_S_right : return GsSymRight ;
        default : return GsSymUnknown ;
    }
}

BipStatus gs_side_unum(Gem* gem) {
    /*   $gs$side_unum(+Player,-Side,-UNum) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,3) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_VAR(vside,gem,icgoal,1) ;
    ARGVAL_VAR(vunum,gem,icgoal,2) ;

    if(unifyWith(gsSideSymbol(player->side),vside,gem) &&
       unifyWith(GaeaInteger(player->unum),vunum,gem)) 
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}

//--------------------------------------------------
// time

BipStatus gs_cycle(Gem* gem) {
    /*   $gs$cycle(+Player,-SCycle,-VCycle,-ACycle,-BCycle,-ECycle) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,6) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_VAR(vscyc,gem,icgoal,1) ;
    ARGVAL_VAR(vvcyc,gem,icgoal,2) ;
    ARGVAL_VAR(vacyc,gem,icgoal,3) ;
    ARGVAL_VAR(vbcyc,gem,icgoal,4) ;
    ARGVAL_VAR(vecyc,gem,icgoal,5) ;

    if(unifyWith(GaeaInteger(player->scycle.count),vscyc,gem) &&
       unifyWith(GaeaInteger(player->vcycle.count),vvcyc,gem) &&
       unifyWith(GaeaInteger(player->acycle.count),vacyc,gem) &&
       unifyWith(GaeaInteger(player->bcycle.count),vbcyc,gem) &&
       unifyWith(GaeaInteger(player->ecycle.count),vecyc,gem))
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}

//--------------------------------------------------
// set/get simulation tic

BipStatus gs_tic(Gem* gem) {
    /*   $gs$tic(+Player,-Tic) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_VAR(vtic,gem,icgoal,1) ;

    if(unifyWith(GaeaInteger(player->simulationTic.msec()),vtic,gem))
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}

BipStatus gs_set_tic(Gem* gem) {
    /*   $gs$set_tic(+Player,+Tic) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_INT(tic,gem,icgoal,1) ;

    player->simulationTic.set(0,tic,0) ;

    return bipSuccess(gem) ;
}

//--------------------------------------------------
// stamina

BipStatus gs_stamina(Gem* gem) {
    /*   $gs$stamina(+Player,-Stamina,-Effort) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,3) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_VAR(vstamina,gem,icgoal,1) ;
    ARGVAL_VAR(veffort,gem,icgoal,2) ;

    player->estimateBodyStateMutex.lock();
    ScFloat stamina = player->estimateBodyState.stamina;
    ScFloat effort = player->estimateBodyState.effort;
    player->estimateBodyStateMutex.unlock();
    
    if(   unifyWith(GaeaFloat(stamina),vstamina,gem)  &&
          unifyWith(GaeaFloat(effort),veffort,gem))
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
    
    //if(unifyWith(GaeaFloat(player->actualBodyState.stamina),vstamina,gem) &&
    //   unifyWith(GaeaFloat(player->actualBodyState.effort),veffort,gem))
    //    return bipSuccess(gem) ;
    //else 
    //    return bipCutFail(gem) ;
}

// 1.04o
GaeaData GsSymActionNone;
GaeaData GsSymActionKick;
GaeaData GsSymActionTurn;
GaeaData GsSymActionDash;
GaeaData GsSymActionCatch;
GaeaData GsSymActionError;

// 1.04o
//--------------------------------------------------
// last action
BipStatus gs_last_action(Gem* gem) {
    //   $gs$last_action(+Player, -ActionType)
    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_VAR(actionType_,gem,icgoal,1) ;

    GaeaData type = NullSymbol;
    switch(player->lastAction) {
        case GSNone:  type=GsSymActionNone;  break;
        case GSKick:  type=GsSymActionKick;  break;
        case GSTurn:  type=GsSymActionTurn;  break;
        case GSDash:  type=GsSymActionDash;  break;
        case GSCatch: type=GsSymActionCatch; break;
        case GSError: type=GsSymActionError; break;
    }

    if (unifyWith(type, actionType_, gem))
        return bipSuccess(gem) ;
    else 
        return bipError(gem,"Can't get last action?!?.") ;
}
        


//--------------------------------------------------
// estimated pos

BipStatus gs_epos(Gem* gem) {
    /*   $gs$epos(+Player,-X,-Y,-BDir,-HDir) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,5) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_VAR(vx,gem,icgoal,1) ;
    ARGVAL_VAR(vy,gem,icgoal,2) ;
    ARGVAL_VAR(vbdir,gem,icgoal,3) ;
    ARGVAL_VAR(vhdir,gem,icgoal,4) ;

    player->eposMutex.lock() ;
    ScPosDir epos = player->eposdir ;
    player->eposMutex.unlock() ;

    if(unifyWith(GaeaFloat(epos.pos.x),vx,gem) &&
       unifyWith(GaeaFloat(epos.pos.y),vy,gem) &&
       unifyWith(GaeaFloat(epos.bdir),vbdir,gem) &&
       unifyWith(GaeaFloat(epos.hdir),vhdir,gem))
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}


    

//--------------------------------------------------
// expected pos

BipStatus gs_xpos(Gem* gem) {
    /*   $gs$xpos(+Player,-X,-Y,-BDir,-HDir) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,5) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_VAR(vx,gem,icgoal,1) ;
    ARGVAL_VAR(vy,gem,icgoal,2) ;
    ARGVAL_VAR(vbdir,gem,icgoal,3) ;
    ARGVAL_VAR(vhdir,gem,icgoal,4) ;

    player->xposMutex.lock() ;
    ScPosDir xpos = player->eposdir ;
    player->xposMutex.unlock() ;

    if(unifyWith(GaeaFloat(xpos.pos.x),vx,gem) &&
       unifyWith(GaeaFloat(xpos.pos.y),vy,gem) &&
       unifyWith(GaeaFloat(xpos.bdir),vbdir,gem) &&
       unifyWith(GaeaFloat(xpos.hdir),vhdir,gem))
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}

//======================================================================
// database
//

//--------------------------------------------------
// goal pos

BipStatus gs_goal_pos(Gem* gem) {
    /*   $gs$goal_pos(+Player,-X,-Y) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,3) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_VAR(vx,gem,icgoal,1) ;
    ARGVAL_VAR(vy,gem,icgoal,2) ;

    if(unifyWith(GaeaFloat(SC_HALF_PITCH_L),vx,gem) &&
       unifyWith(GaeaFloat(0),vy,gem))
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}

BipStatus gs_owngoal_pos(Gem* gem) {
    /*   $gs$owngoal_pos(+Player,-X,-Y) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,3) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_VAR(vx,gem,icgoal,1) ;
    ARGVAL_VAR(vy,gem,icgoal,2) ;

    if(unifyWith(GaeaFloat(-SC_HALF_PITCH_L),vx,gem) &&
       unifyWith(GaeaFloat(0),vy,gem))
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}


//--------------------------------------------------
// ball pos

BipStatus gs_ball_pos(Gem* gem) {
    /*   $gs$ball_pos(+Player,-X,-Y,-VX,-VY,-Cycle) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,6) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_VAR(vx,gem,icgoal,1) ;
    ARGVAL_VAR(vy,gem,icgoal,2) ;
    ARGVAL_VAR(vvx,gem,icgoal,3) ;
    ARGVAL_VAR(vvy,gem,icgoal,4) ;
    ARGVAL_VAR(vcyc,gem,icgoal,5) ;

    player->commandLock(); 
    ScPos bpos = player->ballPosAbs() ;
    ScPos bvel = player->ballVelAbs() ;
    ScInt bcyc = scMax(player->ball().ptime, player->estiBallCycle.count) ;
    player->commandUnlock();

    if(unifyWith(GaeaFloat(bpos.x),vx,gem) &&
       unifyWith(GaeaFloat(bpos.y),vy,gem) &&
       unifyWith(GaeaFloat(bvel.x),vvx,gem) &&
       unifyWith(GaeaFloat(bvel.y),vvy,gem) &&
       unifyWith(GaeaInteger(bcyc),vcyc,gem))
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}

//--------------------------------------------------
// ball pos

BipStatus gs_set_ball_pos_if_need(Gem* gem) {
    /*   $gs$set_ball_pos_if_need(+Player,+X,+Y,+Time) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,4) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_FLT(bx,gem,icgoal,1) ;
    ARGVAL_FLT(by,gem,icgoal,2) ;
    ARGVAL_INT(bt,gem,icgoal,3) ;

    if(player->ball().ptime < bt) {
      player->ball().posdir.pos.x = bx ;
      player->ball().posdir.pos.y = by ;
      cout << "ball pos is updated by teammate information" << endl ;
    }

    return bipSuccess(gem) ;

}

// 1.04o
BipStatus gs_distance_to_ball(Gem* gem) {
    //   $gs$distance_to_ball(+Player,-Dist)

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_VAR(dist_,gem,icgoal,1) ;

    player->commandLock(); 
    ScFloat dist = player->distanceToBall();
    player->commandUnlock();

    if(unifyWith(GaeaFloat(dist),dist_,gem))
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}

// 1.04o
BipStatus gs_ball_in_back_3rd(Gem* gem) {
    //   $gs$ball_in_back_3rd(+Player)
    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;
    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    player->commandLock(); 
    ScPos bpos = player->ballPosAbs() ;
    player->commandUnlock();
    if (bpos.x <= -SC_PITCH_LENGTH / 6.0) 
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}

//--------------------------------------------------
// teammante pos

BipStatus gs_teammate_pos(Gem* gem) {
    /*   $gs$teammate_pos(+Player,+UNum,-X,-Y,-VX,-VY,-BDir,-HDir,-Cycle) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,9) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_INT(unum,gem,icgoal,1) ;
    ARGVAL_VAR(vx,gem,icgoal,2) ;
    ARGVAL_VAR(vy,gem,icgoal,3) ;
    ARGVAL_VAR(vvx,gem,icgoal,4) ;
    ARGVAL_VAR(vvy,gem,icgoal,5) ;
    ARGVAL_VAR(vbdir,gem,icgoal,6) ;
    ARGVAL_VAR(vhdir,gem,icgoal,7) ;
    ARGVAL_VAR(vcyc,gem,icgoal,8) ;
   
    if(unum < 1 || 11 < unum) { return bipError(gem,"Illegal unum") ; } 

    player->objectTableMutex.lock();
    ScPosDir posdir = player->objectTable.getTeammate(unum).posdir ;
    ScPos vel = player->objectTable.getTeammate(unum).vel ;
    ScInt cyc = player->objectTable.getTeammate(unum).ptime ;
    player->objectTableMutex.unlock();

    if(unifyWith(GaeaFloat(posdir.pos.x),vx,gem) &&
       unifyWith(GaeaFloat(posdir.pos.y),vy,gem) &&
       unifyWith(GaeaFloat(vel.x),vvx,gem) &&
       unifyWith(GaeaFloat(vel.y),vvy,gem) &&
       unifyWith(GaeaFloat(posdir.bdir),vbdir,gem) &&
       unifyWith(GaeaFloat(posdir.hdir),vhdir,gem) &&
       unifyWith(GaeaInteger(cyc),vcyc,gem))
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}

//--------------------------------------------------
// opponent pos

BipStatus gs_opponent_pos(Gem* gem) {
    /*   $gs$opponent_pos(+Player,+UNum,-X,-Y,-VX,-VY,-BDir,-HDir,-Cycle) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,9) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_INT(unum,gem,icgoal,1) ;
    ARGVAL_VAR(vx,gem,icgoal,2) ;
    ARGVAL_VAR(vy,gem,icgoal,3) ;
    ARGVAL_VAR(vvx,gem,icgoal,4) ;
    ARGVAL_VAR(vvy,gem,icgoal,5) ;
    ARGVAL_VAR(vbdir,gem,icgoal,6) ;
    ARGVAL_VAR(vhdir,gem,icgoal,7) ;
    ARGVAL_VAR(vcyc,gem,icgoal,8) ;
   
    if(unum < 1 || 11 < unum) { return bipError(gem,"Illegal unum") ; } 

    player->objectTableMutex.lock();
    ScPosDir posdir = player->objectTable.getOpponent(unum).posdir ;
    ScPos vel = player->objectTable.getOpponent(unum).vel ;
    ScInt cyc = player->objectTable.getOpponent(unum).ptime ;
    player->objectTableMutex.unlock();

    if(unifyWith(GaeaFloat(posdir.pos.x),vx,gem) &&
       unifyWith(GaeaFloat(posdir.pos.y),vy,gem) &&
       unifyWith(GaeaFloat(vel.x),vvx,gem) &&
       unifyWith(GaeaFloat(vel.y),vvy,gem) &&
       unifyWith(GaeaFloat(posdir.bdir),vbdir,gem) &&
       unifyWith(GaeaFloat(posdir.hdir),vhdir,gem) &&
       unifyWith(GaeaInteger(cyc),vcyc,gem))
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}

//======================================================================
// ball handling
//

//--------------------------------------------------
// kickable

BipStatus gs_kickable(Gem* gem) {
    /*   $gs$kickable(+Player,+TimeMargin,+Effect) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,3) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_INT(tmargin,gem,icgoal,1) ;
    ARGVAL_FLT(effect,gem,icgoal,2) ;

    // 1.04o added locks
    player->commandLock();
    ScBool r = player->KickEffect((ScUInt)tmargin) > effect;
    player->commandUnlock();
    return (r  ?  bipSuccess(gem)  :  bipCutFail(gem));
}


BipStatus gs_kickable2(Gem* gem) {
    //   $gs$kickable2(+Player)
    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;
    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;

    // 1.04o added locks
    player->commandLock();
    ScBool r = player->KickableP();
    player->commandUnlock();
    return (r  ?  bipSuccess(gem)  :  bipCutFail(gem));
}


// 1.04o
BipStatus gs_dash_kickable(Gem* gem) {
    //   $gs$dash_kickable(+Player)
    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;
    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;

    player->commandLock();
    ScBool r = player->DashKickableP();
    player->commandUnlock();
    return (r  ?  bipSuccess(gem)  :  bipCutFail(gem));
}

// 1.06
BipStatus gs_dash_kickable_by_sensor(Gem* gem) {
    //   $gs$dash_kickable_by_sensor(+Player)
    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;
    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;

    player->commandLock();
    ScBool r = player->dashKickablePbySensor();
    player->commandUnlock();
    return (r  ?  bipSuccess(gem)  :  bipCutFail(gem));
}

//--------------------------------------------------
// goalie functions

// 1.04o
BipStatus gs_catch_ban(Gem *gem) {
    //   $gs$catch_ban(+Player, +FutureSteps)
    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;
    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_INT(timesteps,gem,icgoal,1) ;

    player->commandLock();
    ScBool r = player->catchBanP(timesteps);
    player->commandUnlock();
    return (r  ?  bipSuccess(gem)  :  bipCutFail(gem));
}


// 1.04o
BipStatus gs_catchable_distance(Gem *gem) {
    //   $gs$catchable_distance(+Player)
    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;
    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;

    player->commandLock();
    ScBool r = (player->distanceToBall() < SC_CATCHABLE_AREA_L);
    player->commandUnlock();
    return (r  ?  bipSuccess(gem)  :  bipCutFail(gem));
}


// 1.04o
BipStatus gs_ball_velocity_into_my_goal(Gem *gem) {
    //   $gs$ball_velocity_into_my_goal(+Player)
    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;
    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;

    player->commandLock();
    ScBool r = player->ballVelocityIntoMyGoalP();
    player->commandUnlock();
    return (r  ?  bipSuccess(gem)  :  bipCutFail(gem));
}


// 1.04o
BipStatus gs_ball_freeness_in_goal_area(Gem *gem) {
    //   $gs$ball_freeness_in_goal_area(+Player, -Freeness)
    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;
    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_VAR(freeness_,gem,icgoal,1) ;

    player->commandLock();
    ScFloat freeness = player->BallFreenessInGoalArea();
    player->commandUnlock();
    if(unifyWith(GaeaFloat(freeness),freeness_,gem))
        return bipSuccess(gem);
    else
        return bipCutFail(gem);
}


// 1.04o
BipStatus gs_goalie_critical_action_loop(Gem *gem) {
    //   $gs$goalie_critical_action_loop(+Player)
    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;
    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;

    player->goalieCriticalActionLoop();
    return bipSuccess(gem);
}


// 1.05o
BipStatus gs_goalie_critical_action_loop2(Gem *gem) {
    //   $gs$goalie_critical_action_loop2(+Player)
    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;
    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;

    player->goalieCriticalActionLoop2();
    return bipSuccess(gem);
}


// 1.05o
BipStatus gs_goalie_critical_action_loop3(Gem *gem) {
    //   $gs$goalie_critical_action_loop3(+Player)
    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;
    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;

    player->goalieCriticalActionLoop3();
    return bipSuccess(gem);
}


// 1.05o
BipStatus gs_ball_inside_perimeter(Gem *gem) {
    //   $gs$ball_inside_perimeter(+Player, +PerimeterNum)
    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;
    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_INT(perimeter,gem,icgoal,1) ;

    player->commandLock();
    ScPos bpos = player->ballPosAbs();
    player->commandUnlock();
    if (scDistance(player->GS_OurGoalCenter, bpos)
        < player->getPerimeterDistance(perimeter)) {
        return bipSuccess(gem);
    } else {
        return bipCutFail(gem);
    }
}

// 1.05o
BipStatus gs_ball_inside_center_corridor(Gem *gem) {
    //   $gs$ball_inside_center_corridor(+Player)
    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;
    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;

    player->commandLock();
    ScPos bpos = player->ballPosAbs();
    player->commandUnlock();
    if (player->insideCenterCorridor(bpos)) {
        return bipSuccess(gem);
    } else {
        return bipCutFail(gem);
    }
}

// 1.05o
BipStatus gs_get_goalie_clearing_coordinates(Gem *gem) {

   // V.1.06: comment out
   // cout<<"gs_get_goalie_clearing_coordinates: woohoo"<<endl;
    
    //   $gs$get_goalie_clearing_coordinates(+Player, -GX, -GY, -BX, -BY)
    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,5) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_VAR(gx_,gem,icgoal,1) ;  // goalie position to move to, (x,y)
    ARGVAL_VAR(gy_,gem,icgoal,2) ;
    ARGVAL_VAR(bx_,gem,icgoal,3) ;  // target of kick, (x,y)
    ARGVAL_VAR(by_,gem,icgoal,4) ;

    ScFloat gx, gy, bx, by;
    player->commandLock() ;
    // init version: clear to the side opposite that we are currently on.

    // (gx,gy) is scooted slightly in.
    gx = ((-SC_PITCH_LENGTH / 2.0  +  SC_PENALTY_AREA_LENGTH)
          - SC_CATCHABLE_AREA_L);
    gy = ((SC_PENALTY_AREA_WIDTH / 2.0)
          - SC_CATCHABLE_AREA_L);
    bx = 0;
    by = SC_PITCH_WIDTH / 2.0  -  5.0;   // scoot inwards
    if (player->xposdir.pos.y > 0) {
        gy *= -1;
        by *= -1;
    }
    player->commandUnlock() ;

    // V.1.06 : comment out
    //cout<<"~~clearing coordinates: (gx,gy), (bx, by)"<<endl
    // <<"     ("<<gx<<","<<gy<<"), ("<<bx<<","<<by<<")"<<endl;

    if(unifyWith(GaeaFloat(gx),gx_,gem)
       && unifyWith(GaeaFloat(gy),gy_,gem)
       && unifyWith(GaeaFloat(bx),bx_,gem) 
       && unifyWith(GaeaFloat(by),by_,gem)) 
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}


// 1.05o
BipStatus gs_wait_for_vision(Gem *gem) {
    //   $gs$wait_for_vision(+Player, +FullVisionP)
    //   FullVisionP = [0 | 1].  0 - any vision, 1 - full vision
    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;
    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_INT(fullVisionP,gem,icgoal,1) ;

    if (fullVisionP) {
        player->waitForFullVisualInfo();
    } else {
        player->waitForVisualInfo();
    }
    return bipSuccess(gem);    
}


//--------------------------------------------------
// kickable to absolute dir

BipStatus gs_infront_kickable_to_adir(Gem* gem) {
    /*   $gs$infront_kickable_to_adir(+Player,+ADir) */
    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;
    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_FLT(adir,gem,icgoal,1) ;

    // 1.04o added locks
    player->commandLock();
    ScBool r = player->infrontKickableTo(adir);
    player->commandUnlock();
    return (r  ?  bipSuccess(gem)  :  bipCutFail(gem));
}

//--------------------------------------------------
// kickable to relative pos

BipStatus gs_infront_kickable_to_rpos(Gem* gem) {
    /*   $gs$infront_kickable_to_rpos(+Player,+RX,+RY) */
    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,3) ;
    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_FLT(rx,gem,icgoal,1) ;
    ARGVAL_FLT(ry,gem,icgoal,2) ;

    // 1.04o added locks
    player->commandLock();
    ScBool r = player->infrontKickableToRel(ScPos(rx,ry));
    player->commandUnlock();
    return (r  ?  bipSuccess(gem)  :  bipCutFail(gem));
}

//--------------------------------------------------
// kickable to absolute pos

BipStatus gs_infront_kickable_to_apos(Gem* gem) {
    /*   $gs$infront_kickable_to_apos(+Player,+AX,+AY) */
    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,3) ;
    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_FLT(ax,gem,icgoal,1) ;
    ARGVAL_FLT(ay,gem,icgoal,2) ;

    // 1.04o added locks
    player->commandLock();
    ScBool r = player->infrontKickableToAbs(ScPos(ax,ay));
    player->commandUnlock();
    return (r  ?  bipSuccess(gem)  :  bipCutFail(gem));
}

//--------------------------------------------------
// is my ball in my team

BipStatus gs_closest_team_member_to_ball_p(Gem* gem) {
    /*   $gs$closest_team_member_to_ball_p(+Player) */
    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;
    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;

    // 1.04o added locks
    player->commandLock();
    ScBool r = player->closestTeamMemberToBallP();
    player->commandUnlock();
    return (r  ?  bipSuccess(gem)  :  bipCutFail(gem));
}

//======================================================================
// message handling
//

//--------------------------------------------------
// message buffer is empty

BipStatus gs_msg_is_empty(Gem* gem) {
    /*   $gs$msg_is_empty(+Player) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;

    ScBool r = player->messageRingBuffer.IsEmpty() ;

    if(r)
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}

//--------------------------------------------------
// message buffer is not empty

BipStatus gs_msg_is_not_empty(Gem* gem) {
    /*   $gs$msg_is_not_emtry(+Player) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;

    ScBool r = player->messageRingBuffer.IsEmpty() ;

    if(!r)
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}

//--------------------------------------------------
// sender type data

GaeaData GsSymReferee ;
GaeaData GsSymSelf ;
GaeaData GsSymPlayer ;

GaeaData GsSymOnlineCoachL ;
GaeaData GsSymOnlineCoachR ;

GaeaData& gsSenderTypeSymbol(const ScSenderType sender) {
    switch(sender) {
        case SC_ST_self: return GsSymSelf ;
        case SC_ST_referee: return GsSymReferee ;
        case SC_ST_player: return GsSymPlayer ;
        case SC_ST_onlineCoachL: return GsSymOnlineCoachL ;
        case SC_ST_onlineCoachR: return GsSymOnlineCoachR ;
        default : return GsSymError ;
    }
}

//--------------------------------------------------
// referee message data

GaeaData GsSymBeforeKickOff ;
GaeaData GsSymPlayOn     ;
GaeaData GsSymTimeOver ;
GaeaData GsSymKickOffL ;
GaeaData GsSymKickOffR ;
GaeaData GsSymKickInL    ;
GaeaData GsSymKickInR    ;
GaeaData GsSymFreeKickL ;
GaeaData GsSymFreeKickR ;
GaeaData GsSymCornerKickL ;
GaeaData GsSymCornerKickR ;
GaeaData GsSymGoalKickL ;
GaeaData GsSymGoalKickR ;
GaeaData GsSymGoalL  ;
GaeaData GsSymGoalR  ;
GaeaData GsSymFoulL  ;
GaeaData GsSymFoulR  ;
GaeaData GsSymTimeUpWOATeam ;
GaeaData GsSymTimeUp     ;
GaeaData GsSymTimeExtended ;
GaeaData GsSymHalfTime ;
GaeaData GsSymGoalieCatchL ;
GaeaData GsSymGoalieCatchR ;
GaeaData GsSymDropBall ;
GaeaData GsSymOffSideL ;
GaeaData GsSymOffSideR ;

//--------------------------------------------------
// referee message table

GaeaData& gsRefereeMessageSymbol(ScWord* w) {
    switch(w->id) {
        case SC_WID_BeforeKickOff : return GsSymBeforeKickOff ;   
        case SC_WID_PlayOn : return GsSymPlayOn ;     
        case SC_WID_TimeOver : return GsSymTimeOver ; 
        case SC_WID_KickOffL : return GsSymKickOffL ; 
        case SC_WID_KickOffR : return GsSymKickOffR ; 
        case SC_WID_KickInL : return GsSymKickInL ;       
        case SC_WID_KickInR : return GsSymKickInR ;       
        case SC_WID_FreeKickL : return GsSymFreeKickL ;   
        case SC_WID_FreeKickR : return GsSymFreeKickR ;   
        case SC_WID_CornerKickL : return GsSymCornerKickL ;   
        case SC_WID_CornerKickR : return GsSymCornerKickR ;   
        case SC_WID_GoalKickL : return GsSymGoalKickL ;   
        case SC_WID_GoalKickR : return GsSymGoalKickR ;   
        case SC_WID_GoalL : return GsSymGoalL ;       
        case SC_WID_GoalR : return GsSymGoalR ;       
        case SC_WID_FoulL : return GsSymFoulL ;       
        case SC_WID_FoulR : return GsSymFoulR ;       
        case SC_WID_TimeUpWOATeam : return GsSymTimeUpWOATeam ;   
        case SC_WID_TimeUp : return GsSymTimeUp ;     
        case SC_WID_TimeExtended : return GsSymTimeExtended ; 
        case SC_WID_HalfTime : return GsSymHalfTime ; 
        case SC_WID_GoalieCatchL : return GsSymGoalieCatchL ; 
        case SC_WID_GoalieCatchR : return GsSymGoalieCatchR ; 
        case SC_WID_DropBall : return GsSymDropBall ; 
        case SC_WID_OffSideL : return GsSymOffSideL ; 
        case SC_WID_OffSideR : return GsSymOffSideR ; 
        default : return GsSymUnknown ;
    }
}

//--------------------------------------------------
// get new message

BipStatus gs_get_new_msg(Gem* gem) {
    /*   $gs$get_new_msg(+Player,-Time,-Sender,-[Direction|PMode],-Message) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,5) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_VAR(vtime,gem,icgoal,1) ;
    ARGVAL_VAR(vsender,gem,icgoal,2) ;
    ARGVAL_VAR(vdirpm,gem,icgoal,3) ;
    ARGVAL_VAR(vmsg,gem,icgoal,4) ;

    GSHeardMessage* msg = player->messageRingBuffer.GetBottom() ;
    GaeaData dirpm ;
    if(msg->sender == SC_ST_player) {
        dirpm = GaeaFloat(msg->direction) ;
    } else if (msg->sender == SC_ST_referee) {
        dirpm = gsRefereeMessageSymbol(msg->pmode) ;
    } else {
        dirpm = NullSymbol ;
    }

    if(unifyWith(GaeaInteger(msg->cycle.count),vtime,gem) &&
       unifyWith(gsSenderTypeSymbol(msg->sender),vsender,gem) &&
       unifyWith(dirpm,vdirpm,gem) &&
       unifyWith(GaeaPointer(&(msg->message)),vmsg,gem))
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}

//--------------------------------------------------
// get playmode

BipStatus gs_playmode(Gem* gem) {
    /*   $gs$playmode(+Player,-PlayMode) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_VAR(vpmode,gem,icgoal,1) ;

    if(unifyWith(gsRefereeMessageSymbol(player->playmode),vpmode,gem))
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}

//======================================================================
// useful command
//

//--------------------------------------------------
// turn to absolute dir

BipStatus gs_turn_to_adir(Gem* gem) {
    /*   $gs$turn_to_adir(+Player,+ADir) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_FLT(adir,gem,icgoal,1) ;

    ScBool r = player->turnTo(adir) ;

    if(r)
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}

//--------------------------------------------------
// turn to relative pos

BipStatus gs_turn_to_rpos(Gem* gem) {
    /*   $gs$turn_to_rpos(+Player,+RX,+RY) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,3) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_FLT(rx,gem,icgoal,1) ;
    ARGVAL_FLT(ry,gem,icgoal,2) ;

    ScBool r = player->turnToRel(ScPos(rx,ry)) ;

    if(r)
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}

//--------------------------------------------------
// turn to absolute pos

BipStatus gs_turn_to_apos(Gem* gem) {
    /*   $gs$turn_to_apos(+Player,+AX,+AY) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,3) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_FLT(ax,gem,icgoal,1) ;
    ARGVAL_FLT(ay,gem,icgoal,2) ;

    ScBool r = player->turnToAbs(ScPos(ax,ay)) ;

    if(r)
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}

//--------------------------------------------------
// kick to absolute dir

BipStatus gs_kick_to_adir(Gem* gem) {
    /*   $gs$kick_to_adir(+Player,+Power,+ADir) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,3) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_FLT(power,gem,icgoal,1) ;
    ARGVAL_FLT(adir,gem,icgoal,2) ;

    ScBool r = player->kickTo(power,adir) ;

    if(r)
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}

//--------------------------------------------------
// kick to relative pos

BipStatus gs_kick_to_rpos(Gem* gem) {
    /*   $gs$kick_to_rpos(+Player,+Power,+RX,+RY) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,4) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_FLT(power,gem,icgoal,1) ;
    ARGVAL_FLT(rx,gem,icgoal,2) ;
    ARGVAL_FLT(ry,gem,icgoal,3) ;

    ScBool r = player->kickToRel(power,ScPos(rx,ry)) ;

    if(r)
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}

//--------------------------------------------------
// kick to absolute pos

BipStatus gs_kick_to_apos(Gem* gem) {
    /*   $gs$kick_to_apos(+Player,+Power,+AX,+AY) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,4) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_FLT(power,gem,icgoal,1) ;
    ARGVAL_FLT(ax,gem,icgoal,2) ;
    ARGVAL_FLT(ay,gem,icgoal,3) ;

    ScBool r = player->kickToAbs(power,ScPos(ax,ay)) ;

    if(r)
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}


//======================================================================
// basic play, middle command
//


BipStatus gs_distance_kick(Gem* gem) {
    /*   $gs$distance_kick(+Player, +ADir, +Dist, +Factor, +MaxPower) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,5) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_FLT(adir,gem,icgoal,1) ;
    ARGVAL_FLT(dist,gem,icgoal,2) ;
    ARGVAL_FLT(factor,gem,icgoal,3) ;
    ARGVAL_FLT(maxPower,gem,icgoal,4) ;

    ScBool r = player->distanceKick(adir, dist, factor, maxPower) ;

    if(r)
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}


BipStatus gs_circle_kick(Gem* gem) {
    /*   $gs$circle_kick(+Player, +ADir, +Dir, +StopBallP, -DoneP)
     *       Dir = [ right | left ] 
     */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,5) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_FLT(adir,gem,icgoal,1) ;
    ARGVAL_SYM(dir,gem,icgoal,2) ;
    ARGVAL_INT(stopBallP,gem,icgoal,3) ;
    ARGVAL_VAR(done_,gem,icgoal,4) ;

    ScBool rightP = (dir == SymVal(GsSymRight)) ;
    ScBool doneP ;

    ScBool r = player->circleKick(adir, rightP, stopBallP, &doneP) ;

    if(r  &&  unifyWith(GaeaInteger(doneP),done_,gem))
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}


BipStatus gs_pass_ball_to(Gem* gem) {
    /*   $gs$pass_ball_to(+Player, +KickPow, +ADir, -DoneP)
     */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,4) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_FLT(kickPow,gem,icgoal,1) ;
    ARGVAL_FLT(adir,gem,icgoal,2) ;
    ARGVAL_VAR(done_,gem,icgoal,3) ;

    ScBool doneP ;

    ScBool r = player->passBallTo(kickPow, adir, &doneP) ;

    if(r  &&  unifyWith(GaeaInteger(doneP),done_,gem))
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}


BipStatus gs_pass_ball_to_apos(Gem* gem) { //NODA
    /*   $gs$pass_ball_to_apos(+Player, +KickPow, +X, +Y, -DoneP)
     */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,5) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_FLT(kickPow,gem,icgoal,1) ;
    ARGVAL_FLT(x,gem,icgoal,2) ;
    ARGVAL_FLT(y,gem,icgoal,3) ;
    ARGVAL_VAR(done_,gem,icgoal,4) ;

    ScBool doneP ;

    ScFloat adir = player->xposdir.pos.AngleTo(ScPos(x,y)) ;

    ScBool r = player->passBallTo(kickPow, adir, &doneP) ;

    if(r  &&  unifyWith(GaeaInteger(doneP),done_,gem))
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}


BipStatus gs_multi_pass_ball_to(Gem* gem) {
    /*   $gs$multi_pass_ball_to(+Player, +KickPow, +ADir)  */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,3) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_FLT(kickPow,gem,icgoal,1) ;
    ARGVAL_FLT(adir,gem,icgoal,2) ;

    ScBool r = player->multiPassBallTo(kickPow, adir) ;

    if(r)
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}


BipStatus gs_multi_pass_ball_to_apos(Gem* gem) { //NODA
    /*   $gs$multi_pass_ball_to(+Player, +KickPow, +X, +Y)  */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,4) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_FLT(kickPow,gem,icgoal,1) ;
    ARGVAL_FLT(x,gem,icgoal,2) ;
    ARGVAL_FLT(y,gem,icgoal,3) ;

    ScFloat adir = player->xposdir.pos.AngleTo(ScPos(x,y)) ;

    ScBool r = player->multiPassBallTo(kickPow, adir) ;

    if(r)
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}

// v1.04o
inline GSRotationType GetRotationTypeFromSymbol(Symbol *rotSym) {
    if (rotSym == SymVal(GsSymRotLeft))     return GSRotLeft;
    if (rotSym == SymVal(GsSymRotRight))    return GSRotRight;
    if (rotSym == SymVal(GsSymRotClosest))  return GSRotClosest;
    if (rotSym == SymVal(GsSymRotFurthest)) return GSRotFurthest;
    if (rotSym == SymVal(GsSymRotAvoid))    return GSRotAvoid;
    return GSRotClosest;  // shouldn't ever reach here...
}

// v1.04o
BipStatus gs_multi_smart_kick_to(Gem* gem) {
    //   $gs$multi_smart_kick_to(+Player, +KickPow, +ADir,
    //                           +ControlP, +FaceBallP, +RotType)
    //   RotType = [ rot_right | rot_left | rot_closest |
    //               rot_furthest | rot_avoid ]
    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,6) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_FLT(kickPow,gem,icgoal,1) ;
    ARGVAL_FLT(adir,gem,icgoal,2) ;
    ARGVAL_INT(controlP,gem,icgoal,3) ;
    ARGVAL_INT(faceBallP,gem,icgoal,4) ;
    ARGVAL_SYM(rotSym,gem,icgoal,5) ;

    GSRotationType rot = GetRotationTypeFromSymbol(rotSym);
    ScBool r = player->multiSmartKickTo(kickPow, adir,
                                        controlP, faceBallP, rot) ;
    if(r)
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}

// v1.04o
BipStatus gs_multi_smart_kick_to_apos(Gem* gem) {
    //   $gs$multi_smart_kick_to(+Player, +KickPow, +X, +Y,
    //                           +ControlP, +FaceBallP, +RotType)
    //   RotType = [ rot_right | rot_left | rot_closest |
    //               rot_furthest | rot_avoid ]

   //cout<<"~~gs_multi : start"<<endl;

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,7) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_FLT(kickPow,gem,icgoal,1) ;
    ARGVAL_FLT(x,gem,icgoal,2) ;
    ARGVAL_FLT(y,gem,icgoal,3) ;
    ARGVAL_INT(controlP,gem,icgoal,4) ;
    ARGVAL_INT(faceBallP,gem,icgoal,5) ;
    ARGVAL_SYM(rotSym,gem,icgoal,6) ;

    //cout<<"~~gs_multi : pt 1"<<endl;
    
    player->commandLock();
    ScFloat adir = player->xposdir.pos.AngleTo(ScPos(x,y)) ;
    player->commandUnlock();
    GSRotationType rot = GetRotationTypeFromSymbol(rotSym);

    //cout<<"~~gs_multi : pt 2 : "<<adir<<" : "<<rot<<endl;

    ScBool r = player->multiSmartKickTo(kickPow, adir,
                                        controlP, faceBallP, rot) ;
    //cout<<"~~gs_multi : pt 3 : "<<r<<endl;

    if(r)
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}

// v1.04o
// Like gs_multi_smart_kick_to, but by default choses to do the
// whip-around fast circle kicks.
BipStatus gs_multi_smart_kick_to2(Gem* gem) {
    //   $gs$multi_smart_kick_to(+Player, +KickPow, +ADir, +RotType)
    //   RotType = [ rot_right | rot_left | rot_closest |
    //               rot_furthest | rot_avoid ]
    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,4) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_FLT(kickPow,gem,icgoal,1) ;
    ARGVAL_FLT(adir,gem,icgoal,2) ;
    ARGVAL_SYM(rotSym,gem,icgoal,3) ;

    GSRotationType rot = GetRotationTypeFromSymbol(rotSym);
    ScBool r = player->multiSmartKickTo(kickPow, adir,
                                        ScFalse, ScFalse, rot) ;
    if(r)
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}

// v1.40o
BipStatus gs_multi_smart_kick_to_apos2(Gem* gem) {
    //   $gs$multi_smart_kick_to(+Player, +KickPow, +X, +Y, +RotType)
    //   RotType = [ rot_right | rot_left | rot_closest |
    //               rot_furthest | rot_avoid ]
    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,5) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_FLT(kickPow,gem,icgoal,1) ;
    ARGVAL_FLT(x,gem,icgoal,2) ;
    ARGVAL_FLT(y,gem,icgoal,3) ;
    ARGVAL_SYM(rotSym,gem,icgoal,4) ;

    ScFloat adir = player->xposdir.pos.AngleTo(ScPos(x,y)) ;
    GSRotationType rot = GetRotationTypeFromSymbol(rotSym);
    ScBool r = player->multiSmartKickTo(kickPow, adir,
                                        ScFalse, ScFalse, rot) ;
    if(r)
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}


BipStatus gs_step_goto2(Gem* gem) {
    /*   $gs$step_goto2(+Player,+X,+Y,+Power) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,4) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_FLT(x,gem,icgoal,1) ;
    ARGVAL_FLT(y,gem,icgoal,2) ;
    ARGVAL_FLT(power,gem,icgoal,3) ;

    ScBool r = player->stepGoto2(ScPos(x,y),power,ScTrue) ;

    if(r)
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}


BipStatus gs_step_dash_to_ball2(Gem* gem) {
    /*   $gs$step_dash_to_ball2(+Player,+Power) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_FLT(power,gem,icgoal,1) ;

    ScBool r = player->stepDashToBall2(power,ScTrue) ;

    if(r)
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}


BipStatus gs_get_plan_for_intercept(Gem* gem) {
    //   $gs$get_plan_for_intercept(+Player, +DashPow, +MaxSteps,
    //                              +ExtraSafetyP, -ATurnDir, -NumDashes)

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,6) ;  //v1.04o change

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_FLT(dashPow,gem,icgoal,1) ;
    ARGVAL_INT(maxSteps,gem,icgoal,2) ;
    ARGVAL_INT(xSafetyP,gem,icgoal,3) ;
    ARGVAL_VAR(adir_,gem,icgoal,4) ;
    ARGVAL_VAR(numDash_,gem,icgoal,5) ;

    ScFloat adir;
    ScInt numDash;

    ScBool r = player->getPlanForIntercept(dashPow, maxSteps, xSafetyP,
                                           &adir, &numDash);
    if (  r  &&
          unifyWith(GaeaFloat(adir),adir_,gem)  &&
          unifyWith(GaeaInteger(numDash),numDash_,gem)  )
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}


//V.1.04noda3
#define GS_PLAN_INTERCEPT_MIN_BALL_VEL	0.3
//V.1.04noda3
BipStatus gs_get_plan_for_intercept2(Gem* gem) {
    /*   $gs$get_plan_for_intercept2(+Player, +DashPow, +MaxSteps,
     *                              +ExtraSafetyP, +MinBallVel,
     *			            -ATurnDir, -NumDashes)
     */
    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,7) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_FLT(dashPow,gem,icgoal,1) ;
    ARGVAL_INT(maxSteps,gem,icgoal,2) ;
    ARGVAL_INT(xSafetyP,gem,icgoal,3) ;
    ARGVAL_FLT(minbvel,gem,icgoal,4) ;
    ARGVAL_VAR(adir_,gem,icgoal,5) ;
    ARGVAL_VAR(numDash_,gem,icgoal,6) ;

    ScFloat adir;
    ScInt numDash;

    // V.1.04noda3
    if(player->ballVelValid() &&
       player->ballVelAbs().norm() < minbvel) {
       return bipCutFail(gem) ;
    }

    ScBool r = player->getPlanForIntercept(dashPow, maxSteps, xSafetyP,
                                           &adir, &numDash);
    if (  r  &&
          unifyWith(GaeaFloat(adir),adir_,gem)  &&
          unifyWith(GaeaInteger(numDash),numDash_,gem)  )
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}


// v1.04o
BipStatus gs_get_plan_for_goalie_intercept(Gem* gem) {
    //   $gs$get_plan_for_goalie_intercept(+Player, -ATurnDir,
    //                                     -NumDashes, -AbsCatchDir)

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,4) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_VAR(adir_,gem,icgoal,1) ;
    ARGVAL_VAR(numDash_,gem,icgoal,2) ;
    ARGVAL_VAR(absCatchDir_,gem,icgoal,3) ;

    ScFloat adir;
    ScInt numDash;
    ScFloat absCatchDir;

    ScBool r = player->getPlanForGoalieCatchIntercept(&adir, &numDash,
                                                      &absCatchDir);    
    if (  r  &&
          unifyWith(GaeaFloat(adir),adir_,gem)  &&
          unifyWith(GaeaInteger(numDash),numDash_,gem) &&
          unifyWith(GaeaFloat(absCatchDir),absCatchDir_,gem)  )
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}


// v1.04o
BipStatus gs_set_goalie_max_steps_for_intercept(Gem* gem) {
    //   $gs$set_goalie_max_steps_for_intercept(+Player, +MaxSteps)

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_INT(maxSteps,gem,icgoal,1) ;

    if (maxSteps > 0) {
        player->goalieMaxStepsForIntercept = maxSteps;
        return bipSuccess(gem) ;
    } else {
        return bipCutFail(gem);
    }
}


// v1.04o
BipStatus gs_set_goalie_intercept_dash_power(Gem* gem) {
    //   $gs$set_goalie_intercept_dash_power(+Player, +DashPow)

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_FLT(dashPow,gem,icgoal,1) ;

    if (dashPow > 0  &&  dashPow <= SC_MAXPOWER) {
        player->goalieInterceptDashPower = dashPow;
        return bipSuccess(gem) ;
    } else {
        return bipCutFail(gem);
    }
}


// v1.04o
BipStatus gs_set_goalie_misc_dash_power(Gem* gem) {
    //   $gs$set_goalie_misc_dash_power(+Player, +DashPow)

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_FLT(dashPow,gem,icgoal,1) ;

    if (dashPow > 0  &&  dashPow <= SC_MAXPOWER) {
        player->goalieMiscDashPower = dashPow;
        return bipSuccess(gem) ;
    } else {
        return bipCutFail(gem);
    }
}


// 1.04o
BipStatus gs_set_goalie_free_ball_speed(Gem* gem) {
    //   $gs$set_goalie_free_ball_speed(+Player, +Speed)

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_FLT(speed,gem,icgoal,1) ;

    if (speed >= 0  &&  speed <= SC_BALL_SPEED_MAX) {
        player->goalieFreeBallInP2MaxSpeed = speed;
        return bipSuccess(gem) ;
    } else {
        return bipCutFail(gem);
    }
}


// 1.04o
BipStatus gs_get_goalie_action_values(Gem* gem) {
    //   $gs$get_goalie_action_values(+Player,
    //                                -MaxInterceptSteps,
    //                                -InterceptDashPow
    //                                -MiscDashPow,
    //                                -FreeBallInP2Speed)

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,5) ;
    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_VAR(steps_,gem,icgoal,1) ;
    ARGVAL_VAR(iDashPow_,gem,icgoal,2) ;
    ARGVAL_VAR(mDashPow_,gem,icgoal,3) ;
    ARGVAL_VAR(ballSpeed_,gem,icgoal,4) ;

    ScInt steps;
    ScFloat iDashPow, mDashPow, ballSpeed;
    steps = player->goalieMaxStepsForIntercept;
    iDashPow = player->goalieInterceptDashPower;
    mDashPow = player->goalieMiscDashPower;
    ballSpeed = player->goalieFreeBallInP2MaxSpeed;
    
    if(unifyWith(GaeaInteger(steps),steps_,gem)
       &&  unifyWith(GaeaFloat(iDashPow),iDashPow_,gem)
       &&  unifyWith(GaeaFloat(mDashPow),mDashPow_,gem)
       &&  unifyWith(GaeaFloat(ballSpeed),ballSpeed_,gem)) {
       return bipSuccess(gem) ;
    } else {
       return bipCutFail(gem) ;
    }
}


//--------------------------------------------------
// step goto (deprecated)

BipStatus gs_step_goto(Gem* gem) {
    /*   $gs$step_goto(+Player,+X,+Y,+Power) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,4) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_FLT(x,gem,icgoal,1) ;
    ARGVAL_FLT(y,gem,icgoal,2) ;
    ARGVAL_FLT(power,gem,icgoal,3) ;

    ScBool r = player->stepGoto(ScPos(x,y),power,ScTrue) ;

    if(r)
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}

//--------------------------------------------------
// step swing (deprecated)

BipStatus gs_step_swing(Gem* gem) {
    /*   $gs$step_swing(+Player,+Dir)   Dir = [ right | left ] */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_SYM(dir,gem,icgoal,1) ;

    ScBool r ;
    if(dir == SymVal(GsSymRight)) 
        r = player->stepSwing(ScTrue) ;
    else 
        r = player->stepSwing(ScFalse) ;

    if(r)
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}


//======================================================================
// Misc 
//

//--------------------------------------------------
// check within DISTANCE from POSITION (X,Y)

BipStatus gs_is_within_from(Gem* gem) {  // NODA
    /*   $gs$is_within_from(+Player,+X,+Y,+D) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,4) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_FLT(x,gem,icgoal,1) ;
    ARGVAL_FLT(y,gem,icgoal,2) ;
    ARGVAL_FLT(d,gem,icgoal,3) ;

    player->eposMutex.lock() ;
    ScPos epos = player->eposdir.pos ;
    player->eposMutex.unlock() ;

    if(scDistance(epos,ScPos(x,y)) < d) 
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}

//--------------------------------------------------
// set position parameter

BipStatus gs_set_position_param(Gem* gem) {  // NODA
    /*   $gs$set_position_param(+Player,+IX,+IY,+HX,+HY,+SX,+SY,+X1,+X2,+Y1,+Y2) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,11) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_FLT(ix,gem,icgoal,1) ;
    ARGVAL_FLT(iy,gem,icgoal,2) ;
    ARGVAL_FLT(hx,gem,icgoal,3) ;
    ARGVAL_FLT(hy,gem,icgoal,4) ;
    ARGVAL_FLT(sx,gem,icgoal,5) ;
    ARGVAL_FLT(sy,gem,icgoal,6) ;
    ARGVAL_FLT(ax1,gem,icgoal,7) ;
    ARGVAL_FLT(ax2,gem,icgoal,8) ;
    ARGVAL_FLT(ay1,gem,icgoal,9) ;
    ARGVAL_FLT(ay2,gem,icgoal,10) ;

    player->initPos = ScPos(ix,iy) ;
    player->homePos = ScPos(hx,hy) ;
    player->shiftPos = ScPos(sx,sy) ;
    player->areaLower = ScPos(ax1,ay1) ;
    player->areaUpper = ScPos(ax2,ay2) ;

    return bipSuccess(gem) ;
}

//--------------------------------------------------
// set closest check param

BipStatus gs_set_closest_check_param(Gem* gem) {  // NODA
    /*   $gs$set_closest_check_param(+Player,+TimeMargin) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_INT(tm,gem,icgoal,1) ;

    player->closestCheckTimeMargin = tm ;

    return bipSuccess(gem) ;
}

//--------------------------------------------------
// calc shifted home pos

BipStatus gs_shifted_home_pos(Gem* gem) {  // NODA
    /*   $gs$shifted_home_pos(+Player,-HX,-HY) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,3) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_VAR(vx,gem,icgoal,1) ;
    ARGVAL_VAR(vy,gem,icgoal,2) ;

    player->commandLock(); 
    ScPos bpos = player->ballPosAbs() ;
    player->commandUnlock();

    ScPos hpos = player->homePos ;
    ScPos shift = player->shiftPos ;
    hpos.x += shift.x * (bpos.x / (SC_HALF_PITCH_L)) ;
    hpos.y += shift.y * (bpos.y / (SC_HALF_PITCH_W)) ;

    if(unifyWith(GaeaFloat(hpos.x),vx,gem) &&
       unifyWith(GaeaFloat(hpos.y),vy,gem)) {
       return bipSuccess(gem) ;
    } else {
       return bipCutFail(gem) ;
    }
}

//--------------------------------------------------
// check ball is inside of the area

BipStatus gs_ball_is_in_my_area(Gem* gem) {  // NODA
    /*   $gs$ball_is_in_my_area(+Player) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;

    player->commandLock(); 
    ScPos bpos = player->ballPosAbs() ;
    player->commandUnlock();

    ScPos areaLower = player->areaLower ;
    ScPos areaUpper = player->areaUpper ;

    if(((areaLower.x - bpos.x) * (bpos.x - areaUpper.x) > 0) &&
       ((areaLower.y - bpos.y) * (bpos.y - areaUpper.y) > 0)) {
       return bipSuccess(gem) ;
    } else {
       return bipCutFail(gem) ;
    }
}


// V.1.04noda
//--------------------------------------------------
// find good pass destination

GaeaData GsSymXY ;
GaeaData GsSymAXY ;
GaeaData GsSymHXY ;
GaeaData GsVarX ;
GaeaData GsVarY ;
GaeaData GsTermXY2 ;
GaeaData GsTermAXY2 ;
GaeaData GsTermHXY2 ;

#define GS_PASSDEST_MAX 32
#define GS_SAFETY_PASS_ZONE_X	-25.0
#define GS_SAFETY_PASS_ZONE_Y	10.0
#define GS_TOO_SHORT_PASS_DIST	5.0	
#define GS_BEST_PASS_DIST	20.0
//V.1.04noda3
#define GS_PASS_MINUM_GAIN	8.0

Bool GsXYPosListToPosArray(GSPlayer* player,
			   const InstGaeaData& list,
			   ScPos *array,
			   ScUInt &n) {
   for(InstGaeaData l = fetch(list) ; 
       isConsTerm(l.data) ;
       l = fetch(InstGaeaData(_ConsCdr(l.data),l.binding))) {
      InstGaeaData xy = fetch(InstGaeaData(_ConsCar(l.data),l.binding)) ;
      
      if(!isGaeaTerm(xy.data) ||
	 ((_TermHead(xy.data) != _SymVal(GsSymAXY)) &&
	  (_TermHead(xy.data) != _SymVal(GsSymHXY))) ||
	 (_TermArity(xy.data) != 2)) 
	 return False ;
      
      GaeaData xx = 
	 fetch(InstGaeaData(_TermArgv(xy.data,0),xy.binding)).data ;
      GaeaData yy = 
	 fetch(InstGaeaData(_TermArgv(xy.data,1),xy.binding)).data ;
      array[n].x = FltVal(xx) ;
      array[n].y = FltVal(yy) ;
      if(_TermHead(xy.data) == _SymVal(GsSymHXY)) {
	 //	homepos should be shifted
	 //
	 player->commandLock(); 
	 ScPos bpos = player->ballPosAbs() ;
	 player->commandUnlock();
	 ScPos shift = player->shiftPos ;
	 array[n].x += shift.x * (bpos.x / (SC_HALF_PITCH_L)) ;
	 array[n].y += shift.y * (bpos.y / (SC_HALF_PITCH_W)) ;
      }
      n++ ;
   }
   return True ;
}

BipStatus gs_find_good_pass_dest(Gem* gem) {
    /*   $gs$find_good_pass_dest(+Player,xy(X,Y), [xy(X,Y),xy(X,Y)...]) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,3) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_VAR(vxy,gem,icgoal,1) ;
    ARGVAL_VAR(list,gem,icgoal,2) ;

    //list to array
    ScPos dest[GS_PASSDEST_MAX] ;
    ScUInt n = 0 ;
    if(!GsXYPosListToPosArray(player,list,dest,n)) 
       return bipError(gem,"Illegal pass destination list") ;
    
    //find good direction (simple)
    ScPos d ;
    ScBool findp = ScFalse ;
    ScFloat ddist = 0.0 ;

    player->eposMutex.lock() ;
    ScPos epos = player->eposdir.pos ;
    player->eposMutex.unlock() ;

    for(ScUInt i = 0 ; i < n ; i++) {
       ScPos c = dest[i] ;
       //V.1.04noda3
       if(c.x - epos.x < GS_PASS_MINUM_GAIN) { continue ; }	// backpass 
       if(epos.x < GS_SAFETY_PASS_ZONE_X && 		// in front of goal & 
	  ((epos.y * c.y < 0) ||	// differnt side |
	   (scAbs(c.y) < GS_SAFETY_PASS_ZONE_Y && 	// center area &
	    scAbs(c.y) < scAbs(epos.y)))) { // more center
	  continue ; 
	}
       ScFloat cdist = scDistance(c,epos) ;
       if(cdist < GS_TOO_SHORT_PASS_DIST) { 
	  continue ; 
       } // too close

       if(!findp) { 
	  d = c ;
	  ddist = cdist ;
	  findp = ScTrue ;
	  continue ; 
       }

       if(scAbs(cdist - GS_BEST_PASS_DIST) < scAbs(ddist - GS_BEST_PASS_DIST)){
	  d = c ;
	  ddist = cdist ;
	  continue ; 
       }
    }

    if(!findp) return bipCutFail(gem) ;

    Binding* newbinding = gem->uiheap.newBinding() ;
    if(unify(InstGaeaData(GsTermXY2,newbinding),vxy,&(gem->uiheap)) &&
       unifyWith(GaeaFloat(d.x),InstGaeaData(GsVarX,newbinding),gem) &&
       unifyWith(GaeaFloat(d.y),InstGaeaData(GsVarY,newbinding),gem)) {
       return bipSuccess(gem) ;
    } else {
       return bipCutFail(gem) ;
    }
}

// V.1.04noda
//--------------------------------------------------
// mode 

BipStatus gs_arith_mod(Gem* gem) {
    /*   $gs$arith_mod(+Val,+Mode,-Rem) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,3) ;

    ARGVAL_INT(x,gem,icgoal,0) ;
    ARGVAL_INT(md,gem,icgoal,1) ;
    ARGVAL_VAR(vrem,gem,icgoal,2) ;

    if(unifyWith(GaeaInteger(x%md),vrem,gem)) {
       return bipSuccess(gem) ;
    } else {
       return bipCutFail(gem) ;
    }
}

// V.1.04noda
//--------------------------------------------------
// distance

BipStatus gs_distance(Gem* gem) {
    /*   $gs$distance(+X1,+Y1,+X2,+Y2,-Dist) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,5) ;

    ARGVAL_FLT(x1,gem,icgoal,0) ;
    ARGVAL_FLT(y1,gem,icgoal,1) ;
    ARGVAL_FLT(x2,gem,icgoal,2) ;
    ARGVAL_FLT(y2,gem,icgoal,3) ;
    ARGVAL_VAR(vdist,gem,icgoal,4) ;

    if(unifyWith(GaeaFloat(scDistance(ScPos(x1,y1),ScPos(x2,y2))),
		 vdist,gem)) {
       return bipSuccess(gem) ;
    } else {
       return bipCutFail(gem) ;
    }
}

// V.1.04noda2
//--------------------------------------------------
// which team is closer to the ball

BipStatus gs_closer_team_to_ball(Gem* gem) {
   /*	$gs$closer_team_to_ball(+Player,+TMargin,-DistDiff) 
				DistDiff > 0 : a teammate is nearer
				DistDiff < 0 : an opponent is nearer */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,3) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_INT(tmargin,gem,icgoal,1) ;
    ARGVAL_VAR(vdistdiff,gem,icgoal,2) ;

    ScPos ballPos = player->ballPosAbs() ;
    
    GSMovingObject *teammate = 
       player->objectTable.nearestTeammateTo(ballPos,
					     player->scycle.count,
					     tmargin) ;
    GSMovingObject *opponent = 
       player->objectTable.nearestOpponentTo(ballPos,
					     player->scycle.count,
					     tmargin) ;

    ScFloat diff =  0.0 ;

    //V.1.05n28
    //if(scIsNullPtr(teammate) && scIsNullPtr(opponent)) {
    if(ScFalse) {
       return bipCutFail(gem) ;
    } else if (scIsNullPtr(opponent)) {
       diff = 3 ;
    } else if (scIsNullPtr(teammate)) {
       diff = (scDistance(opponent->posdir.pos,ballPos) -
	       scDistance(player->eposdir.pos,ballPos)) ;
    } else {
       diff = (scDistance(opponent->posdir.pos,ballPos) -
	       scDistance(teammate->posdir.pos,ballPos)) ;
    }

    if(unifyWith(GaeaFloat(diff),vdistdiff,gem)) 
       return bipSuccess(gem) ;
    else
       return bipCutFail(gem) ;
}

// V.1.04noda2
//--------------------------------------------------
// ball is going to opponent

BipStatus gs_ball_go_forward(Gem* gem) {
   /*	$gs$ball_go_forward(+Player,+Margin) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_FLT(margin,gem,icgoal,1) ;

    if(!player->ballVelValid()) 
       return bipCutFail(gem) ;

    if(player->ballVelAbs().x > margin)
       return bipSuccess(gem) ;
    else
       return bipCutFail(gem) ;
}

// V.1.04noda2
//--------------------------------------------------
// ball is going to own goal

BipStatus gs_ball_go_backward(Gem* gem) {
   /*	$gs$ball_go_backward(+Player,+Margin) */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_FLT(margin,gem,icgoal,1) ;

    if(!player->ballVelValid()) 
       return bipCutFail(gem) ;

    if(player->ballVelAbs().x < -margin)
       return bipSuccess(gem) ;
    else
       return bipCutFail(gem) ;
}


// v1.04o
//--------------------------------------------------
// kicks the ball towards the sidelines at mid-field.
BipStatus gs_kick_ball_to_near_mid_sideline(Gem *gem) {
    /*   $gs$kick_ball_to_near_mid_sideline(+Player, +kickPower) */
    
    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,2) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_FLT(power,gem,icgoal,1) ;

    player->commandLock() ;
    if (!player->KickableP()) {
        player->commandUnlock() ;
        return bipCutFail(gem) ;
    }

    ScPos bpos = player->ballPosAbs();
    player->commandUnlock();

    ScPos relTarget(0,0);
    // scoot value towards middle slightly
    relTarget.y = (bpos.y > 0) ? SC_PITCH_WIDTH/2 -5 : -SC_PITCH_WIDTH/2 +5;
    relTarget -= player->xposdir.pos;

    ScBool r;
    if (bpos.x < -SC_PITCH_LENGTH/2) {
        // allow whip-type kicks
        r = player->multiSmartKickTo(power, relTarget.Angle(), ScFalse,
                                     ScFalse, GSRotClosest);
    } else {
        // kick slower
        r = player->multiPassBallTo(power, relTarget.Angle());
    }
    
    if (r) {
       return bipSuccess(gem) ;
    } else {
       return bipCutFail(gem) ;
    }
}


// v1.04o
// Note: fails if no nearest opponent.
BipStatus gs_get_nearest_opponent(Gem* gem) {
    //   $gs$get_nearest_opponent(+Player, +Target, -UNum, -Dist)
    //   Target = [ball | me]

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_SYM(target, gem, icgoal, 1);
    ARGVAL_VAR(unum,gem,icgoal,2) ;
    ARGVAL_VAR(dist,gem,icgoal,3) ;

    ScBool fromMeP;

    if (target == SymVal(GsSymMe)) {
        fromMeP = ScTrue;
    } else if (target == SymVal(GsSymBall)) {
        fromMeP = ScFalse;
    } else {
        return bipCutFail(gem) ;
    }

    player->commandLock();
    ScPos targetPos = (fromMeP) ? player->xposdir.pos : player->ballPosAbs();
    GSMovingObject* opp = player->objectTable.
        nearestOpponentTo(targetPos,
                          player->scycle.count,
                          player->closestCheckTimeMargin);
    if (scIsNullPtr(opp)) {
        player->commandUnlock();
        return bipCutFail(gem);
    }
    ScFloat distance = opp->distanceTo(targetPos);
    player->commandUnlock();

    if(unifyWith(GaeaInteger(opp->id.unum),unum,gem) &&
       unifyWith(GaeaFloat(distance),dist,gem))
        return bipSuccess(gem) ;
    else 
        return bipCutFail(gem) ;
}


//V.1.05noda
//======================================================================
// misc

BipStatus gs_out_of_field(Gem* gem) {
    /*   $gs$out_of_field(-Player)  */

    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,1) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;

    if(player->outOfFieldP) {
       return bipSuccess(gem) ;
    } else {
       return bipCutFail(gem) ;
    }
}

BipStatus gs_should_not_chase_ball(Gem* gem) {
   /*	$gs$should_not_chase_ball(+Player,+MsgCycle,+TMargin,
					+ODist,+DistMargin,
					+OStamina,+StamMargin) */
    PICKUP_CURRENT_GOAL_CHECK_ARITY(icgoal,cgoal,gem,7) ;

    ARGVAL_EXTOBJ(GSPlayer*,player,gem,icgoal,0,isGSPlayer,
                  "Illegal GSPlayer object") ;
    ARGVAL_INT(mcycle,gem,icgoal,1) ;
    ARGVAL_INT(tmargin,gem,icgoal,2) ;
    ARGVAL_FLT(odist,gem,icgoal,3) ;
    ARGVAL_FLT(dmargin,gem,icgoal,4) ;
    ARGVAL_FLT(ostam,gem,icgoal,5) ;
    ARGVAL_FLT(smargin,gem,icgoal,6) ;

    //cout << mcycle << "," << tmargin << "," << odist << ","
    //	 << dmargin << "," << ostam << "," << smargin << endl ;

    if(tmargin < player->scycle.count - mcycle)
       return bipCutFail(gem) ;

    if(dmargin >= player->distanceToBall() - odist)
       return bipCutFail(gem) ;
    
    if(smargin <= player->estimateBodyState.stamina - ostam)
       return bipCutFail(gem) ;

    return bipSuccess(gem) ;
}

//======================================================================
// Initialize (internal)
//

static Bool bipInit() {
   
    _SymVal(GaeaSymbol("$gs$create_player"))->setfunc(gs_create_player) ;
    _SymVal(GaeaSymbol("$gs$send_init"))->setfunc(gs_send_init) ;
    _SymVal(GaeaSymbol("$gs$send_ginit"))->setfunc(gs_send_ginit) ;
    _SymVal(GaeaSymbol("$gs$half_time_reset"))->setfunc(gs_half_time_reset) ;

    _SymVal(GaeaSymbol("$gs$recv_scan_sinfo"))->setfunc(gs_recv_scan_sinfo) ;

    _SymVal(GaeaSymbol("$gs$move"))->setfunc(gs_move) ;
    //V.1.06
    _SymVal(GaeaSymbol("$gs$change_view"))->setfunc(gs_change_view) ;

    _SymVal(GaeaSymbol("$gs$turn"))->setfunc(gs_turn) ;
    _SymVal(GaeaSymbol("$gs$dash"))->setfunc(gs_dash) ;
    _SymVal(GaeaSymbol("$gs$kick"))->setfunc(gs_kick) ;
    _SymVal(GaeaSymbol("$gs$catch"))->setfunc(gs_catch) ;
    _SymVal(GaeaSymbol("$gs$make_catch"))->setfunc(gs_make_catch) ;  //1.04o
    _SymVal(GaeaSymbol("$gs$turn_neck"))->setfunc(gs_turn_neck) ;
    _SymVal(GaeaSymbol("$gs$say"))->setfunc(gs_say) ;
    _SymVal(GaeaSymbol("$gs$say_list"))->setfunc(gs_say_list) ;
    _SymVal(GaeaSymbol("$gs$scan_msg_list"))->setfunc(gs_scan_msg_list) ;

    _SymVal(GaeaSymbol("$gs$side_unum"))->setfunc(gs_side_unum) ;
    _SymVal(GaeaSymbol("$gs$cycle"))->setfunc(gs_cycle) ;
    _SymVal(GaeaSymbol("$gs$tic"))->setfunc(gs_tic) ;
    _SymVal(GaeaSymbol("$gs$set_tic"))->setfunc(gs_set_tic) ;
    _SymVal(GaeaSymbol("$gs$stamina"))->setfunc(gs_stamina) ;
    _SymVal(GaeaSymbol("$gs$last_action"))->setfunc(gs_last_action) ;  //1.04o
    _SymVal(GaeaSymbol("$gs$epos"))->setfunc(gs_epos) ;
    _SymVal(GaeaSymbol("$gs$xpos"))->setfunc(gs_xpos) ;

    _SymVal(GaeaSymbol("$gs$goal_pos"))->setfunc(gs_goal_pos) ;
    _SymVal(GaeaSymbol("$gs$owngoal_pos"))->setfunc(gs_owngoal_pos) ;

    _SymVal(GaeaSymbol("$gs$ball_pos"))->setfunc(gs_ball_pos) ;
    _SymVal(GaeaSymbol("$gs$set_ball_pos_if_need"))->setfunc(gs_set_ball_pos_if_need) ;

    // 1.04n
    _SymVal(GaeaSymbol("$gs$ball_distance"))->setfunc(gs_distance_to_ball) ;
    //1.04o
    _SymVal(GaeaSymbol("$gs$distance_to_ball"))->setfunc(gs_distance_to_ball) ;
    _SymVal(GaeaSymbol("$gs$ball_in_back_3rd"))->setfunc(gs_ball_in_back_3rd) ;
    
    _SymVal(GaeaSymbol("$gs$teammate_pos"))->setfunc(gs_teammate_pos) ;
    _SymVal(GaeaSymbol("$gs$opponent_pos"))->setfunc(gs_opponent_pos) ;

    _SymVal(GaeaSymbol("$gs$kickable"))->setfunc(gs_kickable) ;
    _SymVal(GaeaSymbol("$gs$kickable2"))->setfunc(gs_kickable2) ;
    //1.04o
    _SymVal(GaeaSymbol("$gs$dash_kickable"))->setfunc(gs_dash_kickable) ;
    //1.06
    _SymVal(GaeaSymbol("$gs$dash_kickable_by_sensor"))->setfunc(gs_dash_kickable_by_sensor) ;

    //1.04o
    _SymVal(GaeaSymbol("$gs$catch_ban"))->setfunc(gs_catch_ban) ;
    _SymVal(GaeaSymbol("$gs$catchable_distance"))->setfunc(gs_catchable_distance) ;
    _SymVal(GaeaSymbol("$gs$ball_velocity_into_my_goal"))->setfunc(gs_ball_velocity_into_my_goal) ;
    _SymVal(GaeaSymbol("$gs$ball_freeness_in_goal_area"))->setfunc(gs_ball_velocity_into_my_goal) ;
    _SymVal(GaeaSymbol("$gs$goalie_critical_action_loop"))->setfunc(gs_goalie_critical_action_loop) ;

    //1.05o
    _SymVal(GaeaSymbol("$gs$goalie_critical_action_loop2"))->setfunc(gs_goalie_critical_action_loop2) ;
    _SymVal(GaeaSymbol("$gs$goalie_critical_action_loop3"))->setfunc(gs_goalie_critical_action_loop3) ;
    _SymVal(GaeaSymbol("$gs$ball_inside_perimeter"))->setfunc(gs_ball_inside_perimeter) ;
    _SymVal(GaeaSymbol("$gs$ball_inside_center_corridor"))->setfunc(gs_ball_inside_center_corridor) ;
    _SymVal(GaeaSymbol("$gs$get_goalie_clearing_coordinates"))->setfunc(gs_get_goalie_clearing_coordinates) ;
    _SymVal(GaeaSymbol("$gs$wait_for_vision"))->setfunc(gs_wait_for_vision) ;

    _SymVal(GaeaSymbol("$gs$infront_kickable_to_adir"))->setfunc(gs_infront_kickable_to_adir) ;
    _SymVal(GaeaSymbol("$gs$infront_kickable_to_rpos"))->setfunc(gs_infront_kickable_to_rpos) ;
    _SymVal(GaeaSymbol("$gs$infront_kickable_to_apos"))->setfunc(gs_infront_kickable_to_apos) ;
    _SymVal(GaeaSymbol("$gs$closest_team_member_to_ball_p"))->setfunc(gs_closest_team_member_to_ball_p) ;

    _SymVal(GaeaSymbol("$gs$msg_is_empty"))->setfunc(gs_msg_is_empty) ;
    _SymVal(GaeaSymbol("$gs$msg_is_not_empty"))->setfunc(gs_msg_is_not_empty) ;
    _SymVal(GaeaSymbol("$gs$get_new_msg"))->setfunc(gs_get_new_msg) ;
    _SymVal(GaeaSymbol("$gs$playmode"))->setfunc(gs_playmode) ;

    _SymVal(GaeaSymbol("$gs$turn_to_adir"))->setfunc(gs_turn_to_adir) ;
    _SymVal(GaeaSymbol("$gs$turn_to_rpos"))->setfunc(gs_turn_to_rpos) ;
    _SymVal(GaeaSymbol("$gs$turn_to_apos"))->setfunc(gs_turn_to_apos) ;
    _SymVal(GaeaSymbol("$gs$kick_to_adir"))->setfunc(gs_kick_to_adir) ;
    _SymVal(GaeaSymbol("$gs$kick_to_rpos"))->setfunc(gs_kick_to_rpos) ;
    _SymVal(GaeaSymbol("$gs$kick_to_apos"))->setfunc(gs_kick_to_apos) ;
   
    _SymVal(GaeaSymbol("$gs$distance_kick"))->setfunc(gs_distance_kick) ;
    _SymVal(GaeaSymbol("$gs$circle_kick"))->setfunc(gs_circle_kick) ;
    _SymVal(GaeaSymbol("$gs$pass_ball_to"))->setfunc(gs_pass_ball_to) ;
    _SymVal(GaeaSymbol("$gs$pass_ball_to_apos"))->setfunc(gs_pass_ball_to_apos) ; //NODA
    _SymVal(GaeaSymbol("$gs$multi_pass_ball_to_apos"))->setfunc(gs_multi_pass_ball_to_apos) ; //NODA

    // 1.04o
    _SymVal(GaeaSymbol("$gs$multi_smart_kick_to"))->setfunc(gs_multi_smart_kick_to) ; 
    _SymVal(GaeaSymbol("$gs$multi_smart_kick_to_apos"))->setfunc(gs_multi_smart_kick_to_apos) ; 
    _SymVal(GaeaSymbol("$gs$multi_smart_kick_to2"))->setfunc(gs_multi_smart_kick_to2) ; 
    _SymVal(GaeaSymbol("$gs$multi_smart_kick_to_apos2"))->setfunc(gs_multi_smart_kick_to_apos2) ; 
    
    
    _SymVal(GaeaSymbol("$gs$step_goto2"))->setfunc(gs_step_goto2) ;
    _SymVal(GaeaSymbol("$gs$step_dash_to_ball2"))->setfunc(gs_step_dash_to_ball2) ;
    _SymVal(GaeaSymbol("$gs$get_plan_for_intercept"))->setfunc(gs_get_plan_for_intercept) ;

    //V.1.04noda3
    _SymVal(GaeaSymbol("$gs$get_plan_for_intercept2"))->setfunc(gs_get_plan_for_intercept2) ;

    // 1.04o
    _SymVal(GaeaSymbol("$gs$get_plan_for_goalie_intercept"))->setfunc(gs_get_plan_for_goalie_intercept) ;
    _SymVal(GaeaSymbol("$gs$set_goalie_max_steps_for_intercept"))->setfunc(gs_set_goalie_max_steps_for_intercept) ;
    _SymVal(GaeaSymbol("$gs$set_goalie_intercept_dash_power"))->setfunc(gs_set_goalie_intercept_dash_power) ;
    _SymVal(GaeaSymbol("$gs$set_goalie_misc_dash_power"))->setfunc(gs_set_goalie_misc_dash_power) ;
    _SymVal(GaeaSymbol("$gs$set_goalie_free_ball_speed"))->setfunc(gs_set_goalie_free_ball_speed) ;
    _SymVal(GaeaSymbol("$gs$get_goalie_action_values"))->setfunc(gs_get_goalie_action_values) ;
    
    
    _SymVal(GaeaSymbol("$gs$step_goto"))->setfunc(gs_step_goto) ;
    _SymVal(GaeaSymbol("$gs$step_swing"))->setfunc(gs_step_swing) ;

    _SymVal(GaeaSymbol("$gs$is_within_from"))->setfunc(gs_is_within_from) ;
    _SymVal(GaeaSymbol("$gs$set_position_param"))->setfunc(gs_set_position_param) ;
    _SymVal(GaeaSymbol("$gs$set_closest_check_param"))->setfunc(gs_set_closest_check_param) ;
    _SymVal(GaeaSymbol("$gs$shifted_home_pos"))->setfunc(gs_shifted_home_pos) ;
    _SymVal(GaeaSymbol("$gs$ball_is_in_my_area"))->setfunc(gs_ball_is_in_my_area) ;
    //V.1.04noda
    _SymVal(GaeaSymbol("$gs$find_good_pass_dest"))->setfunc(gs_find_good_pass_dest) ;
    //V.1.04noda
    _SymVal(GaeaSymbol("$gs$arith_mod"))->setfunc(gs_arith_mod) ;
    //V.1.04noda
    _SymVal(GaeaSymbol("$gs$distance"))->setfunc(gs_distance) ;
    //V.1.04noda2
    _SymVal(GaeaSymbol("$gs$closer_team_to_ball"))->setfunc(gs_closer_team_to_ball) ;
    //V.1.04noda2
    _SymVal(GaeaSymbol("$gs$ball_go_forward"))->setfunc(gs_ball_go_forward) ;
    _SymVal(GaeaSymbol("$gs$ball_go_backward"))->setfunc(gs_ball_go_backward) ;

    // v1.04o
    _SymVal(GaeaSymbol("$gs$kick_ball_to_near_mid_sideline"))->setfunc(gs_kick_ball_to_near_mid_sideline) ;
    // v1.04o
    _SymVal(GaeaSymbol("$gs$_get_nearest_opponent"))->setfunc(gs_get_nearest_opponent) ;

    //V.1.05noda
    _SymVal(GaeaSymbol("$gs$out_of_field"))->setfunc(gs_out_of_field) ;
    _SymVal(GaeaSymbol("$gs$should_not_chase_ball"))->setfunc(gs_should_not_chase_ball) ;
    
    // _SymVal(GaeaSymbol(""))->setfunc() ;

    EOT_GSPlayer = DefExtObjType("gsplayer") ;

    GsSymError = GaeaSymbol("error") ;

    // v.1.04o
    GsSymMe = GaeaSymbol("me");
    GsSymBall = GaeaSymbol("ball");

    GsSymSee = GaeaSymbol("see") ;
    GsSymHear = GaeaSymbol("hear") ;
    GsSymSenseBody = GaeaSymbol("sense_body") ;
    

    GsSymLeft = GaeaSymbol("left") ;
    GsSymRight = GaeaSymbol("right") ;
    GsSymUnknown = GaeaSymbol("unknown") ;

    // v1.04o
    GsSymRotLeft = GaeaSymbol("rot_left") ;
    GsSymRotRight = GaeaSymbol("rot_right") ;
    GsSymRotClosest = GaeaSymbol("rot_closest") ;
    GsSymRotFurthest = GaeaSymbol("rot_furthest") ;
    GsSymRotAvoid = GaeaSymbol("rot_avoid") ;
    //1.04o
    GsSymActionNone  = GaeaSymbol("action_none");
    GsSymActionKick  = GaeaSymbol("action_kick");
    GsSymActionTurn  = GaeaSymbol("action_turn");
    GsSymActionDash  = GaeaSymbol("action_dash");
    GsSymActionCatch = GaeaSymbol("action_catch");
    GsSymActionError = GaeaSymbol("action_error");
    
    GsSymReferee = GaeaSymbol("referee");
    GsSymSelf = GaeaSymbol("self");
    GsSymPlayer = GaeaSymbol("player") ;
    GsSymOpponent = GaeaSymbol("opponent") ;
    GsSymTeammate = GaeaSymbol("teammate") ;

    GsSymOnlineCoachL = GaeaSymbol("online_coach_left") ;
    GsSymOnlineCoachR = GaeaSymbol("online_coach_right") ;

    GsSymBeforeKickOff   = GaeaSymbol("before_kick_off") ;   
    GsSymPlayOn      = GaeaSymbol("play_on") ;   
    GsSymTimeOver    = GaeaSymbol("time_over") ; 
    GsSymKickOffL    = GaeaSymbol("kick_off_l") ;
    GsSymKickOffR    = GaeaSymbol("kick_off_r") ;
    GsSymKickInL     = GaeaSymbol("kick_in_l") ;
    GsSymKickInR     = GaeaSymbol("kick_in_r") ;
    GsSymFreeKickL   = GaeaSymbol("free_kick_l") ;
    GsSymFreeKickR   = GaeaSymbol("free_kick_r") ;
    GsSymCornerKickL     = GaeaSymbol("corner_kick_l") ;
    GsSymCornerKickR     = GaeaSymbol("corner_kick_r") ;
    GsSymGoalKickL   = GaeaSymbol("goal_kick_l") ;
    GsSymGoalKickR   = GaeaSymbol("goal_kick_r") ;
    GsSymGoalL       = GaeaSymbol("goal_l") ;
    GsSymGoalR       = GaeaSymbol("goal_r") ;
    GsSymFoulL       = GaeaSymbol("foul_l") ;
    GsSymFoulR       = GaeaSymbol("foul_r") ;
    GsSymTimeUpWOATeam   = GaeaSymbol("time_up_without_a_team") ;
    GsSymTimeUp      = GaeaSymbol("time_up") ;
    GsSymTimeExtended    = GaeaSymbol("time_extended") ;
    GsSymHalfTime    = GaeaSymbol("half_time") ;
    GsSymGoalieCatchL    = GaeaSymbol("goalie_catch_ball_l") ;
    GsSymGoalieCatchR    = GaeaSymbol("goalie_catch_ball_r") ;
    GsSymDropBall    = GaeaSymbol("drop_ball") ;
    GsSymOffSideL    = GaeaSymbol("offside_l") ;
    GsSymOffSideR    = GaeaSymbol("offside_r") ;

    //V.1.04noda3
    GsSymXY = GaeaSymbol("xy") ;
    //V.1.04noda
    GsSymAXY = GaeaSymbol("axy") ;
    GsSymHXY = GaeaSymbol("hxy") ;
    GsVarX = GaeaVariable("X") ;
    GsVarY = GaeaVariable("Y") ;
    GsTermXY2 = GaeaTerm(GsSymXY,GsVarX,GsVarY) ;
    GsTermAXY2 = GaeaTerm(GsSymAXY,GsVarX,GsVarY) ;
    GsTermHXY2 = GaeaTerm(GsSymHXY,GsVarX,GsVarY) ;

    //V.1.06: change view

    GsSymHigh = GaeaSymbol("high") ;
    GsSymLow = GaeaSymbol("low") ;
    GsSymWide = GaeaSymbol("wide") ;
    GsSymNormal = GaeaSymbol("normal") ;
    GsSymNarrow = GaeaSymbol("narrow") ;
    
    return True ;
} ;



