// -*- Mode:C++ -*-
//
//======================================================================
// for test
//======================================================================

#include "GSPlayer.h"

enum Mode {
    M_before_kick_off,
    M_play_on,
    M_tired,
    M_error,
} ;


void waitSeeInfo(GSPlayer& player, ScUInt n)
{
    for(ScUInt c = 0 ; c < n ;) {
        ScBool r = player.recvScanSInfo() ;
        if(r && player.sinfo.type == SC_SIT_see) c++ ;
    }
}

//----------------------------------------------------------------------
// sensor loop

void sensorLoop(GSPlayer& player)
{
    ScBool r;
    while(ScTrue) {
        r = player.recvScanSInfo() ;
    }
}


void* sensorLoopEntry(void* p)
{
    sensorLoop(*(GSPlayer*)p) ;
    return NULL;
}


//----------------------------------------------------------------------
// action loops


void kickTest(GSPlayer& player)
{
    Mode mode = M_before_kick_off ;

    scRandomizeByTime() ;

    ScInt pTime ;
    pTime = -10000 ;

    
    while(ScTrue) {
        if (mode == M_before_kick_off) {
            cout << "================================" << endl;
            if(player.scycle.count > 0) {
                mode = M_play_on ;
                continue;
            }
            player.move(0,-.7) ;
            scSleep(1,0);
        }


        else if (mode == M_play_on) {
            if (player.scycle.count - player.ecycle.count > 20) {
                cout << "================================" << endl;
                cout << "\n" << player << "I lost my position." << endl ;
                for(ScUInt i = 0 ; i < 6 ; i++) {
                    player.turn(80) ;
                }
            }

            else if (player.actualBodyState.stamina < 1000) {
                cout << "================================" << endl;
                cout << "\n" << player << "I am tired." << endl ;
                mode = M_tired ;
            }

            else if (player.DashKickableP(3)) {
                cout << "================================" << endl;
                cout << "\n###Kick!!!###. " << endl;
                player.printPositionInfo("(a)");

                // kick rightward for distance of 14.02/20.0 (goalwidth).
                player.distanceKick(0, 14.02/20.0, 1.0);

                player.printPositionInfo("(b)");

                player.waitForVisualInfo();
                
                player.printPositionInfo("(c)");
                
                cout << player << "\n###Nap!!!###. " << endl;
                scSleep(3,0); // nap for 2 secs.

            }

            else {
                cout << "================================" << endl;

                cout << "D " << flush;
                player.stepDashToBall2(40);
            }

        }

        else if (mode == M_tired) {
            cout << "================================" << endl;

            player.turn(10) ;
            if(player.actualBodyState.stamina > 3000) {
                cout << player << "I'm refleshed!!" << endl ;
                mode = M_play_on ;
            }
        }
    }

}


void dashTestLoop0(GSPlayer& player)
{
    Mode mode = M_before_kick_off ;
    ScInt actionCount = 0;
    ScPos dest(40, -.7);

    const int cStop = 3;
    
    while(actionCount < cStop) {
        if (mode == M_before_kick_off) {
            if(player.scycle.count > 0) {
                mode = M_play_on ;
                continue;
            }
            player.move(0,-.7) ;
            scSleep(1,0);
        }

        else if (mode == M_play_on) {

            if (actionCount == 0) {
                actionCount++;
                player.distanceKick(45, 10, 1);
            }
            
            else if (actionCount > 0  &&  actionCount < cStop-1) {
                cout << "==================================="<<endl;
                cout << actionCount-1 << endl;
                player.printPositionInfo("(a) ");
                cout << "NumDashes: "
                     << player.actualBodyState.countDash<<endl;
                player.stepGoto2(dest, 40, ScFalse);
                player.printPositionInfo("(b) ");
                ++actionCount;
            }

            else {
                cout << "\n\nLet vision catch up: (1),(2),(3)." << endl;
                player.printPositionInfo("(1) ");
                player.waitForVisualInfo();
                player.printPositionInfo("(2) ");
                player.waitForVisualInfo();
                player.printPositionInfo("(3) ");
                ++actionCount;
            }
        } // end M_play_on
    }
}


void dashTestLoop1(GSPlayer& player)
{
    Mode mode = M_before_kick_off ;
    ScInt actionCount = 0;
    ScPos dest(40, -.7);

    const int cStop = 50;
    
    while(actionCount < cStop) {
        if (mode == M_before_kick_off) {
            player.move(0,-.7) ;
            scSleep(1,0);
            if(player.scycle.count > 0) mode = M_play_on ;
        }

        else if (mode == M_tired) {
            player.turn(10) ;
            if(player.actualBodyState.stamina > 3000) {
                cout << player << "I'm refleshed!!" << endl ;
                mode = M_play_on ;
            }
        }
        
        else if (mode == M_play_on) {
            if (player.scycle.count - player.ecycle.count > 20) {
                cout << "\n" << player << "I lost my position." << endl ;
                for(ScUInt i = 0 ; i < 6 ; i++) {
                    player.turn(80) ;
                }
            }

            else if (player.actualBodyState.stamina < 1000) {
                cout << "\n" << player << "I am tired." << endl ;
                mode = M_tired ;
            }
            
            else if (actionCount == 0) {
                actionCount++;
                player.distanceKick(45, 10, 1);
            }
            
            else if (actionCount > 0  &&  actionCount < cStop-1) {
                cout << "==================================="<<endl;
                cout << actionCount-1 << endl;
                player.printPositionInfo("(a) ");
                cout << "NumDashes: "
                     << player.actualBodyState.countDash<<endl;
                player.stepGoto2(dest, 40, ScFalse);
                player.printPositionInfo("(b) ");
                ++actionCount;
            }

            else {
                cout << "\n\nLet vision catch up: (1),(2),(3)." << endl;
                player.waitForVisualInfo();
                player.printPositionInfo("(1) ");
                player.waitForVisualInfo();
                player.printPositionInfo("(2) ");
                player.waitForVisualInfo();
                player.printPositionInfo("(3) ");
                ++actionCount;
            }
        } // end M_play_on

        
    }
}


void passingTestLoop(GSPlayer& player)
{
    Mode mode = M_before_kick_off ;

    ScInt pTime ;
    pTime = -10000 ;

    cout << "Passing loop." << endl;
    
    while(ScTrue) {
        if (mode == M_before_kick_off) {
            cout << "================================" << endl;
            if(player.scycle.count > 0) {
                mode = M_play_on ;
                continue;
            }
            player.move(-20,0) ;
            scSleep(1,0);
        }


        else if (mode == M_play_on) {
            if (player.scycle.count - player.ecycle.count > 20) {
                cout << "================================" << endl;
                cout << "\n" << player << "I lost my position." << endl ;
                for(ScUInt i = 0 ; i < 6 ; i++) {
                    player.turn(80) ;
                }
            }

            else if (player.actualBodyState.stamina < 1000) {
                cout << "================================" << endl;
                cout << "\n" << player << "I am tired." << endl ;
                mode = M_tired ;
            }

            else if (player.DashKickableP(3)) {
                cout << "================================" << endl;
                cout << "\n###Pass!!!###. " << endl;
                player.printPositionInfo("(a)");

                ScBool pDoneP;
                ScFloat completionFlag = ScFalse;
                
                while (player.KickableP(3)) {
                    if (player.passBallTo(100, -180, &pDoneP)  &&  pDoneP) {
                        completionFlag = ScTrue;
                    }
                }
                
                if (completionFlag) {
                    cout << "###Neat!###."<<endl;
                }
                        
                //player.printPositionInfo("(b)");

                //player.waitForVisualInfo();
                //player.waitForVisualInfo();
                
                //player.printPositionInfo("(c)");
                
                //cout << player << "\n###Nap!!!###. " << endl;
                //scSleep(3,0); // nap for 2 secs.

                exit(1);

            }

            else {
                cout << "================================" << endl;

                cout << "D " << flush;
                player.stepDashToBall2(40);
            }

        }

        else if (mode == M_tired) {
            cout << "================================" << endl;

            player.turn(10) ;
            if(player.actualBodyState.stamina > 3000) {
                cout << player << "I'm refleshed!!" << endl ;
                mode = M_play_on ;
            }
        }
    }

}
    


void dribbleLoop(GSPlayer& player)
{
    Mode mode = M_before_kick_off ;

    scRandomizeByTime() ;

    ScPos destPos ;
    ScInt pTime ;
    pTime = -10000 ;

    ScPos dest[5];
    dest[0].x = 20;
    dest[0].y = 0;
    dest[1].x = 0;
    dest[1].y = 0;
    dest[2].x = -20;
    dest[2].y = 0;

    int destIndex = 0;

    destPos = dest[destIndex++];
    
    while(ScTrue) {
        if(mode == M_before_kick_off) {
            //
            // move to some random spot
            //
            //ScFloat x = scFltRand(-SC_HALF_PITCH_L,0) ;
            //ScFloat y = scFltRand(-SC_HALF_PITCH_W,SC_HALF_PITCH_W) ;
            //player.move(x,y) ;
            player.move(-20, 0);
                
            scSleep(0,500) ;
            player.turn(scFltRand(-90,90)) ; 
            scSleep(0,500) ;
            if(player.scycle.count > 0) mode = M_play_on ;

        } else if (mode == M_play_on) {
            if(player.scycle.count - player.ecycle.count > 20) {
                cout << player << "I lost my position." << endl ;
                for(ScUInt i = 0 ; i < 6 ; i++) {
                    player.turn(80) ;
                }

            } else if(player.actualBodyState.stamina < 1000) {
                cout << player << "I am tired." << endl ;
                mode = M_tired ;

            } else if (scDistance(player.xposdir.pos, destPos) < 2.0) {
                cout << player << "I'm there, dribble to new spot: " ;

                if (destIndex > 1) {
                    destPos.x = scFltRand(-SC_HALF_PITCH_L,SC_HALF_PITCH_L) ;
                    destPos.y = scFltRand(-SC_HALF_PITCH_W,SC_HALF_PITCH_W) ;
                } else {
                    destPos = dest[destIndex++];
                }
                
                cout << destPos << endl;                    

            } else {

                // !!!
                
                //cout << player << "Dribbling." << endl ;
                player.simpleDribble(destPos, 0, 20, 50, 10);
            }

        } else if (mode == M_tired) {
            player.turn(10) ;
            if(player.actualBodyState.stamina > 3000) {
                cout << player << "I'm refleshed!!" << endl ;
                mode = M_play_on ;
            }
        }
    }
} 


void stepSwingTestLoop(GSPlayer& player)
{
    ScPos goal;
    goal.x = SC_HALF_PITCH_L;
    goal.y = 0;

    ScBool inSpotP = false;
    ScFloat absAngle = 120;

    Mode mode = M_before_kick_off ;

    cout << "stepSwingTestLoop" << endl;

    while (ScTrue) {

        //cout << "should see this" << endl;

        if (mode == M_before_kick_off) {
            cout << "==================================="<<endl;
            cout << "stepSwingTestLoop: Pre-kickoff." << endl;
            player.move(-.5,0);
            scSleep(2,0);
            player.turnToAbs(goal);
            if(player.scycle.count > 0) mode = M_play_on ;

        } else {
            //cout << "game" << endl;

            if(player.scycle.count - player.ecycle.count > 20) {
                cout << "==================================="<<endl;
                cout << "stepSwingTestLoop: "<< player <<
                    " I lost my position." << endl ;
                for(ScUInt i = 0 ; i < 6 ; i++) {
                    player.turn(80) ;
                }
                continue;
            }

            while (!player.DashKickableP(3)) {
                cout << "==================================="<<endl;
                cout << "swingStepTestLoop: dash ("<<absAngle<<")"<<endl;
                cerr << "D (" << absAngle <<")"<<endl;
                player.stepDashToBall(40);
            }

            inSpotP = false;
            while (!inSpotP  &&  player.KickableP(3)) {
                cout << "==================================="<<endl;
                cout << "swingStepTestLoop: CircleKick to " << absAngle << endl;

                if (player.circleKick(absAngle,    // where to turn
                                      ScTrue,      // turn right-ways
                                      ScTrue,      // stop ball when done
                                      &inSpotP)) { // marker
                    cout << "ck " << endl;
                } else {
                    cout << "-- " << endl;
                }
            }

            if (inSpotP) {
                cout << "==================================="<<endl;
                cout << endl << "swingStepTestLoop: Done!"<<endl;
                cerr << endl << "swingStepTestLoop: Done!"<<endl;
                cout << "swingStepTestLoop: Desired angle: "<<absAngle<<endl;
                cerr << "swingStepTestLoop: Desired angle: "<<absAngle<<endl;
                cout << "swingStepTestLoop: Actual angle: "<<
                    player.ballAngleRel();
                cerr << "swingStepTestLoop: Actual angle: "<<
                    player.ballAngleRel();
                absAngle = scNormalizeAngle(absAngle + 120);
                cout << endl <<
                    "swingStepTestLoop: New absAngle: "
                     << absAngle << " (nap)" << endl;
                cerr << endl <<
                    "swingStepTestLoop: New absAngle: "
                     << absAngle << " (nap)" << endl;
                scSleep(2,0);
            } else {
                cout << "==================================="<<endl;
                cout << endl << "swingStepTestLoop: Lost Ball!" << endl;
            }
        }
    }

    cout << "hmm, bad mojo" << endl;
}


//----------------------------------------------------------------------
// option

#define DEFAULT_HOST		"localhost"
#define DEFAULT_PORT		6000
#define DEFAULT_TEAMNAME	"Japan"
#define DEFAULT_PROTO_VERSION	5.0

char* SHost = DEFAULT_HOST ;
ScInt SPort = DEFAULT_PORT ;
char* TeamName = DEFAULT_TEAMNAME ;
ScFloat ProtocolVersion = DEFAULT_PROTO_VERSION ;
ScBool GoalieP = ScFalse ;

ScBool getOption(int argc, char** argv) {
    for(int i = 1 ; i < argc ; i++) {
        if        (!strcmp(argv[i],"-host")) {
            i++ ; SHost = argv[i] ;
        } else if (!strcmp(argv[i],"-port")) {
            i++ ; SPort = atol(argv[i]) ;
        } else if (!strcmp(argv[i],"-team")) {
            i++ ; TeamName = argv[i] ;
        } else if (!strcmp(argv[i],"-protocol")) {
            i++ ; ProtocolVersion = atof(argv[i]) ;
        } else if (!strcmp(argv[i],"-goalie")) {
            GoalieP = ScTrue ;
        } else {
            if (strcmp(argv[i],"-help")) {
                cerr << "Illegal option:" << argv[i] << endl ;
            } 
            cerr << "Usage:" << endl ;
            cerr << "\t % gstest [options]" << endl ;
            cerr << "\t     [options]" << endl ;
            cerr << "\t\t-host HOST" << endl ;
            cerr << "\t\t-port PORT" << endl ;
            cerr << "\t\t-team TEAM" << endl ;
            cerr << "\t\t-protocol PROTOCOL" << endl ;
            cerr << "\t\t-goalie GOALIE" << endl ;
            exit(1) ;
        }
    }
    return ScTrue ;
}

	    

//----------------------------------------------------------------------
// main for test

main(int argc, char** argv) {

    getOption(argc,argv) ;

    // cout << "Try to open socket to host:port=" 
    // << SHost << ":" << SPort << endl ;

    GSPlayer player ;
    player.init(ScString(TeamName),SHost,SPort,ScFalse) ;
   
    // cout << "Succeed to open socket." << endl ;

    if(!player.sendInit(ProtocolVersion,GoalieP)) {
        cerr << "Can't send init command" << endl ;
        exit(1) ;
    }

    // cout << "Start sensor loop." << endl ;

    pthread_t sensor_thread ;
    pthread_create(&sensor_thread, SC_NULLPTR, sensorLoopEntry,
                   (void*)&player) ;

    // cout << "Enter simple command loop." << endl ;

    //dashTestLoop0(player);
    //dashTestLoop1(player);
    //kickTest(player) ;
    passingTestLoop(player);
    //stepSwingTestLoop(player);
    //dribbleLoop(player) ;
}

