%% -*- mode: prolog -*-
%% 
%%

%%======================================================================
%% load library

binload("libsclient/libsclient.so").
binload("bipred_gsc.so").

%%======================================================================
%% basic util

$rule(mem(Item,[Item|_]),!).
$rule(mem(Item,[_|Rest]),
	$and(!,mem(Item,Rest))).

$rule(my_team_has_free_kick(Player), $and(
    $gs$side_unum(Player, left, _),
    $gs$playmode(Player, free_kick_l),
    !)).
$rule(my_team_has_free_kick(Player), $and(
    $gs$side_unum(Player, right, _),
    $gs$playmode(Player, free_kick_r),
    !)).

%$rule(write_myname(_),!).
$rule(write_myname, $and(player(Player),write_myname(Player))).
$rule(write_myname(Player), $and(
	$gs$side_unum(Player,Side,UNum),
	$write(Side),$write(UNum),$write_special_char(':'),
	!)).

%$rule(dlog(_,_),!).
%$rule(dlog(_),!).

$rule(dlog(Msg), $and(
	player(Player),
	dlog(Player,Msg),
	!)).

$rule(dlog(Player,Msg),$and(not(goalie),!)).

$rule(dlog(Player,Msg), $and(
	write_myname(Player),
	$write(Msg),$nl,
	!)).


%%======================================================================
%% cell list

newcell(gsinit).

newcell(sensor).
newcell(listener).
newcell(motor).
newcell(instinct).
newcell(planner).

newcell(planner0).

newcell(planner_before_kickoff).

newcell(subsumed_by_planner).
newcell(subsumed_by_instinct).

newcell(base_com).
newcell(base_com_move).

newcell(stay_init_pos0).
newcell(dango0).
newcell(dango).
newcell(dango_goalie).   % 1.04o

%% V.1.04noda3
newcell(with_say_chase).

%% V.1.04noda
newcell(fixed_pass_dest).

%V.1.05noda
%newcell(pn_home_pos_base).
%newcell(pn_stay_home_base).

newcell(confirm_pos).
%%V.1.05noda
newcell(look_at_ball).
newcell(find_ball).
%%V.1.05noda
newcell(back_to_field).

newcell(goto_home).
newcell(hurry_to_home).

%%V.1.04noda
newcell(pass_request).

newcell(stay_and_look_around).
newcell(stay_and_look_ball).
newcell(clear_ball_to_side).

newcell(goalie_caught_ball).  % 1.04o

newcell(xxposition_table).
newcell(position_table).
newcell(position_base).
newcell(myposition).

newcell(pnbase_home_pos).
newcell(pnbase_stay_home).
newcell(pnbase_hurry_home).

%%V.1.05noda
newcell(set_play_kicker).

%%======================================================================
%% default parameter

push(gsinit).

$fact(teamname("KappaII")).
$fact(shost("localhost")).
%$fact(shost("genji")).
$fact(sport(6000)).
%$fact(sport(6010)).

%%V.1.04noda2
$rule(player_setup(Team,Host,Port), $and(
	$write(teamname(Team)),$nl,
	$write(shost(Host,Port)),$nl,
	$asserta_fact(teamname(Team),gsinit),
	$asserta_fact(shost(Host),gsinit),
	$asserta_fact(sport(Port),gsinit),
	!)).
$rule(run_sleep(Team,Host,Port), $and(
	player_setup(Team,Host,Port),
	foo,
	sleep_forever)).
$rule(grun_sleep(Team,Host,Port), $and(
	player_setup(Team,Host,Port),
	gfoo,
	sleep_forever)).

pop(_).

%%======================================================================
%% test run

push(gsinit).

$rule(foo,
	$and(	init,
		$sleep(1),
		% with(base_com_move,move_random),
		with(base_com_move,decide_position_by_occupancy),
		$sleep(1),
		start_sensor,
		start_listener,
		start_instinct,
		start_planner,
		start_motor)).

$rule(gfoo,
	$and(	ginit,
		$sleep(1),
		% with(base_com_move,move_random),
		with(base_com_move,decide_position_by_occupancy),
		$sleep(1),
		start_sensor,
		start_listener,
		start_instinct,
		start_planner,
		start_motor)).

$rule(sleep_forever, $sleep_forever).

pop(_).

%%======================================================================
%% real toplevel run

%%======================================================================
%% initialize

push(gsinit).

$rule(init,
	$and(	teamname(Team),
		shost(SHost),sport(SPort),
		$gs$create_player(Player,Team,SHost,SPort),
		$gs$send_init(Player),
		$gensym(Work,player),
		$newcell(Work),$push(Work),
		$gs$set_closest_check_param(Player,10),
		aa(work(Work)),
		aa(player(Player)))).

$rule(ginit,
	$and(	teamname(Team),
		shost(SHost),sport(SPort),
		$gs$create_player(Player,Team,SHost,SPort),
		$gs$send_ginit(Player),
		%% V.1.06 : goalie always uses narrow view
		$msleep(0,300),
		$gs$change_view(Player,narrow,high),
		$gensym(Work,player),
		$newcell(Work),$push(Work),
		$gs$set_closest_check_param(Player,10),
		aa(work(Work)),
		aa(goalie),
		aa(player(Player)))).

pop(_).

%%======================================================================
%% fork process

push(gsinit).

$rule(gsfork(Name,SecondEnv,RemoveEnv,Goal),
	$and(	$fork($and(pushsecond_env_list(SecondEnv),
			   remove_env_list(RemoveEnv),!,
			   Goal),
		      PID),
		aa(pid(Name,PID)))).
$rule(pushsecond_env_list([]),!).
$rule(pushsecond_env_list([Cell|Rest]),
	$and(	pushsecond_env_list(Rest),
		!,
		once(pushsecond(Cell)))).

$rule(remove_env_list([]),!).
$rule(remove_env_list([Cell|Rest]),
	$and(	once($remove_cell(Cell)),
		!,
		once(remove_env_list(Rest)))).

pop(_).

%%======================================================================
%% sensor

push(sensor).

$rule(cycle(Player),
	$and(	$gs$recv_scan_sinfo(Player,Type),
		$signal(Type),
		!)).

$rule(main_loop,
	$and(	player(Player),
		$repeat,
		cycle(Player),
		$fail)).

pop(_).

push(gsinit).
$rule(start_sensor,
	gsfork(sensor,[sensor],[gsinit],main_loop)).
pop(gsinit).

%%======================================================================
%% listener

push(listener).

$rule(cycle(Player),
	$and(	$wait_signal(Signal,[hear]),!,
		cycle(Player,Signal),
		!)).

$rule(cycle(Player,hear),
	$and(	$while($gs$msg_is_not_empty(Player)),
		$gs$get_new_msg(Player,Time,Sender,DirPMode,Msg),
		cycle(Player,Time,Sender,DirPMode,Msg),
		$fail)).
$rule(cycle(Player,hear),!).

$rule(cycle(Player,Time,referee,DirPMode,_),
	$and(	$signal(referee),
	%	$write(message_from_referee(Time,DirPMode)),$nl,
		!)).
$rule(cycle(Player,Time,player,DirPMode,Msg),
	$and(	$gs$scan_msg_list(Player,Msg,MList),
		reaction(Player,Time,DirPMode,MList),
		!)).

$rule(cycle(Player,Time,self,DirPMode,Msg),!).
	
$rule(cycle(Player,Time,Sender,DirPMode,Msg),
	$and(	$write(Time),$write_special_char(','),
		$write(Sender),$write_special_char(','),
		$write(DirPMode),$write_special_char(','),
		$gs$scan_msg_list(Player,Msg,MList),
		$write(MList),$nl,
		!)).

%%V.1.04noda2
$rule(reaction(Player,Time,Dir,opponent),!).

$rule(reaction(Player,Time,Dir,[routine,UNum,Position,PX,PY,BX,BY,BTime]),
	$and(	
	%	$gs$set_ball_pos_if_need(Player,BX,BY,BTime),
		with(position_table,set_occupancy(Position,1.0)),
	%	$write(rt(UNum,Position)),$nl,
		!)).
$rule(reaction(Player,Time,Dir,[position_change,UNum,Position]),
	$and(	with(position_table,set_occupancy(Position,1.0)),
	%	$write(cp(UNum,Position)),$nl,
		!)).

%% V.1.04noda
$rule(reaction(Player,Time,Dir,[pass_request,UNum,SCycle,DX,DY]),
	$and(	$cv_put(fixed_pass_dest,cycle,SCycle,force),
		$cv_put(fixed_pass_dest,destx,DX,force),
		$cv_put(fixed_pass_dest,desty,DY,force),
	%	$write(cp(UNum,Position)),$nl,
		!)).

%% V.1.04noda3
$rule(reaction(Player,Time,Dir,[chase,UNum,SC,Dist,Stamina,Effort]), $and(
	$cv_put(with_say_chase,unum,UNum,force),
	$cv_put(with_say_chase,cycle,SC,force),
	$cv_put(with_say_chase,distance,Dist,force),
	$cv_put(with_say_chase,stamina,Stamina,force),
	$cv_put(with_say_chase,effort,Effort,force),
%	$write(chase(UNum,SC,Dist)),$nl,
	!)).

%% V.1.04noda
$rule(reaction(Player,Time,Dir,MSG),
	$and(	
		$write(unknown(Time,Dir,MSG)),$nl,
		!)).

$rule(main_loop,
	$and(	player(Player),
		$repeat,
		cycle(Player),
		$fail)).
pop(_).

push(gsinit).
$rule(start_listener,
	gsfork(listener,[listener],[gsinit],main_loop)).	
pop(_).

%%======================================================================
%% instinct

push(instinct).

$rule(cycle(Player),
	$and(	$timedwait_signal(Signal,[never],1,0),
		cycle(Player,Signal))).

%%V.1.05noda
$rule(cycle(Player,_), $and(
	$gs$out_of_field(Player),
	$gs$side_unum(Player,Side,UNum),
	write_myname(Player),
	$write("I'm out side of the field"),$nl,
	needs(back_to_field,subsumed_by_instinct),
	$fail)).

$rule(cycle(Player,_),
	$and(	$gs$cycle(Player,S,V,A,B,E),
		$is(Diff,sub(S,E)),
		$comp_gt(Diff,7),
		needs(confirm_pos,subsumed_by_instinct),
		$fail)).

$rule(cycle(Player,_),
	$and(	$gs$cycle(Player,S,V,A,B,E),
		$gs$ball_pos(Player,X,Y,VX,VY,BTime),
		$is(Diff,sub(S,BTime)),
		% V.1.06
		$comp_gt(Diff,7),
		needs(find_ball,subsumed_by_instinct),
		%% V.1.05noda
		needs(look_at_ball,subsumed_by_instinct),
		$fail)).

$rule(cycle(Player,_),
	$and(	$gs$cycle(Player,S,V,A,B,E),
		$gs$side_unum(Player,Side,UNum),
		$gs$epos(Player,PX,PY,PBDir,PHDir),
		$gs$ball_pos(Player,BX,BY,BVX,BVY,BTime),
		%  $gs$say_list(Player,[Side,UNum,S,PX,PY,E,BX,BY,BTime]),
		with(myposition,position(Position)),
		$is(MS,irand(100)),$msleep(0,MS),
		$gs$say_list(Player,[routine,UNum,Position,PX,PY,BX,BY,BTime]),
		$fail)).

$rule(main_loop,
	$and(	player(Player),
		$repeat,
		cycle(Player),
		$fail)).

pop(_).

push(gsinit).
$rule(start_instinct,
	gsfork(instinct,[instinct],[gsinit],main_loop)).
pop(_).

%%======================================================================
%% planner with planning

push(gsinit).
$rule(start_planner,
	gsfork(planner,[planner],[gsinit],main_loop)).
%	gsfork(planner,	[planner_before_kickoff,planner],
%		[gsinit],
%		main_loop)).
pop(_).

load("strategy.g").

push(planner).
push(planning_base,planner).

$rule(main_loop, $and(toploop)).

%% for competition
$rule(write_node,!).
$rule(write_node,$and(
	node(Node),
	player(Player),
	$gs$side_unum(Player,Side,UNum),
	$write(Side),$write(UNum),$write_special_char(':'),
	$write(Node),$nl,
	!)).

pop(_).

%%----------------------------------------------------------------------
%% define strategy nodes

push(strategy).

$cv_put(planning_base,topnode,pn_root,force).

%%--------------------------------------------------
%% root

define_node(pn_root,10000000,[],[pn_top],0.0).
push(pn_root).
$rule(condition(0.0),!).
$rule(activate, !).
$rule(deactivate,!).
pop(_).

%%--------------------------------------------------
%% top

%1.04o
define_node(pn_top,500,[],[pn_before_kickoff,pn_play,pn_goalie],0.1).
%define_node(pn_top,500,[],[pn_before_kickoff,pn_play],0.1).
push(pn_top).
$rule(condition(1.0),!).
$rule(activate, $and(
	write_node,
	$set_environment1([],subsumed_by_planner),
	!)).
$rule(deactivate,!).
pop(_).

%%--------------------------------------------------
%% before kickoff

define_node(pn_before_kickoff,1000,[referee],[pn_change_pos],0.1).
push(pn_before_kickoff).
$rule(condition(1.0),$and(
	player(Player),
	$gs$playmode(Player,PM),
	member(PM,[before_kick_off,goal_l,goal_r]),
	!)).
$rule(condition(0.0),!).
$rule(activate, $and(
	write_node,
	$set_environment1([stay_and_look_around],subsumed_by_planner),
	!)).
$rule(deactivate,!).
pop(_).

%%--------------------------------------------------
%% change_pos

define_node(pn_change_pos,300,[],[],0.0).
push(pn_change_pos).
$rule(condition(0.0),$and(
	goalie, 
	%% V.1.06
	$save_cutlevel(CL),
	with(base_com_move,moveto_initpos),
	$restore_cutlevel(CL),
	!)).
$rule(condition(0.0),$and(
	$save_cutlevel(CL),
	with(position_table,decay_all_occupancy(0.9)),
	with(position_table,find_minimum_occupancy(NewPos,NewOcc)),
	with(base_com_move,change_position(NewPos)),
	with(myposition,position(OldPos)),
	$restore_cutlevel(CL),
	$not_unify(NewPos,OldPos),
	$write(p(OldPos,NewPos)),$nl,
	$fail)).
$rule(condition(0.0),$and(
	$save_cutlevel(CL),
	with(base_com_move,moveto_initpos),
	$restore_cutlevel(CL),
	!)).
$rule(activate, $and(
	write_node,
	!)).
$rule(deactivate,!).
pop(_).

%%--------------------------------------------------
%% play on

% v1.04o
%define_node(pn_play,10000,[],[pn_normal],0.1).
define_node(pn_play,10000,[],[pn_normal,pn_clear_ball_to_side],0.1).

push(pn_play).
$rule(condition(0.5),!).
$rule(activate, $and(
	write_node,
	$set_environment1([],subsumed_by_planner),
	!)).
$rule(deactivate,!).
pop(_).

%%--------------------------------------------------
%% normal mode

define_node(pn_normal,300,[referee],[pn_chase_ball, pn_home_pos],0.1).
push(pn_normal).
$rule(condition(0.3),!).
$rule(activate, $and(
	write_node,
	$set_environment1([],subsumed_by_planner),
	!)).
$rule(deactivate,!).
pop(_).

%%--------------------------------------------------
%% home position mode

%%V.1.04noda2
define_node(pn_home_pos,150,[referee],[pn_stay_home,pn_hurry_home],0.1).

%%V.1.04noda3
push(pnbase_home_pos,pn_home_pos).


push(pnbase_home_pos).
$rule(condition(0.8),$and(
	player(Player),
	$gs$playmode(Player,PMode),
	$gs$side_unum(Player,Side,UNum),
	is_stay_mode(Side,PMode),
	!)).
$rule(condition(0.5),!).

$rule(activate, $and(
	write_node,
	$set_environment1([goto_home],subsumed_by_planner),
	!)).
$rule(deactivate,!).

$rule(is_stay_mode(left,PMode),
	$and(mem(PMode,[before_kick_off,time_over,
			kick_off_r,kick_in_r,free_kick_r,
			corner_kick_r,goal_kick_r,
			time_up,time_up_without_a_team]),
	     !)).
$rule(is_stay_mode(right,PMode),
	$and(mem(PMode,[before_kick_off,time_over,
			kick_off_l,kick_in_l,free_kick_l,
			corner_kick_l,goal_kick_l,
			time_up,time_up_without_a_team]),
	     !)).

pop(_).


%%--------------------------------------------------
%% stay home position mode

define_node(pn_stay_home,150,[referee],[],0.1).
push(pnbase_stay_home,pn_stay_home).

push(pnbase_stay_home).
$rule(condition(0.5),$and(
	player(Player),
%	with(myposition,home_pos(X,Y)),
	$gs$shifted_home_pos(Player,X,Y),
	$gs$is_within_from(Player,X,Y,2),
	!)).
$rule(condition(0.0),!).
$rule(activate, $and(
	write_node,
	%% V.1.04noda
	$set_environment1([pass_request,stay_and_look_ball],
			  subsumed_by_planner),
	!)).
$rule(deactivate,!).

pop(_).


%%V.1.04noda2
%%--------------------------------------------------
%% stay home position mode

define_node(pn_hurry_home,1000,[],[],0.1).
push(pnbase_hurry_home,pn_hurry_home).

push(pnbase_hurry_home).
%%V.1.05n28
$rule(condition(0.3),$and(
	player(Player),
	$gs$closer_team_to_ball(Player,30,DistDiff),
	%%V.1.05n28
	%%$write(cttb(DistDiff)),$nl,
	hurry_up_situation(Player,DistDiff),
	!)).
$rule(condition(0.0),!).

%% V.1.05noda29
$rule(hurry_up_situation(Player,DistDiff), $and( %% attacking phase
	$comp_gt(DistDiff,1.0),
	$gs$ball_go_forward(Player,0.1),
%	$gs$ball_go_forward(Player,0.5),
	!)).		

$rule(hurry_up_situation(Player,DistDiff), $and( %% defending phase
	$comp_lt(DistDiff,-1.0),
	$gs$ball_go_backward(Player,0.1),
%	$gs$ball_go_backward(Player,0.5),
	!)).

$rule(activate, $and(
	write_node,
	%% V.1.04noda
	$set_environment1([hurry_to_home],
			  subsumed_by_planner),
	!)).
$rule(deactivate,!).

pop(_).

%%--------------------------------------------------
%% chase ball

%%V.1.05noda28
%%V.1.05noda
define_node(pn_chase_ball,150,[referee],[pn_home_pos2,pn_set_play_kick,pn_direct_shoot],0.1).
push(pn_chase_ball).
$rule(condition(0.7),$and(
	player(Player),
	$gs$ball_is_in_my_area(Player),
	!)).
$rule(condition(0.7),$and(
	player(Player),
	$gs$closest_team_member_to_ball_p(Player),
	!)).
$rule(condition(0.0),!).
$rule(activate, $and(
	write_node,
	%% V.1.04noda3
	%% V.1.05noda
	$set_environment1([with_say_chase,fixed_pass_dest],
		          subsumed_by_planner),
	!)).
$rule(deactivate,!).
pop(_).


%%V.1.04noda3
%%--------------------------------------------------
%% home pos in chase
define_node(pn_home_pos2,150,[referee],[pn_stay_home2,pn_hurry_home2],0.1).
push(pnbase_home_pos,pn_home_pos2).

push(pn_home_pos2).
%%V.1.04noda3
$rule(xxcondition(0.5),$and(
	player(Player),
	$gs$cycle(Player,SC,_,_,_,_),
	$cv_get(with_say_chase,cycle,CC,no,no),
	$is(DiffC,sub(SC,CC)),$comp_le(DiffC,4),
	$cv_get(with_say_chase,distance,CDist,no,no),
	$gs$ball_distance(Player,BDist),
	$is(DiffDist,sub(BDist,CDist)),$comp_gt(DiffDist,2),
	$cv_get(with_say_chase,stamina,CStamina,no,no),
	$gs$stamina(Player,Stamina,Effort),
	$is(DiffStam,sub(Stamina,CStamina)),$comp_lt(DiffStam,500),
%	$write("give up chasing"),$nl,
	!)).
$rule(condition(0.5),$and(
	player(Player),
	$cv_get(with_say_chase,cycle,CC,no,no),
	$cv_get(with_say_chase,distance,CDist,no,no),
	$cv_get(with_say_chase,stamina,CStamina,no,no),
	$gs$should_not_chase_ball(Player,CC,4,CDist,2,CStamina,500),
%	$gs$side_unum(Player,S,U),$write(x(S,U,makashita)),$nl,
	!)).

$rule(condition(0.0),!).
pop(_).

define_node(pn_stay_home2,150,[referee],[],0.1).
push(pnbase_stay_home,pn_stay_home2).

define_node(pn_hurry_home2,1000,[],[],0.1).
push(pnbase_hurry_home,pn_hurry_home2).


%%V.1.05noda
%%--------------------------------------------------
%% set play
define_node(pn_set_play_kick,10000,[referee],[],0.1).
push(pn_set_play_kick).
$rule(condition(0.3),$and(
	player(Player),
	$gs$playmode(Player,PMode),
	$gs$side_unum(Player,Side,UNum),
	is_set_play_mode(Side,PMode),
	!)).
$rule(condition(0.0),!).
$rule(activate,$and(
	write_node,
	$set_environment1([set_play_kicker],subsumed_by_planner),
	!)).
$rule(deactivate,!).

%%V.1.05noda
$rule(is_set_play_mode(left,PMode), $and(
	mem(PMode,[kick_off_l,kick_in_l,free_kick_l,
	    corner_kick_l,goal_kick_l]),
	!)).
$rule(is_set_play_mode(right,PMode), $and(
	mem(PMode,[kick_off_r,kick_in_r,free_kick_r,
	    corner_kick_r,goal_kick_r]),
	!)).
			
pop(_).


%%V.1.05noda28
%%--------------------------------------------------
%% direct shoot

define_node(pn_direct_shoot,1000,[see],[],0.1).
push(pn_direct_shoot).
$rule(condition(0.2),$and(
	player(Player),
	$gs$ball_pos(Player,BX,BY,_,_,_),
	$comp_gt(BX,40),
	!)).
$rule(condition(0.0),!).
$rule(activate,$and(
	write_node,
	$set_environment1([],subsumed_by_planner),
	!)).
$rule(deactivate,!).

pop(_).

% v1.04o
%%--------------------------------------------------
%% clear ball to side  
%%   pn_root.pn_top.pn_play.
%%              pn_clear_ball_to_side
%%   siblings:
%%              pn_normal
%%
define_node(pn_clear_ball_to_side,150,[],[],0.1).

push(pn_clear_ball_to_side).

$rule(condition(0.5), $and(
    player(Player),
    $gs$dash_kickable(Player),
    $gs$ball_pos(Player, X, Y, _, _, _),
    %% V.1.05noda29
    %% V.1.05noda
    $comp_lt(X, -25), $comp_gt(Y,-25),$comp_lt(Y,25),
    !)).

$rule(condition(0.0),!).

$rule(activate, $and(
    write_node,
    $set_environment1([clear_ball_to_side], subsumed_by_planner),
    !)).

$rule(deactivate,!).

pop(_).


% v1.04o
%%--------------------------------------------------
%%   pn_root.pn_top.
%%              pn_goalie
%%   siblings:
%%              pn_before_kickoff
%%              pn_play
%% --------------
%% goalie 

%%V.1.06
define_node(pn_goalie,10000,[referee],
            [pn_not_in_our_half, pn_in_our_half, pn_caught, pn_our_freekick],0.1).
push(pn_goalie).
$rule(condition(0.8), $and(
    goalie,
    !)).
$rule(condition(0.0),!).

% this activate should only be called once.
$rule(activate, $and(
	write_node,
	$set_environment1([],subsumed_by_planner),
	!)).
$rule(deactivate,!).
pop(_).


% 1.04o
%%--------------------------------------------------
%%   pn_root.pn_top.pn_goalie.
%%              pn_not_in_our_half
%%   siblings:
%%              pn_in_our_half
%%              pn_caught
%% --------------
%% goalie, ball is far away, so just hang out
define_node(pn_not_in_our_half, 1000, [], [pn_goalie_home_pos], 0.1).
push(pn_not_in_our_half).
$rule(condition(0.2), $and(
    player(Player),
    $gs$ball_pos(Player, X, Y, _, _, _),
    $comp_gt(X, 0),
    !)).
$rule(condition(0.0), !).
    
$rule(activate, $and(
    write_node,
    player(Player),
    $gs$set_goalie_intercept_dash_power(Player, 100), %1.05o
    $gs$set_goalie_misc_dash_power(Player, 10), %1.05o
	%% V.1.06: remove comment out
	$set_environment1([],subsumed_by_planner),
    !)).
$rule(deactivate,!).
pop(_).

define_node(pn_goalie_home_pos,150,[referee],[pn_goalie_stay_home],0.1).
push(pnbase_home_pos, pn_goalie_home_pos).

define_node(pn_goalie_stay_home,150,[referee],[],0.1).
push(pnbase_stay_home, pn_goalie_stay_home).


%%--------------------------------------------------
%%   pn_root.pn_top.pn_goalie
%%              pn_in_our_half
%%   siblings:
%%              pn_not_in_our_half
%%              pn_caught
%% --------------
%% goalie, ball is near our goal, give control to dango_goalie
define_node(pn_in_our_half, 150, [], [], 0.1).
push(pn_in_our_half).

$rule(condition(0.3), $and(
    player(Player),
    $gs$ball_pos(Player, X, Y, _, _, _),
    $comp_lt(X, 0),
    !)).
$rule(condition(0.2),!).

$rule(activate, $and(
    write_node,
    player(Player),
    $gs$set_goalie_intercept_dash_power(Player, 100), %1.05o
    $gs$set_goalie_misc_dash_power(Player, 30), %1.05o
	$set_environment1([],subsumed_by_planner),
    !)).
$rule(deactivate,!).

pop(_).

%%--------------------------------------------------
%%   pn_root.pn_top.pn_goalie
%%              pn_caught
%%   siblings:
%%              pn_in_our_half
%%              pn_not_in_our_half
%% --------------
%% goalie, caught ball.
define_node(pn_caught, 100, [referee], [], 0.1).
push(pn_caught).

$rule(condition(0.5), $and(
    % $write("Caught conditional."), $nl,
    player(Player),
    $gs$ball_pos(Player, X, Y, _, _, _),
    $comp_lt(X, 0),
    $gs$catchable_distance(Player),
    my_team_has_free_kick(Player),
    % $write("Caught conditional should be on."), $nl,
    !)).
$rule(condition(0.0),!).

$rule(activate, $and(
    write_node,
    $set_environment1([goalie_caught_ball], subsumed_by_planner),
    !)).
$rule(deactivate,!).

% V.1.06
%%--------------------------------------------------
%%   pn_root.pn_top.pn_goalie
%%              pn_our_freekick
%%   siblings:
%%              pn_in_our_half
%%              pn_not_in_our_half
%% --------------
%% goalie, go to home

define_node(pn_our_freekick, 100, [referee], [pn_our_freekick_stay_home], 0.1).
push(pn_our_freekick).
$rule(condition(0.4),$and(
	player(Player),
	$gs$playmode(Player,PMode),
	$gs$side_unum(Player,Side,UNum),
	is_stay_mode(Side,PMode),
	!)).

$rule(activate, $and(
	write_node,
	$set_environment1([goto_home],subsumed_by_planner),
	!)).
$rule(deactivate,!).

$rule(is_stay_mode(left,PMode),
	$and(mem(PMode,[kick_off_l,kick_in_l,free_kick_l,
			corner_kick_l,goal_kick_l]),
	     !)).
$rule(is_stay_mode(right,PMode),
	$and(mem(PMode,[kick_off_r,kick_in_r,free_kick_r,
			corner_kick_r,goal_kick_r]),
	     !)).

define_node(pn_our_freekick_stay_home, 100, [referee], [], 0.1).
push(pnbase_stay_home,pn_our_freekick_stay_home).

%%--------------------
pop(_). %% strategy

%%======================================================================
%% planner version 0

push(planner0).

$rule(is_stay_mode(left,PMode),
	$and(mem(PMode,[before_kick_off,time_over,
			kick_off_r,kick_in_r,free_kick_r,
			corner_kick_r,goal_kick_r,
			time_up,time_up_without_a_team]),
	     !)).
$rule(is_stay_mode(right,PMode),
	$and(mem(PMode,[before_kick_off,time_over,
			kick_off_l,kick_in_l,free_kick_l,
			corner_kick_l,goal_kick_l,
			time_up,time_up_without_a_team]),
	     !)).

$rule(cycle(Player),
	$and(	$timedwait_signal(Signal,[referee],0,200),
		cycle(Player,Signal))).

$rule(cycle(Player,_),
	$and(	check_stay_mode(Player),
		$fail)).

$rule(check_stay_mode(Player),
	$and(	$gs$playmode(Player,PMode),
		$gs$side_unum(Player,Side,UNum),
		is_stay_mode(Side,PMode),
		needs(stay_and_look_around,subsumed_by_planner),
		!)).
$rule(check_stay_mode(Player),
	$and(	$remove_cell(stay_and_look_around,subsumed_by_planner),
		needs(goto_home,subsumed_by_planner),
		!)).

$rule(main_loop,
	$and(	player(Player),
		$repeat,
		cycle(Player),
		$fail)).

pop(_).

push(planner_before_kickoff).

$rule(cycle(Player,_),
	$and(	$gs$playmode(Player,PMode),
		$not_unify(PMode,before_kick_off),
		$remove_this_cell,
		!)).
$rule(cycle(Player,_),
	$and(	$save_cutlevel(CL),
		with(position_table,decay_all_occupancy(0.9)),
		with(position_table,find_minimum_occupancy(NewPos,NewOcc)),
		with(base_com_move,change_position(NewPos)),
		with(myposition,position(OldPos)),
		$not_unify(NewPos,OldPos),
		$write(p(OldPos,NewPos)),$nl,
		$restore_cutlevel(CL),
		$fail)).
$rule(cycle(Player,_),
	$and(	$save_cutlevel(CL),
		with(base_com_move,moveto_initpos),
		$restore_cutlevel(CL),
		$fail)).

pop(_).



%%======================================================================
%% motor process

push(motor).

% 1.04o
%push(dango,motor).
%push(subsumed_by_planner,motor).
%push(subsumed_by_instinct,motor).

$rule(main_loop,
	$and(	player(Player),
		$repeat,
		cycle(Player),
		$fail)).

pop(_).

push(gsinit).

% 1.04o, moved environment pushes into this predicate
$rule(start_motor, $and(
    goalie,
    !,
    push(dango_goalie,motor),
    push(subsumed_by_planner,motor),
    push(subsumed_by_instinct,motor),
	gsfork(motor,[motor,base_com],[gsinit],main_loop))).
$rule(start_motor, $and(
    push(dango,motor),
    push(subsumed_by_planner,motor),
    push(subsumed_by_instinct,motor),
	gsfork(motor,[motor,base_com],[gsinit],main_loop))).
pop(_).

%%======================================================================
%% basic_command

push(base_com).

%%--------------------------------------------------
%% look around

$rule(look_around,
	$and(	player(P),
		$for(I,0,10,1),
		$gs$turn(P,45),
		$wait_signal(_,see),
		$fail)).
$rule(look_around,!).

pop(_).


%%--------------------------------------------------
%% move random

push(base_com_move).

$rule(move_random,
	$and(	$is(X,frand(-53,0)),
		$is(Y,frand(-34,34)),
		player(P),
		$gs$move(P,X,Y),
		!)).

%%--------------------------------------------------
%% decide position by occupancy

$rule(change_position(NewPos),
	$and(	clear_position,
		push(NewPos,myposition),
		moveto_initpos,
		player(Player),$gs$side_unum(Player,_,UNum),
		with(myposition,$and(
			init_pos(IX,IY),
			home_pos(HX,HY),
			shift(SX,SY),
			area(X1,X2,Y1,Y2),
			$gs$set_position_param(Player,
				IX,IY,HX,HY,SX,SY,X1,X2,Y1,Y2))),
		$is(MS,irand(100)),$msleep(0,MS),
		$gs$say_list(Player,[position_change,UNum,NewPos]),
		!)).

$rule(decide_position_by_occupancy,
    $and(	goalie,
            change_position(goalie),
            !)).

$rule(decide_position_by_occupancy,
	$and(	with(position_table,find_minimum_occupancy(Pos,Occ)),
		change_position(Pos))).

$rule(moveto_initpos,
	$and(	with(myposition,init_pos(X,Y)),
		player(P),
		$gs$move(P,X,Y),
		!)).

$rule(clear_position,
	$and(pop(_,myposition),clear_position)).
$rule(clear_position,!).
		
pop(_).

%%======================================================================
%% stay init pos

push(stay_init_pos0).

$rule(cycle(Player),
	$and(	$gs$playmode(Player,Mode),
		$not_unify(Mode,before_kick_off),
		$remove_this_cell,
		!)).

$rule(cycle(Player),
	$and(	cycle(Player,random_move),
		cycle(Player,look_around),
		$gs$epos(Player,X,Y,BDir,HDir),
%%		$write(epos(X,Y)),$nl,
		!)).

$rule(cycle(Player,random_move),	
	$and(	$is(X,frand(-53,0)),
		$is(Y,frand(-34,34)),
		$gs$move(Player,X,Y),
%%		$write(moveto(X,Y)),$nl,
		$msleep(200),
		!)).

$rule(cycle(Player,look_around),
	$and(look_around,!)).
pop(_).

%%======================================================================
%% only chasing (dango0)

push(dango0).

$rule(target(Player,X,Y),
	$and($gs$goal_pos(Player,X,Y),!)).

$rule(cycle(Player),
	$and(	$gs$kickable(Player,3,0.7),
		target(Player,X,Y),
		$gs$infront_kickable_to_apos(Player,X,Y),
		$gs$kick_to_apos(Player,100,X,Y),
		!)).

$rule(cycle(Player),
	$and(	$gs$kickable(Player,3,0.7),
		$gs$step_swing(Player,right),
		!)).

$rule(cycle(Player),
	$and(	!,
		$gs$ball_pos(Player,BX,BY,BVX,BVY,BCycle),
		$gs$step_goto(Player,BX,BY,30))).

pop(_).

%%======================================================================
%% only chasing (new, using Damon's functions)

push(dango).

%%V.1.05noda28
$rule(target(Player,X,Y1), $and(
    $gs$goal_pos(Player,X,Y),
    $is(Y1,add(Y,frand(-7,7))),
    !)).

$rule(cycle(Player), $and(
	%$gs$kickable2(Player),
	% V.1.06
    $gs$dash_kickable(Player), % 1.04o
    % $gs$dash_kickable_by_sensor(Player), % 1.04o
    target(Player,X,Y),
    $gs$multi_pass_ball_to_apos(Player,100,X,Y),
    !)).

%%V.1.04noda3
$rule(xcycle(Player),$and(
	$gs$get_plan_for_intercept2(Player,80,5,0,0.2,ADir,NDash),
	turn_and_dash(Player,ADir,80.0,NDash),
	!)).

$rule(turn_and_dash(Player,ADir,Power,NDash), $and(
	$gs$turn_to_adir(Player,ADir),
%%	$write(plan(ADir,NDash,Power)),$nl,
	$for(I,0,NDash,1),
	$gs$dash(Player,Power),
	$fail)).
$rule(turn_and_dash(Player,ADir,Power,NDash), !).


$rule(cycle(Player), $and(
	$nop,
	$gs$step_dash_to_ball2(Player,100),
    !)).

pop(_).

%% push(gsinit).
%% $rule(start_dango,
%%	gsfork(dango,[stay_init_pos0,dango],[gsinit],main_loop)).
%% pop(_).


%% dango goalie 1.04o
push(dango_goalie).

%% V.1.06: older version
$rule(xxcycle(Player), $and(
    %$write("Goalie: entering loop."), $nl,
    $gs$goalie_critical_action_loop(Player))).

%% V.1.06: newer version
$rule(cycle(Player), $and(
    adjust_goalie_dash_powers(Player), %1.05o
			%% V.1.06:  should use loop3 ???
			%% loop2 is best [i.noda]
    % $gs$goalie_critical_action_loop3(Player))).  %1.06
    $gs$goalie_critical_action_loop2(Player))).  %1.05o
    % $gs$goalie_critical_action_loop(Player))).  %1.05o : comment out

% 1.05o : adjust dashing powers based on ball position.
$rule(adjust_goalie_dash_powers(Player), $and(
    not($gs$ball_inside_perimeter(Player, 4)),
    !,
    %$write("Setting goalie dash: 10"),$nl,
    $gs$set_goalie_intercept_dash_power(Player, 100),
    $gs$set_goalie_misc_dash_power(Player, 10))).
$rule(adjust_goalie_dash_powers(Player), $and(
    not($gs$ball_inside_perimeter(Player, 3)),
    !,
    %$write("Setting goalie dash: 20"),$nl,
    $gs$set_goalie_intercept_dash_power(Player, 100),
    $gs$set_goalie_misc_dash_power(Player, 20))).
$rule(adjust_goalie_dash_powers(Player), $and(
    not($gs$ball_inside_perimeter(Player, 2)),
    !,
    %$write("Setting goalie dash: 40"),$nl,
    $gs$set_goalie_intercept_dash_power(Player, 100),
    $gs$set_goalie_misc_dash_power(Player, 40))).
$rule(adjust_goalie_dash_powers(Player), $and(
    !,
    %$write("Setting goalie dash: 50"),$nl,
    $gs$set_goalie_intercept_dash_power(Player, 100),
    $gs$set_goalie_misc_dash_power(Player, 50))).


pop(_).


%%V.1.04noda3
%%======================================================================
%% with_say_chase

push(with_say_chase).
$new_cellvar(with_say_chase,unum,0).
$new_cellvar(with_say_chase,cycle,0).
$new_cellvar(with_say_chase,distance,0.0).
$new_cellvar(with_say_chase,stamina,0.0).
$new_cellvar(with_say_chase,effort,0.0).

$rule(cycle(Player), $and(
	$gs$cycle(Player,SC,VC,AC,BC,EC),
	% V.1.05noda
	$gs$arith_mod(SC,2,0),
	say_chase(Player),
	$fail)).

$rule(say_chase(Player), $and(
	$gs$ball_distance(Player,Dist),
	$gs$cycle(Player,SC,_,_,_,_),
	$gs$side_unum(Player,_,UNum),	
	$gs$stamina(Player,Stamina,Effort),
	$gs$say_list(Player,[chase,UNum,SC,Dist,Stamina,Effort]),
	!)).

pop(_).

%% V.1.04noda
%%--------------------------------------------------
%% fixed pass dest

push(fixed_pass_dest).

$new_cellvar(fixed_pass_dest,cycle,-100).
$new_cellvar(fixed_pass_dest,destx,0.0).
$new_cellvar(fixed_pass_dest,desty,0.0).

$rule(position_to_pass_dest([],[]),!).
$rule(position_to_pass_dest([MyPos | Rest],DestList),$and(
	with(myposition,posiition(MyPos)),
	!)).
$rule(position_to_pass_dest([Pos | Rest],[hxy(X,Y) | DRest]),$and(
	with(Pos,home_pos(X,Y)),
	position_to_pass_dest(Rest,DRest),
	!)).

%% V.1.05noda28
$rule(pass_dest_list(Player,[axy(GX,-5), axy(GX,5) | DestList]), $and(
	$cv_get(position_table,position_list,PosList,no,yes),
	position_to_pass_dest(PosList,DestList),
	$gs$goal_pos(Player,GX,_),
	!)).

$rule(target(Player,X,Y), $and(
	$cv_get(fixed_pass_dest,cycle,RC,no,yes),
	$gs$cycle(Player,SC,V,A,B,E),
	$is(Diff,sub(SC,RC)),
	$comp_lt(Diff,4),
	!,
	$cv_get(fixed_pass_dest,destx,X,no,yes),
	$cv_get(fixed_pass_dest,desty,Y,no,yes),
	$write(pass_request(X,Y)),$nl,
	!)).

$rule(target(Player,X,Y), $and(
	pass_dest_list(Player,DestList),
	$gs$find_good_pass_dest(Player,xy(X,Y),DestList),
	!)).

pop(_).

%% end V.1.04noda


%%======================================================================
%% confirm_pos

push(confirm_pos).

$rule(cycle(Player),
	$and(	goalie,
		$timedwait_signal(_,[see],1),
		$fail)).
$rule(cycle(Player),
	$and(	$gs$change_view(Player,narrow,high),
		$timedwait_signal(_,[see],1),
		$gs$change_view(Player,normal,high),
		$fail)).
$rule(cycle(Player),
	$and(	$gs$cycle(Player,S,V,A,B,E),
		$is(Diff,sub(S,E)),
		%% V.1.05noda
		$comp_lt(Diff,4),
		$remove_this_cell(subsumed_by_instinct),
		!)).
$rule(cycle(Player), $and(	
	% dlog(Player,"in confirm_pos"),
	$gs$turn(Player,45),
	!)).

pop(_).

%%V.1.05noda
%%======================================================================
%% look at ball

push(look_at_ball).

$rule(cycle(Player), $and(
	goalie,
	$gs$ball_pos(Player,BX,BY,_,_,_),
	$gs$turn_to_apos(Player,BX,BY),
	$timedwait_signal(_,[see],10),
	$gs$turn_to_apos(Player,BX,BY),
	$remove_this_cell(subsumed_by_instinct),
	$fail)).
$rule(cycle(Player), $and(
	$gs$ball_pos(Player,BX,BY,_,_,_),
	$gs$turn_to_apos(Player,BX,BY),
	$gs$change_view(Player,narrow,high),
	$timedwait_signal(_,[see],10),
	$gs$change_view(Player,normal,high),
	$gs$turn_to_apos(Player,BX,BY),
	$remove_this_cell(subsumed_by_instinct),
	$fail)).
pop(_).

%%======================================================================
%% find ball

push(find_ball).

$rule(cycle(Player),
	$and(	goalie,
		$timedwait_signal(_,[see],1),
		$fail)).
$rule(cycle(Player),
	$and(	$gs$change_view(Player,narrow,high),
		$timedwait_signal(_,[see],1),
		$gs$change_view(Player,normal,high),
		$fail)).
$rule(cycle(Player),
	$and(	$gs$cycle(Player,S,V,A,B,E),
		$gs$ball_pos(Player,X,Y,VX,VY,BTime),
		$is(Diff,sub(S,BTime)),
		%% V.1.05noda
		$comp_lt(Diff,4),
		$remove_this_cell(subsumed_by_instinct),
		!)).
$rule(cycle(Player), $and(
	goalie,
	% dlog(Player,"lose ball in find_ball (goalie)"),
	$gs$turn(Player,40),
	!)).

$rule(cycle(Player), $and(
	% dlog(Player,"lose ball in find_ball"),
	$gs$turn(Player,80),
	!)).

pop(_).

%% V.1.05noda
%%======================================================================
%% come back to field from outside

push(back_to_field).

$rule(cycle(Player), $and(
	not($gs$out_of_field(Player)),
%	$write("I'm recovered."),$nl,
	$remove_this_cell(subsumed_by_instinct),
	!)).

$rule(cycle(Player), $and(
	$gs$cycle(Player,S,V,A,B,E),
	$is(Diff,sub(S,E)),
	$comp_lt(Diff,4),
%	$write("Goto center."),$nl,
	$gs$turn_to_apos(Player,0,0),
	% V.1.06
	$gs$dash(Player,15),
	!)).

$rule(cycle(Player), $and(
	goalie,
	$gs$turn(Player,30),
	% V.1.06
	$gs$dash(30),
%	$write("where am I?"),$nl,
	$timedwait_signal(_,[see],1),
	!)).

$rule(cycle(Player), $and(
	$gs$turn(Player,60),
	% V.1.06
	$gs$dash(30),
%	$write("where am I?"),$nl,
	$gs$change_view(Player,narrow,high),
	$timedwait_signal(_,[see],1),
	$gs$change_view(Player,normal,high),
	!)).

pop(_).

%%======================================================================
%% stay and look around

push(stay_and_look_around).

$rule(cycle(Player),
	$and(	$timedwait_signal(_,[see],0,500),
		$fail)).
$rule(cycle(Player),$and(	
	goalie,
	% dlog(Player,"in stay_and_look_around (goalie)"),
	$gs$turn(Player,40),
	!)).
$rule(cycle(Player),$and(	
	% dlog(Player,"in stay_and_look_around"),
	$gs$turn(Player,80),
	!)).

pop(_).

%% V.1.04noda
%%======================================================================
%% say pass request           %% V.1.04noda3 (comment only)

push(pass_request).

$rule(cycle(Player), $and(
	$gs$cycle(Player,SC,VC,AC,BC,EC),
	$gs$arith_mod(SC,4,0),
	say_pass_request(Player),
	$fail)).

$rule(say_pass_request(Player),$and(
	$gs$epos(Player,PX,PY,BDir,HDir),
	$gs$ball_pos(Player,BX,BY,BVX,BVY,BallC),
	$comp_gt(BX,-20),
	$is(DiffX,sub(PX,BX)),
	$comp_gt(DiffX,5),
	$gs$distance(PX,PY,BX,BY,Dist),
	$comp_lt(Dist,30),
	!,
	$gs$cycle(Player,SC,VC,AC,BC,EC),
	$gs$side_unum(Player,_,UNum),	
	$gs$say_list(Player,[pass_request,UNum,SC,PX,PY]),
	!)).

pop(_).


%%======================================================================
%% stay and look the ball

push(stay_and_look_ball).

$rule(cycle(Player),
	$and(	goalie,
		$timedwait_signal(_,[see],0,150),
		$fail)).
$rule(cycle(Player),
	$and(	$gs$change_view(Player,narrow,high),
		$timedwait_signal(_,[see],0,150),
		$gs$change_view(Player,normal,high),
		$fail)).
$rule(cycle(Player),
	$and(	$gs$cycle(Player,S,V,A,B,E),
		$gs$ball_pos(Player,X,Y,VX,VY,BTime),
		$is(Diff,sub(S,BTime)),
		$comp_lt(Diff,3),
		!)).
$rule(cycle(Player),$and(	
	goalie,
	% dlog(Player,"lose ball in stay_and_look_ball (goalie)"),
	$gs$turn(Player,40),
	!)).
$rule(cycle(Player),$and(	
	% dlog(Player,"lose ball in stay_and_look_ball"),
	$gs$turn(Player,80),
	!)).
pop(_).

% v1.04o
%%======================================================================
%% clear ball to side
push (clear_ball_to_side).

$rule(cycle(Player), $and(
    $gs$kick_ball_to_near_mid_sideline(Player, 100),
    !)).

pop(_).

% 1.04o
%%======================================================================
push(goalie_caught_ball).

%% right now, goalie just moves within kicking distance and clears
%% ball.

$rule(cycle(Player), $and(
    not($gs$dash_kickable(Player)),
    $gs$step_dash_to_ball2(Player,10))).

%1.05o
%V.1.06: comment out
$rule(cycle(Player), $and(
    $gs$dash_kickable(Player),
    % $write("cgaea: ~~~~ goalie caught ball, clearing now."),$nl,
    $gs$get_goalie_clearing_coordinates(Player, GX, GY, BX, BY),
    % $write("cgaea: clearing A."),$nl,
    $gs$move(Player, GX, GY),
    % V.1.06
    $msleep(0,50),
    $gs$move(Player, GX, GY),
    $msleep(0,50),
    $gs$move(Player, GX, GY),
    % $write("cgaea: clearing B."),$nl,
    $gs$wait_for_vision(Player, 0),
    % $write("cgaea: clearing C."),$nl,
    $gs$wait_for_vision(Player, 1),
    $gs$wait_for_vision(Player, 1),
    % $write("cgaea: clearing D."),$nl,
    $gs$multi_smart_kick_to_apos(Player, 100, BX, BY, 0, 0, rot_closest),
    % $write("cgaea: clearing E."),$nl,
    % do extra kick, to give ball time to clear catchable range
    $gs$kick_to_adir(Player, 100, 0),
    $gs$kick_to_adir(Player, 100, 0),
    $gs$kick_to_adir(Player, 100, 0),
    % $write("cgaea: clearing F."),$nl,
	% V.1.06:  go back to goal position.
	% this does not work ???
    % rush_back_to_own_goal(Player,10),
	% here is alternative
    $gs$owngoal_pos(Player,OGX,OGY),
    $is(OGX1,add(OGX,3)),
    $msleep(0,100),
    !,
    $for(I,0,10,1),
    $gs$step_goto2(Player,OGX1,OGY,100),
    $fail,
    !)).

$rule(cycle(Player), $and(
    $gs$dash_kickable(Player),
    % $write("cgaea: goalie caught ball, simple clearing."),$nl,
    $gs$kick_ball_to_near_mid_sideline(Player, 100),
    % do 2nd kick, to give ball time to clear catchable range
    $gs$kick_to_adir(Player, 100, 0))).

%% V.1.06  ??? it does not work ???
$rule(rush_back_to_own_goal(Player,N), $and(
	% $write(bbb(Player,N)),$nl,
	$gs$owngoal_pos(Player,OGX,OGY),
	$is(OGX1,add(OGX,3)),
	$for(I,0,N,1),
	$gs$step_goto2(Player,OGX1,OGY,100),
	$fail,
	!)).
$rule(rush_back_to_own_goal(Player,N), !).

pop(_).


%%======================================================================
%% goto pos

push(goto_home).

$rule(xxcycle(Player),
	$and(	$push(myposition),home_pos(X,Y),shift(SX,SY),$pop(_),!,
		$gs$ball_pos(Player,BX,BY,BVX,BVY,BC),
		$is(SHX,add(X,mul(SX,div(BX,53)))),
		$is(SHY,add(Y,mul(SY,div(BY,34)))),
		%V.1.05noda
		%V.1.05noda28
		%$gs$step_goto(Player,SHX,SHY,20),
		$gs$step_goto2(Player,SHX,SHY,20),
		%V.1.05noda28
		!)).
$rule(cycle(Player),
	$and(	$gs$shifted_home_pos(Player,SHX,SHY),
		%V.1.05noda
		%V.1.05noda28
		%$gs$step_goto(Player,SHX,SHY,20),
		$gs$step_goto2(Player,SHX,SHY,20),
		!)).
pop(_).


%%V.1.04noda2
%%======================================================================
%% goto pos

push(hurry_to_home).

$rule(cycle(Player),
	$and(	$gs$shifted_home_pos(Player,SHX,SHY),
		%V.1.05noda
		%V.1.05noda28
		%$gs$step_goto(Player,SHX,SHY,20),
		%$gs$step_goto(Player,SHX,SHY,20),
		$gs$step_goto2(Player,SHX,SHY,20),
		$gs$step_goto2(Player,SHX,SHY,20),
		!)).
pop(_).

%%V.1.05noda
%%======================================================================
%% set play kicker

push(set_play_kicker).

$rule(cycle(Player),$and(	%% already good position
	$gs$epos(Player,PX,PY,_,_),
	$gs$ball_pos(Player,BX,BY,_,_,_),
	% V.1.06
	$is(DX,sub(BX,PX)),
	$comp_gt(DX,0),
	$is(DY,sub(BY,PY)),
	$comp_lt(DY,0.5),$comp_gt(DY,-0.5),
%	$write("ready to kick."),$nl,
	$call_super(cycle(Player)),
	!)).
$rule(cycle(Player),$and(	%% still not good position
	$gs$ball_pos(Player,BX,BY,_,_,_),
	$is(BX1,sub(BX,2)),
%	$write("aproach to kick."),$nl,
	$gs$turn_to_apos(Player,BX1,BY),
	$gs$dash(Player,20),
	!)).
pop(_).


%%======================================================================
%% position table

%%--------------------------------------------------
%% trashed table

push(xxposition_table).
$rule(defpos(_,_,_,_,_,_),!).
$rule(collect_position(_),!).
pop(_).

%%--------------------------------------------------
%% table

push(position_table).

$new_cellvar(position_table,position_work).
$new_cellvar(position_table,occupancy_work).

$rule(defpos(Name,InitPos,HomePos,AreaX,AreaY,Shift),
	$and(newcell(Name),
	     push(position_base,Name),
	     $new_cellvar(Name,init_pos,InitPos),
	     $new_cellvar(Name,home_pos,HomePos),
	     $new_cellvar(Name,area_x,AreaX),
	     $new_cellvar(Name,area_y,AreaY),
	     $new_cellvar(Name,shift,Shift),
	     % V.1.05noda28
	     $is(Occ,frand(0,0.1)),
	     $new_cellvar(Name,occupancy,Occ),
	     asserta_in(position(Name),Name),
	     assertz_in(pos(Name),position_table),
	     !)).

$rule(decay_occupancy(Name,Rate),
	$and(	$cv_get(Name,occupancy,OldVal,yes,yes),
		$is(NewVal,mul(OldVal,Rate)),
		$cv_put(Name,occupancy,NewVal,force),!)).
$rule(decay_all_occupancy(Rate),
	$and(	pos(Position),
		decay_occupancy(Position,Rate),
		$fail)).
$rule(decay_all_occupancy(Rate),!).

$rule(inc_occupancy(Name,Inc),
	$and(	$cv_get(Name,occupancy,OldVal,yes,yes),
		$is(NewVal,add(OldVal,Inc)),
		$cv_put(Name,occupancy,NewVal,force),!)).
$rule(set_occupancy(Name,NewVal),
	$and(	$cv_get(Name,occupancy,OldVal,yes,yes),
		$cv_put(Name,occupancy,NewVal,force),!)).

$rule(minval_pair(X,XVal,Y,YVal,X,XVal),
	$and(!,$comp_lt(XVal,YVal))).
$rule(minval_pair(X,XVal,Y,YVal,Y,YnVal),
	!).

$rule(find_minimum_occupancy(Position,Occupancy),
	$and(	$cv_put(position_table,position_work,[],force),
		$cv_put(position_table,occupancy_work,100000.0,force),
		!,
		find_minimum_occupancy_body,
		$cv_get(position_table,position_work,Position,no,yes),
		$cv_get(position_table,occupancy_work,Occupancy,no,yes),
		!)).
		
$rule(find_minimum_occupancy_body,
	$and(	pos(NewPos),
		$cv_get(NewPos,occupancy,NewOcc,no,yes),
%		$cv_get(position_table,position_work,OldPos,no,yes),
		$cv_get(position_table,occupancy_work,OldOcc,no,yes),
		$comp_le(NewOcc,OldOcc),
		$cv_put(position_table,position_work,NewPos,force),
		$cv_put(position_table,occupancy_work,NewOcc,force),
		$fail)).
$rule(find_minimum_occupancy_body,!).
	
%% V.1.04noda
%%------------------------------
%% collect all position list

$new_cellvar(position_table,position_list,[]).

$rule(collect_position(L), $and(
	$cv_put(position_table,position_list,[],force),
	collect_position_sub,
	$cv_get(position_table,position_list,L,no,yes),
	!)).
$rule(collect_position_sub, $and(
	pos(Pos),
	$cv_get(position_table,position_list,List,yes,yes),
	$cv_put(position_table,position_list,[Pos | List],force),
	fail)).
$rule(collect_position_sub,!).

pop(_).

%%--------------------------------------------------
%% positioin base

push(position_base).
$rule(init_pos(X,Y),
	$and(!,
		position(Name),
		$cv_get(Name,init_pos,xy(X,Y),no,yes))).
$rule(home_pos(X,Y),
	$and(!,
		position(Name),
		$cv_get(Name,home_pos,xy(X,Y),no,yes))).
$rule(area(X1,X2,Y1,Y2),
	$and(!,
		position(Name),
		$cv_get(Name,area_x,xx(X1,X2),no,yes),
		$cv_get(Name,area_y,yy(Y1,Y2),no,yes))).
$rule(shift(X,Y),
	$and(!,
		position(Name),
		$cv_get(Name,shift,xy(X,Y),no,yes))).

pop(_).

%%--------------------------------------------------
%% positioin   (3-5-2) (version 0)

push(xxposition_table).

defpos(goalie,	   xy(-52,  0),xy(-51,  0),xx(-55,-45),yy(-10, 10),xy( 0, 5)).
defpos(sweeper,	   xy(-48,  0),xy(-20,  0),xx(-54,-25),yy(-15, 15),xy(15, 5)).
defpos(stopper_l,  xy(-43,-10),xy(-15,-10),xx(-54,-15),yy(-25,  2),xy(15, 5)).
defpos(stopper_r,  xy(-43, 10),xy(-15, 10),xx(-54,-15),yy( -2, 25),xy(15, 5)).
defpos(wingback_l, xy(-25,-20),xy(  0,-20),xx(-55, 55),yy(-35,-20),xy(15, 5)).
defpos(wingback_r, xy(-25, 20),xy(  0, 20),xx(-55, 55),yy( 20, 35),xy(15, 5)).
defpos(midfield_c, xy(-35,  0),xy( -5,  0),xx(-20, 10),yy(-20, 20),xy(15, 5)).
defpos(midfield_l, xy(-30,-12),xy( 10,-12),xx( -5, 30),yy(-25,  2),xy(15, 5)).
defpos(midfield_r, xy(-30, 12),xy( 10, 12),xx( -5, 30),yy( -2, 25),xy(15, 5)).
defpos(forward_l,  xy(-10,-10),xy( 20,-10),xx( 15, 55),yy(-25,  3),xy(15, 5)).
defpos(forward_r,  xy(-10, 10),xy( 20, 10),xx( 15, 55),yy( -3, 25),xy(15, 5)).

%% V.1.04n
collect_position(L).	

pop(_).

%% V.1.06
%%--------------------------------------------------
%% positioin   (3-5-2) (version 1)

push(position_table).

defpos(goalie,	   xy(-51,  0),xy(-50,  0),xx(-55,-45),yy(-10, 10),xy( 0, 5)).
defpos(sweeper,	   xy(-25,  0),xy(-25,  0),xx(-54,-25),yy(-15, 15),xy(25, 5)).
defpos(stopper_l,  xy(-25,-15),xy(-25,-15),xx(-54,-15),yy(-25,  2),xy(25, 5)).
defpos(stopper_r,  xy(-25, 15),xy(-25, 15),xx(-54,-15),yy( -2, 25),xy(25, 5)).
defpos(wingback_l, xy( -5,-20),xy(  0,-25),xx(-55, 55),yy(-35,-20),xy(30, 5)).
defpos(wingback_r, xy( -5, 20),xy(  0, 25),xx(-55, 55),yy( 20, 35),xy(30, 5)).
defpos(midfield_c, xy(-18,  0),xy(-13,  0),xx(-30, 10),yy(-20, 20),xy(25, 5)).
defpos(midfield_l, xy(-10,-12),xy( -5,-13),xx(-10, 30),yy(-25,  2),xy(25, 5)).
defpos(midfield_r, xy(-10, 12),xy( -5, 13),xx(-10, 30),yy( -2, 25),xy(25, 5)).
defpos(forward_l,  xy( -4, -1),xy( 10, -7),xx( 20, 55),yy(-25,  3),xy(30, 5)).
defpos(forward_r,  xy( -1,  9),xy( 10,  7),xx( 20, 55),yy( -3, 25),xy(30, 5)).

%% V.1.04n
collect_position(L).	

pop(_).


%% V.1.04noda2
%%======================================================================
%% draw position  (generate tgif objects)

push(gsinit).

$rule(gen_pos_fig, gen_pos_fig("position_figure.obj")).
$rule(gen_pos_fig(FileName), $and(
	$system("cp","tgif.field.header.obj",FileName),
	$fopen(Stream,FileName,append),
	$tell(Stream),
	draw_position,
	$told(Stream),
	$fclose(Stream,append),
	!)).
	
$rule(draw_position, $and(
	$cv_get(position_table,position_list,PosList,no,yes),
	draw_position(PosList),
	!)).
$rule(draw_position([]), !).
$rule(draw_position([Pos|Rest]), $and(
	draw_position1(Pos),
	draw_position(Rest),
	!)).
$rule(draw_position1(Pos),$and(
	with(Pos,area(X1,X2,Y1,Y2)),
	$is(DX1,add(100,mul(5,add(X1,52.5)))),
	$is(DX2,add(100,mul(5,add(X2,52.5)))),
	$is(DY1,add(100,mul(5,add(Y1,34.0)))),
	$is(DY2,add(100,mul(5,add(Y2,34.0)))),
	with(Pos,home_pos(HX,HY)),
	$is(DHX,add(100,mul(5,add(HX,52.5)))),
	$is(DHY,add(100,mul(5,add(HY,34.0)))),
	draw_box(black,DX1,DY1,DX2,DY2,0,2),
	draw_line(red,DX1,DY1,DX2,DY2,0,3),
	draw_line(red,DX1,DY2,DX2,DY1,0,3),
	draw_circle(blue,DHX,DHY,10,1,3),
	!)).

$rule(draw_box(Color,DX1,DY1,DX2,DY2,Fill,Width), $and(
	$write_special_char("box('"),
	$write(Color),  $write_special_char("',"),
	$write(DX1),	$write_special_char(","),
	$write(DY1),	$write_special_char(","),
	$write(DX2),	$write_special_char(","),
	$write(DY2),	$write_special_char(","),
	$write(Fill),	$write_special_char(","),
	$write(Width),	$write_special_char(","),
	$write_special_char("1,22,0,0,0,0,0,'1',["), $nl,
	$write_special_char("])."), $nl,
	!)).
$rule(draw_line(Color,DX1,DY1,DX2,DY2,Arrow,Width), $and(
	$write_special_char("poly('"),
	$write(Color),  $write_special_char("',2,["),$nl,
	$write_special_char(tab),
	$write(DX1),	$write_special_char(","),
	$write(DY1),	$write_special_char(","),
	$write(DX2),	$write_special_char(","),
	$write(DY2),	$write_special_char(","),
	$write(Arrow),	$write_special_char(","),
	$write(Width),	$write_special_char(","),
	$write_special_char("1,26,0,0,0,0,8,3,0,0,0,'1','8','3',"), $nl,
	$write_special_char("    "),$write("0"),
	$write_special_char(",["), $nl,
	$write_special_char("])."), $nl,
	!)).

$rule(draw_circle(Color,DX,DY,DR,Fill,Width), $and(
	$write_special_char("oval('"),
	$write(Color),  $write_special_char("',"),
	$is(DX1,sub(DX,DR)), $write(DX1),$write_special_char(","),
	$is(DY1,sub(DY,DR)), $write(DY1),	$write_special_char(","),
	$is(DX2,add(DX,DR)), $write(DX2),	$write_special_char(","),
	$is(DY2,add(DY,DR)), $write(DY2),	$write_special_char(","),
	$write(Fill),	$write_special_char(","),
	$write(Width),	$write_special_char(","),
	$write_special_char("1,23,0,0,0,0,0,'1',["), $nl,
	$write_special_char("])."), $nl,
	!)).

pop(gsinit).


%%======================================================================
%% initial setup

push(gsinit).

$write(hona),$nl.
