// -*- Mode:C++ -*-
//
//======================================================================
// for test
//======================================================================
//
//  Kicks ball in random directions.

#include "GSPlayer.h"

enum Mode {
    M_before_kick_off,
    M_play_on,
    M_tired,
    M_error,
} ;


//----------------------------------------------------------------------
// action loop(s)


void PreGameAction(GSPlayer& player, Mode *pMode)
{
    cout << "\n================================" << endl;
    if(player.scycle.count > 0) {
        *pMode = M_play_on ;
        return;
    }
    ScFloat x = scFltRand(-SC_HALF_PITCH_L,0) ;
    ScFloat y = scFltRand(-SC_HALF_PITCH_W,SC_HALF_PITCH_W) ;
    player.move(x,y) ;
    cout << player << "move(" << x << ", " << y << ")"<<endl;
    scSleep(0,444) ;
    player.turn(scFltRand(-90,90)) ; 
    scSleep(0,444) ;
}


void GameAction(GSPlayer& player)
{
    static ScPos destPos(scFltRand(-SC_HALF_PITCH_L,SC_HALF_PITCH_L),
                         scFltRand(-SC_HALF_PITCH_W,SC_HALF_PITCH_W));
    
    if (player.DashKickableP()) {
        ScFloat dir = scFltRand(-90,90) ;
        player.kick(100,dir) ;
        cout << "\n================================" << endl;
        cout << player << "Kick(30," << dir << ")" << endl ;
    }
    else if(player.ClosestTeamMemberToBallP()) {
        //cout << player << "My Ball!!!" << endl ;
        player.stepDashToBall(50);
    }
    else if(scDistance(player.xposdir.pos,destPos) < 2.0) {
        destPos.x = scFltRand(-SC_HALF_PITCH_L,SC_HALF_PITCH_L) ;
        destPos.y = scFltRand(-SC_HALF_PITCH_W,SC_HALF_PITCH_W) ;
        cout << "\n================================" << endl;
        cout << player << "I change mind." << destPos << endl;
    } else {
        player.stepGoto(destPos, 50);
    }
}


void ActionLoop(GSPlayer& player)
{
    Mode mode = M_before_kick_off;

    cout << "\nInterceptTestLoop: kicker\n" << endl;
    
    while(ScTrue) {
        if (mode == M_before_kick_off) {
            PreGameAction(player, &mode);
        }

        else if (mode == M_play_on) {
            if (player.scycle.count - player.ecycle.count > 20) {
                cout << "\n================================" << endl;
                cout << player << "I lost my position." << endl;
                for(ScUInt i = 0; i < 6; i++) {
                    player.turn(80);
                }
            }
            else if (player.actualBodyState.stamina < 1000) {
                cout << "\n================================" << endl;
                cout << player << "I am tired." << endl;
                mode = M_tired;
            }
            else if((player.xposdir.pos.x > SC_HALF_PITCH_L &&
                     scAbs(player.eposdir.bdir) <= 90) ||
                    (player.xposdir.pos.x < -SC_HALF_PITCH_L &&
                     scAbs(player.eposdir.bdir) >= 90) ||
                    (player.xposdir.pos.y > SC_HALF_PITCH_W &&
                     player.xposdir.bdir >= 0) ||
                    (player.xposdir.pos.y < -SC_HALF_PITCH_W &&
                     player.xposdir.bdir <= 0)) {
                cout << "\n================================" << endl;
                cout << player << "I am out of field." << endl ;
                player.turn(scFltRand(-180,180)) ;
            }
            else {
                GameAction(player);
            }
        }

        else if (mode == M_tired) {
            cout << "\n================================" << endl;
            player.turn(10);
            if(player.actualBodyState.stamina > 3000) {
                cout << player << "I'm refleshed!!" << endl;
                mode = M_play_on;
            }
        }

    }  // end while loop
}
    

//----------------------------------------------------------------------
// sensor loop

void sensorLoop(GSPlayer& player)
{
    ScBool r;
    while(ScTrue) {
        r = player.recvScanSInfo();
    }
}


void* sensorLoopEntry(void* p)
{
    sensorLoop(*(GSPlayer*)p);
    return NULL;
}


//----------------------------------------------------------------------
// option

#define DEFAULT_HOST		"localhost"
#define DEFAULT_PORT		6000
#define DEFAULT_TEAMNAME	"Japan"
#define DEFAULT_PROTO_VERSION	5.0

char* SHost = DEFAULT_HOST;
ScInt SPort = DEFAULT_PORT;
char* TeamName = DEFAULT_TEAMNAME;
ScFloat ProtocolVersion = DEFAULT_PROTO_VERSION;
ScBool GoalieP = ScFalse;

ScBool getOption(int argc, char** argv) {
    for(int i = 1; i < argc; i++) {
        if        (!strcmp(argv[i],"-host")) {
            i++; SHost = argv[i];
        } else if (!strcmp(argv[i],"-port")) {
            i++; SPort = atol(argv[i]);
        } else if (!strcmp(argv[i],"-team")) {
            i++; TeamName = argv[i];
        } else if (!strcmp(argv[i],"-protocol")) {
            i++; ProtocolVersion = atof(argv[i]);
        } else if (!strcmp(argv[i],"-goalie")) {
            GoalieP = ScTrue;
        } else {
            if (strcmp(argv[i],"-help")) {
                cerr << "Illegal option:" << argv[i] << endl;
            } 
            cerr << "Usage:" << endl;
            cerr << "\t % gstest [options]" << endl;
            cerr << "\t     [options]" << endl;
            cerr << "\t\t-host HOST" << endl;
            cerr << "\t\t-port PORT" << endl;
            cerr << "\t\t-team TEAM" << endl;
            cerr << "\t\t-protocol PROTOCOL" << endl;
            cerr << "\t\t-goalie GOALIE" << endl;
            exit(1);
        }
    }
    return ScTrue;
}

	    

//----------------------------------------------------------------------
// main for test

main(int argc, char** argv) {
    scRandomizeByTime() ;
    getOption(argc,argv);

    // cout << "Try to open socket to host:port=" 
    // << SHost << ":" << SPort << endl;

    GSPlayer player;
    player.init(ScString(TeamName),SHost,SPort,ScFalse);
   
    // cout << "Succeed to open socket." << endl;

    if(!player.sendInit(ProtocolVersion,GoalieP)) {
        cerr << "Can't send init command" << endl;
        exit(1);
    }

    // cout << "Start sensor loop." << endl;

    pthread_t sensor_thread;
    pthread_create(&sensor_thread, SC_NULLPTR, sensorLoopEntry,
                   (void*)&player);

    // cout << "Enter simple command loop." << endl;

    ActionLoop(player);
}
