// -*- Mode:C++ -*-
//Header:
//File: libsclient.cc
//Author: NDOA, Itsuki
//Date: 1999/04/15
//

//ModifyHistory:
// 1999/04/15: Create this file
//EndModifyHistory:

#include "libsclient.h"

void scNop() { } ;  // for debug

//======================================================================
// basic type

//--------------------------------------------------
// sleep

#ifdef THREADED
pthread_cond_t ScNeverAwakeCV ;
pthread_mutex_t ScNeverAwakeMX ;
#endif


//======================================================================
// class ScSocket

//----------------------------------------------------------------------
// init

ScBool ScSocket::init(char* host, ScInt port, ScBool nonblockp) {

   _host = host ; 
   _port = port ;
   
   //--------------------------------------------------
   // setup serv_addr

   ///// clean up serv_addr 
   bzero((char *)(&serv_addr),sizeof(serv_addr)) ;

   ///// set inet family flag 
   serv_addr.sin_family = AF_INET ;

   ///// find srever host address 
   unsigned long int saddr = inet_addr(host) ;
   if(saddr == -1) {
      struct hostent *h = (struct hostent *)gethostbyname(host) ;
      if(scIsNullPtr(h)) {
	 cerr << "Wrong host name: " << host << endl ;
	 return ScFalse ;
      }

      saddr = inet_addr(inet_ntoa(*(struct in_addr *)(h->h_addr_list[0]))) ;
   } 
   serv_addr.sin_addr.s_addr = saddr ;

   ///// set port number
   serv_addr.sin_port = htons(port) ;


   //--------------------------------------------------
   // setup socketfd

   ///// get socketfd
   socketfd = socket(AF_INET,SOCK_DGRAM,0) ;
   if(socketfd < 0) {
      cerr << "Can't open socket." << endl ;
      return ScFalse ;
   }

   ///// setup non blocking mode
   setBlockingMode(!nonblockp) ;

   //--------------------------------------------------
   // setup socketfd

   struct sockaddr_in cli_addr ;
   bzero((char*)&cli_addr,sizeof(cli_addr)) ;
   cli_addr.sin_family = AF_INET ;
   cli_addr.sin_addr.s_addr = htonl(INADDR_ANY) ;
   cli_addr.sin_port = htons(0) ;
      
   if(bind(socketfd,(struct sockaddr*)&cli_addr,sizeof(cli_addr)) < 0) {
      cerr << "Can't bind local address." << endl ;
      return ScFalse ;
   }

   //--------------------------------------------------
   // initialize mutex

   sendMutex.init() ;
   recvMutex.init() ;

   return ScTrue ;
}

ScBool ScSocket::checkBlockingMode() const {  return blockp ; }

ScBool ScSocket::setBlockingMode(const ScBool& blockingp) {
   unsigned int flag ;
   if(blockingp) { 
#ifdef THREADED
      flag = fcntl(socketfd,F_GETFL,0) ;
      flag &= ~O_NONBLOCK ;
      if(fcntl(socketfd,F_SETFL,flag) == -1) {
	 cerr << "Can't fcntl on socket." << endl ;
	 return ScFalse ;
      }
#else
      flag = 0 ;
      if(ioctl(socketfd,FIONBIO,&flag) < 0) {
	 cerr << "Can't ioctl on socket." << endl ;
	 return ScFalse ;
      }
#endif
      blockp = ScTrue ;
   } else {
#ifdef THREADED
      flag = fcntl(socketfd,F_GETFL,0) ;
      flag |= O_NONBLOCK ;
      if(fcntl(socketfd,F_SETFL,flag) == -1) {
	 cerr << "Can't fcntl on socket." << endl ;
	 return ScFalse ;
      }
#else
      flag = 1 ;
      if(ioctl(socketfd,FIONBIO,&flag) < 0) {
	 cerr << "Can't ioctl on socket." << endl ;
	 return ScFalse ;
      }
#endif
      blockp = ScFalse ;
   }
   return ScTrue ;
}


//----------------------------------------------------------------------
// send

ScBool ScSocket::send(const ScString& buf) {
   sendMutex.lock() ;
   ScInt l = buf.strlength() ;
   send_errno = 0 ;
   ScBool r = ScFalse ;
   if (sendto(socketfd, (char*)buf, l, 0,(struct sockaddr *)&serv_addr,
	      sizeof(serv_addr)) != l) {
      send_errno = errno ;
      r = ScFalse ;
   }  else {
      r = ScTrue ;
   }
   sendMutex.unlock() ;
   return r ;
}

//----------------------------------------------------------------------
// recv

ScBool ScSocket::recv(ScString& buf, ScBool adjustportp) {
   recvMutex.lock() ;

   struct sockaddr_in saddr ;

   ///// receive
   int servlen = sizeof(serv_addr) ;
   ScInt r = recvfrom(socketfd, (char*)buf, buf.length(), 0,
		      (struct sockaddr *)&saddr, &servlen) ;
   ///// check error
   recv_errno = 0 ;
   if (r < 0) {
      recv_errno = errno ;
      if (errno != ENOTSOCK && errno != EWOULDBLOCK && errno != 0) {
	 cerr << "Warning: recvfrom Error.[" << errno << "]:"  ;
	 switch(errno) {
	    case EINTR : { 
	       cerr << "Interrupted system call" << endl ;  break ; }
	    case EIO : {
	       cerr << "I/O error" << endl ; break ; }
	    case ECONNREFUSED : {
	       cerr << "Connection refused" << endl ; break ; }
	    default : {
	       cerr << "?" << endl ; break ; }
	 }
      }
      r = 0 ; 
      goto _finish_ ;
   } 

   ///// adjust server port
   if (adjustportp) {
      serv_addr.sin_port = saddr.sin_port ;
   }

   //// terminate buffer
   buf.head[r] = '\0' ;

   ///// return

  _finish_:

   recvMutex.unlock() ;

   if(r == 0)
      return ScFalse ;
   else
      return ScTrue ;
}

//----------------------------------------------------------------------
// close

void ScSocket::close() {
   ::close(socketfd) ;
}

//----------------------------------------------------------------------
// pipe

ScBool ScSocket::pipe(istream& sin, ostream& sout) {
#ifdef THREADED
   int sin_id = sin.rdbuf()->_fileno ;
#else
   int sin_id = fileno(sin.rdbuf()) ;
#endif   

   ///// setup select readfds mask
   fd_set readfds ;
   FD_ZERO(&readfds) ;
   FD_SET(sin_id,&readfds) ;
   FD_SET(socketfd,&readfds) ;
   int max_fd = scMax((ScInt)socketfd,(ScInt)sin_id) + 1;
   
   while(ScTrue) {
      ScString buf(SC_BUFSIZE) ;

      ///// reset mask
      fd_set rfds ; rfds = readfds ;

      ///// select
      int r = select(max_fd, &rfds, SC_NULLPTR, SC_NULLPTR, SC_NULLPTR) ;

      ///// check error
      if(r < 0) {
	 cerr << "Error: fail select operation." << endl ;
	 return ScFalse ;
      }

      ///// skip no selection
      if(r == 0) continue ;

      ///// stdin -> socket
      if(FD_ISSET(sin_id,&rfds)) {
	 sin.getline((char*)buf,buf.length()) ;
	 if(!send(buf)) {
	    cerr << "Error: fail to send." << endl ;
	    return ScFalse ;
	 }
      }
      
      ///// socket -> stdout
      if(FD_ISSET(socketfd,&rfds)) {
	 if(recv(buf)) {
	    sout << buf ;
	    sout.flush() ;
	 }
      }
   }
   return ScTrue ;  /* never come here */
}
   
//======================================================================
// Field Parameters
//

ScWord ScWordTable[] = SC_WORD_TABLE ;
ScLandmark ScLandmarkTable[] = SC_LANDMARK_TABLE ;

ScWord* scString2Word(const ScSubString& msg) {
   ScUInt i ;
   for(i = 0 ; i < SC_WID__MAX ; i++) {
      if(msg.hasPrefix(ScWordTable[i].name)) break ;
   }
   return &(ScWordTable[i]) ;
} ;

//======================================================================
// Token
//

#define SC_OPEN_PAREN_CHR   '('
#define SC_CLOSE_PAREN_CHR  ')'
#define SC_SPACE_CHR        ' '

//--------------------------------------------------
// check delimiter

inline ScBool isOpenParen(const ScSubString& buffer) {
   return buffer.isFirst(SC_OPEN_PAREN_CHR) ;
} ;
inline ScBool isCloseParen(const ScSubString& buffer) {
   return buffer.isFirst(SC_CLOSE_PAREN_CHR) ;
} ;
inline ScBool isSpace(const ScSubString& buffer) {
   return buffer.isFirst(SC_SPACE_CHR) ;
} ;
inline ScBool isTerminated(const ScSubString& buffer) {
   return (buffer.isFirst(SC_NULLCHR) || buffer.tailp());
} ;

//--------------------------------------------------
// skip space/paren

inline ScBool ignoreSpace(ScSubString& buffer) {
   while(!isTerminated(buffer)) {
      if(!isSpace(buffer)) {
	 return ScTrue ;
      } 
      buffer.ignoreChar()  ;
   }
   return ScFalse ;
}

inline ScBool skipOpenParen(ScSubString& buffer, ScInt n = 1) {
   for(ScInt i = 0 ; i < n ; i++) {
      if(!isOpenParen(buffer)) return ScFalse ;
      if(!ignoreSpace(buffer)) return ScFalse ;
      buffer.ignoreChar() ;
   }
   return ScTrue ;
}

inline ScBool skipCloseParen(ScSubString& buffer, ScInt n = 1) {
   for(ScInt i = 0 ; i < n ; i++) {
      if(!isCloseParen(buffer)) return ScFalse ;
      if(!ignoreSpace(buffer)) return ScFalse ;
      buffer.ignoreChar() ;
   }
   return ScTrue ;
}

//--------------------------------------------------
// convert Int,UInt,Float value

inline ScBool getIntVal(ScSubString& buffer, ScInt& i) {
   ignoreSpace(buffer) ;
   return buffer.popIntVal(i) ;
}

inline ScBool getUIntVal(ScSubString& buffer, ScUInt& i) {
   ignoreSpace(buffer) ;
   return buffer.popUIntVal(i) ;
}

inline ScBool getFloatVal(ScSubString& buffer, ScFloat& v) {
   ignoreSpace(buffer) ;
   return buffer.popFloatVal(v) ;
}

//--------------------------------------------------
// push char (unget)

inline void pushChar(const char c, ScSubString& buffer) {
   buffer.pushChar(c) ;
} ;

//--------------------------------------------------
// pop word

inline ScSubString popWord(ScSubString& buffer) {

   ignoreSpace(buffer) ;

   ScSubString word ;
   word.head = buffer.head ;

   while(ScTrue) {
      if(isTerminated(buffer) || isSpace(buffer) 
	 || isOpenParen(buffer) || isCloseParen(buffer)) {
	 word.tail = buffer.head ;
	 return word ;
      }
      buffer.ignoreChar() ;
   }
}

inline ScBool ignoreWord(ScSubString& buffer) {
   ScSubString w = popWord(buffer) ;
   return !w.isNullStr() ;
} 
  
//--------------------------------------------------
// length to close

inline ScUInt lengthToCloseParen(const ScSubString& buffer) {
   ScUInt l = 0 ;
   ScSubString s = buffer ;
   while(!isTerminated(s)) {
      if(isCloseParen(s)) break ;
      l++ ; s.ignoreChar() ;
   }
   return l ; 
}

inline ScBool ignoreToCloseParen(ScSubString& buffer) {
   while(!isTerminated(buffer)) {
      if(isCloseParen(buffer)) return ScTrue ;
      buffer.ignoreChar() ;
   } ;
   return ScFalse ;
}
   
//--------------------------------------------------
// get S expression (not used now)

static ScSubString popSexp(ScSubString& buffer) {
   ScInt level = 0 ;

   if(!ignoreSpace(buffer)) return ScSubString() ;

   ScSubString sexp = buffer ;

   while(ScTrue) {
      if(isOpenParen(buffer)) {
	 level++ ;
      } else if(isCloseParen(buffer)) {
	 if(level == 0) {    buffer.ignoreChar() ; break ; }
	 else           {level-- ; }
      } else if(isSpace(buffer) && level == 0) {
	 break ;
      } else if(isTerminated(buffer)) {
	 break ;
      }
      buffer.ignoreChar() ;
   }
   sexp.tail = buffer.head ;
   return sexp ;
}

//======================================================================
// scan seeinfo
//

//--------------------------------------------------
// init info

ScBool ScInitInfo::scan(ScSubString& buffer) {

   ///// skip "("
   if(!skipOpenParen(buffer)) return ScFalse ;
   
   ///// tag (init/reconnect)
   ScSubString tag = popWord(buffer) ;
   if      (tag == SC_WORDSTR(Reconnect)) reconnectp = ScTrue ;
   else if (tag == SC_WORDSTR(Init))      reconnectp = ScFalse ;
   else                         return ScFalse ;

   ///// infos

   ScSubString sidestr ;
   ScSubString playmodestr ;

   if(!ignoreSpace(buffer)) return ScFalse ;

   if(reconnectp) {
      sidestr = popWord(buffer) ;
      ignoreSpace(buffer) ;
      playmodestr = popWord(buffer) ;
   } else {
      sidestr = popWord(buffer) ;
      ignoreSpace(buffer) ;
      getIntVal(buffer,unum) ;
      ignoreSpace(buffer) ;
      playmodestr = popWord(buffer) ;
   }

   side = scChar2Side(*(sidestr.head)) ;
   playmode = scString2Word(playmodestr) ;

   return ScTrue ;
}
   

//--------------------------------------------------
// scan object info

ScBool ScSeenObj::scan(ScSubString& buffer, const ScString& teamname) {
   
   ///// skip "("
   if(!skipOpenParen(buffer)) return ScFalse ;
   
   ///// Object Name
   if(!scanObjName(buffer, teamname)) return ScFalse ;

   quality = SC_SQ_none ;

   ///// dist
   ignoreSpace(buffer) ;
   if(!getFloatVal(buffer,dist)) goto _exit_ ;
   quality = SC_SQ_dist ;

   ///// dir
   ignoreSpace(buffer) ;
   if(!getFloatVal(buffer,dir)) goto _exit_ ;
   quality = SC_SQ_dir ;

   ///// dist_chng
   ignoreSpace(buffer) ;
   if(!getFloatVal(buffer,dist_chng)) goto _exit_ ;
   quality = SC_SQ_dist_chng ;

   ///// dir_chng
   ignoreSpace(buffer) ;
   if(!getFloatVal(buffer,dir_chng)) goto _exit_ ;
   quality = SC_SQ_dir_chng ;

   ///// body_dir
   ignoreSpace(buffer) ;
   if(!getFloatVal(buffer,body_dir)) goto _exit_ ;
   quality = SC_SQ_body_dir ;

   ///// head_dir
   ignoreSpace(buffer) ;
   if(!getFloatVal(buffer,head_dir)) goto _exit_ ;
   quality = SC_SQ_head_dir ;

  _exit_:

   ///// skip ")"
   ignoreSpace(buffer) ;
   if(!skipCloseParen(buffer)) return ScFalse ;

   return ScTrue ;
}

//--------------------------------------------------
// convert relative to absolute position

ScBool ScSeenObj::rel2absPos(const ScPosDir& origin, ScPosDir& abspos) const {
   if(know_dir()&&know_dist()) {
      ScFloat adir = origin.hdir + dir ;
      abspos.pos.x = origin.pos.x + scCos(adir) ;
      abspos.pos.y = origin.pos.y + scSin(adir) ;
      if(know_body_dir()) abspos.bdir = origin.hdir + body_dir ;
      if(know_head_dir()) abspos.hdir = origin.hdir + head_dir ;
      return ScTrue ;
   } else {
      return ScFalse ;
   } 
} 

//--------------------------------------------------
// convert relative to absolute velocity

ScBool ScSeenObj::rel2absVel(const ScPosDir& orgpos, 
                             const ScPos& orgvel, 
                             ScPos& absvel) const {
   if(know_dir_chng()&&know_dist_chng()) {
      absvel = orgvel ;
      ScFloat adir = orgpos.hdir + dir ;
      absvel.x +=
	 dist_chng * scCos(adir) -
	 dist * scDeg2Rad(dir_chng) * scSin(adir) ;
      absvel.y +=
	 dist_chng * scSin(adir) +
	 dist * scDeg2Rad(dir_chng) * scCos(adir) ;
      return ScTrue ;
   } else {
      return ScFalse ;
   } 
} 


//--------------------------------------------------
// scan object name

ScBool ScSeenObj::scanObjName(ScSubString& buffer, const ScString& teamname) {
   
   ///// skip "("
   if(!skipOpenParen(buffer)) return ScFalse ;

   ///// object type
   ScSubString objtype = popWord(buffer) ;

   if        (objtype == SC_WORDSTR(Player)) { ///// player /////
      type = SC_OT_player ; inViewConeP = ScTrue ;
      if(!ignoreSpace(buffer)) return ScFalse ;
      ScSubString team = popWord(buffer) ;
      if(team == teamname) {
	 id.player.team = SC_WT_our ;
      } else if(team.nullp()) {
	 id.player.team = SC_WT_unknown ; 
	 id.player.unum = SC_UNUM_unknown ; 
	 goto _exit_ ;
      } else {
	 id.player.team = SC_WT_opponent ;
      }
      if(!ignoreSpace(buffer)) return ScFalse ;
      if(!getUIntVal(buffer,id.player.unum)) { 
	 id.player.unum = SC_UNUM_unknown ; 
      }  

   } else if (objtype == SC_WORDSTR(PlayerOutOfView)) { ///// Player /////
      type = SC_OT_player ; inViewConeP = ScFalse ;
      id.player.team = SC_WT_unknown ;
      id.player.unum = SC_UNUM_unknown ;

   } else if (objtype == SC_WORDSTR(Flag)) { ///// flag /////
      type = SC_OT_flag ; inViewConeP = ScTrue ;
      if(!ignoreSpace(buffer)) return ScFalse ;
      ScSubString flagname =
	 buffer.popSubString(lengthToCloseParen(buffer)) ;
      if(flagname.nullp()) {
	 id.flag = &(ScLandmarkTable[SC_LMID_Unknown]) ;
      } else {
	 ScInt i ;
	 for(i = 0 ; i < SC_LMID_Unknown ; i++) {
	    if(flagname == ScLandmarkTable[i].namestr) {
	       break ;
	    }
	 }
	 id.flag = &(ScLandmarkTable[i]) ;
      }

   } else if (objtype == SC_WORDSTR(FlagOutOfView)) { ///// Flag /////
      type = SC_OT_flag ; inViewConeP = ScFalse ;
      id.flag = &(ScLandmarkTable[SC_LMID_Unknown]) ;
      
   } else if (objtype == SC_WORDSTR(Goal)) { ///// goal /////
      type = SC_OT_goal ; inViewConeP = ScTrue ;
      if(!ignoreSpace(buffer)) return ScFalse ;
      if(scChar2Side(buffer.first()) == SC_S_left) {
	 id.flag = &(ScLandmarkTable[SC_LMID_GL]) ;
	 buffer.ignoreChar() ;
      } else if (scChar2Side(buffer.first()) == SC_S_right) {
	 id.flag = &(ScLandmarkTable[SC_LMID_GR]) ;
	 buffer.ignoreChar() ;
      } else {
	 id.flag = &(ScLandmarkTable[SC_LMID_Unknown]) ;
      }

   } else if (objtype == SC_WORDSTR(GoalOutOfView)) { ///// Goal /////
      type = SC_OT_goal ; inViewConeP = ScFalse ;
      id.flag = &(ScLandmarkTable[SC_LMID_Unknown]) ;

   } else if (objtype == SC_WORDSTR(Line)) { ///// line /////
      type = SC_OT_line ; inViewConeP = ScTrue ;
      if(!ignoreSpace(buffer)) return ScFalse ;
      if        (buffer.isFirst('l')) {
	 id.line = SC_LID_l ;
	 buffer.ignoreChar() ;
      } else if (buffer.isFirst('r')) {
	 id.line = SC_LID_r ;
	 buffer.ignoreChar() ;
      } else if (buffer.isFirst('t')) {
	 id.line = SC_LID_t ;
	 buffer.ignoreChar() ;
      } else if (buffer.isFirst('b')) {
	 id.line = SC_LID_b ;
	 buffer.ignoreChar() ;
      } else {
	 id.flag = &(ScLandmarkTable[SC_LMID_Unknown]) ;
      }

   } else if (objtype == SC_WORDSTR(LineOutOfView)) { ///// Line /////
      type = SC_OT_line ; inViewConeP = ScFalse ;
      id.flag = &(ScLandmarkTable[SC_LMID_Unknown]) ;

   } else if (objtype == SC_WORDSTR(Ball)) { ///// ball /////
      type = SC_OT_ball ; inViewConeP = ScTrue ;

   } else if (objtype == SC_WORDSTR(BallOutOfView)) { ///// Ball /////
      type = SC_OT_ball ; inViewConeP = ScFalse ;
   } else {
      cerr << "Warning: unknown object name: " << objtype 
	   << " ; rest buffer = " << buffer << endl ;
      return ScFalse ;
   }
   
  _exit_:
    
   // skip ")"
   ignoreSpace(buffer) ;
   skipCloseParen(buffer) ;
   
   return ScTrue ;
}

//--------------------------------------------------
// scan visual info

ScBool ScSeeInfo::scan(ScSubString& buffer, const ScString& teamname) {

   ///// skip "(see" 
   if(!skipOpenParen(buffer)) return ScFalse ;
   if(!ignoreWord(buffer)) return ScFalse ;
   
   ///// time
   if(!ignoreSpace(buffer)) return ScFalse ;
   if(!getIntVal(buffer,time)) return ScFalse ;
   
   ///// object list
   n = 0 ;
   while(!buffer.tailp()) {
      ignoreSpace(buffer) ;
      if(isCloseParen(buffer)) break ;
      if(isTerminated(buffer)) return ScFalse ;
      if(!(obj[n].scan(buffer,teamname))) return ScFalse ;
      n++ ;
   }

   ///// skip ")"
   // ignoreSpace(buffer) ;
   // skipCloseParen(buffer) ;
   return ScTrue ;
} ;
      
//======================================================================
// scan hearinfo
//

ScBool ScHearInfo::scan(ScSubString& buffer) {
   ///// skip "(hear"
   if(!skipOpenParen(buffer)) return ScFalse ;
   if(!ignoreWord(buffer)) return ScFalse ;

   ///// time
   if(!ignoreSpace(buffer)) return ScFalse ;
   if(!getIntVal(buffer,time)) return ScFalse ;

   ///// sender
   if(!ignoreSpace(buffer)) return ScFalse ;
   ScSubString sendertype = popWord(buffer) ;
   if(sendertype.nullp()) return ScFalse ;
   
   if     (sendertype == SC_SENDERNAME_REFEREE) { sender = SC_ST_referee ; }
   else if(sendertype == SC_SENDERNAME_SELF)    { sender = SC_ST_self ; }
   else if(sendertype == SC_SENDERNAME_OLCOACHL){ sender = SC_ST_onlineCoachL;}
   else if(sendertype == SC_SENDERNAME_OLCOACHR){ sender = SC_ST_onlineCoachR;}
   else { sender = SC_ST_player ; getFloatVal(sendertype,aux.direction) ; }

   if(!ignoreSpace(buffer)) return ScFalse ;
   
   ScSubString m = buffer.popSubString(lengthToCloseParen(buffer)) ;
   ScSubString msg = message() ;
   msg.copy(m,ScFalse) ;
   if(sender == SC_ST_referee) {
      aux.pmode = scString2Word(msg) ;
   }

   return ScTrue ;
} 

//======================================================================
// scan sense body
//

ScBool ScSenseBodyInfo::scan(ScSubString& buffer) {
   ///// skip "(sense_body"
   if(!skipOpenParen(buffer)) return ScFalse ;
   if(!ignoreWord(buffer)) return ScFalse ;

   ///// time
   if(!ignoreSpace(buffer)) return ScFalse ;
   if(!getIntVal(buffer,time)) return ScFalse ;

   ///// other
   if(!ignoreSpace(buffer)) return ScFalse ;
   while(!(isCloseParen(buffer) || isTerminated(buffer))) {
      ScSubString slot = popSexp(buffer) ;
      if(slot.nullp()) break ;

      if(!skipOpenParen(slot)) return ScFalse ;
      ScSubString slotName = popWord(slot) ;
      if        (slotName == SC_WORDSTR(ViewMode)) {
	 ScSubString q = popWord(slot) ;
	 if      (q == SC_WORDSTR(High)) { viewQuality = SC_VQ_high ; }
	 else if (q == SC_WORDSTR(Low)) { viewQuality = SC_VQ_low ; }
	 else { 
	    cerr <<
	       "Warning: illegal view quality: " << q << slot << endl ;
	    continue ;
	 }
	 ScSubString w = popWord(slot) ;
	 if      (w == SC_WORDSTR(Narrow)) { viewWidth = SC_VW_narrow ; }
	 else if (w == SC_WORDSTR(Normal)) { viewWidth = SC_VW_normal ; }
	 else if (w == SC_WORDSTR(Wide))   { viewWidth = SC_VW_wide ; }
	 else {
	    cerr << "Warning: illegal view width: " << w << slot << endl ;
	    continue ;
	 }
      } else if (slotName == SC_WORDSTR(Stamina)) {
	 if(!getFloatVal(slot,stamina)) {
	    cerr <<
	       "Warning [sense_body]: illegal stamina: " << slot << endl ;
	    continue ;
	 }
	 ignoreSpace(slot) ;
	 if(!getFloatVal(slot,effort)) {
	    cerr <<
	       "Warning [sense_body]: illegal stamina: " << slot << endl ;
	    continue ;
	 }
      } else if (slotName == SC_WORDSTR(Speed)) {
	 if(!getFloatVal(slot,speed)) {
	    cerr <<
	       "Warning [sense_body]: illegal speed: " << slot << endl ;
	    continue ;
	 }
      } else if (slotName == SC_WORDSTR(HeadAngle)) {
	 if(!getFloatVal(slot,headAngle)) {
	    cerr << "Warning [sense_body]: illegal head_angle: " 
		 << slot << endl ;
	    continue ;
	 }
	 headAngle = 0.0 ;
	 //*** This shuold be removed after fixed server bug.

      } else if (slotName == SC_WORDSTR(Kick)) {
	 if(!getUIntVal(slot,countKick)) {
	    cerr <<
	       "Warning [sense_body]: illegal kick #: " << slot << endl ;
	    continue ;
	 }
      } else if (slotName == SC_WORDSTR(Dash)) {
	 if(!getUIntVal(slot,countDash)) {
	    cerr <<
	       "Warning [sense_body]: illegal dash #: " << slot << endl ;
	    continue ;
	 }
      } else if (slotName == SC_WORDSTR(Turn)) {
	 if(!getUIntVal(slot,countTurn)) {
	    cerr <<
	       "Warning [sense_body]: illegal turn #: " << slot << endl ;
	    continue ;
	 }
      } else if (slotName == SC_WORDSTR(Say)) {
	 if(!getUIntVal(slot,countSay)) {
	    cerr <<
	       "Warning [sense_body]: illegal say #: " << slot << endl ;
	    continue ;
	 }
      } else if (slotName == SC_WORDSTR(Catch)) {
	 if(!getUIntVal(slot,countCatch)) {
	    cerr <<
	       "Warning [sense_body]: illegal catch #: " << slot << endl ;
	    continue ;
	 }
      } else if (slotName == SC_WORDSTR(TurnNeck)) {
	 if(!getUIntVal(slot,countTurnNeck)) {
	    cerr << "Warning [sense_body]: illegal turn_neck #: " 
		 << slot << endl ;
	    continue ;
	 }
      } else {
	 cerr << "Warning: unknown sense_body info: " << slotName
	      << " (rest_slot=" << slot 
	      << ", rest=" << buffer << ")" << endl ;
      } 
   }
   return ScTrue ;
}

//======================================================================
// scan sensor info
//

ScBool ScSensorInfo::scan(ScString& buffer, ScBool rewindp) {
   ScSubString work = buffer ;

   ///// skip "("
   if(!skipOpenParen(work)) return ScFalse ;
   
   ScSubString tag = popWord(work) ;
   work = buffer ;
   ScBool r ;
   if        (tag == SC_WORDSTR(See)) {
      type = SC_SIT_see ;
      r = info.see.scan(work,teamname) ;
   } else if (tag == SC_WORDSTR(Hear)) {
      type = SC_SIT_hear ;
      r = info.hear.scan(work) ;
   } else if (tag == SC_WORDSTR(SenseBody)) {
      type = SC_SIT_senseBody ;
      r = info.senseBody.scan(work) ;
   } else if (tag == SC_WORDSTR(Error)) {
      type = SC_SIT_error ;
      cerr << "Warning: " << (char*)buffer << endl ; 
      scNop() ;
      return ScFalse ;
   } else {
      cerr << "Warning: unknown response from the server: " 
	   << (char*)buffer << endl ;
      return ScFalse ;
   }
   
   if(!rewindp) buffer = work ;
}

//======================================================================
// Position Estimation
//

//----------------------------------------------------------------------
// estimate pos dir vec (1) in abstruct orientation

//V.1.05noda
ScBool scEstimatePosDirVelAbs1(ScPosDir& posdir, ScPos& vel, 
			       ScBool& outOfFieldP,
                               const ScSeeInfo& sinf) {
   ScFloat dist = 2.0 * SC_PITCH_LENGTH ;
   const ScSeenObj *line = (ScSeenObj *)SC_NULLPTR;
   const ScSeenObj *flag = (ScSeenObj *)SC_NULLPTR;

   outOfFieldP = ScTrue ;

   for(ScUInt i = 0 ; i < sinf.n ; i++) {
      const ScSeenObj& obj = sinf.obj[i] ;
      if(obj.type == SC_OT_flag || obj.type == SC_OT_goal) {
	 if(!obj.inViewConeP) continue ;
	 if(scAbs(dist) > scAbs(obj.dist)) {
	    flag = &obj ;
	    dist = obj.dist ;
	 }
      } else if (obj.type == SC_OT_line) {
	 if(scIsNullPtr(line)) {
	    outOfFieldP = ScFalse ;
	    line = &obj ; 
	 } else {
	    outOfFieldP = ScTrue ;
	 }
	 i++ ;
      }
   }

   if(scIsNullPtr(line)) return ScFalse ;
   if(scIsNullPtr(flag)) return ScFalse ;

//   cerr << flag->id.flag->namestr << ":" << line->id.line ;
   
   ScPos flag_pos = flag->id.flag->pos ;

   switch(line->id.line) {
      case SC_LID_t: {
	 if (outOfFieldP) {
	    posdir.hdir = -line->dir - ((line->dir > 0) ? -180.0 : 0.0) ;
	    posdir.pos.y = (- SC_HALF_PITCH_W 
			    - scAbs(line->dist * scSin(line->dir))) ;
	 } else {
	    posdir.hdir = -line->dir - ((line->dir < 0) ?  180.0 : 0.0) ;
	    posdir.pos.y = (- SC_HALF_PITCH_W 
			    + scAbs(line->dist * scSin(line->dir))) ;
	 }

	 ScFloat diff_x = sqrt(scAbs(scSq(flag->dist) -
				     scSq(flag_pos.y - posdir.pos.y))) ;
	 ScFloat obj_dir = scNormalizeAngle(flag->dir + posdir.hdir) ;
     
	 if(scAbs(obj_dir) < 90.0)
	    posdir.pos.x = flag_pos.x - diff_x ;
	 else
	    posdir.pos.x = flag_pos.x + diff_x ;
	 break ;
      }
      case SC_LID_b: {
	 if (outOfFieldP) {
	    posdir.hdir = -line->dir - ((line->dir < 0) ?  180.0 : 0.0) ;
	    posdir.pos.y = (  SC_HALF_PITCH_W 
			      + scAbs(line->dir * scSin(line->dir))) ;
	 } else {
	    posdir.hdir = -line->dir - ((line->dir > 0) ? -180.0 : 0.0) ;
	    posdir.pos.y = (  SC_HALF_PITCH_W 
			      - scAbs(line->dir * scSin(line->dir))) ;
	 }

	 ScFloat diff_x = sqrt(scAbs(scSq(flag->dist) -
				     scSq(flag_pos.y - posdir.pos.y))) ;
	 ScFloat obj_dir = scNormalizeAngle(flag->dir + posdir.hdir) ;
      
	 if(scAbs(obj_dir) < 90.0)
	    posdir.pos.x = flag_pos.x - diff_x ;
	 else
	    posdir.pos.x = flag_pos.x + diff_x ;
	 break ;
      }
      case SC_LID_l: {
	 if (outOfFieldP) {
	    posdir.hdir = -line->dir - ((line->dir > 0) ?  -90.0 : 90.0) ;
	    posdir.pos.x = (- SC_HALF_PITCH_L 
			    - scAbs(line->dist * scSin(line->dir))) ;
	 } else {
	    posdir.hdir = -line->dir - ((line->dir < 0) ?  -90.0 : 90.0) ;
	    posdir.pos.x = (- SC_HALF_PITCH_L 
			    + scAbs(line->dist * scSin(line->dir))) ;
	 }

	 ScFloat diff_y = sqrt(scAbs(scSq(flag->dist) -
				     scSq(flag_pos.x - posdir.pos.x))) ;
	 ScFloat obj_dir = scNormalizeAngle(flag->dir + posdir.hdir) ;
      
	 if(obj_dir > 0.0)
	    posdir.pos.y = flag_pos.y - diff_y ;
	 else
	    posdir.pos.y = flag_pos.y + diff_y ;
	 break ;
      }
      case SC_LID_r: {
	 if (outOfFieldP) {
	    posdir.hdir = -line->dir - ((line->dir < 0) ?  -90.0 : 90.0) ;
	    posdir.pos.x = (  SC_HALF_PITCH_L 
			      + scAbs(line->dist * scSin(line->dir))) ;
	 } else {
	    posdir.hdir = -line->dir - ((line->dir > 0) ?  -90.0 : 90.0) ;
	    posdir.pos.x = (  SC_HALF_PITCH_L 
			      - scAbs(line->dist * scSin(line->dir))) ;
	 }

	 ScFloat diff_y = sqrt(scAbs(scSq(flag->dist) -
				     scSq(flag_pos.x - posdir.pos.x))) ;
	 ScFloat obj_dir = scNormalizeAngle(flag->dir + posdir.hdir) ;
      
	 if(obj_dir > 0.0)
	    posdir.pos.y = flag_pos.y - diff_y ;
	 else
	    posdir.pos.y = flag_pos.y + diff_y ;
	 break ;
      }
      default: return ScFalse ;
   }

   return ScTrue ;
}

//V.1.05noda
ScBool scEstimatePosDirVelTeam1(ScPosDir& posdir, ScPos& vel,
				ScBool& outOfFieldP,
                                const ScSeeInfo& sinf, const ScSide side) {
   if(scEstimatePosDirVelAbs1(posdir,vel,outOfFieldP,sinf)) {
      scOrientationAbs2Team(posdir,vel,side) ;
      return ScTrue ;
   } else {
      return ScFalse ;
   } 
}


//----------------------------------------------------------------------
// estimate pos dir vec (2) in abstruct orientation

inline ScFloat confidentFactor(const ScFloat dist) { 
   return (1.0 / scSq(dist)) ;
} ;

//V.1.05noda
ScBool scEstimatePosDirVelAbs2(ScPosDir& posdir, ScFloat& posConf, 
                               ScPos& vel, ScFloat& velConf,
			       ScBool& outOfFieldP,
                               const ScSeeInfo& sinf) {
   ScPos p ; p.x = 0.0 ; p.y = 0.0 ;
   ScPos pN ; pN.x = 0.0 ; pN.y = 0.0 ;
   ScPos v ; v.x = 0.0 ; v.y = 0.0 ;
   ScFloat vN = 0.0 ;

   const ScSeenObj *line = (ScSeenObj *)SC_NULLPTR;
   
   outOfFieldP = ScTrue ;

   for(ScUInt i = 0 ; i < sinf.n ; i++) {
      if(sinf.obj[i].type == SC_OT_line) {
	 if(scIsNullPtr(line)) { 
	    line = &(sinf.obj[i]) ; 
	    outOfFieldP = ScFalse ;
	 } else { 
	    outOfFieldP = ScTrue ;
	 }
      }
   }
   
   if(scIsNullPtr(line)) return ScFalse ;

   switch(line->id.line) {
      case SC_LID_t :{
	 ScFloat ey ;
	 if(outOfFieldP) {
	    posdir.hdir = -line->dir - ((line->dir > 0) ? -180.0 : 0.0) ;
	    ey = (- SC_HALF_PITCH_W -
		  scAbs(line->dist * scSin(line->dir))) ;
	 } else {
	    posdir.hdir = -line->dir - ((line->dir < 0) ?  180.0 : 0.0) ;
	    ey = (- SC_HALF_PITCH_W +
		  scAbs(line->dist * scSin(line->dir))) ;
	 }
	 ScFloat w = confidentFactor(line->dist) ;
	 p.y += ey * w ; pN.y += w ;
	 break ;
      }
      case SC_LID_b : {
	 ScFloat ey ;
	 if(outOfFieldP) {
	    posdir.hdir = -line->dir - ((line->dir < 0) ?  180.0 : 0.0) ;
	    ey = (  SC_HALF_PITCH_W -
		    scAbs(line->dist * scSin(line->dir))) ;
	 } else {
	    posdir.hdir = -line->dir - ((line->dir > 0) ? -180.0 : 0.0) ;
	    ey = (  SC_HALF_PITCH_W +
		    scAbs(line->dist * scSin(line->dir))) ;
	 }
	 break ;
	 ScFloat w = confidentFactor(line->dist) ;
	 p.y += ey * w ; pN.y += w ;
      }
      case SC_LID_l : {
	 ScFloat ex ;
	 if(outOfFieldP) {
	    posdir.hdir = -line->dir - ((line->dir > 0) ?  -90.0 : 90.0) ;
	    ex = (- SC_HALF_PITCH_L -
		  scAbs(line->dist * scSin(line->dir))) ;
	 } else {
	    posdir.hdir = -line->dir - ((line->dir < 0) ?  -90.0 : 90.0) ;
	    ex = (- SC_HALF_PITCH_L +
		  scAbs(line->dist * scSin(line->dir))) ;
	 }
	 ScFloat w = confidentFactor(line->dist) ;
	 p.x += ex * w ; pN.x += w ;
	 break ;
      }
      case SC_LID_r : {
	 ScFloat ex ;
	 if(outOfFieldP) {
	    posdir.hdir = -line->dir - ((line->dir < 0) ?  -90.0 : 90.0) ;
	    ex = (  SC_HALF_PITCH_L +
		    scAbs(line->dist * scSin(line->dir))) ;
	 } else {
	    posdir.hdir = -line->dir - ((line->dir > 0) ?  -90.0 : 90.0) ;
	    ex = (  SC_HALF_PITCH_L -
		    scAbs(line->dist * scSin(line->dir))) ;
	 }
	 ScFloat w = confidentFactor(line->dist) ;
	 p.x += ex * w ; pN.x += w ;
	 break ;
      }
      default: return ScFalse ;
   }   
   for(ScUInt i = 0 ; i < sinf.n ; i++) {
      const ScSeenObj& obj = sinf.obj[i] ;

      if(!(obj.type == SC_OT_flag || obj.type == SC_OT_goal)) continue ;
      if(!(obj.inViewConeP)) continue ;

      ScPos flag_pos = obj.id.flag->pos ;

      ScFloat obj_dir = obj.dir + posdir.hdir ;
      ScFloat obj_dist = obj.dist ;
      ScPos ep = ScPos(flag_pos.x - obj_dist * scCos(obj_dir),
		       flag_pos.y - obj_dist * scSin(obj_dir)) ;
      ScFloat w = confidentFactor(obj_dist) ;

      p.x += w * ep.x ; p.y += w * ep.y ; 
      pN.x += w ; pN.y += w ;

      if(obj.know_dist_chng() && obj.know_dir_chng()) {
	 ScFloat r = obj.dist_chng ;
	 ScFloat a = scDeg2Rad(obj.dir_chng) * obj_dist ;

	 v.x += -(w * (r * scCos(obj_dir) - a * scSin(obj_dir))) ;
	 v.y += -(w * (r * scSin(obj_dir) + a * scCos(obj_dir))) ;
	 vN += w ;
      }
   }

   if(pN.x <= 0.0 || pN.y <= 0.0) return ScFalse ;

   posdir.pos.set((p.x / pN.x), (p.y / pN.y)) ;
   posConf = scMin(pN.x,pN.y) ;
   //   cerr << p << pN << endl ;

   if(vN < 0.00001) { vel.set(0.0,0.0) ; } 
   else             { vel.set(v.x/vN, v.y/vN) ; }
   velConf = vN ;

   return ScTrue ;
}

//V.1.05noda
ScBool scEstimatePosDirVelTeam2(ScPosDir& posdir, ScFloat& posConf,
                                ScPos& vel, ScFloat& velConf,
				ScBool& outOfFieldP,
                                const ScSeeInfo& sinf, const ScSide side) {
   if(scEstimatePosDirVelAbs2(posdir,posConf,vel,velConf,outOfFieldP,sinf)) {
      scOrientationAbs2Team(posdir,vel,side) ;
      return ScTrue ;
   } else {
      return ScFalse ;
   } 
}

//======================================================================
// Class Player
//

ScPlayer::~ScPlayer() { }

//----------------------------------------------------------------------
// basic command

//--------------------------------------------------
// init

ScBool ScPlayer::sendInit(ScFloat version, ScBool goaliep) {
   char* buffer = (char*)sendBuffer ;

   if(goaliep) {
      sprintf(buffer,"(%s %s (version %lf) (goalie))\n\000", 
	      SC_WORDSTR(Init), (char*)teamname, version) ;
   } else {
      sprintf(buffer,"(%s %s (version %lf))\n\000", 
	      SC_WORDSTR(Init), (char*)teamname, version) ;
   }
   if (!socket.send(sendBuffer)) return ScFalse ;

   ScBool original_block_mode = socket.checkBlockingMode() ;
   socket.setBlockingMode(ScTrue) ;
   if (!socket.recv(recvBuffer)) return ScFalse ;
   socket.setBlockingMode(original_block_mode) ;

   ScSubString work = recvBuffer ;
   if(!iinfo.scan(work)) return ScFalse ;
   side = iinfo.side ;
   unum = iinfo.unum ;
   playmode = iinfo.playmode ;

   return ScTrue ;
} ;

//--------------------------------------------------
// reconnect

ScBool ScPlayer::sendReconnect(ScUInt unumber) {
   char* buffer = (char*)sendBuffer ;

   if(unumber == 0) unumber = unum ;

   sprintf(buffer,"(%s %s %ld)\n\000", 
	   SC_WORDSTR(Reconnect), (char*)teamname, unumber) ;

   if(!socket.send(sendBuffer)) return ScFalse ;

   if (!socket.recv(recvBuffer)) return ScFalse ;

   ScSubString work = recvBuffer ;
   if(!iinfo.scan(work)) return ScFalse ;
   side = iinfo.side ;

   return ScTrue ;
} ;

//--------------------------------------------------
// move

ScBool ScPlayer::sendMove(ScFloat x, ScFloat y) {
   char* buffer = (char*)sendBuffer ;
   sprintf(buffer,"(%s %lf %lf)\n\000",SC_WORDSTR(Move),x,y) ;
   return send() ;
} ;

//--------------------------------------------------
// turn

ScBool ScPlayer::sendTurn(ScFloat pow) {
   char* buffer = (char*)sendBuffer ;
   sprintf(buffer,"(%s %lf)\n\000",SC_WORDSTR(Turn),pow) ;

   if(!send()) return ScFalse ;

   estimateBodyState.countTurn++ ;
    
   return ScTrue ;
} ;

//--------------------------------------------------
// dash

ScBool ScPlayer::sendDash(ScFloat pow) {
   char* buffer = (char*)sendBuffer ;
   sprintf(buffer,"(%s %lf)\n\000",SC_WORDSTR(Dash),pow) ;

   if(!send()) return ScFalse ;

   estimateBodyState.countDash++ ;

   return ScTrue ;
} ;

//--------------------------------------------------
// kick

ScBool ScPlayer::sendKick(ScFloat pow, ScFloat dir) {
   char* buffer = (char*)sendBuffer ;
   sprintf(buffer,"(%s %lf %lf)\n\000",SC_WORDSTR(Kick),pow,dir) ;

   if(!send()) return ScFalse ;

   estimateBodyState.countKick++ ;

   return ScTrue ;
} ;

//--------------------------------------------------
// catch

ScBool ScPlayer::sendCatch(ScFloat dir) {
   char* buffer = (char*)sendBuffer ;
   sprintf(buffer,"(%s %lf)\n\000",SC_WORDSTR(Catch),dir) ;

   if(!send()) return ScFalse ;

   estimateBodyState.countCatch++ ;

   return ScTrue ;
} ;

//--------------------------------------------------
// turn_neck

ScBool ScPlayer::sendTurnNeck(ScFloat pow) {
   char* buffer = (char*)sendBuffer ;
   sprintf(buffer,"(%s %lf)\n\000",SC_WORDSTR(TurnNeck),pow) ;

   if(!send()) return ScFalse ;

   estimateBodyState.countTurnNeck++ ;

   return ScTrue ;
} ;

//--------------------------------------------------
// say

ScBool ScPlayer::sendSay(ScString msg) {
   char* buffer = (char*)sendBuffer ;
   sprintf(buffer,"(%s %s)\n\000",SC_WORDSTR(Say),(char*)msg) ;

   if(!send()) return ScFalse ;

   estimateBodyState.countSay++ ;

   return ScTrue ;
} ;

//--------------------------------------------------
// sense body

ScBool ScPlayer::sendSenseBody() {
   char* buffer = (char*)sendBuffer ;
   sprintf(buffer,"(%s)\n\000",SC_WORDSTR(SenseBody)) ;

   if(!send()) return ScFalse ;

   estimateBodyState.countSenseBody++ ;

   return ScTrue ;
} ;

//--------------------------------------------------
// change view

ScBool ScPlayer::sendChangeView(ScViewWidth vw, ScViewQuality vq) {
   char* buffer = (char*)sendBuffer ;
   char* quality_name ; char *width_name ;

   width_name = SC_VW_NAME(vw) ;
   quality_name = SC_VQ_NAME(vq) ;

   sprintf(buffer,"(%s %s %s)\n\000",
	   SC_WORDSTR(ChangeView),width_name, quality_name) ;

   if(!send()) return ScFalse ;

   estimateBodyState.countChangeView++ ;

   return ScTrue ;
} ;

//----------------------------------------------------------------------
// sensor info

ScBool ScPlayer::recvScanSInfo() {
   if(socket.recv(recvBuffer)) {
      sinfo.scan(recvBuffer,ScTrue) ;
      switch(sinfo.type) {
	 case SC_SIT_see:{
	    scycle.set(sinfo.info.see.time) ;
	    vcycle.set(sinfo.info.see.time) ;
	    // if(estimatePosDirVel1()) {
	    if(estimatePosDirVel2()) {
	       eposMutex.lock() ;
	       ecycle.set(sinfo.info.see.time) ;
	       //if (  (vcycle.rtime.MSec() - xcycle.rtime.MSec())
	       //      < simulationTic.MSec()) {
	       // keep estimate info; the latest visual info
	       // does not reflect our last action
	       // xcycle = vcycle ; // should we?
	       //} else {
	       xposdir = eposdir ; xvel = evel ;
	       xcycle = vcycle ;
	       //}
	       eposMutex.unlock() ;
	    }
	    break ; 
	 }
	 case SC_SIT_hear: {
	    scycle.set(sinfo.info.hear.time) ;
	    acycle.set(sinfo.info.hear.time) ;
	    if(sinfo.info.hear.sender == SC_ST_referee) 
	       playmode = sinfo.info.hear.aux.pmode ;
	    break ;
	 }
	 case SC_SIT_senseBody: {
	    adjustBodyState(sinfo) ;
	    break ;
	 }
	 case SC_SIT_error: {
	    cerr << "Warning: previous command is " 
		 << (char*) sendBuffer << endl ;
	    break ;
	 }
	 default: {
	    return ScFalse ;
	 }
      }
      return ScTrue ;
   } else {
      return ScFalse ;
   }
}

//----------------------------------------------------------------------
// basic command

ScBool ScPlayer::waitNextFootCommandTic() {
   ScTimeVal now ; now.update();
   ScTimeVal diff = now - fcycle.rtime ;
   if(diff < simulationTic) {
      scSleep(simulationTic - diff) ;
      return ScTrue ;
   } 
   return ScFalse ;
}


ScBool ScPlayer::move(const ScFloat& x, const ScFloat& y) {
   //waitNextFootCommandTic() ;
   xposMutex.lock() ;
   ScBool r = sendMove(x,y) ;
   if(r) {
      xposdir.pos.x = x ;
      xposdir.pos.y = y ;
      xcycle.update() ;
   }
   xposMutex.unlock() ;
   return r ;
}


ScBool ScPlayer::turn(const ScFloat& pow) {
   if(scAbs(pow) < 1.0) return ScTrue ;
   waitNextFootCommandTic() ;
   xposMutex.lock() ;
   ScBool r = sendTurn(pow) ;
   if(r) {
      fcycle.update() ;
      xposdir.bdir = scNormalizeAngle(xposdir.bdir + pow) ;
      xcycle.update() ;
   } 
   xposMutex.unlock() ;
   return r ;
}


ScBool ScPlayer::dash(const ScFloat& pow) {
   waitNextFootCommandTic() ;
   xposMutex.lock() ;
   ScBool r = sendDash(pow) ;
   if(r) {
      fcycle.update() ;
      xposdir.pos.x += (dashPowerEffect
			* pow * scCos(xposdir.bdir)) ;
      xposdir.pos.y += (dashPowerEffect
			* pow * scSin(xposdir.bdir)) ;
      xcycle.update() ;
   }
   xposMutex.unlock() ;
   return r ;
}


ScBool ScPlayer::kick(const ScFloat& pow, const ScFloat& dir) {
   waitNextFootCommandTic() ;
   ScBool r = sendKick(pow,dir) ;
   if(r) fcycle.update() ;

   return r ;
}


ScBool ScPlayer::catchBall(const ScFloat& dir) {
   ScBool r = sendCatch(dir) ;
   if(r) fcycle.update() ;  // 1.04o
   return r ;
}


ScBool ScPlayer::turnNeck(const ScFloat& pow) {
   ScBool r = sendTurnNeck(pow) ;
   return r ;
}


ScBool ScPlayer::say(ScString msg) {
   ScBool r = sendSay(msg) ;
   return r ;
}


// 1.05
ScBool ScPlayer::turnTo(const ScFloat& dir) {
    return turn(scNormalizeAngle(dir - xposdir.bdir)) ;
}

