// -*- Mode:C++ -*-
//Header:
//File: libsclient.h
//Author: NDOA, Itsuki
//Date: 1999/04/15
//

//ModifyHistory:
// 1999/04/15: Create this file
//EndModifyHistory:

#ifndef _libsclient_h_
#define _libsclient_h_
////////////////////////////////////////////////////////////////////////

//======================================================================
// Includes
//

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

extern "C" {

#include <math.h>

#ifdef THREADED
#  include <pthread.h>
#endif

#include <stdlib.h>
#include <errno.h>
}

//==================================================
// for configure [#include <unistd.h>]
//
#if HAVE_UNISTD_H
# include <sys/types.h>
# include <unistd.h>
#endif
#ifdef _POSIX_VERSION
/* Code for POSIX.1 systems.  */
#endif
//--------------------------------------------------

//==================================================
// for configure [#include <string.h>]
//
#if STDC_HEADERS
# include <string.h>
#else
# ifndef HAVE_STRCHR
#  define strchr index
#  define strrchr rindex
# endif
char *strchr (), *strrchr ();
# ifndef HAVE_MEMCPY
#  define memcpy(d, s, n) bcopy ((s), (d), (n))
#  define memmove(d, s, n) bcopy ((s), (d), (n))
# endif
#endif
//--------------------------------------------------

#include <netdb.h>
#include <errno.h>

#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#ifdef Solaris
#include <sys/filio.h>
#else
#include <sys/ioctl.h>
#endif
#include <fcntl.h>


#include <iostream.h>

void scNop() ; // for debug

//======================================================================
// General Types
//

typedef long ScInt ;
typedef unsigned long ScUInt ;
typedef double ScFloat ;

typedef int ScBool ;
const ScBool ScFalse = 0 ;
const ScBool ScTrue = 1 ;

#define SC_NULLPTR  NULL
#define SC_NULLCHR  '\000'

inline ScBool scIsNullPtr(const void* pt) { return (pt == SC_NULLPTR) ; } ;
inline ScBool scIsNullChr(const char c) { return (c == SC_NULLCHR) ; } ;

//--------------------------------------------------
// string

class ScSubString {
public:
    char* head ;
    char* tail ;

    //--------------------------------------------------
    // constructor
    ScSubString(const ScUInt size) { alloc(size) ; } ;
    ScSubString(char* h, char* t) { set(h,t) ; } ;
    ScSubString(char* h, const ScUInt l) { set(h,l) ; } ;
    ScSubString(char* str = SC_NULLPTR) { set(str) ; } ;
    ScSubString(const ScSubString& src) { set(src) ; } ;
    ScSubString(const ScSubString& src, 
                const ScUInt hidx, const ScUInt tidx) {
        set(src,hidx,tidx) ;
    } ;

    //--------------------------------------------------
    // operator=

    ScSubString& operator= (const ScSubString& s) { 
        set(s) ; return *this ;
    } ;

    //--------------------------------------------------
    // operator==

    ScBool eq(const ScSubString& r) const { 
        return (head == r.head) && (tail == r.tail) ;
    } ;

    ScBool operator== (const ScSubString& r) const { 
        ScUInt l = length() ;
        return (l == r.length()) && !strncmp(head,r.head,l) ;
    } ;

    ScBool operator== (const char* ref) const { 
        ScUInt l = length() ;
        return !strncmp(head,ref,l) ;
    } ;

    ScBool isPrefixOf(const ScSubString& ref) const {
        ScUInt l = length() ;
        return l <= ref.length() && !strncmp(head,ref.head,length()) ;
    } ;
    ScBool isPrefixOf(const char* ref) const {
        return isPrefixOf(ScSubString((char*)ref)) ;
    } ;

    ScBool hasPrefix(const ScSubString& prefix) const {
        return prefix.isPrefixOf(*this) ;
    } ;
    ScBool hasPrefix(const char* ref) const {
        return hasPrefix(ScSubString((char*)ref)) ;
    } ;

    //--------------------------------------------------
    // set

    void set(char* h, char* t) { head = h ;  tail = t ; } ;
    void set(char* h, const ScUInt l) { 
        head = h ; 
        setLength(l) ;
    } ;
    void set(char* str = SC_NULLPTR) { 
        if(scIsNullPtr(str)) set(SC_NULLPTR,(char*)SC_NULLPTR) ;
        else set(str,strlen(str)) ;
    } ;
    void set(const ScSubString& src) { set(src.head,src.tail) ; } ;
    void set(const ScSubString& src, const ScUInt hidx, const ScUInt tidx) {
        head = &(src.head[hidx]) ;
        tail = &(src.head[tidx]) ;
    } ;
    void set(const ScSubString& src, const ScUInt hidx) {
        head = &(src.head[hidx]) ;
        tail = src.tail ;
    } ;

    //--------------------------------------------------
    // alloc
    ScBool alloc(const ScUInt l) {
        head = new char[l] ;
        if(scIsNullPtr(head)) {
            tail = SC_NULLPTR ;
            return ScFalse ;
        } else {
            set(head,l) ;
            return ScTrue ;
        }
    } ;

    //--------------------------------------------------
    // length

    ScUInt length() const { return (tail - head) ; } ;
    ScUInt size() const { return length() + 1 ; } ;

    void setLength(ScUInt l) { tail = &(head[l]) ; } ;

    ScUInt strlength() const { 
        ScUInt l = 0 ;
        for(char* s = head ; s < tail ; s++) {
            if(scIsNullChr(*s)) return l ;
            l++ ;
        } 
        return l+1 ;
    } ;

    //--------------------------------------------------
    // index

    char* find(char c, ScBool nullreturnp = ScFalse) {
        for(char* s = head ; s < tail ; s++) {
            if(c == *s) return s ;
        }
        if(nullreturnp) return SC_NULLPTR ;
        return tail ;
    } ;

    //--------------------------------------------------
    // copy

    ScBool copy(const ScSubString& src, ScBool allocp = ScFalse) {
        if(allocp) alloc(src.length()+1) ;
        tail = head ;
        for(char* s = (char*)src ; s < src.tail ; s++) {
            *tail = *s ; tail++ ;
        }
        terminate() ;
        return ScTrue;
    } ;

    ScBool copy(char* h, char* t, ScBool allocp = ScFalse) {
        return copy(ScSubString(h,t),allocp) ;
    } ;

    //--------------------------------------------------
    // converter

    operator char* () const { return head ; } ;

    //--------------------------------------------------
    // terminate
    void terminate() { *tail = SC_NULLCHR ; } ;

    ScBool terminatedp() { return strlength() <= length() ; } ; 

    ScBool isNullStr() { return head == tail ; } ;

    //--------------------------------------------------
    // basic operation

    char& first() const { return *head ; } ;

    ScBool isFirst(const char c) const { return first() == c ; } ;
      
    ScBool tailp() const { return head >= tail ; } ;
    ScBool nullp() const { return head >= tail ; } ;

    char popChar() { char c = first() ; head++ ; return c ;} ;
    void pushChar(const char c) { 
        head-- ;
        first() = c ;
    } ;

    void ignoreChar() { head++ ; } ;
    ScBool skipNChar(ScUInt n = 1) {
        if(n > length()) return ScFalse ;
        head = &(head[n]) ;
        return ScTrue ;
    } ;

    ScSubString sub(ScUInt length) const { return sub(0, length) ; }
    ScSubString sub(ScUInt offset, ScUInt length) const {
        return ScSubString(&(head[offset]),&(head[offset+length])) ;
    } ;

    ScSubString popSubString(ScUInt length) { 
        return popSubString(0,length) ; 
    } ;
    ScSubString popSubString(ScUInt offset, ScUInt length) {
        ScSubString s = sub(offset,length) ;
        head = s.tail ;
        return s ;
    } ;

    ScSubString popUntil(char delimiter) {
        char* delpos = find(delimiter,ScFalse) ;
        ScSubString s(head,delpos) ;
        head = delpos ;
        return s ;
    } ;
      
    //--------------------------------------------------
    // conversion

    ScBool popIntVal(ScInt& i) {
        int l ;
        if(1 > sscanf(head, "%ld%n", &i, &l)) return ScFalse ;
        return skipNChar(l) ;
    } ;

    ScBool popUIntVal(ScUInt& i) {
        int l ;
        if(1 > sscanf(head, "%ld%n", &i, &l)) return ScFalse ;
        return skipNChar(l) ;
    } ;

    ScBool popFloatVal(ScFloat& v) {
        int l ;
        if(1 > sscanf(head, "%lf%n", &v, &l)) return ScFalse ;
        return skipNChar(l) ;
    } ;

    //--------------------------------------------------
    // describe

    void describe(ostream& o, ScBool detailp = ScTrue, 
                  ScBool newlinep = ScTrue) const {
        if(detailp) {
            o << "#ScSubString[" << head << "-" << tail
              << "(len=" << length() << "):" ;
        } 
        if(scIsNullPtr(head)) {
            o << "(null)" ;
        } else {
            for(char* s = head ; s < tail ; s++) {
                o << *s ;
            }
        }
        if(detailp) {
            o << "]" ;
        }
        if(newlinep) o << endl ;
    } ;

    friend ostream& operator<< (ostream& o, const ScSubString& s) {
        s.describe(o,ScFalse,ScFalse) ;
        return o ;
    } ;
} ;

typedef ScSubString ScString ;

//--------------------------------------------------
// time

class ScTimeVal {
public:
    struct timeval tval ;

    //--------------------------------------------------
    // constructor

    inline ScTimeVal(ScInt sec = 0, ScInt msec = 0, ScInt usec = 0) {
        ScInt us = msec * 1000 + usec ;
        tval.tv_usec = us % 1000000 ;
        tval.tv_sec = sec + (us / 1000000) ;
    } ;
    inline ScTimeVal(const ScTimeVal& src) {
        tval = src.tval ;
    } ;
    inline ScTimeVal(const ScFloat sec) {
        set(sec) ;
    } ;

    //--------------------------------------
    // get value

    ScInt sec() const { return tval.tv_sec ; } ;
    ScInt usec() const { return tval.tv_usec ; } ;
    ScInt msec() const { return tval.tv_usec / 1000 ; } ;
    ScInt rusec() const { return tval.tv_usec % 1000 ; } ;
    ScUInt MSec() const { return msec() + (sec() * 1000) ; } ;
    ScUInt USec() const { return usec() + (sec() * 1000000) ; } ;

    ScFloat MSecFloat() const { return msec() + (sec() * 1000.0) ; } ;

    ScFloat fltSec() const { 
        return (((ScFloat)sec()) + (((ScFloat)usec())/1000000.0)) ;
    } ;

    //--------------------------------------
    // Extract 
    void extract(ScInt& s, ScInt& ms, ScInt& us) {
        s = sec() ;
        ms = msec() ;
        us = rusec() ;
    } ;

    //--------------------------------------
    // set value

    ScTimeVal& getTimeOfDay() {
        gettimeofday(&tval,NULL) ;
        return *this;
    } ;

    ScTimeVal& update() { return getTimeOfDay() ; } ;

    ScTimeVal& set(ScFloat sec) {
        tval.tv_sec = (ScInt)floor(sec) ;
        tval.tv_usec = 
            (ScInt)floor((sec - (ScFloat)tval.tv_sec) * 1000000.0) ;
        return *this;
    } ;

    ScTimeVal& set(ScUInt sec = 0, ScUInt msec = 0, ScUInt usec = 0) {
        ScUInt us = msec * 1000 + usec ;
        tval.tv_usec = us % 1000000 ;
        tval.tv_sec = sec + (us / 1000000) ;
        return *this ;
    } ;

    ScTimeVal operator=(const ScTimeVal& tv) {
        tval = tv.tval ;
        return *this ;
    } ;

    //--------------------------------------
    // Inc / dec / diff /mul / div

    ScTimeVal& inc(const ScInt sec, const ScInt msec = 0, 
                   const ScInt usec = 0) {
        ScUInt us = tval.tv_usec + (msec * 1000) + usec ;
        tval.tv_usec = us % 1000000 ;
        tval.tv_sec += sec + (us / 1000000) ;
        return *this ;
    } ;
    
    ScTimeVal& inc(const ScTimeVal& x) {
        return inc(x.sec(),0,x.usec()) ;
    } ;

    ScTimeVal& dec(const ScInt sec, const ScInt msec = 0, 
                   const ScInt usec = 0) {
        ScUInt d = (msec * 1000) + usec ;
        ScUInt over=0 ;
        if(d > (ScUInt)tval.tv_usec) over++ ;
        over += d/1000000 ;
        tval.tv_usec += (over * 1000000 - d) ;
        tval.tv_sec -= (over + sec) ;
        return *this ;
    } ;
    
    ScTimeVal& dec(const ScTimeVal& x) {
        return dec(x.sec(),0,x.usec()) ;
    } ;
      
    ScTimeVal diff(const ScTimeVal& x) const {
        ScTimeVal v = x ;
        return v.dec(*this) ;
    } ;

    ScUInt diffSec(const ScTimeVal& x) const {
        ScTimeVal v = diff(x) ;
        return v.sec() ;
    } ;
    ScUInt diffMSec(const ScTimeVal& x) const {
        ScTimeVal v = diff(x) ;
        return v.MSec() ;
    } ;
    ScUInt diffUSec(const ScTimeVal& x) const {
        ScTimeVal v = diff(x) ;
        return v.USec() ;
    } ;

    ScTimeVal& mul(const ScFloat m) {
        ScFloat s = fltSec() ;
        set(s * m) ;
        return *this ;
    } ;

    ScTimeVal& div(const ScFloat m) {
        ScFloat s = fltSec();
        set(s / m) ;
        return *this ;
    } ;

    //--------------------------------------
    // add / del / mul / div

    ScTimeVal operator+(const ScTimeVal& tv) const {
        ScTimeVal r = *this ;
        return r.inc(tv) ;
    } ;

    ScTimeVal operator-(const ScTimeVal& tv) const {
        ScTimeVal r = *this ;
        return r.dec(tv) ;
    } ;

    ScTimeVal operator*(const ScFloat m) const {
        ScTimeVal r = *this ;
        return r.mul(m) ;
    } ;


    ScTimeVal operator/(const ScFloat m) const {
        ScTimeVal r = *this ;
        return r.div(m) ;
    } ;

    ScFloat operator/(const ScTimeVal& tv) const {
        return fltSec() / tv.fltSec() ;
    } ;

    //--------------------------------------
    // Comparison

    ScBool operator> (const ScTimeVal& tv) const {
        return (sec() > tv.sec() || 
                (sec() == tv.sec() && usec() > tv.usec())) ;
    } ;
    ScBool operator>= (const ScTimeVal& tv) const {
        return (sec() > tv.sec() || 
                (sec() == tv.sec() && usec() >= tv.usec())) ;
    } ;
    ScBool operator< (const ScTimeVal& tv) const {
        return (sec() < tv.sec() || 
                (sec() == tv.sec() && usec() < tv.usec())) ;
    } ;
    ScBool operator<= (const ScTimeVal& tv) const {
        return (sec() < tv.sec() || 
                (sec() == tv.sec() && usec() <= tv.usec())) ;
    } ;
    ScBool operator== (const ScTimeVal& tv) const {
        return (sec() == tv.sec() && usec() == tv.usec()) ;
    } ;

    //--------------------------------------
    // output

    inline void describe(ostream& ostr, ScBool detailp = ScTrue, 
                         ScBool newlinep = ScTrue) 
        const {
        ostr << "#ScTimeVal[" << sec() << "s" << msec() << "ms" 
             << rusec() << "]"  ;
        if(newlinep) ostr << endl ;
    } ;
      
    friend ostream& operator<< (ostream& o, const ScTimeVal& tv) {
        tv.describe(o,ScFalse,ScFalse) ;
        return o ;
    } ;

} ;

//--------------------------------------------------
// random

#ifdef Solaris
extern "C" {
    long random();
  //    int srandom(unsigned seed);
    void srandom(unsigned seed);
}
#endif

inline void scRandomizeByTime() {
    ScTimeVal ctime ;
    ctime.getTimeOfDay() ;
    ScUInt seed = ctime.usec() + ctime.sec() ;
    srandom(seed) ;
} ;

inline ScInt scIntRand(ScInt max) { return (random() % max) ; } ;
inline ScInt scIntRand(ScInt min, ScInt max) { 
    return scIntRand(max - min) + min ; } ;

#define SC_FLOATRANDBASE (RAND_MAX/2) 
inline ScFloat scFltRand() { 
    return ((ScFloat)scIntRand(SC_FLOATRANDBASE))/(ScFloat)SC_FLOATRANDBASE ;
} ;
inline ScFloat scFltRand(ScFloat max) { return max * scFltRand() ; } ;
inline ScFloat scFltRand(ScFloat min, ScFloat max) {
    return scFltRand(max - min) + min ;
} ;

//--------------------------------------------------
// sleep

#ifdef THREADED
extern pthread_cond_t  ScNeverAwakeCV ;
extern pthread_mutex_t ScNeverAwakeMX ;
#endif

inline void scSleep(const ScUInt sec, 
                    const ScUInt msec = 0, 
                    const ScUInt usec = 0) {
#ifdef THREADED
    struct timeval ctime ;
    gettimeofday(&ctime,SC_NULLPTR) ;

    struct timespec etime ;
    ScUInt eusec = msec * 1000 + usec + ctime.tv_usec ;
    etime.tv_nsec = (eusec % 1000000) * 1000 ;
    etime.tv_sec = sec + ctime.tv_sec + (eusec / 1000000) ;

    pthread_mutex_lock(&ScNeverAwakeMX) ;
    pthread_cond_timedwait(&ScNeverAwakeCV,&ScNeverAwakeMX,&etime) ;
    pthread_mutex_unlock(&ScNeverAwakeMX) ;
#else
    sleep(sec) ;
    usleep(msec * 1000 + usec) ;
#endif
}

inline void scMSleep(const ScUInt msec) { scSleep(0,msec,0) ; }
inline void scUSleep(const ScUInt usec) { scSleep(0,0,usec) ; }
   
inline void scSleep(const ScTimeVal& tv) { scSleep(tv.sec(),0,tv.usec()) ; } 

//--------------------------------------------------
// mutex

class ScMutex {
public:
#ifdef THREADED
    pthread_mutex_t entity ;
#endif

    inline ScMutex() { init() ; } ;

    inline ~ScMutex() { destroy() ; } ;

    inline void init() {
#ifdef THREADED
        pthread_mutex_init(&entity,SC_NULLPTR) ;
#endif
    } ;

    inline void destroy() {
#ifdef THREADED
        pthread_mutex_destroy(&entity) ;
#endif
    } ;

    inline void lock() {
#ifdef THREADED
        pthread_mutex_lock(&entity) ;
#endif
    } ;

    inline void unlock() {
#ifdef THREADED
        pthread_mutex_unlock(&entity) ;
#endif
    } ;
} ;


//--------------------------------------------------
// my Cond Var

class ScCondVar {
public:
#ifdef THREADED
    pthread_cond_t entity ;
#else
    Int dummy ;
#endif

    inline ScCondVar() { init() ; } ;
    inline ~ScCondVar() { destroy() ; } ;

    inline void init() {
#ifdef THREADED
        pthread_cond_init(&entity,SC_NULLPTR) ;
#endif
    } ;

    inline void destroy() {
#ifdef THREADED
        pthread_cond_destroy(&entity) ;
#endif
    } ;

    inline void signal() {
#ifdef THREADED
        pthread_cond_signal(&entity) ;
#endif
    } ;

    inline void broadcast() {
#ifdef THREADED
        pthread_cond_broadcast(&entity) ;
#endif
    } ;

    inline ScInt wait(ScMutex& mutex) {
#ifdef THREADED
        return pthread_cond_wait(&entity,&(mutex.entity)) ;
#else
        return 0 ;
#endif
     
    } ;

    inline ScBool timedwait(ScMutex& mutex,
                            ScUInt sec, ScUInt msec = 0, ScUInt usec = 0) 
        {
#ifdef THREADED
            struct timeval ctime ;

            gettimeofday(&ctime,SC_NULLPTR) ;

            struct timespec etime ;

            ScUInt eusec = msec * 1000 + usec + ctime.tv_usec ;
            etime.tv_nsec = (eusec % 1000000) * 1000 ;
            etime.tv_sec = sec + ctime.tv_sec + (eusec / 1000000) ;

            ScInt r = pthread_cond_timedwait(&entity,&(mutex.entity),&etime) ;

            if(!r) return ScTrue ;
            else return ScFalse ;
#else
            return ScTrue ;
#endif
        } ;
};
     

//--------------------------------------------------
// ScMutexCondVar
//

class ScMutexCondVar {
public:
    ScMutex mutex ;
    ScCondVar cond ;

    inline ScMutexCondVar() : mutex(), cond() {} ;
    inline ~ScMutexCondVar() { destroy() ; } ;

    inline void init() { mutex.init() ; cond.init() ; } ;

    inline void destroy() {
        mutex.destroy() ;
        cond.destroy() ;
    } ;

    inline void lock() { mutex.lock() ; } ;

    inline void unlock() { mutex.unlock() ; } ;
      
    inline ScInt wait() { return cond.wait(mutex) ; } ;

    inline ScBool timedwait(ScUInt sec, ScUInt msec = 0, ScUInt usec = 0) {
        return cond.timedwait(mutex,sec,msec,usec) ;
    } ;

    inline ScBool timedlockwait(ScUInt sec, ScUInt msec = 0, 
                                ScUInt usec = 0) {
        mutex.lock() ;
        ScBool r = timedwait(sec,msec,usec) ;
        mutex.unlock() ;
        return r ;
    } ;

    inline void lockwait() {
        mutex.lock() ;
        cond.wait(mutex) ;
        mutex.unlock() ;
    } ;

    inline void signal() { cond.signal() ; } ;
      
    inline void broadcast() { cond.broadcast() ; } ;

} ;

//======================================================================
// Constants
//

//--------------------------------------------------
// Misc

#define SC_BUFSIZE      1024
#define SC_NAMEMAX      256
#define SC_COMMAX       16

#define SC_RAD2DEG  180.0 / M_PI
#define SC_DEG2RAD  M_PI / 180.0

#ifdef HAVE_MEMSET
#define bzero(a, b) memset(a, 0, b)
#endif

//======================================================================
// Utility Inline Functions
//

inline ScFloat scMax(const ScFloat& x, const ScFloat& y) { return (x>y)?x:y; }
inline ScFloat scMin(const ScFloat& x, const ScFloat& y) { return (x<y)?x:y; }
inline ScInt scMax(const ScInt& x, const ScInt& y) { return (x>y)?x:y ; } ;
inline ScInt scMin(const ScInt& x, const ScInt& y) { return (x<y)?x:y ; } ;

inline ScFloat scAbs(const ScFloat& x) {  return ((x > 0.0)?x:(-x)) ; } ;
inline ScInt scAbs(const ScInt& x) {  return ((x > 0)?x:(-x)) ; } ;

inline ScFloat scSq(const ScFloat& x) { return x * x ; } ;

inline ScFloat scRad2Deg(const ScFloat& x) { return x * SC_RAD2DEG ; } ;
inline ScFloat scDeg2Rad(const ScFloat& x) { return x * SC_DEG2RAD ; } ;

inline ScFloat scSin(const ScFloat& x) { return sin(scDeg2Rad(x)) ; } ;
inline ScFloat scCos(const ScFloat& x) { return cos(scDeg2Rad(x)) ; } ;
inline ScFloat scTan(const ScFloat& x) { return tan(scDeg2Rad(x)) ; } ;

inline ScFloat scASin(const ScFloat& x) { return scRad2Deg(asin(x)) ; } ;
inline ScFloat scACos(const ScFloat& x) { return scRad2Deg(acos(x)) ; } ;

inline ScFloat scATan(const ScFloat& y) { return scRad2Deg(atan(y)) ; } ;
inline ScFloat scATan(const ScFloat& y, const ScFloat& x) {
    if (x == 0 && y == 0) return 0;  // 1.04o
    return scRad2Deg(atan2(y,x));
} ;


inline ScFloat scNormalizeAngle(ScFloat a) {
    while(a >  180.0) { a -= 360.0 ; } 
    while(a < -180.0) { a += 360.0 ; }
    return a ;
} ;


//======================================================================
// Socket 
//

class ScSocket {
public:
    char* _host ;
    ScInt _port ;
    int socketfd ;
    struct sockaddr_in serv_addr ;
    ScBool blockp ;
    int recv_errno ;
    int send_errno ;
    ScMutex sendMutex ;
    ScMutex recvMutex ;

    //--------------------------------------------------
    // constructor

    ScSocket() { _host = SC_NULLPTR ; _port = 0 ; } ;
    ScSocket(char* host, ScInt port, ScBool nonblockp = ScTrue) {
        if(!init(host,port,nonblockp)) {
            cerr << "Can not create ScSocket(" << host 
                 << ":" << port << ",nonblock=" << nonblockp
                 << ")" << endl ;
            exit(1) ;
        }
    } ;
      
    //--------------------------------------------------
    // init

    ScBool init(char* host, ScInt port, ScBool nonblockp = ScTrue) ;

    //--------------------------------------------------
    // blocking mode

    ScBool checkBlockingMode() const ;

    ScBool setBlockingMode(const ScBool& blockingp) ;

    //--------------------------------------------------
    // send & recv

    ScBool send(const ScString& buf) ;
    ScBool send(char* buf) { return send(ScString(buf)) ; } ;

    ScBool recv(ScString& buf, ScBool adjustportp = ScTrue) ;
    ScBool recv(char* buf, ScUInt size, ScBool adjustportp = ScTrue) {
        ScString b = ScString(buf,size) ;
        return recv(b,adjustportp) ;
    } ;

    //--------------------------------------------------
    // close

    void close() ;

    //--------------------------------------------------
    // pipe

//      ScBool pipe(streambuf* sin, streambuf* sout) ;   
//      /* = pipe(FILE* fpin, FILE* fpout) */

//      ScBool pipe(istream& istr, ostream& ostr) {
//   return pipe(istr.rdbuf(),ostr.rdbuf()) ;
//      } ;
    ScBool pipe(istream& istr, ostream& ostr) ;

    //--------------------------------------------------
    // describe

    void describe(ostream& o, ScBool detailp = ScTrue,
                  ScBool newlinep = ScTrue) const {
        o << "#ScSocket[" << socketfd << "| host=" 
          << _host << ":" << _port <<"/udp" << "]" ;
        if(newlinep) o << endl ;
    } ;
    friend ostream& operator<< (ostream& o, const ScSocket& sock) {
        sock.describe(o,ScFalse,ScFalse) ;
        return o ;
    } ;

} ;

//======================================================================
// Count & Time
//

class ScCycleStep {
public:
    ScUInt count ;
    ScTimeVal rtime;

    //--------------------------------------------------
    // constructor

    ScCycleStep(ScUInt c = 0) { init(c) ; } ;
    ScCycleStep(const ScCycleStep& src) {
        set(src) ;
    } ;

    //--------------------------------------------------
    // init

    void init(ScUInt c = 0) {
        set(c) ;
    } ;

    //--------------------------------------------------
    // update rtime

    void update() { rtime.update() ; } ;

    //--------------------------------------------------
    // set

    ScCycleStep& set(ScUInt c) {
        count = c ;
        update() ;
        return *this ;
    } ;
    ScCycleStep& set(const ScCycleStep& src) {
        count = src.count ;
        rtime = src.rtime ;
        return *this ;
    } ;

    ScCycleStep& operator=(const ScCycleStep& src) { return set(src) ; } ;

    ScCycleStep& setupdate(const ScCycleStep& src) {
        return set(src.count) ;
    } ;

    ScBool operator==(const ScCycleStep& ref) { 
        return count == ref.count ; } ;

    //--------------------------------------------------
    // describe
    void describe(ostream& o, ScBool detailp = ScTrue, 
                  ScBool newlinep = ScTrue) const {
        o << "#ScCycleStep[" << count << ":" ;
        rtime.describe(o,detailp,newlinep) ;
        o << "]" ;
        if(newlinep) o << endl ;
    } ;

    friend ostream& operator<<(ostream& o, const ScCycleStep& c) {
        c.describe(o,ScFalse,ScFalse) ;
        return o ;
    } ;
} ;


//======================================================================
// XY Position
//

class ScPos {
public:
    ScFloat x ;
    ScFloat y ;

    // 1.04o const correctness added
    ScPos operator -() { return ScPos(-x, -y); }
    ScPos operator +(const ScPos &a) const { return ScPos(x + a.x, y + a.y); }
    ScPos operator -(const ScPos &a) const { return ScPos(x - a.x, y - a.y); }
    ScPos operator *(const ScFloat &a) const { return ScPos(x * a, y * a); }
    ScPos operator *(const ScPos &a) const { return ScPos(x * a.x, y * a.y); }
    ScPos operator /(const ScFloat &a) const { return ScPos(x / a, y / a); }
    ScPos operator /(const ScPos &a) const { return ScPos(x / a.x, y / a.y); }
    void operator =(const ScFloat &a) { x = a; y = a; }
    void operator +=(const ScPos &a) { x += a.x;  y += a.y; }
    void operator +=(const ScFloat &a) { x += a;  y += a; }
    void operator -=(const ScPos &a) { x -= a.x;  y -= a.y; }
    void operator -=(const ScFloat &a) { x -= a;  y -= a; }
    void operator *=(const ScFloat &a) { x *= a;  y *= a; }
    void operator /=(const ScFloat &a) { x /= a;  y /= a; }
    operator !=(const ScPos &a) const { return (x != a.x) || (y != a.y); } 
    operator !=(const ScFloat &a) const { return (x != a) || (y != a); }
    operator ==(const ScPos &a) const { return (x == a.x) && (y == a.y); } 


    //--------------------------------------------------
    // constructor

    ScPos(const ScFloat xi = 0.0, const ScFloat yi = 0.0) { 
        x = xi; y = yi ;
    } ;

    //--------------------------------------------------
    // set value

    ScPos& set(const ScFloat sx, const ScFloat sy) { 
        x = sx ; y = sy ; 
        return *this ;
    } ;

    ScPos& set(const ScPos& src) { return set(src.x,src.y) ; } ;

    //--------------------------------------------------
    // converter

    ScFloat norm() const { return sqrt(scSq(x) + scSq(y)) ;} ;
    ScFloat angle() const { return atan2(y,x) ; } ;
    ScFloat Angle() const { return scNormalizeAngle(scATan(y,x)) ; } ;
    ScFloat Angle(const ScFloat& refdir) const {
        return scNormalizeAngle(Angle() - refdir) ;
    } ;

    ScFloat AngleTo(const ScPos& target) const {
        return relativePosTo(target).Angle() ;
    } ;
    ScFloat AngleTo(const ScPos& target, const ScFloat& refdir) const {
        return relativePosTo(target).Angle(refdir) ;
    } ;

    ScFloat AngleFrom(const ScPos& target) const {
        return relativePosFrom(target).Angle() ;
    } ;
    ScFloat AngleFrom(const ScPos& target, const ScFloat& refdir) const {
        return relativePosFrom(target).Angle(refdir) ;
    } ;

    //--------------------------------------------------
    // basic calculations

    ScPos negative() const { return ScPos(-x,-y) ; } ;
    ScPos multiple(ScFloat a) const { return ScPos(a * x, a * y) ; } ;

    ScPos relativePosFrom(const ScPos& org) const {
        return ScPos(x - org.x, y - org.y) ;
    } ;
    ScPos relativePosTo(const ScPos& target) const {
        return ScPos(target.x - x, target.y - y) ;
    } ;

    ScPos rotate(ScFloat angle) const {  /* radian */
        ScPos r ;
        r.x = x * cos(angle) - y * sin(angle) ;
        r.y = x * sin(angle) + y * cos(angle) ;
        return r ;
    } ;

    ScPos Rotate(ScFloat angle) const {  /* degree */
        ScPos r ;
        r.x = x * scCos(angle) - y * scSin(angle) ;
        r.y = x * scSin(angle) + y * scCos(angle) ;
        return r ;
    } ;

    //--------------------------------------------------
    // unit Vector

    ScPos unit(ScFloat len = 1.0) const {
        ScFloat l = norm() ;
        if(l <= 0.0) l = len ;
        return multiple(len/l) ; 
    } ;

    ScPos normalize(ScFloat len = 1.0) {
        return *this = unit(len) ;
    } ;

    //--------------------------------------------------
    // describe

    void describe(ostream& o, ScBool detailp = ScTrue,
                  ScBool newlinep = ScTrue) const {
        o << "#ScPos(" << x << "," << y << ")" ;
        if(newlinep) o << endl ;
    } ;
    friend ostream& operator<< (ostream& o, const ScPos& p) {
        p.describe(o,ScFalse,ScFalse) ;
        return o ;
    } ;

} ;

inline ScPos Polar2ScPos(ScFloat radius, ScFloat ang) {
    return ScPos(radius * scCos(ang), radius * scSin(ang));
}

inline ScPos scPosDiff(const ScPos& from, const ScPos& to) {
    return to.relativePosFrom(from) ;
}

inline ScFloat scDistance(const ScPos& from, const ScPos& to) {
    return scPosDiff(from,to).norm() ;
} 

inline ScFloat scInterProduct(const ScPos& p, const ScPos& q) {
    return p.x * q.x + p.y * q.y ; 
} 

//--------------------------------------------------
// XY Position + Angle

class ScPosDir {
public:
    ScPos pos ;
    ScFloat bdir ;    // body direction
    ScFloat hdir ;    // head direction

    //--------------------------------------------------
    // describe

    void describe(ostream& o, ScBool detailp = ScTrue,
                  ScBool newlinep = ScTrue) const {
        o << "#ScPosDir[" << pos << "," << hdir << "/"<< bdir << "]" ;
        if(newlinep) o << endl ;
    } ;
    friend ostream& operator<< (ostream& o, const ScPosDir& pd) {
        pd.describe(o,ScFalse,ScFalse) ;
        return o ;
    } ;

} ;

//======================================================================
// Field Parameters
//

//--------------------------------------------------
// Number of Objects

#define SC_N_PLAYER_PER_TEAM    11
#define SC_N_PLAYER     (SC_N_PLAYER_PER_TEAM * 2) 
#define SC_N_LINE       4 
#define SC_N_GOAL       2 
#define SC_N_FLAG       (6 + 6 + 4 + 22 + 14 + 1) 
/* corner, penalty, goal post, touchline, goalline, center mark */

#define SC_SEENOBJMAX   (1 + SC_N_PLAYER + SC_N_LINE + SC_N_GOAL + SC_N_FLAG)

//--------------------------------------------------
// Field Param

#define SC_PITCH_LENGTH         105.0 
#define SC_PITCH_WIDTH          68.0 
#define SC_PENALTY_AREA_LENGTH  16.5
#define SC_PENALTY_AREA_WIDTH   40.32 
#define SC_GOAL_WIDTH           14.02

//--------------------------------------------------
// Player Param

#define SC_PLAYER_SIZE          0.3
#define SC_PLAYER_DECAY         0.4
#define SC_PLAYER_RAND          0.1
#define SC_PLAYER_WEIGHT        60.0
#define SC_PLAYER_SPEED_MAX     1.0

#define SC_STAMINA_MAX          3500.0
#define SC_STAMINA_INC_MAX      20.0

#define SC_RECOVER_DEC_THR      0.3
#define SC_RECOVER_DEC          0.002
#define SC_RECOVER_MIN          0.5

#define SC_EFFORT_DEC_THR       0.3
#define SC_EFFORT_DEC           0.005
#define SC_EFFORT_INC_THR       0.6
#define SC_EFFORT_INC           0.01
#define SC_EFFORT_MIN           0.6

#define SC_HEAR_MAX             2
#define SC_HEAR_INC             1
#define SC_HEAR_DECAY           2

#define SC_INERTIA_MOMENT       5.0

#define SC_SENSE_BODY_STEP      100

// 1.04o (renamed cacheable -> catchable
#define SC_CATCHABLE_AREA_L     2.0
#define SC_CATCHABLE_AREA_W     1.0
#define SC_CATCH_PROBABILITY    1.0
#define SC_CATCH_BAN_CYCLE      5

#define SC_BALL_SIZE            0.085
#define SC_BALL_DECAY           0.94
#define SC_BALL_RAND            0.05
#define SC_BALL_WEIGHT          0.2
#define SC_BALL_SPEED_MAX       2.7

#define SC_WIND_FORCE           0.0
#define SC_WIND_DIR             0.0
#define SC_WIND_RAND            0.0

#define SC_KICKABLE_MARGIN      0.7
#define SC_KICKABLE_AREA (SC_BALL_SIZE + SC_PLAYER_SIZE + SC_KICKABLE_MARGIN)
#define SC_CKICK_MARGIN         1.0

#define SC_DASH_POWER_RATE      0.006
#define SC_KICK_POWER_RATE      0.016

#define SC_MAXPOWER             100.0
#define SC_MINPOWER             (-100.0)

#define SC_MAXMOMENT            180.0
#define SC_MINMOMENT            (-180.0)





//--------------------------------------------------
// Simulation Param

#define SC_DEFAULT_BODY_COMMAND_TIC_MSEC    100
#define SC_ESTIMATE_TIC_ALPHA           0.7
#define SC_ESTIMATE_TIC_CUTOFF          1.5

#define SC_DEFAULT_DASH_POW_POS_EFFECT  0.02
#define SC_DASH_POWER_RATE 0.006
#define SC_KICK_POWER_RATE 0.016


//--------------------------------------------------
// Side

enum ScSide {
    SC_S_unknown,
    SC_S_left,
    SC_S_right
} ;

inline ScBool scIsLeftSide(const ScSide side){ return (side == SC_S_left);};
inline ScBool scIsRightSide(const ScSide side){ return (side == SC_S_right);};

inline ScInt scSide2Sign(const ScSide side) {
    if(scIsLeftSide(side)) return 1 ;
    if(scIsRightSide(side)) return -1 ;
    return 0 ;
} ;

inline ScSide scChar2Side(const char c) {
    if(c == 'l') return SC_S_left ;
    if(c == 'r') return SC_S_right ;
    return SC_S_unknown ;
} ;

inline char scSide2Char(const ScSide side) {
    if      (scIsLeftSide(side))  return 'l' ;
    else if (scIsRightSide(side)) return 'r' ;
    else                          return '?' ;
} ;


//--------------------------------------------------
// Object type

enum ScObjectType {
    SC_OT_ball,
    SC_OT_player,
    SC_OT_goal,
    SC_OT_flag,
    SC_OT_line
} ;

//--------------------------------------------------
// play mode, flag

#include <vocabulary.h>
extern ScWord ScWordTable[] ;

extern ScWord* scString2Word(const ScSubString& msg) ;
 
#include <flagtable.h>
extern ScLandmark ScLandmarkTable[] ;

//--------------------------------------------------
// Line Info

enum ScLineID {
    SC_LID_l = SC_LMID_LineL,
    SC_LID_t = SC_LMID_LineT,
    SC_LID_b = SC_LMID_LineB,
    SC_LID_r = SC_LMID_LineR,
} ;

//--------------------------------------------------
// view mode

enum ScViewWidth {
    SC_VW_narrow = SC_WID_Narrow,
    SC_VW_normal = SC_WID_Normal,
    SC_VW_wide = SC_WID_Wide
} ;

#define SC_VW_NAME(wmode) ScWordTable[wmode].name

enum ScViewQuality {
    SC_VQ_high = SC_WID_High,
    SC_VQ_low = SC_WID_Low
} ;

#define SC_VQ_NAME(qmode) ScWordTable[qmode].name


//--------------------------------------------------
// player ID

enum ScWhichTeam {
    SC_WT_opponent = -1,
    SC_WT_unknown = 0,
    SC_WT_our = 1,
    SC_WT_other = 2   // used for misc purpose
} ;

const ScUInt SC_UNUM_unknown = 0 ;

class ScPlayerID {
public:
    ScWhichTeam team ;
    ScUInt unum ;

    //--------------------------------------------------
    // check

    ScBool isTeammate() const { return team == SC_WT_our ; } ;
    ScBool isTeammate(ScUInt un) const { 
        return isTeammate() && unum == un ; } ;
    ScBool isOpponent() const { return team == SC_WT_opponent ; } ;
    ScBool isOpponent(ScUInt un) const { 
        return isOpponent() && unum == un ; } ;
    ScBool isUnknownTeam() const { return team == SC_WT_unknown ; } ;
    ScBool isUnknownUNum() const { return unum == SC_UNUM_unknown ; } ;
    ScBool isUnknown() const { return isUnknownTeam() || isUnknownUNum();};

    //--------------------------------------------------
    // describe

    void describe(ostream& o, ScBool detailp = ScTrue,
                  ScBool newlinep = ScTrue) const {
        o << "#ScPlayerID[" ;
        if      (team == SC_WT_opponent) o << "-" << unum ;
        else if (team == SC_WT_our)      o << "+" << unum ;
        else                             o << "?" ;
        o << "]" ;
        if(newlinep) o << endl ;
    } ;
    friend ostream& operator<< (ostream& o, const ScPlayerID& pid) {
        pid.describe(o,ScFalse,ScFalse) ;
        return o ;
    } ;

} ;

//======================================================================
// information fron the server
//

//--------------------------------------------------
// init info

class ScInitInfo {
public:
    ScBool reconnectp ;
    ScSide side ;
    ScInt unum ;
    ScWord* playmode ;

    //--------------------------------------------------
    // scan

    ScBool scan(ScString& buffer) ;
      
    //--------------------------------------------------
    // describe

    void describe(ostream& o, ScBool detailp = ScTrue,
                  ScBool newlinep = ScTrue) const {
        o << "#ScInitInfo[side=" << scSide2Char(side) 
          << ",unum=" << unum 
          << ",playmode=" << playmode->name << "]" ;
        if(newlinep) o << endl ;
    } ;
    friend ostream& operator<< (ostream& o, const ScInitInfo& iinfo) {
        iinfo.describe(o,ScFalse,ScFalse) ;
        return o ;
    } ;



} ;

//--------------------------------------------------
// seen obj info

enum ScSeenQuality {
    SC_SQ_none,
    SC_SQ_dist,
    SC_SQ_dir,
    SC_SQ_dist_chng,
    SC_SQ_dir_chng,
    SC_SQ_body_dir,
    SC_SQ_head_dir
} ;

class ScSeenObj {
public:
    ScObjectType type ;
    ScBool inViewConeP ;
    union {
        ScPlayerID player ;
        ScLandmark* flag ;
        ScLineID line ;
    } id ;
    ScSeenQuality quality ;
    ScFloat dist ;
    ScFloat dir ;
    ScFloat dist_chng ;
    ScFloat dir_chng ;
    ScFloat body_dir ;
    ScFloat head_dir ;

    //--------------------------------------------------
    // type check
      
    ScBool isBall() const { return type == SC_OT_ball ; } ;
    ScBool isPlayer() const { return type == SC_OT_player ; } ;
    ScBool isGoal() const { return type == SC_OT_goal ; } ;
    ScBool isFlag() const { return type == SC_OT_flag ; } ;
    ScBool isLine() const { return type == SC_OT_line ; } ;

    //--------------------------------------------------
    // seen quality

    ScBool know_dist() const { return quality >= SC_SQ_dist ; }
    ScBool know_dir() const { return quality >= SC_SQ_dir ; }
    ScBool know_dist_chng() const { return quality >= SC_SQ_dist_chng ; }
    ScBool know_dir_chng() const { return quality >= SC_SQ_dir_chng ; }
    ScBool know_body_dir() const { return quality >= SC_SQ_body_dir ; }
    ScBool know_head_dir() const { return quality >= SC_SQ_head_dir ; }

    //--------------------------------------------------
    // scan

    ScBool scan(ScString& buffer, const ScString& teamname) ;
    ScBool scanObjName(ScString& buffer, const ScString& teamname) ;

    //--------------------------------------------------
    // relative to absolute position

    ScBool rel2absPos(const ScPosDir& origin, ScPosDir& abspos) const ;

    ScPosDir rel2absPos(const ScPosDir& origin) const {
        ScPosDir posdir ;
        rel2absPos(origin,posdir) ;
        return posdir ;
    } ;

    //--------------------------------------------------
    // relative to absolute velocity

    ScBool rel2absVel(const ScPosDir& orgpos, 
                      const ScPos& orgvel,
                      ScPos& absvel) const ;

    ScPos rel2absVel(const ScPosDir& orgpos, 
                     const ScPos& orgvel) const {
        ScPos absvel ;
        rel2absVel(orgpos,orgvel,absvel) ;
        return absvel ;
    } ;

    //--------------------------------------------------
    // describe

    void describe(ostream& o, ScBool detailp = ScTrue,
                  ScBool newlinep = ScTrue) const {
        o << "#ScSeenObj[" ;
        if (!inViewConeP) o << "?" ;
        if        (type == SC_OT_ball) {
            o << "ball" ;
        } else if (type == SC_OT_player) {
            o << id.player ;
        } else if (type == SC_OT_goal) {
            o << "goal_" << id.flag->namestr ;
        } else if (type == SC_OT_flag) {
            o << "flag(" << id.flag->namestr << ")" ;
        } else if (type == SC_OT_line) {
            o << "line_" ;
            switch(id.line) { 
                case SC_LID_l : o << "l" ; break ;
                case SC_LID_r : o << "r" ; break ;
                case SC_LID_t : o << "t" ; break ;
                case SC_LID_b : o << "b" ; break ;
                default :       o << "?" ; break ;
            }
        }

        o << ":dst=" ;
        if(know_dist()) { o << dist ; } else { o << "?" ; }
        if(know_dist_chng()) { o << "(" << dist_chng  << ")" ; }

        o << ",dir=" ;
        if(know_dir()) { o << dir ; } else { o << "?" ; }
        if(know_dir_chng()) { o << "(" << dir_chng  << ")" ; }

        if(type == SC_OT_player && know_body_dir() && know_head_dir()) {
            o << ",ang=(B:" << body_dir << ",H:" << head_dir << ")" ;
        }

        o << "]" ;
        if(newlinep) o << endl ;
    } ;

    friend ostream& operator<< (ostream& o, const ScSeenObj& sobj) {
        sobj.describe(o,ScFalse,ScFalse) ;
        return o ;
    } ;

} ;

//--------------------------------------------------
// see info

class ScSeeInfo {
public:
    ScInt time ;
    ScInt n ;
    ScSeenObj obj[SC_SEENOBJMAX] ;

    //--------------------------------------------------
    // scan

    ScBool scan(ScString& buffer, const ScString& teamname) ;

    //--------------------------------------------------
    // describe

    void describe(ostream& o, ScBool detailp = ScTrue,
                  ScBool newlinep = ScTrue) const {
        o << "#ScSeeInfo[t=" << time << ", n=" << n << "]" ;
        if(detailp) {
            if(newlinep) o << endl ;
            for(ScInt i = 0 ; i <  n ; i++) {
                o << "\t" ; 
                obj[i].describe(o,detailp,newlinep) ;
            }
        }
    } ;
    friend ostream& operator<< (ostream& o, const ScSeeInfo& sinfo) {
        sinfo.describe(o,ScFalse,ScFalse) ;
        return o ;
    } ;

} ;

//--------------------------------------------------
// hear info

enum ScSenderType {
    SC_ST_self = SC_WID_Self,
    SC_ST_referee = SC_WID_Referee,
    SC_ST_player = SC_WID_Player,
    SC_ST_onlineCoachL = SC_WID_OnlineCoachL,
    SC_ST_onlineCoachR = SC_WID_OnlineCoachR
} ;

#define SC_SENDERNAME_SELF  SC_WORDSTR(Self)
#define SC_SENDERNAME_REFEREE   SC_WORDSTR(Referee)
#define SC_SENDERNAME_PLAYER    SC_WORDSTR(Player)
#define SC_SENDERNAME_OLCOACHL	SC_WORDSTR(OnlineCoachL)
#define SC_SENDERNAME_OLCOACHR	SC_WORDSTR(OnlineCoachR)

inline const char* scSenderName(const ScSenderType& type) {
    switch(type) {
        case SC_ST_self :    return SC_SENDERNAME_SELF ; 
        case SC_ST_referee : return SC_SENDERNAME_REFEREE ; 
        case SC_ST_player :  return SC_SENDERNAME_PLAYER ;
        case SC_ST_onlineCoachL :  return SC_SENDERNAME_OLCOACHL ;
        case SC_ST_onlineCoachR :  return SC_SENDERNAME_OLCOACHR ;
        default : return SC_WORDSTR(Error) ;
    }
} ;


class ScHearInfo {
public:
    ScInt time ;
    ScSenderType sender ;
    union {
        ScFloat direction ;
        ScWord* pmode ;
    } aux ;
    char _dummy_[SC_BUFSIZE] ;

    //--------------------------------------------------
    // scan

    ScString message() const { 
        return ScString((char *)_dummy_,SC_BUFSIZE) ; 
    } ;

    //--------------------------------------------------
    // scan

    ScBool scan(ScString& buffer) ;

    //--------------------------------------------------
    // describe

    void describe(ostream& o, ScBool detailp = ScTrue,
                  ScBool newlinep = ScTrue) const {
        o << "#ScHearInfo[t=" << time << ", from=" 
          << scSenderName(sender) ;
        if(sender == SC_ST_player) 
            o << "(" << aux.direction << ")" ; 
        o << ",msg=(" << (char*)message() << ")" << "]" ;
        if(newlinep) o << endl ;
    } ;
    friend ostream& operator<< (ostream& o, const ScHearInfo& hinfo) {
        hinfo.describe(o,ScFalse,ScFalse) ;
        return o ;
    } ;

} ;

//--------------------------------------------------
// sense body

class ScSenseBodyInfo {
public:
    ScInt time ;
    ScViewQuality viewQuality ;
    ScViewWidth viewWidth ;
    ScFloat stamina ;
    ScFloat effort ;
    ScFloat speed ;
    ScFloat headAngle ;
    ScUInt countKick ;
    ScUInt countDash ;
    ScUInt countTurn ;
    ScUInt countSay ;
    ScUInt countCatch ;
    ScUInt countTurnNeck ;
    ScUInt countChangeView ;
    ScUInt countSenseBody ;

    //--------------------------------------------------
    // reset

    void reset() {
        time = 0 ; 
        viewQuality = SC_VQ_high ; viewWidth = SC_VW_normal ;
        stamina = 0.0 ; effort = 1.0 ; 
        speed = 0.0 ; headAngle = 0.0 ;
        countKick = 0 ; countDash = 0 ; 
        countTurn = 0 ; countCatch = 0 ; 
        countTurnNeck = 0 ; countSay = 0 ;
        countChangeView = 0 ;countSenseBody = 0 ;
    } ;

    //--------------------------------------------------
    // set

    void set(const ScSenseBodyInfo& src) {
        time = src.time ; 
        viewQuality = src.viewQuality ; viewWidth = src.viewWidth ;
        stamina = src.stamina ; effort = src.effort ; 
        speed = src.speed ; headAngle = src.headAngle ;
        countKick = src.countKick ; countDash = src.countDash ; 
        countTurn = src.countTurn ; countCatch = src.countCatch ; 
        countTurnNeck = src.countTurnNeck ; countSay = src.countSay ;
        countChangeView = src.countChangeView ; 
        countSenseBody = src.countSenseBody ;
    } ;

    //--------------------------------------------------
    // scan

    ScBool scan(ScString& buffer) ;

    //--------------------------------------------------
    // describe

    void describe(ostream& o, ScBool detailp = ScTrue,
                  ScBool newlinep = ScTrue) const {
        o << "#ScSenseBodyInfo[t=" << time ;

        o << ":view=(" ; 
        switch(viewQuality) {
            case SC_VQ_high: o << "high" ; break ;
            case SC_VQ_low:  o << "low" ; break ;
        }
        o << "," ;
        switch(viewWidth) {
            case SC_VW_wide: o << "wide" ; break ;
            case SC_VW_normal: o << "normal" ; break ;
            case SC_VW_narrow: o << "narrow" ; break ;
        }
        o << "), stamina=" << stamina ;
        o << ", effort=" << effort ;
        o << ", speed=" << speed ;
        o << ", head=" << headAngle ;
        o << ", kick=" << countKick ;
        o << ", dash=" << countDash ;
        o << ", turn=" << countTurn ;
        o << ", say=" << countSay ;
        o << ", catch=" << countCatch ;
        o << ", turn_neck=" << countTurnNeck ;
        o << ", change_view=" << countChangeView ;
        o << ", sense_body=" << countSenseBody ;
        o << "]" ;
        if(newlinep) o << endl ;
    } ;
    friend ostream& operator<< (ostream& o, const ScSenseBodyInfo& sbinfo) {
        sbinfo.describe(o,ScFalse,ScFalse) ;
        return o ;
    } ;

} ;

//--------------------------------------------------
// all sensor info

enum ScSensorInfoType {
    SC_SIT_see = SC_WID_See,
    SC_SIT_hear = SC_WID_Hear,
    SC_SIT_senseBody = SC_WID_SenseBody,
    SC_SIT_error = SC_WID_Error
} ;

class ScSensorInfo {
public:
    ScSensorInfoType type ;
    union {
        ScSeeInfo see ;
        ScHearInfo hear ;
        ScSenseBodyInfo senseBody ;
    } info ;
    ScString teamname ;

    //--------------------------------------------------
    // constructor

    ScSensorInfo(char* team = SC_NULLPTR) { 
        teamname.set(team) ; 
    } ;

    //--------------------------------------------------
    // set teamname
    void setTeamname(char* team) { teamname.set(team) ; } ;
     
    //--------------------------------------------------
    // scan

    ScBool scan(ScString& buffer, ScBool rewindep) ;

    //--------------------------------------------------
    // describe

    void describe(ostream& o, ScBool detailp = ScTrue,
                  ScBool newlinep = ScTrue) const {
        o << "#ScSensorInfo[" ;
        switch(type) {
            case SC_SIT_see: info.see.describe(o,detailp,newlinep) ; break ;
            case SC_SIT_hear: info.hear.describe(o,detailp,newlinep) ; break ;
            case SC_SIT_senseBody: 
                info.senseBody.describe(o,detailp,newlinep) ; break ;
            default: break;
        }
        o << "]" ;
        if(newlinep) o << endl ;
    } ;
    friend ostream& operator<< (ostream& o, const ScSensorInfo& info) {
        info.describe(o,ScFalse,ScFalse) ;
        return o ;
    } ;

} ;

//======================================================================
// Position Estimation
//

inline void scOrientationAbs2Team(ScPos& pos, const ScSide side) {
    if(scIsRightSide(side)) { pos.x = -pos.x ; pos.y = -pos.y ; } 
} 
      
inline void scOrientationAbs2Team(ScPosDir& posdir, const ScSide side) {
    if(scIsRightSide(side)) { 
        posdir.pos.x = -posdir.pos.x ; posdir.pos.y = -posdir.pos.y ; 
        posdir.hdir += ((posdir.hdir > 0.0) ? -180.0 : 180.0) ;
        posdir.bdir += ((posdir.bdir > 0.0) ? -180.0 : 180.0) ;
    } 
}
   
inline void scOrientationAbs2Team(ScPosDir& posdir, ScPos& vel, 
                                  const ScSide side) {
    if(scIsRightSide(side)) {
        posdir.pos.x = -posdir.pos.x ; posdir.pos.y = -posdir.pos.y ; 
        posdir.hdir += ((posdir.hdir > 0.0) ? -180.0 : 180.0) ;
        posdir.bdir += ((posdir.bdir > 0.0) ? -180.0 : 180.0) ;
        vel.x = -vel.x ; vel.y = -vel.y ; 
    } 
}

//V.1.05noda
extern ScBool scEstimatePosDirVelAbs1(ScPosDir& posdir, ScPos& vel,
				      ScBool& outOfFieldP,
                                      const ScSeeInfo& sinfo) ;
//V.1.05noda
extern ScBool scEstimatePosDirVelTeam1(ScPosDir& posdir, ScPos& vel,
				       ScBool& outOfFieldP,
                                       const ScSeeInfo& sinfo, 
                                       const ScSide side) ;
                
//V.1.05noda
extern ScBool scEstimatePosDirVelAbs2(ScPosDir& posdir, ScFloat& posConf,
                                      ScPos& vel, ScFloat& velConf,
				      ScBool& outOfFieldP,
                                      const ScSeeInfo& sinfo) ;
//V.1.05noda
extern ScBool scEstimatePosDirVelTeam2(ScPosDir& posdir, ScFloat& posConf,
                                       ScPos& vel, ScFloat& velConf,
				       ScBool& outOfFieldP,
                                       const ScSeeInfo& sinfo, 
                                       const ScSide side) ;
                
//======================================================================
// Player Class
//

class ScPlayer {
public:
    ScString teamname ;
    ScSide side ;
    ScUInt unum ;

    ScSocket socket ;
    ScString sendBuffer ;
    ScString recvBuffer ;

    ScInitInfo iinfo ;
    ScSensorInfo sinfo ;

    ScWord* playmode ;

    ScSenseBodyInfo estimateBodyState ;
    ScSenseBodyInfo actualBodyState ;

    ScTimeVal rtime ;
    ScCycleStep scycle ;  /* sserver time */
    ScCycleStep vcycle ;  /* most recent visual infomation time */
    ScCycleStep acycle ;  /* most recent auditorial informaiton time */
    ScCycleStep bcycle ;  /* most recent sense_body time */
    ScCycleStep fcycle ;  /* most recent foot command cycle */
    ScTimeVal simulationTic ;  /* estimated simulation tic by sense body */

    ScMutex   eposMutex ;
    ScPosDir eposdir ;    /* current pos dir estimated by visual info */
    ScFloat eposConf ;    /* confident factor of eposdir */
    ScPos evel ;      /* current velocity estimated by visual info */
    ScFloat evelConf ;    /* confident factor of evel */
    ScCycleStep ecycle ;  /* time to estimate pos */
//V.1.05noda
    ScBool outOfFieldP ; /* player is inside of field or not */

    ScMutex   xposMutex ;
    ScPosDir xposdir ;    /* expected pos dir */
    ScPos xvel ;      /* expected velocity */
    ScCycleStep xcycle ;  /* expected time to estimate pos */
    ScFloat dashPowerEffect ;

    //--------------------------------------------------
    // constructor

    ScPlayer() { 
        init(ScString(), SC_NULLPTR, 0) ;
    } ;
    ScPlayer(ScString team, char* host, ScUInt port, 
             ScBool nonblockp = ScTrue) {
        init(team,host,port,nonblockp) ;
    } ;

    virtual ~ScPlayer();

    //--------------------------------------------------
    // init

    ScBool init(ScString team, char* host, ScUInt port, 
                ScBool nonblockp = ScTrue) {
        teamname = ScString(team) ;
        sinfo.teamname = teamname ;

        estimateBodyState.reset() ;
        actualBodyState.reset() ;

        scycle.update() ; vcycle.update() ; acycle.update() ; 
        bcycle.update() ; fcycle.update() ; 
        simulationTic.set(0,SC_DEFAULT_BODY_COMMAND_TIC_MSEC,0) ;

        dashPowerEffect = SC_DASH_POWER_RATE ;

        if (!setupBuffer()) return ScFalse ;
        if (!scIsNullPtr(host)) {
            return setupSocket(host,port,nonblockp) ;
        } else {
            return ScTrue ;
        }
    } ;

    ScBool setupSocket(char* host, ScUInt port, ScBool nonblockp = ScTrue) {
        return socket.init(host,port,nonblockp) ;
    } ;

    ScBool setupBuffer(ScUInt bufsize = SC_BUFSIZE) {
        return sendBuffer.alloc(bufsize) && recvBuffer.alloc(bufsize) ;
    } ;

    //--------------------------------------------------
    // adjust body state

    void adjustSimulationTic() {
        ScFloat dcount = (ScFloat)(scycle.count - bcycle.count) ;
        if(dcount <= 1) return ;

        ScFloat oldtic = simulationTic.fltSec() ;
        ScFloat newtic = (scycle.rtime - bcycle.rtime).fltSec()/dcount ;

        if(newtic/oldtic > SC_ESTIMATE_TIC_CUTOFF || 
           oldtic/newtic > SC_ESTIMATE_TIC_CUTOFF) return ;

        simulationTic.set(SC_ESTIMATE_TIC_ALPHA * oldtic +
                          (1.0 - SC_ESTIMATE_TIC_ALPHA) * newtic) ;
    } ;

    void adjustBodyState(const ScSensorInfo& sinfo) {
        actualBodyState.set(sinfo.info.senseBody) ;
        
        estimateBodyState.set(actualBodyState) ;

        scycle.set(sinfo.info.senseBody.time) ;
        adjustSimulationTic() ;
        bcycle.set(sinfo.info.senseBody.time) ;
    } ;

    //--------------------------------------------------
    // send command (basic)

    ScBool send() { return socket.send(sendBuffer) ; } ;

    ScBool sendInit(ScFloat version, ScBool goaliep) ;
    ScBool sendReconnect(ScUInt unumber = 0) ;

    ScBool sendMove(ScPos p) { return sendMove(p.x,p.y) ;} ;
    ScBool sendMove(ScFloat x, ScFloat y) ; 

    ScBool sendTurn(ScFloat pow) ;
    ScBool sendDash(ScFloat pow) ;
    ScBool sendKick(ScFloat pow, ScFloat dir) ;

    ScBool sendCatch(ScFloat dir) ;

    ScBool sendTurnNeck(ScFloat pow) ;
      
    ScBool sendSay(ScString msg) ;
    ScBool sendSay(char* msg) { return sendSay(ScString(msg)) ; } ;

    ScBool sendSenseBody() ;
    ScBool sendChangeView(ScViewWidth vw, ScViewQuality vq) ;

    //--------------------------------------------------
    // receive and scan sensor info

    ScBool recvScanSInfo() ;

//V.1.05noda
    ScBool estimatePosDirVel1() {
        if(scEstimatePosDirVelTeam1(eposdir, evel, outOfFieldP,
				    sinfo.info.see, side)) {
            eposdir.bdir = eposdir.hdir - actualBodyState.headAngle ;
            return ScTrue ;
        } else {
            return ScFalse ;
        }
    } ;
//V.1.05noda
    ScBool estimatePosDirVel2() {
        if(scEstimatePosDirVelTeam2(eposdir, eposConf,
                                    evel, evelConf,
				    outOfFieldP,
                                    sinfo.info.see, side)) {
            eposdir.bdir = eposdir.hdir - actualBodyState.headAngle ;
            return ScTrue ;
        } else {
            return ScFalse ;
        }
    } ;

    //--------------------------------------------------
    // basic command

    virtual ScBool waitNextFootCommandTic() ;

    ScBool move(const ScPos& p) {
        return move(p.x,p.y) ;
    } ;
    
    virtual ScBool move(const ScFloat& x, const ScFloat& y) ;

    virtual ScBool turn(const ScFloat& pow) ;

    virtual ScBool dash(const ScFloat& pow) ;

    virtual ScBool kick(const ScFloat& pow, const ScFloat& dir) ;

    virtual ScBool catchBall(const ScFloat& dir) ;

    virtual ScBool turnNeck(const ScFloat& pow) ;
    
    ScBool say(char* msg) {
        return say(ScString(msg)) ;
    } ;
    
    virtual ScBool say(ScString msg) ;

    //--------------------------------------------------
    // turn to absolute dir/pos

    // 1.05
    virtual ScBool turnTo(const ScFloat& dir) ;

    ScBool turnToRel(const ScPos& rpos) {
        return turnTo(rpos.Angle()) ;
    } ;

    ScBool turnToAbs(const ScPos& apos) {
        return turnToRel(xposdir.pos.relativePosTo(apos)) ;
    } ;

    //--------------------------------------------------
    // kick to absolute dir/pos

    ScBool kickTo(const ScFloat& power, const ScFloat& dir) {
        return kick(power, scNormalizeAngle(dir - xposdir.bdir)) ;
    } ;

    ScBool kickToRel(const ScFloat& power, const ScPos& rpos) {
        return kickTo(power, rpos.Angle()) ;
    } ;

    ScBool kickToAbs(const ScFloat& power, const ScPos& apos) {
        return kickToRel(power, xposdir.pos.relativePosTo(apos)) ;
    } ;

    //--------------------------------------------------
    // describe

    void describe(ostream& o, ScBool detailp = ScTrue,
                  ScBool newlinep = ScTrue) const {
        o << "#ScPlayer[" << teamname 
          << ":" << unum << "/" << scSide2Char(side) << "]" ;
        if(detailp) {
            socket.describe(o,detailp,newlinep) ;
            sinfo.describe(o,detailp,newlinep) ;
            o << "sendBuffer:" << sendBuffer ;
            if(newlinep) o << endl ;
            o << "recvBuffer:" << sinfo ;
        }
        if(newlinep) o << endl ;
    } ;

    friend ostream& operator<< (ostream& o, const ScPlayer& p) {
        p.describe(o,ScFalse,ScFalse) ;
        return o ;
    } ;

} ;
      
////////////////////////////////////////////////////////////////////////
#endif
